/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

        acitapi.h

    Purpose:

        Header file for declaring all ACIT library functions and data 
        structures that are exported to the OSM. This is the main file
        that OSMs include to interface with the ACIT library.
        

--*/

/* Debug message level */
#define ACIT_DBGMSG_TRACE                       (1 << 0)
#define ACIT_DBGMSG_WARNING                     (1 << 1)
#define ACIT_DBGMSG_ERROR                       (1 << 2)
#define ACIT_DBGMSG_FATAL                       (1 << 3)
#define ACIT_DBGMSG_UCODE                       (1 << 4)
#define ACIT_DBGMSG_MGMT                        (1 << 5)

/*
 * Definitions for hardware buses supported by ACIT Library
 */
#define ACIT_DEVICE_BUS_PCI                     (0x01)

/*
 * Definition for ACIT Operating Memory Level
 */
#define ACIT_MEM_LEVEL_MINIMUM                      (0)
#define ACIT_MEM_LEVEL_1                            (1)
#define ACIT_MEM_LEVEL_2                            (2)
#define ACIT_MEM_LEVEL_3                            (3)
#define ACIT_MEM_LEVEL_MAXIMUM                      (4)

#define ACIT_MEM_LEVEL_DEFAULT                      (2)

/*
 * Definition for Memory alignment (8-byte) required by ACIT.
 */
#define ACIT_MEM_ALIGNMENT_MASK                 (0x7)


/*
 * Definitions for Flag values for ACIT_PARAMETER
 */
/* These definitions should be a bit-filed */
#define ACIT_PFLAG_PARAMETER_STRING             (0x1)
#define ACIT_PFLAG_PARAMETER_NUMERIC            (0x2)
#define ACIT_PFLAG_PARAMETER_POWEROF2           (0x4)
/* Definitions for OSM Parameter flags */
#define ACIT_OSM_PFLAG_DEFAULT_VALUE            (8)

/*
 * Definition for Target Table flags maintained by OSMs
 */
#define ACIT_OSM_TFLAG_IN_USE                   (1 << 0)
#define ACIT_OSM_TFLAG_NAILED                   (1 << 1)

/*
 * Definition for Target Table flags maintained by ACIT
 */
#define ACIT_TFLAG_IN_USE                       (1 << 0)
#define ACIT_TFLAG_NEW                          (1 << 1)
#define ACIT_TFLAG_REMOVE_PENDING               (1 << 2)
#define ACIT_TFLAG_BUSY                         (1 << 3) /* don't start I/O */
#define ACIT_TFLAG_ACTIVE                       (1 << 4) /* I/Os are pending */




#define ACIT_IOCTL_FLAG_POLLED 			(1 << 0) 

/*
 * Definition for Flags for Register Mapping
 */
#define ACIT_RFLAG_LITTLE_ENDIAN                (1U << 0)
#define ACIT_RFLAG_BIG_ENDIAN                   (1U << 1)


/*
 * Definitions for the ACIT_CONFIG_PARAM structure.
 */
#define ACIT_MAX_CONFIG_NAME_LENGTH             (22)
#define ACIT_MAX_CONFIG_PARAMETER_LENGTH        (40)


/*
 * Definition for Memory types for ACIT_MEM_PARAMETER
 */
#define ACIT_MEM_CATEGORY_CACHED                (1U << 0)
#define ACIT_MEM_CATEGORY_NON_CACHED            (1U << 1)
#define ACIT_MEM_CATEGORY_NON_VOLATILE          (1U << 2)
#define ACIT_MEM_CATEGORY_BELOW_4GB             (1U << 3)

/*
 * Definition for alignment types.
 */
#define ACIT_ALIGN_1BYTE                        (0x00000000)
#define ACIT_ALIGN_2BYTE                        (0x00000001)
#define ACIT_ALIGN_4BYTE                        (0x00000003)
#define ACIT_ALIGN_8BYTE                        (0x00000007)
#define ACIT_ALIGN_16BYTE                       (0x0000000f)
#define ACIT_ALIGN_32BYTE                       (0x0000001f)
#define ACIT_ALIGN_64BYTE                       (0x0000003f)
#define ACIT_ALIGN_128BYTE                      (0x0000007f)
#define ACIT_ALIGN_256BYTE                      (0x000000ff)
#define ACIT_ALIGN_512BYTE                      (0x000001ff)
#define ACIT_ALIGN_1024BYTE                     (0x000003ff)
#define ACIT_ALIGN_2048BYTE                     (0x000007ff)
#define ACIT_ALIGN_4096BYTE                     (0x00000fff)

/*
 * Definition for intialization types
 */
#define ACIT_INIT_TYPE_NORMAL                   (1)
#define ACIT_INIT_TYPE_POWER_MGMT               (2)
#define ACIT_INIT_TYPE_HIBERNATION              (3)
#define ACIT_INIT_TYPE_CRASHDUMP                (4)

#define ACIT_STOP_TYPE_NORMAL                   (1)
#define ACIT_STOP_TYPE_POWER_MGMT               (2)

/*
 * Definition for target discovery opcodes.
 */
#define ACIT_DISCOVER_OPCODE_SYNCHRONOUS        (1)
#define ACIT_DISCOVER_OPCODE_ASYNCHRONOUS       (2)

/*
 * Definition for ACIT Async Class and Event codes.
 */
#define ACIT_EVENT_CLASS_CONTROLLER             (1)
#define ACIT_EVENT_CLASS_TARGET                 (2)
#define ACIT_EVENT_CLASS_LUN                    (3)

#define ACIT_EVENT_CONTROLLER_BUSY              (1)
#define ACIT_EVENT_CONTROLLER_READY             (2)
#define ACIT_EVENT_CONTROLLER_LINK_DOWN         (3)
#define ACIT_EVENT_CONTROLLER_LINK_UP           (4)
#define ACIT_EVENT_CONTROLLER_DISCOVERY_COMPLETE 	(5)
#define ACIT_EVENT_CONTROLLER_LINKDOWN_TIMEOUT  (6)

#define ACIT_EVENT_TARGET_ADD                   (16)
#define ACIT_EVENT_TARGET_REMOVE                (17)
#define ACIT_EVENT_TARGET_BUSY                  (18)
#define ACIT_EVENT_TARGET_READY                 (19)
#define ACIT_EVENT_TARGET_MISSING               (20)
#define ACIT_EVENT_TARGET_RECONNECTED           (21)
#define ACIT_EVENT_TARGET_REMOVE_PENDING        (22)
#define ACIT_EVENT_TARGET_RESCAN_REQUIRED       (23)

/*
 * Definition for ACIT TMF function codes 
 */
#define ACIT_TMF_ABORT_TASK                      (1)
#define ACIT_TMF_ABORT_TASK_SET                  (2)
#define ACIT_TMF_CLEAR_ACA                       (3)
#define ACIT_TMF_CLEAR_TASK_SET                  (4)
#define ACIT_TMF_LUN_RESET                       (5)
#define ACIT_TMF_TARGET_WARM_RESET               (6)
#define ACIT_TMF_TARGET_COLD_RESET               (7)
#define ACIT_TMF_TASK_REASSIGN                   (8)
#define ACIT_TMF_ABORT_MULTIPLE_IOS              (9)

/*
 * Definition for ACIT Operating modes
 */
#define ACIT_OP_MODE_NORMAL                     (1)
#define ACIT_OP_MODE_RESTRICTED                 (2)

/*
 * Definition for ACIT API version.
 */
#define ACIT_API_HI_VERSION                     (2)
#define ACIT_API_LO_VERSION                     (2)

/* 
 * API table version definitions 
 */
#define ACIT_OSM_API_TABLE_HI_VERSION           (1)
#define ACIT_OSM_API_TABLE_LO_VERSION           (2)

typedef PACIT_VOID ACIT_REGISTER_HANDLE, *PACIT_REGISTER_HANDLE;
struct _ACIT_CONTROLLER_KEY;
struct _ACIT_IO_REQUEST;
struct _ACIT_IOCTL_REQUEST;
struct _ACIT_TMF_REQUEST;

typedef struct _ACIT_KEY
{
    PACIT_VOID      acit_key;
    PACIT_VOID      osm_key;
} ACIT_KEY, *PACIT_KEY;

typedef ACIT_KEY                    ACIT_CONTROLLER_KEY, *PACIT_CONTROLLER_KEY;
typedef ACIT_KEY                    ACIT_TARGET_KEY,     *PACIT_TARGET_KEY;
typedef ACIT_KEY                    ACIT_LUN_KEY,        *PACIT_LUN_KEY;
typedef ACIT_KEY                    ACIT_IO_KEY,         *PACIT_IO_KEY;
typedef ACIT_KEY                    ACIT_TMF_KEY,        *PACIT_TMF_KEY;

typedef ACIT_UEXACT32 (* PACIT_IO_REQUEST_POST)(PACIT_CONTROLLER_KEY,
                                                struct _ACIT_IO_REQUEST *);

typedef ACIT_UEXACT32 (* PACIT_IOCTL_REQUEST_POST) (PACIT_CONTROLLER_KEY,
                                                struct _ACIT_IOCTL_REQUEST *);

typedef ACIT_UEXACT32 (* PACIT_TMF_REQUEST_POST) (PACIT_CONTROLLER_KEY,
                                                struct _ACIT_TMF_REQUEST *);


typedef enum _ACIT_STATUS
{
    ACIT_STATUS_SUCCESS = 0,
    ACIT_STATUS_ERROR,
    ACIT_STATUS_INVALID_PARAMETER,  /* 2 */
    ACIT_STATUS_INCORRECT_VERSION,
    ACIT_STATUS_INSUFFICIENT_RESOURCE,
    ACIT_STATUS_INSUFFICIENT_TIMEOUT,
    ACIT_STATUS_LINK_DOWN_TIMEOUT,
    ACIT_STATUS_TARGET_TIMEOUT,
    ACIT_STATUS_NO_DEVICE,
    ACIT_STATUS_UNKNOWN_INTERRUPT,
    ACIT_STATUS_MORE_MEMORY,
    ACIT_STATUS_PENDING,
    ACIT_STATUS_DUPLICATE_ENTRY,
    ACIT_STATUS_RESET_IN_PROGRESS
} ACIT_STATUS, *PACIT_STATUS;


typedef enum _ACIT_IO_STATUS
{
    ACIT_IO_STATUS_SUCCESS = 0,
    ACIT_IO_STATUS_ERROR,
    ACIT_IO_STATUS_TRANSPORT_ERROR,
    ACIT_IO_STATUS_CRC_ERROR,
    ACIT_IO_STATUS_TARGET_DELETED,
    ACIT_IO_STATUS_ABORTED,
    ACIT_IO_STATUS_IO_TIMEOUT,
    ACIT_IO_STATUS_LINK_DOWN_TIMEOUT,
    ACIT_IO_STATUS_TARGET_TIMEOUT,
    ACIT_IO_STATUS_TARGET_ERROR
} ACIT_IO_STATUS, *PACIT_IO_STATUS;

typedef enum _ACIT_TMF_STATUS
{
    ACIT_TMF_STATUS_SUCCESS = 0,
    ACIT_TMF_STATUS_ERROR = 81,
    ACIT_TMF_STATUS_NOT_SUPPORTED,
    ACIT_TMF_STATUS_FUNCTION_REJECTED,
    ACIT_TMF_STATUS_INVALID_TASK,
    ACIT_TMF_STATUS_IN_PROGRESS,
    ACIT_TMF_STATUS_TIMEOUT,
    ACIT_TMF_STATUS_LINK_DOWN_TIMEOUT,
    ACIT_TMF_STATUS_TARGET_TIMEOUT
} ACIT_TMF_STATUS, *PACIT_TMF_STATUS;

/*
 * SCSI Device types.
 */

typedef enum _ACIT_SCSI_DEVTYPE
{
    ACIT_SCSI_DEVTYPE_DISK = 0x00,
    ACIT_SCSI_DEVTYPE_TAPE = 0x01,
    ACIT_SCSI_DEVTYPE_PRINTER = 0x02,
    ACIT_SCSI_DEVTYPE_PROCESSOR = 0x03,
    ACIT_SCSI_DEVTYPE_WORM = 0x04,
    ACIT_SCSI_DEVTYPE_CDROM = 0x05,
    ACIT_SCSI_DEVTYPE_SCANNER = 0x06,
    ACIT_SCSI_DEVTYPE_OPTICAL = 0x07,
    ACIT_SCSI_DEVTYPE_MEDIUMCHANGER = 0x08,
    ACIT_SCSI_DEVTYPE_COMMUNICATION = 0x09,
    ACIT_SCSI_DEVTYPE_RAIDCTRL = 0x0c,
    ACIT_SCSI_DEVTYPE_SES = 0x0d,
    ACIT_SCSI_DEVTYPE_RBC = 0x0e,
    ACIT_SCSI_DEVTYPE_OCRW = 0x0f,
    ACIT_SCSI_LU_NOT_PRESENT_DEVICE = 0x7f,
    ACIT_SCSI_DEVTYPE_UNKNOWN = 0xff

} ACIT_SCSI_DEVTYPE, *PACIT_SCSI_DEVTYPE;

#define    ACIT_SCSI_MAX_CDB_LENGTH                    (16)

/*
 * Definitions for SCSI Completion Status as defined by SAM-2
 */
#define ACIT_SCSI_STATUS_GOOD                       (0x00)
#define ACIT_SCSI_STATUS_CHECK_CONDITION            (0x02)
#define ACIT_SCSI_STATUS_CONDITION_MET              (0x04)
#define ACIT_SCSI_STATUS_BUSY                       (0x08)
#define ACIT_SCSI_STATUS_INTERMEDIATE               (0x10)
#define ACIT_SCSI_STATUS_INTERMEDIATE_CONDITION_MET (0x14)
#define ACIT_SCSI_STATUS_RESERVATION_CONFLICT       (0x18)
#define ACIT_SCSI_STATUS_OBSELETE                   (0x22)
#define ACIT_SCSI_STATUS_TASK_SET_FULL              (0x28)
#define ACIT_SCSI_STATUS_ACA_ACTIVE                 (0x30)
#define ACIT_SCSI_STATUS_TASK_ABORTED               (0x40)

/*
 * Task attribute definitions
 */
#define ACIT_TASK_ATTRIBUTE_UNTAGGED                (0)       
#define ACIT_TASK_ATTRIBUTE_SIMPLE                  (1)       
#define ACIT_TASK_ATTRIBUTE_ORDERED                 (2)       
#define ACIT_TASK_ATTRIBUTE_HEAD_OF_QUEUE           (3)     

typedef struct _ACIT_LUN
{
    union
    {
        ACIT_UEXACT8            lun_bytes[8];
        ACIT_UEXACT16           lun_leveled[4];
        ACIT_U64                lun_complete;
    } u;
} ACIT_LUN, *PACIT_LUN;


typedef struct _ACIT_VERSION
{
    ACIT_UEXACT16           hi_ver;
    ACIT_UEXACT16           lo_ver;
} ACIT_VERSION, *PACIT_VERSION;


typedef struct _ACIT_DEVICE
{
    ACIT_VERSION                acit_version;
    ACIT_UEXACT8                vendor_id_str[8];
    ACIT_UEXACT8                device_id_str[8];
    ACIT_UEXACT16               vendor_id;
    ACIT_UEXACT16               device_id;
    ACIT_UEXACT16               subsys_id;
    ACIT_UEXACT16               subsys_vendor_id;
    ACIT_UEXACT8                revision_id;
    ACIT_UEXACT8                reserved;
    ACIT_UEXACT16               bus_type;
} ACIT_DEVICE, *PACIT_DEVICE;


typedef struct _ACIT_TARGET_ENTRY
{
    PACIT_TARGET_KEY            ptarget_key;
    ACIT_UEXACT32               session_handle;
    ACIT_UEXACT16               osm_tflags;
    ACIT_UEXACT16               acit_tflags;
    ACIT_UEXACT8                domain;
    ACIT_UEXACT8                is_login_pending;
    ACIT_UEXACT8                reserved1[2];
} ACIT_TARGET_ENTRY, *PACIT_TARGET_ENTRY;


typedef struct _ACIT_TARGET_TABLE
{
    ACIT_UEXACT32               total_entries;
    ACIT_UEXACT32               occupied_entries;
    ACIT_UEXACT32               change_count;
    ACIT_UEXACT32               reserved;
    ACIT_TARGET_ENTRY           target_entry[1];
} ACIT_TARGET_TABLE, *PACIT_TARGET_TABLE;


#ifdef _WINISCSI_INBOX

typedef struct _ACIT_SG_ELEMENT
{
    ACIT_BUS_ADDRESS            address;
    ACIT_UEXACT32               length;
    ACIT_U32_64                 reserved;
} ACIT_SG_ELEMENT, *PACIT_SG_ELEMENT;

typedef struct _ACIT_SG_LIST
{
    ACIT_UEXACT32               number_elements;
    ACIT_U32_64                 reserved;
    ACIT_SG_ELEMENT             sglist[1];
} ACIT_SG_LIST, *PACIT_SG_LIST;

#else

typedef struct _ACIT_SG_ELEMENT
{
    ACIT_BUS_ADDRESS            address;
    ACIT_UEXACT32               length;
    ACIT_UEXACT32               reserved;
} ACIT_SG_ELEMENT, *PACIT_SG_ELEMENT;

typedef struct _ACIT_SG_LIST
{
    ACIT_UEXACT32               number_elements;
    ACIT_UEXACT32               reserved;
    ACIT_SG_ELEMENT             sglist[1];
} ACIT_SG_LIST, *PACIT_SG_LIST;

#endif // WINISCSI_INBOX

typedef struct _ACIT_MEM_DESCRIPTOR
{
    ACIT_UEXACT32       index;          /* Index of this memory parameter */
    ACIT_UEXACT32       category;       /* type indicates cached/non-cached*/
    ACIT_BUS_ADDRESS    bus_address;    /* Bus address of location */
    PACIT_VOID          virtual_address;/* virtual address to the location */
    ACIT_UEXACT32       num_elements;   /* number of elements in this descriptor */
    ACIT_UEXACT32       alignment_mask; /* Alignment mask for this block */
    ACIT_UEXACT32       size_in_bytes;           /* Size required by memory block */
} ACIT_MEM_DESCRIPTOR, *PACIT_MEM_DESCRIPTOR;


typedef struct _ACIT_CONFIG_PARAM 
{
    ACIT_UEXACT8        osm_config_param[ACIT_MAX_CONFIG_NAME_LENGTH];
    ACIT_UEXACT8        acit_pflag;
    ACIT_UEXACT8        osm_pflag;
    ACIT_UEXACT16       reserved;
    ACIT_UEXACT32       min_value;
    ACIT_UEXACT32       max_value;
    ACIT_UEXACT32       numeric_value;
    ACIT_UEXACT8        string_value[ACIT_MAX_CONFIG_PARAMETER_LENGTH];
} ACIT_CONFIG_PARAM, *PACIT_CONFIG_PARAM;


typedef struct _ACIT_DISCOVER_PARAM
{
    ACIT_UEXACT32       opcode;
    ACIT_UEXACT32       info;
    PACIT_TARGET_TABLE  *pptarget_table;
    ACIT_UEXACT32       result;
    ACIT_UEXACT32       reserved;
} ACIT_DISCOVER_PARAM, *PACIT_DISCOVER_PARAM;

typedef struct _ACIT_IO_REQUEST
{
    PACIT_IO_KEY            pio_key;               /* I/O Key for this IO */
    PACIT_LUN_KEY           plun_key;              /* LUN Key for this IO */
    PACIT_TARGET_KEY        ptarget_key;           /* Target Key for  IO */
    PACIT_CONTROLLER_KEY    pcontroller_key;       /* Controller Key for IO */
                                                
    ACIT_UEXACT8            cdb[ACIT_SCSI_MAX_CDB_LENGTH];/* CBD for this IO */
    ACIT_BUS_ADDRESS        psense_buffer_physical;/* Ptr to physical address
                                                      of Sense Buffer */
    PACIT_SG_LIST           psg_list;              /* Pointer to the SG List */
    PACIT_VOID              psense_buffer_virtual; /* Ptr to Virtual address
                                                      of Sense Buffer */
    PACIT_IO_REQUEST_POST   ppost_routine;         /* Post routine called when 
                                                      this I/O completes */
    PACIT_VOID              pdata_buffer;          /* Ptr to Virtual address
                                                      of Data Buffer for I/O */
    ACIT_LUN                lun;                   /* LUN for this I/O */
    struct                                      
    {                                           
        ACIT_UEXACT32       inbound_data:1;        /* If set, Indicates data is
                                                      coming into host */
        ACIT_UEXACT32       outbound_data:1;       /* If set, Indicates data is
                                                      going out from host */
        ACIT_UEXACT32       reserved1:30;          /* Reserved to hold other 
                                                      unused bits of Flag */
    } flags;                                       /* Other indications */
    PACIT_VOID              posm_reference;        /* OS I/P req ptr for debug 
                                                      purpose */
    ACIT_UEXACT32           target_handle;         /* target for this I/O */
    ACIT_UEXACT32           function;              /* Opcode of IO Request */
    ACIT_UEXACT32           data_transfer_length;  /* Length of data transfer 
                                                      for this I/O */
    ACIT_UEXACT32           cdb_length;            /* Length of CDB */
    ACIT_UEXACT32           sense_buffer_length;   /* Length of SenseBuffer */
    ACIT_UEXACT32           timeout;               /* OSM Provided value to 
                                                      complete IO in given 
                                                      seconds */
    ACIT_UEXACT32           io_status;             /* Status of IO Completion */
    ACIT_UEXACT32           scsi_status;           /* SCSI Completion Status */
    ACIT_UEXACT32           auto_sense_valid;      /* Set to TRUE to indicate if
                                                      ACIT copied Sense Info */
    ACIT_UEXACT32           task_attribute;        /* Queue Action for I/O */
    ACIT_UEXACT32           reserved2;             /* Maybe used to return ITT
                                                      information to OSMs */
    ACIT_UEXACT32           underrun_length;       /* Size of underrun in bytes
                                                      if any */
    ACIT_UEXACT32           overrun_length;        /* Size of overrun in bytes
                                                      if any */
    ACIT_UEXACT32           channel;               /* Submission channel number */
    ACIT_UEXACT32           message_id;            /* msi-x redirection hint */

} ACIT_IO_REQUEST, *PACIT_IO_REQUEST;


typedef struct _ACIT_IOCTL_REQUEST
{
    IOCTL_HEADER                ioctl_hdr;
    PACIT_VOID                  posm_reference;
    PACIT_VOID                  pdata_buffer;
    ACIT_UEXACT32               data_length;
    ACIT_UEXACT32               ioctl_flag;
    PACIT_IOCTL_REQUEST_POST    ppost_routine;
} ACIT_IOCTL_REQUEST, *PACIT_IOCTL_REQUEST;

typedef struct _ACIT_EVENT_LOG_ENTRY {

    ACIT_UEXACT32               size_in_bytes;  // of this structure           
    ACIT_UEXACT32               uec;            // Universal Error Code
    ACIT_UEXACT32               dump_data_size; // size in bytes of dump data
    PACIT_VOID                  dump_data_ptr;  // pointer to dump data
    ACIT_UEXACT32               num_addi_dwords;// number of additional dwords
    PACIT_UEXACT32              addi_dwords_ptr;// pointer to array of additional dwords

} ACIT_EVENT_LOG_ENTRY, *PACIT_EVENT_LOG_ENTRY;

typedef struct _ACIT_TMF_REQUEST
{
    ACIT_UEXACT8            tmf_function;          /* Type of TMF Request */
    ACIT_UEXACT8            acit_flag;
    ACIT_UEXACT16           osm_flag;
    ACIT_UEXACT32           tmf_status;            /* Status of TMF Completion */
    PACIT_CONTROLLER_KEY    pcontroller_key;       /* Controller Key for TMF */
    PACIT_TARGET_KEY        ptarget_key;           /* Target Key for  TMF */
    PACIT_LUN_KEY           plun_key;              /* LUN Key for this TMF */
    PACIT_TMF_KEY           ptmf_key;              /* TMF Key for this TMF */
    ACIT_LUN                lun;                   /* LUN address used in LUN reset */
    PACIT_IO_REQUEST        pio_request;           /* I/O Request Pointer */
    PACIT_TMF_REQUEST_POST  ppost_routine;         /* Post routine called when 
                                                      this TMF completes */
    PACIT_VOID              posm_reference;        /* OS I/P req ptr for */ 
    ACIT_UEXACT32           timeout;               /* OSM Provided value to 
                                                      complete TMF in given 
                                                      seconds */
} ACIT_TMF_REQUEST, *PACIT_TMF_REQUEST;

typedef struct _ACIT_LUN_INFO
{
    ACIT_UEXACT8        device_type;
    ACIT_UEXACT8        rsvd_unused[3];
} ACIT_LUN_INFO, *PACIT_LUN_INFO;

typedef struct _ACIT_CAPABILITIES
{
    ACIT_UEXACT32           num_cxn;
    ACIT_UEXACT32           num_ios;
    ACIT_UEXACT32           interrupt_redirection;

} ACIT_CAPABILITIES, *PACIT_CAPABILITIES;


/* Define typedefs for all OSM API procedures */
typedef ACIT_UEXACT32 (OSM_GET_CONFIG_VALUE_T)  (PACIT_CONTROLLER_KEY,
                                                 PACIT_CONFIG_PARAM);
typedef ACIT_UEXACT32 (OSM_COMPLETE_IO_T)       (PACIT_CONTROLLER_KEY,
                                                 PACIT_IO_REQUEST);
typedef ACIT_UEXACT32 (OSM_COMPLETE_TMF_T)      (PACIT_CONTROLLER_KEY,
                                                 PACIT_TMF_REQUEST);
typedef ACIT_UEXACT32 (OSM_COMPLETE_IOCTL_T)    (PACIT_CONTROLLER_KEY,
                                                 PACIT_IOCTL_REQUEST);
typedef ACIT_UEXACT32 (OSM_ASYNC_EVENT_T)       (PACIT_CONTROLLER_KEY,
                                                 ACIT_UEXACT32,
                                                 ACIT_UEXACT32,
                                                 ACIT_UEXACT32);
typedef ACIT_UEXACT8  (OSM_READ8_T)             (PACIT_CONTROLLER_KEY,
                                                 ACIT_REGISTER_HANDLE,
                                                 ACIT_UEXACT32);
typedef ACIT_UEXACT16 (OSM_READ16_T)            (PACIT_CONTROLLER_KEY,
                                                 ACIT_REGISTER_HANDLE,
                                                 ACIT_UEXACT32);
typedef ACIT_UEXACT32 (OSM_READ32_T)            (PACIT_CONTROLLER_KEY,
                                                 ACIT_REGISTER_HANDLE,
                                                 ACIT_UEXACT32);
typedef ACIT_UEXACT32 (OSM_READ_BUFFER_T)       (PACIT_CONTROLLER_KEY,
                                                 ACIT_REGISTER_HANDLE,
                                                 ACIT_UEXACT32,
                                                 PACIT_VOID,
                                                 ACIT_UEXACT32);
typedef ACIT_VOID     (OSM_WRITE8_T)            (PACIT_CONTROLLER_KEY,
                                                 ACIT_REGISTER_HANDLE,
                                                 ACIT_UEXACT32,
                                                 ACIT_UEXACT8);
typedef ACIT_VOID     (OSM_WRITE16_T)           (PACIT_CONTROLLER_KEY,
                                                 ACIT_REGISTER_HANDLE,
                                                 ACIT_UEXACT32,
                                                 ACIT_UEXACT16);
typedef ACIT_VOID     (OSM_WRITE32_T)           (PACIT_CONTROLLER_KEY,
                                                 ACIT_REGISTER_HANDLE,
                                                 ACIT_UEXACT32,
                                                 ACIT_UEXACT32);
typedef ACIT_UEXACT32 (OSM_WRITE_BUFFER_T)      (PACIT_CONTROLLER_KEY,
                                                 ACIT_REGISTER_HANDLE,
                                                 ACIT_UEXACT32,
                                                 PACIT_VOID,
                                                 ACIT_UEXACT32);
typedef ACIT_UEXACT32 (OSM_READ_CONFIG8_T )     (PACIT_CONTROLLER_KEY,
                                                 ACIT_UEXACT32,
                                                 PACIT_UEXACT8);
typedef ACIT_UEXACT32 (OSM_READ_CONFIG16_T )    (PACIT_CONTROLLER_KEY,
                                                 ACIT_UEXACT32,
                                                 PACIT_UEXACT16);
typedef ACIT_UEXACT32 (OSM_READ_CONFIG32_T )    (PACIT_CONTROLLER_KEY,
                                                 ACIT_UEXACT32,
                                                 PACIT_UEXACT32);
typedef ACIT_UEXACT32 (OSM_READ_CONFIG_BUFFER_T )(PACIT_CONTROLLER_KEY);
typedef ACIT_UEXACT32 (OSM_WRITE_CONFIG8_T )    (PACIT_CONTROLLER_KEY,
                                                 ACIT_UEXACT32,
                                                 ACIT_UEXACT8);
typedef ACIT_UEXACT32 (OSM_WRITE_CONFIG16_T )   (PACIT_CONTROLLER_KEY,
                                                 ACIT_UEXACT32,
                                                 ACIT_UEXACT16);
typedef ACIT_UEXACT32 (OSM_WRITE_CONFIG32_T )   (PACIT_CONTROLLER_KEY,
                                                 ACIT_UEXACT32,
                                                 ACIT_UEXACT32);
typedef ACIT_UEXACT32 (OSM_WRITE_CONFIG_BUFFER_T )(PACIT_CONTROLLER_KEY);
typedef ACIT_BUS_ADDRESS(OSM_GET_BUS_ADDRESS_T) (PACIT_CONTROLLER_KEY,
                                                 PACIT_VOID,
                                                 PACIT_VOID,
                                                 PACIT_UEXACT32);
typedef ACIT_UEXACT32 (OSM_DELAY_TIMER_T)       (PACIT_CONTROLLER_KEY,
                                                 ACIT_UEXACT32);
typedef ACIT_UEXACT32 (OSM_MAP_IO_HANDLE_T)     (PACIT_CONTROLLER_KEY,
                                                 ACIT_UEXACT32,
                                                 ACIT_UEXACT32,
                                                 ACIT_UEXACT32,
                                                 PACIT_BUS_ADDRESS,
                                                 PACIT_VOID *,
                                                 PACIT_REGISTER_HANDLE);
typedef ACIT_UEXACT32 (OSM_RELEASE_IO_HANDLE_T) (PACIT_CONTROLLER_KEY,
                                                 ACIT_REGISTER_HANDLE);
typedef ACIT_VOID     (OSM_LOG_ERROR_T)         (PACIT_CONTROLLER_KEY,
                                                 ACIT_UEXACT32);
typedef ACIT_VOID     (OSM_LOG_EVENT_T)         (PACIT_CONTROLLER_KEY,
                                                 PACIT_EVENT_LOG_ENTRY);

/* Table of OSM API procedures */
typedef struct _ACIT_OSM_API_TABLE{
    ACIT_VERSION                osm_api_table_version;
    OSM_GET_CONFIG_VALUE_T      *posm_get_config_value;
    OSM_COMPLETE_IO_T           *posm_complete_io;
    OSM_COMPLETE_IOCTL_T        *posm_complete_ioctl;
    OSM_COMPLETE_TMF_T          *posm_complete_tmf;
    OSM_ASYNC_EVENT_T           *posm_async_event;
    
    OSM_READ8_T                 *posm_read8;
    OSM_READ16_T                *posm_read16;
    OSM_READ32_T                *posm_read32;
    OSM_READ_BUFFER_T           *posm_read_buffer;
    OSM_WRITE8_T                *posm_write8;
    OSM_WRITE16_T               *posm_write16;
    OSM_WRITE32_T               *posm_write32;
    OSM_WRITE_BUFFER_T          *posm_write_buffer;
    OSM_READ_CONFIG8_T          *posm_read_config8;
    OSM_READ_CONFIG16_T         *posm_read_config16;
    OSM_READ_CONFIG32_T         *posm_read_config32;
    OSM_READ_CONFIG_BUFFER_T    *posm_read_config_buffer;
    OSM_WRITE_CONFIG8_T         *posm_write_config8;
    OSM_WRITE_CONFIG16_T        *posm_write_config16;
    OSM_WRITE_CONFIG32_T        *posm_write_config32;
    OSM_WRITE_CONFIG_BUFFER_T   *posm_write_config_buffer;
    OSM_GET_BUS_ADDRESS_T       *posm_get_bus_address;
    OSM_DELAY_TIMER_T           *posm_delay_timer;
    OSM_MAP_IO_HANDLE_T         *posm_map_io_handle;
    OSM_RELEASE_IO_HANDLE_T     *posm_release_io_handle;
    OSM_LOG_ERROR_T             *posm_log_error;
    OSM_LOG_EVENT_T             *posm_log_event;
} ACIT_OSM_API_TABLE, *PACIT_OSM_API_TABLE;

/*
 * Typedefs for ACIT API functions
 */
typedef ACIT_STATUS (ACIT_GET_API_VERSION_T)(
    PACIT_VERSION               pacit_api_version);

typedef ACIT_STATUS (ACIT_GET_SUPPORTED_DEVICE_T)(
    ACIT_UEXACT32               index,
    PACIT_DEVICE                pacit_device);

typedef ACIT_STATUS (ACIT_CHECK_SUPPORTED_DEVICE_T)(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_DEVICE                pacit_device);

typedef ACIT_STATUS (ACIT_GET_WS_SIZES_T)(
    PACIT_UEXACT32              pcontroller_ws_size,
    PACIT_UEXACT32              ptarget_ws_size,
    PACIT_UEXACT32              plun_ws_size,
    PACIT_UEXACT32              pio_ws_size,
    PACIT_UEXACT32              ptmf_ws_size);

typedef ACIT_STATUS (ACIT_INIT_CONTROLLER_WS_T)(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_VOID                  pcontroller_ws);

typedef ACIT_STATUS (ACIT_INIT_TARGET_WS_T)(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_TARGET_KEY            ptarget_key,
    PACIT_VOID                  ptarget_ws);

typedef ACIT_STATUS (ACIT_INIT_LUN_WS_T)(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_TARGET_KEY            ptarget_key,
    PACIT_LUN_KEY               plun_key,
    PACIT_VOID                  plun_ws);

typedef ACIT_STATUS (ACIT_INIT_IO_WS_T)(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_IO_REQUEST            pio_request,
    PACIT_IO_KEY                pio_key,
    PACIT_VOID                  pio_ws);

typedef ACIT_STATUS (ACIT_INIT_TMF_WS_T)(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_TMF_REQUEST           ptmf_request,
    PACIT_TMF_KEY               ptmf_key,
    PACIT_VOID                  ptmf_ws);

typedef ACIT_STATUS (ACIT_SET_OSM_API_TABLE_T)(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_OSM_API_TABLE         posm_api_table,
    ACIT_UEXACT32               table_length);

typedef ACIT_STATUS (ACIT_INIT_CONFIGURABLE_PARAMS_T)(
    PACIT_CONTROLLER_KEY        pcontroller_key);

typedef ACIT_STATUS (ACIT_INIT_CONTROLLER_T)(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    ACIT_UEXACT32               init_type,
    PACIT_VOID                  osm_handle);

typedef ACIT_STATUS (ACIT_STOP_CONTROLLER_T)(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    ACIT_UEXACT32               stop_type);

typedef ACIT_STATUS (ACIT_DISCOVER_TARGETS_T)(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_DISCOVER_PARAM        pdiscover_param);

typedef ACIT_STATUS (ACIT_ENABLE_IRQ_T)(
    PACIT_CONTROLLER_KEY        pcontroller_key);

typedef ACIT_STATUS (ACIT_DISABLE_IRQ_T)(
    PACIT_CONTROLLER_KEY        pcontroller_key);

typedef ACIT_STATUS (ACIT_PROCESS_IRQ_T)(
    PACIT_CONTROLLER_KEY        pcontroller_key);

typedef ACIT_STATUS (ACIT_PROCESS_MSI_IRQ_T)(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    ACIT_UEXACT32               message_id);

typedef ACIT_STATUS (ACIT_START_IO_T)(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_IO_REQUEST            pio_request);

typedef ACIT_STATUS (ACIT_START_TMF_T)(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_TMF_REQUEST           ptmf_request);

typedef ACIT_STATUS (ACIT_START_IOCTL_T)(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_IOCTL_REQUEST         pioctl_request);

typedef ACIT_STATUS (ACIT_PROCESS_TIMER_TICK_T)(
    PACIT_CONTROLLER_KEY        pcontroller_key);

typedef ACIT_STATUS (ACIT_SET_MEM_PARAMS_T)(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    ACIT_UEXACT32               operating_mem_level);

typedef ACIT_STATUS (ACIT_GET_MEM_PARAMS_T)(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_UEXACT32              poperating_mem_level);

typedef ACIT_STATUS (ACIT_GET_MEM_DESCRIPTOR_T)(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_DEVICE                pacit_device,
    ACIT_UEXACT32               index,
    PACIT_MEM_DESCRIPTOR        pmem_descriptor);

typedef ACIT_STATUS (ACIT_SET_MEM_DESCRIPTOR_T)(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    ACIT_UEXACT32               index,
    PACIT_MEM_DESCRIPTOR        pmem_descriptor);

typedef ACIT_STATUS (ACIT_UPDATE_LUN_INFO_T) (
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_TARGET_KEY            ptarget_key,
    PACIT_LUN_KEY               plun_key,
    PACIT_LUN_INFO              plun_info);

typedef ACIT_STATUS (ACIT_GET_CAPABILITIES_T)(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_CAPABILITIES          pcapabilities);


typedef struct _ACIT_API_TABLE
{
    ACIT_VERSION                    api_table_version;
    ACIT_GET_API_VERSION_T          *pacit_get_api_version;
    ACIT_INIT_CONTROLLER_WS_T       *pacit_init_controller_ws;
    ACIT_INIT_TARGET_WS_T           *pacit_init_target_ws;
    ACIT_INIT_LUN_WS_T              *pacit_init_lun_ws;
    ACIT_INIT_IO_WS_T               *pacit_init_io_ws;
    ACIT_INIT_TMF_WS_T              *pacit_init_tmf_ws;
    ACIT_SET_OSM_API_TABLE_T        *pacit_set_osm_api_table;
    ACIT_INIT_CONFIGURABLE_PARAMS_T *pacit_init_configurable_params;
    ACIT_SET_MEM_PARAMS_T           *pacit_set_mem_params;
    ACIT_GET_MEM_PARAMS_T           *pacit_get_mem_params;
    ACIT_GET_MEM_DESCRIPTOR_T       *pacit_get_mem_descriptor;
    ACIT_SET_MEM_DESCRIPTOR_T       *pacit_set_mem_descriptor;
    ACIT_INIT_CONTROLLER_T          *pacit_init_controller;
    ACIT_STOP_CONTROLLER_T          *pacit_stop_controller;
    ACIT_DISCOVER_TARGETS_T         *pacit_discover_targets;
    ACIT_ENABLE_IRQ_T               *pacit_enable_irq;
    ACIT_DISABLE_IRQ_T              *pacit_disable_irq;
    ACIT_PROCESS_IRQ_T              *pacit_process_irq;
    ACIT_PROCESS_MSI_IRQ_T          *pacit_process_msi_irq;
    ACIT_START_IO_T                 *pacit_start_io;
    ACIT_START_IOCTL_T              *pacit_start_ioctl;
    ACIT_START_TMF_T                *pacit_start_tmf;
    ACIT_PROCESS_TIMER_TICK_T       *pacit_process_timer_tick;
    ACIT_UPDATE_LUN_INFO_T          *pacit_update_lun_info;
    ACIT_GET_CAPABILITIES_T         *pacit_get_capabilities;

} ACIT_API_TABLE, *PACIT_API_TABLE;





/*
 * Functions Exported from ACIT Library
 */
ACIT_STATUS
acit_initialize(
    ACIT_UEXACT32               operating_mode,
    PACIT_API_TABLE             pacit_api_table,
    ACIT_UEXACT32               table_length);

ACIT_STATUS
acit_get_api_version(
    PACIT_VERSION               pacit_api_version);

ACIT_STATUS
acit_get_supported_device(
    ACIT_UEXACT32               index,
    PACIT_DEVICE                pacit_device);

ACIT_STATUS
acit_check_supported_device(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_DEVICE                pacit_device);

ACIT_STATUS
acit_get_ws_sizes(
    PACIT_VERSION               pacit_api_version,
    PACIT_UEXACT32              pcontroller_ws_size,
    PACIT_UEXACT32              ptarget_ws_size,
    PACIT_UEXACT32              plun_ws_size,
    PACIT_UEXACT32              pio_ws_size,
    PACIT_UEXACT32              ptmf_ws_size);

ACIT_STATUS
acit_init_controller_ws(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_VOID                  pcontroller_ws);

ACIT_STATUS
acit_init_target_ws(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_TARGET_KEY            ptarget_key,
    PACIT_VOID                  ptarget_ws);

ACIT_STATUS
acit_init_lun_ws(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_TARGET_KEY            ptarget_key,
    PACIT_LUN_KEY               plun_key,
    PACIT_VOID                  plun_ws);

ACIT_STATUS
acit_init_io_ws(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_IO_REQUEST            pio_request,
    PACIT_IO_KEY                pio_key,
    PACIT_VOID                  pio_ws);

ACIT_STATUS
acit_init_tmf_ws(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_TMF_REQUEST           ptmf_request,
    PACIT_TMF_KEY               ptmf_key,
    PACIT_VOID                  ptmf_ws);

ACIT_STATUS
acit_set_osm_api_table(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_OSM_API_TABLE         posm_api_table,
    ACIT_UEXACT32               table_length);

ACIT_STATUS
acit_init_configurable_params(
    PACIT_CONTROLLER_KEY        pcontroller_key);

ACIT_STATUS
acit_get_mem_params(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_UEXACT32              poperating_mem_level);

ACIT_STATUS
acit_set_mem_params(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    ACIT_UEXACT32               operating_mem_level);

ACIT_STATUS
acit_get_mem_descriptor(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_DEVICE                pacit_device,
    ACIT_UEXACT32               index,
    PACIT_MEM_DESCRIPTOR        pmem_descriptor);

ACIT_STATUS
acit_set_mem_descriptor(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    ACIT_UEXACT32               index,
    PACIT_MEM_DESCRIPTOR        pmem_descriptor);

ACIT_STATUS
acit_init_controller(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    ACIT_UEXACT32               init_type,
    PACIT_VOID                  osm_handle);

ACIT_STATUS
acit_stop_controller(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    ACIT_UEXACT32               stop_type);

ACIT_STATUS
acit_discover_targets(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_DISCOVER_PARAM        pdiscover_param);

ACIT_STATUS
acit_enable_irq(
    PACIT_CONTROLLER_KEY        pcontroller_key);

ACIT_STATUS
acit_disable_irq(
    PACIT_CONTROLLER_KEY        pcontroller_key);

ACIT_STATUS
acit_process_irq(
    PACIT_CONTROLLER_KEY        pcontroller_key);

ACIT_STATUS
acit_process_msi_irq(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    ACIT_UEXACT32               message_id);

ACIT_STATUS
acit_start_io(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_IO_REQUEST            pio_request);

ACIT_STATUS
acit_start_tmf(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_TMF_REQUEST           ptmf_request);

ACIT_STATUS
acit_start_ioctl(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_IOCTL_REQUEST         pioctl_request);

ACIT_STATUS
acit_process_timer_tick(
    PACIT_CONTROLLER_KEY        pcontroller_key);


ACIT_STATUS
acit_update_lun_info(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_TARGET_KEY            ptarget_key,
    PACIT_LUN_KEY               plun_key,
    PACIT_LUN_INFO              plun_info);

ACIT_STATUS
acit_get_capabilities(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_CAPABILITIES          pcapabilities);

//
//ACIT_STATUS
//acit_process_stat_ioctl(
//    PACIT_CONTROLLER_KEY        pcontroller_key,
//    PDRVR_STATS_REQUEST         pstats_hdr);
//

