/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

    iscsiapi.h - iSCSI Layer Interface Routines

    Purpose:

        The iSCSI layer interface routines for ACIT and management layer.
        

--*/


#define ISID_LENGTH                 6
/*
 * This is used for the size of array of pointers to
 * connection ws in ISCSI_TARGET_WS
 */
#define MAX_CXNS_PER_SESSION        1

/*
 * This is used for timeout on 0-window updates for a target.
 * This is only used if ETO is 0
 */
#define DEFAULT_PING_WAIT_TICKS     50

typedef enum _ISCI_STATUS
{
    ISCSI_STATUS_SUCCESS = 0,
    ISCSI_STATUS_FAILED,
    ISCSI_STATUS_NO_CONNECTIONS,  /* 2 */
}
ISCI_STATUS;

typedef enum _ISCSI_LOGOUT_REASON
{
    ISCSI_LOGOUT_REASON_CLOSE_SESSION = 0,
    ISCSI_LOGOUT_REASON_CLOSE_CONNECTION,
    ISCSI_LOGOUT_REASON_REMOVE_CONNECTION_FOR_RECOVERY
}
ISCSI_LOGOUT_REASON;

/* Flag values for ISCSI_CONTROLLER_WS */
#define FLAG_ISCSI_ALL_TGTS_SESSION_RECOVERY    (1U << 0)

/* iSCSI connection status flags */

#define ISCSI_CONN_STATUS_VALID                 (1U << 0)
#define ISCSI_CONN_STATUS_TCP_RESET_PENDING     (1U << 1)
#define ISCSI_CONN_STATUS_RESET                 (1U << 2)
#define ISCSI_CONN_STATUS_INVALIDATE_PENDING    (1U << 3)
#define ISCSI_CONN_STATUS_INVALIDATED           (1U << 4)
#define ISCSI_CONN_STATUS_UPLOADED              (1U << 5)
#define ISCSI_CONN_STATUS_NULL                  (1U << 6)
#define ISCSI_CONN_STATUS_INITIALIZING          (1U << 7)
#define ISCSI_CONN_STATUS_WRITE_REQ_LIMIT       (1U << 8)
#define ISCSI_CONN_CQ_ERROR                     (1U << 9)
#define ISCSI_CONN_OFFLOAD_ERROR                (1U << 10)

// Following are the resource status for connections
#define ISCSI_CONN_STATUS_NEED_INVALIDATION     (1U << 16)
#define ISCSI_CONN_STATUS_NEED_UPLOAD           (1U << 17)
#define ISCSI_CONN_STATUS_NEED_RESET            (1U << 18)

#define ISCSI_CONN_LOGIN_OPTIONS_IPSEC_ENABLED  (1U << 0)
#define ISCSI_CONN_LOGIN_OPTIONS_HEADER_DIGEST  (1U << 1)
#define ISCSI_CONN_LOGIN_OPTIONS_DATA_DIGEST    (1U << 2)

#define ISCSI_GET_TARGET_PORT(ptarget)          ((GET_ISCSI_TARGET_WS(ptarget))->port_num)
#define ISCSI_GET_CMD_SN(ptarget)               ((GET_ISCSI_TARGET_WS(ptarget))->cmd_sn)
#define ISCSI_IS_IMM_DATA(flags) ((flags & ISCSI_SESSION_LOGIN_OPTIONS_IMMEDIATE_DATA) != 0)
#define ISCSI_IS_INIT_R2T(flags) ((flags & ISCSI_SESSION_LOGIN_OPTIONS_INITIAL_R2T) != 0)

#define SCSI_SVC_RESP_GOOD              (0)
#define SCSI_SVC_RESP_TARGET_ERROR      (1)

#define REJECT_ACTION_INVALIDATE_CMD    (1)
#define REJECT_ACTION_INVALIDATE_CXN    (2)
#define REJECT_ACTION_DO_NOTHING        (3)

typedef struct _ISCSI_CONN_LOGIN_OPTIONS
{
    ACIT_UEXACT32    flags;
    ACIT_UEXACT32    max_recv_datasegment_len_ini;
    ACIT_UEXACT32    max_recv_datasegment_len_tgt;

    struct {

        ACIT_UEXACT16   header;
        ACIT_UEXACT16   data;

    } digest;

    /* use these when needed */
    /*
    ACIT_UEXACT32    max_frame_size;
    ACIT_UEXACT32    tcp_windows_size;
    */

} ISCSI_CONN_LOGIN_OPTIONS, *PISCSI_CONN_LOGIN_OPTIONS;

typedef struct _ISCSI_CONNECTION
{
    struct _ISCSI_TARGET_WS *piscsi_target_ws;   /* back pointer to session */
    ACIT_UEXACT32       iscsi_cmd_pdus;
    ACIT_UEXACT32       iscsi_rsp_pdus;
    ACIT_UEXACT16       num_of_active_cmds; /* total number of commands active */
    ACIT_UEXACT16       num_of_active_writes;/* number of writes active */
    
    ACIT_UEXACT16       availability;       /* If this connection is free or available */    
    ACIT_UEXACT16       port_num;           /* The port num this connection is on */
    ACIT_UEXACT32       total_pending_bytes;/* bytes to be transmitted */
    ACIT_UEXACT32       pending_xfer_bytes; /* bytes to be transmitted */
    ACIT_UEXACT32       pending_read_bytes; /* bytes to be read from target */
    ACIT_UEXACT32       total_bytes_xferred;/* total bytes transferred */
    ACIT_UEXACT32       total_bytes_read;   /* total bytes read till now */
    ACIT_UEXACT32       connection_handle;  /* ARM fw connection identifier */
    ACIT_UEXACT32       status;
    ACIT_UEXACT32       busy;
    ACIT_UEXACT32       num_tmfs;

    ACIT_UEXACT16       cid;           /* Offload cid - hardware dependent */
    ACIT_UEXACT16       iscsi_cid;     /* iscsi cid use it for logout pdu */
    ACIT_UEXACT32       cri;           /* Cxn Resource Index - 0 based number */

    ACIT_UEXACT32       async_pending; /* Set when no target is assigned to this connection */

    ACIT_UEXACT32       digest_errors;
    ACIT_UEXACT32       timeout_errors;
    ACIT_UEXACT32       format_errors;

    ISCSI_CONN_LOGIN_OPTIONS login_options;

}ISCSI_CONNECTION, *PISCSI_CONNECTION;

#define ISCSI_CXN_DESC_SZ      (sizeof(struct _ISCSI_CONNECTION) * BE_MAX_CXNS)
#define ISCSI_MEM_NUM_CXN       ACIT_CEIL(ISCSI_CXN_DESC_SZ, ACIT_CFG_MAX_CONTIGUOUS_MEM)

#define GET_ISCSI_CXN(pc, cri)  ((GET_ISCSI_CONTROLLER_WS(pc))->connection_table[cri])

typedef struct _ISCSI_CONTROLLER_WS
{
    PISCSI_CONNECTION         connections_pool[ISCSI_MEM_NUM_CXN];
    struct _ISCSI_CONNECTION* connection_table[BE_MAX_CXNS];
    PACIT_VOID                mgmt_ioctl_request;
    ACIT_UEXACT32             flags;
    ACIT_UEXACT32             cxn_mem_index;
    ACIT_UEXACT32             max_sgl_per_descriptor;
    ACIT_UEXACT32             max_cxnpool_per_desc;
    ACIT_UEXACT32             num_uploads_pending;
    ACIT_UEXACT32             async_pending;

}ISCSI_CONTROLLER_WS, *PISCSI_CONTROLLER_WS;


#define ISCSI_SESSION_LOGIN_OPTIONS_INITIAL_R2T         (1U << 0)
#define ISCSI_SESSION_LOGIN_OPTIONS_IMMEDIATE_DATA      (1U << 1)      
#define ISCSI_SESSION_LOGIN_OPTIONS_DATAPDU_INORDER     (1U << 2)
#define ISCSI_SESSION_LOGIN_OPTIONS_DATASEQ_INORDER     (1U << 3)

typedef struct _ISCSI_SESSION_LOGIN_OPTIONS
{
    UEXACT8     flags;
    UEXACT8     erl;
    UEXACT16    reserved;
    UEXACT32    first_burst_length;
    UEXACT32    max_burst_length;
    UEXACT16    max_connections;
    UEXACT16    max_outstanding_r2t;
    UEXACT16    default_time2wait;
    UEXACT16    default_time2retain;
} ISCSI_SESSION_LOGIN_OPTIONS, *PISCSI_SESSION_LOGIN_OPTIONS;

// iscsi_target_ws flags
#define ISCSI_DATA_DIGEST_ERR       (1)

typedef struct _ISCSI_TARGET_WS
{
    struct _ACIT_TARGET_WS *    pacit_ws;

    ACIT_UEXACT32       session_handle;     /* firmware's session handle */
    ACIT_UEXACT32       session_status;     /* Session status. See MGMT_SESS_STATUS_* definitions in mgmttype.h */
    ACIT_UEXACT32       session_type;       /* normal or discovery */
    ACIT_UEXACT32       cmd_sn;             /* current Cmd Ref Num */
    
    ACIT_UEXACT32       exp_cmd_sn;         /* next expected CmdSN */
    ACIT_UEXACT32       max_cmd_sn;         /* current queue depth of target */
    //ACIT_UEXACT32       min_load;           /* used for the load balancing. it avoids loop */

    ACIT_UEXACT32       flags;
    ACIT_UEXACT16       port_num;           /* physical ports this target can be reached from */
    ACIT_UEXACT16       last_cxn;           /* cxn index on which last cmd was sent; used for round robin algo */
    ACIT_UEXACT16       max_ios_per_cxn;
    ACIT_UEXACT16       curr_cid;           /* current CID for new connection */
    ACIT_UEXACT16       num_connections_invalidated;
    ACIT_UEXACT16       num_connections_reset;
    ACIT_UEXACT16       num_connections_reset_needed;
    ACIT_UEXACT16       num_connections_uploaded;
    ACIT_UEXACT8        num_connections;    /* number of connections allocated currently */
    ACIT_UEXACT8        max_connections;    /* max connections allowed on this session */

    PACIT_VOID          mgmt_ioctl_request;
    
    PISCSI_CONNECTION   pcxn_list[MAX_CXNS_PER_SESSION]; /* connections */

    ISCSI_SESSION_LOGIN_OPTIONS login_options;

}ISCSI_TARGET_WS, *PISCSI_TARGET_WS;

typedef struct _ISCSI_TMF_WS
{
    ACIT_UEXACT16       connection_index;
    ACIT_UEXACT16       reserved;
    ACIT_UEXACT32       unused;

}ISCSI_TMF_WS, *PISCSI_TMF_WS;

//typedef struct _ISCSI_IO_WS
//{
//}
//ISCSI_IO_WS, *PISCSI_IO_WS;

#define GET_ISCSI_CONTROLLER_WS(pcontroller)    ((PISCSI_CONTROLLER_WS)pcontroller->piscsi_ws)
#define GET_ISCSI_TARGET_WS(ptarget)            ((PISCSI_TARGET_WS)ptarget->piscsi_ws)
#define GET_ISCSI_IO_WS(pio)                    ((PISCSI_IO_WS)pio->piscsi_ws)

#define ISCSI_GET_PDU_TEMPLATE_ADDRESS(pc, pa) {\
    pa->lo = pc->mem_descriptors[ISCSI_MEM_GLOBAL_HEADER].bus_address.u.a32.address_lo;  \
    pa->hi = pc->mem_descriptors[ISCSI_MEM_GLOBAL_HEADER].bus_address.u.a32.address_hi;  \
}

ACIT_UEXACT32
iscsi_get_ws_sizes(
    PACIT_UEXACT32          pcontroller_ws_size,
    PACIT_UEXACT32          ptarget_ws_size,
    PACIT_UEXACT32          plun_ws_size,
    PACIT_UEXACT32          pio_ws_size,
    PACIT_UEXACT32          ptmf_ws_size);

ACIT_UEXACT32
iscsi_init_mem_descriptors(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
iscsi_init_controller_ws(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
iscsi_init_controller(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
iscsi_init_target_ws(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_UEXACT16          pconnection_handle,
    PACIT_UEXACT16          pcid,
    PPHYS_ADDR              ptemplate_address);

ACIT_UEXACT32
iscsi_reinit_target_ws(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_UEXACT16          pconnection_handle,
    PACIT_UEXACT16          pcid,
    PPHYS_ADDR              ptemplate_address);

ACIT_UEXACT32
iscsi_update_session_info(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    POFFLOAD_SESSION_OUT    psession);

ACIT_UEXACT32
iscsi_update_session_info_from_async(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           session_handle,
    ACIT_UEXACT32           connection_handle);

ACIT_UEXACT32
iscsi_init_hiber_target_ws(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           session_handle,
    ACIT_UEXACT16           connection_handle);

PISCSI_CONNECTION
iscsi_get_cxn_from_cxn_handle(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget, 
    ACIT_UEXACT32           connection_handle,
    PACIT_UEXACT32          pis_leading_cxn);
    
ACIT_UEXACT32
iscsi_recover_all_sessions(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_VOID              mgmt_ioctl_request);

ACIT_UEXACT32
iscsi_process_connection_event(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           connection_handle,
    ACIT_UEXACT32           session_handle);

ACIT_UEXACT32
iscsi_get_cid_from_connection_handle(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           connection_handle);

// No MCS support
//ACIT_UEXACT32
//iscsi_close_connection(
//    PACIT_CONTROLLER_WS     pcontroller,
//    PACIT_TARGET_WS         ptarget,
//    PACIT_VOID              mgmt_ioctl_request);

ACIT_UEXACT32
iscsi_check_and_close_session(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           recover);

ACIT_UEXACT32
iscsi_logout_session(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_VOID              mgmt_ioctl_request);

ACIT_UEXACT32
iscsi_process_target_event(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           async);

ACIT_UEXACT32
iscsi_init_lun_ws(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun);

ACIT_UEXACT32
iscsi_init_io_ws(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_WS             pio);

ACIT_UEXACT32
iscsi_init_tmf_ws(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_TMF_WS            ptmf);

ACIT_UEXACT32
iscsi_start_io(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_IO_HANDLE         pio_handle);

ACIT_UEXACT32
iscsi_start_logout(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_LOGOUT_HANDLE     plogout_handle);

ACIT_UEXACT32
iscsi_alloc_tmf_resources(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_TARGET_WS         ptarget,
    PACIT_TMF_HANDLE        ptmf_handle);

ACIT_VOID
iscsi_free_tmf_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_HANDLE        ptmf_handle);

ACIT_UEXACT32
iscsi_start_tmf(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_HANDLE        ptmf_handle);

ACIT_UEXACT32
iscsi_start_nopout(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_NOP_HANDLE        pnop_handle,
    PPDU_NOP_IN             pdu,
    PACIT_NOP_HANDLE        poptional_handle);

ACIT_UEXACT32
iscsi_alloc_io_resources(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_TARGET_WS         ptarget,
    PACIT_IO_HANDLE         pio_handle);

ACIT_UEXACT32
iscsi_free_io_resources(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_IO_HANDLE         pio_handle);

INLINE
ACIT_VOID
iscsi_update_io_counters(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle);

ACIT_UEXACT32
iscsi_complete_io(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle,
    PI_SOL_CQE              psol,
    ACIT_UEXACT32           invalidated);

ACIT_UEXACT32
iscsi_complete_tmf(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_HANDLE        ptmf_handle,
    ACIT_UEXACT8            iresponse,
    ACIT_UEXACT32           max_cmd_sn_offset,
    ACIT_UEXACT32           exp_cmd_sn);

PISCSI_CONNECTION
iscsi_alloc_cxn (
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_VOID
iscsi_free_cxn (
    PACIT_CONTROLLER_WS     pcontroller,
    PISCSI_CONNECTION       pcxn);


ACIT_UEXACT32
iscsi_rst_send_notify(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           cid);

ACIT_UEXACT32
iscsi_recover_session(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_VOID              mgmt_ioctl_request);

ACIT_UEXACT32
iscsi_rst_recv_notify(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           cid);

ACIT_UEXACT32
iscsi_invalidate_connection(
    PACIT_CONTROLLER_WS     pcontroller,
    PISCSI_TARGET_WS        piscsi_target,
    PISCSI_CONNECTION       pcxn,
    ACIT_UEXACT16           issue_reset);

ACIT_UEXACT32
iscsi_invalidate_all_connections(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           force_rst);

ACIT_UEXACT32
iscsi_upload_connection(
    PACIT_CONTROLLER_WS     pcontroller, 
    PISCSI_TARGET_WS        ptarget, 
    ACIT_UEXACT32           cri);

ACIT_UEXACT32
iscsi_connection_upload_notify(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           cid);

ACIT_UEXACT32
iscsi_connection_invalidate_notify(
    PACIT_CONTROLLER_WS     pcontroller, 
    ACIT_UEXACT16           cid);

STATIC INLINE
ACIT_VOID
iscsi_update_io_request(
    PACIT_IO_REQUEST        pio_request,
    ACIT_UEXACT8            response,
    ACIT_UEXACT8            status,
    ACIT_UEXACT8            flag,
    ACIT_UEXACT32           residual_len)
{
    switch(response)
    {
    case SCSI_SVC_RESP_GOOD:
        pio_request->io_status = ACIT_IO_STATUS_SUCCESS;
        pio_request->scsi_status = status;

        if(IS_PDU_STATUS_UNDERRUN(flag))
        {
            pio_request->underrun_length = residual_len;
            pio_request->overrun_length = 0;
        }
        else if (IS_PDU_STATUS_OVERRUN(flag))
        {
            pio_request->overrun_length = residual_len;
            pio_request->underrun_length = 0;
        }
        else
        {
            pio_request->underrun_length = 0;
            pio_request->overrun_length = 0;
        }

        break;

    case SCSI_SVC_RESP_TARGET_ERROR:
        pio_request->io_status = ACIT_IO_STATUS_TARGET_ERROR;
        break;

    default:
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Unknown target response for IO 0x%x\n", response));
        pio_request->io_status = ACIT_IO_STATUS_ERROR;
    }
}

ACIT_VOID
iscsi_write_logout_pdu (
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_LOGOUT_HANDLE     plogout_handle);

ACIT_VOID
iscsi_write_cmd_pdu(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle);

ACIT_VOID
iscsi_write_tmf_pdu (
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_HANDLE        ptmf_handle);

ACIT_VOID
iscsi_write_nop_pdu (
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_NOP_HANDLE        pnop_handle);

ACIT_VOID
iscsi_complete_logout(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           cri,
    ACIT_UEXACT8            iresponse,
    ACIT_UEXACT32           reason_code,
    ACIT_UEXACT8            timeout);

ACIT_STATUS
iscsi_process_async_pdu (
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT16           cid,
    PPDU_BASE               pdu,
    ACIT_UEXACT32           pdu_len,
    PVOID                   buffer,
    ACIT_UEXACT32           buf_len);

ACIT_UEXACT32
iscsi_check_async_pdu (
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           cri,
    PPDU_BASE               ppdu,
    PACIT_UEXACT32          pbytes_needed);

ACIT_STATUS
iscsi_process_nop_in (
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_NOP_HANDLE        pnop_handle, // optional, only exists if this is a reply
    PISCSI_CONNECTION       pcxn,
    PPDU_NOP_IN             pdu,
    ACIT_UEXACT32           pdu_len);

ACIT_STATUS
iscsi_process_async_event (
    PACIT_CONTROLLER_WS     pcontroller,
    PISCSI_CONNECTION       pcxn,
    PPDU_ASYNC_MSG          pasync_msg,
    ACIT_UEXACT32           async_msg_len,
    PVOID                   pbuffer,
    ACIT_UEXACT32           buf_len);

ACIT_STATUS
iscsi_process_reject (
    PACIT_CONTROLLER_WS     pcontroller,
    PISCSI_CONNECTION       pcxn,
    PPDUH_REJECT            preject_msg,
    ACIT_UEXACT32           reject_msg_len,
    PPDU_BASE               ppdu_hdr,
    ACIT_UEXACT32           pdu_hdr_len);

ACIT_VOID
iscsi_update_statistics(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           cid,
    ACIT_UEXACT32           code);

ACIT_VOID
iscsi_flag_connection(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           cri,
    ACIT_UEXACT32           cxn_reason,
    ACIT_UEXACT32           is_ready);

ACIT_UEXACT32
iscsi_process_linkdown_event(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           port_num);

ACIT_UEXACT32
iscsi_process_linkup_event(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           port_num);

STATIC INLINE
ACIT_UEXACT32
iscsi_is_session_open(
    PACIT_TARGET_WS         ptarget)
{
    PISCSI_TARGET_WS        piscsi_target = ptarget->piscsi_ws;

    OSM_ASSERT(piscsi_target->pcxn_list[0]);

    if(ACIT_IS_BIT_SET(piscsi_target->pcxn_list[0]->status, ISCSI_CONN_STATUS_INITIALIZING) ||
           piscsi_target->pcxn_list[0]->status == ISCSI_CONN_STATUS_VALID)
    {
           return 1;
    }

    return 0;
}

PACIT_TARGET_WS
iscsi_get_target_from_cid(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           cid);

ACIT_UEXACT32
iscsi_start_cmd_window_timer(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget);

ACIT_UEXACT32
iscsi_stop_cmd_window_timer(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget);

ACIT_UEXACT32
iscsi_cleanup_fw_owned_session(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget);

MGMT_STATUS
iscsi_process_ws_init_failure(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_TARGET_WS         ptarget, 
    MGMT_STATUS             mgmt_status);

STATIC
INLINE
ACIT_UEXACT32
iscsi_target_has_active_cmds(
    PISCSI_TARGET_WS        piscsi_target)
{
    ACIT_UEXACT32           i = 0;

    for(i=0; i<MAX_CXNS_PER_SESSION; i++)
    {
        if(piscsi_target->pcxn_list[i] &&
           piscsi_target->pcxn_list[i]->num_of_active_cmds)
        {
            return ACIT_TRUE;
        }
    }
    
    return ACIT_FALSE;
}

/*
#define ISCSI_IS_ABORTIVE_CLEANUP(pc, pacit_t, piscsi_t)                     \
               ((piscsi_t->num_connections_reset)                        ||  \
               (piscsi_t->num_connections_invalidated)                  ||  \
               (iscsi_target_has_active_cmds(piscsi_t))                 ||  \
               (piscsi_t->num_connections_reset_needed)                 ||  \
               (ACIT_IS_TARGET_SUBSTATE(pacit_t, TARGET_SUB_STATE_TMF_TIMEOUT)) || \
               (ACIT_IS_TARGET_SUBSTATE(pacit_t, TARGET_SUB_STATE_TMF_ERROR)) || \
               (ACIT_IS_TARGET_SUBSTATE(pacit_t, TARGET_SUB_STATE_IO_ERROR))  || \
               (ACIT_IS_TARGET_SUBSTATE(pacit_t, TARGET_SUB_STATE_ASYNC_MSG_ERROR)))
*/

STATIC INLINE
ACIT_UEXACT32
iscsi_is_abortive_cleanup(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PISCSI_TARGET_WS        piscsi_target)
{
    ACIT_UEXACT32           rc = 0;
    //
    // Add this for UNH certification. This requires the initiator to issue a
    // logout when the target has issued an async logout request. Earlier, there
    // were some cases where an async logout would cause an abortive cleanup. Most
    // common one would be when some commands were pending on that connection
    //
    if(ACIT_IS_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_ASYNC_MSG_LOGOUT))
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Return 0 since async logout received\n"));
        return 0;
    }

    rc = ((piscsi_target->num_connections_reset)                            ||
          (piscsi_target->num_connections_invalidated)                      ||
          (iscsi_target_has_active_cmds(piscsi_target))                     ||
          (piscsi_target->num_connections_reset_needed)                     ||
          (ACIT_IS_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_TMF_TIMEOUT))  ||
          (ACIT_IS_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_TMF_ERROR))    ||
          (ACIT_IS_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_IO_ERROR))     ||
          (ACIT_IS_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_ASYNC_MSG_ERROR)));
    
    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Return %d\n", rc));  
    return rc;
}

ACIT_VOID
iscsi_process_unsol_data_digest_err_notify(
    PACIT_CONTROLLER_WS     pcontroller, 
    ACIT_UEXACT32           cid);
