/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
										

/*++

    Filename:

	acit_iface.c

    Purpose:

	This file implements the different routines that are needed by the
	ACIT to function in an OS environment.


--*/

#include "acitcfg.h"
#include <linux/string.h>
#include <linux/kernel.h>
#include "acitdefs.h"
#include "acitscsi.h"
//#include "mgmtdefs.h"
//#include "ioctl.h"
//#include "hostioct.h"
#include "ioctl_top_bmap.h"
#include "ioctl_top_enum.h"
#include "acitapi.h"
#include <linux/version.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/list.h>
#include <asm/unistd.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
#include <asm/semaphore.h>
#else
#include <linux/semaphore.h>
#endif
#include <linux/genhd.h>
#include <linux/blkdev.h>
#include <linux/spinlock.h>
#if defined(LINUX_2_4)
#include <linux/tqueue.h>
#include "scsi.h"
#include "hosts.h"
#else
#include <linux/interrupt.h>
#include <scsi/scsi_host.h>
#include "scsi/scsi_cmnd.h"
#endif
#if !defined(__VMKERNEL_MODULE__) && !defined(__VMKLNX__) && !defined(LINUX_2_4)
#include <linux/nmi.h>
#endif
#include "be2iscsi_common.h"
#include "be2iscsi_debug.h"
#include <linux/pci.h>
#include "be2iscsi_osm.h"
#include "acit_iface.h"
#include "uerr.h"
#include <linux/delay.h>	/* for udelay */

#ifndef page_to_bus
#define page_to_bus(page) (page_to_phys(page))
#endif
#define VALID_PARAM(index, pconfig_param) \
	(pconfig_param->min_value <= se_param[index].val && \
	 pconfig_param->max_value >= se_param[index].val)

#define MAX_BEDISPSTR 25 

unsigned long gudel_jiffies = 0;

#define MAX_ERROR_RETURN	4
unsigned int	gerrmaxcnt = 0;
unsigned int	gerrignore = 0;
unsigned int	gerrcnt = 0;
unsigned int	gerrvar[MAX_ERROR_RETURN];
unsigned char	*gerrstring[MAX_ERROR_RETURN];

/*++
Function:
	be2iscsi_log_error

Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	error_code:   [in]   
				that describes the config parameter.

Return:
	ACIT_STATUS_SUCCESS - no errors

 Context:

Description:
	This routine is used by ACIT to log error.

--*/

ACIT_VOID 
be2iscsi_acit_error(
unsigned		uec_sub_module,
unsigned		uec_level,
unsigned		uec_phase,
unsigned		uec_additional_length,
unsigned		uec_code
)
{
	switch (uec_code) {

	case ACIT_EC_CONTROLLER_FAILURE:
		gerrstring[0] = "Hardware Initialization Failed. Either "
			"Hardware/Firmware is not initialized or is "
			"malfunctioning";
		break;

	case ACIT_EC_FIRMWARE_MISMATCH:
		gerrstring[0] = "Driver version does not match Firmware. "
			"Please run Installer";
		break;

	case ACIT_EC_LINK_DOWN:
	case ACIT_EC_LINK_UP:
		gerrignore = 1;
		/*
		 * These are handled in se_handle_contoller_event
		 */
		break;

	case ACIT_EC_LDTO_EXPIRY:
		gerrstring[0] = "LinkDown Timeout Expired. Please check the "
			"Physical Link to BladeEngine";
		break;

	case ACIT_EC_TARGET_ETO_EXPIRY:
		gerrstring[0] = "Extended Timeout Expired. Loss of connection "
			"to target exceeded ETO limits";
		gerrstring[1] = "Session ID";
		break;

	case ACIT_EC_UNHANDLED_TMF_REQUEST:
		gerrstring[0] = "Received unsupported Task Management Function";
		gerrstring[1] = "Task Management Function code";
		break;

	case ACIT_EC_TASK_INVALID_FOR_OSM_ABORT:
		gerrstring[0] = "Received a TMF Abort for an IO that is not "
			"present with the driver";
		break;

	case ACIT_EC_FW_CONFIG_FAILED:
		// RST - need error message
		break;

	case ACIT_EC_ISCSI_SESSION_RESET:
		/* NOT USED */
		break;

	case ACIT_EC_TCP_CONNECTION_RESET:
		/* NOT USED */
		break;

	case ACIT_EC_ISCSI_CMDSN_ERROR:
		gerrstring[0] = "Received invalid iSCSI Command Sequence "
			"Number update from Target";
		gerrstring[1] = "Session Handle";
		gerrstring[2] = "MaxCmdSN";
		gerrstring[3] = "ExpCmdSN";
		break;

	case ACIT_EC_PING_WAIT_TIMEOUT:
		gerrstring[0] = "Did not receive an iSCSI Command window "
			"update from Target for at least 25 Secs";
		gerrstring[1] = "Session Handle";
		break;

	case ACIT_EC_ARM_INIT_STATUS:
		/* NOT USED */
		break;

	case ACIT_EC_OFFLOAD_ERROR:
		gerrstring[0] = "Error during iSCSI offload";
 		gerrstring[1] = "Sess Handle(16bit) | (Fw Error code stat(8bit) | Add stat(8bit))";
		break;

	case ACIT_EC_IOCTL_HANDLE_STATUS:
		/* NOT USED */
		break;

	case ACIT_EC_MCC_SGE_EXCEEDED:
		gerrstring[0] = "Rejected IOCTL since buffer size limit "
			"exceeded";
		gerrstring[1] = "Subsystem / Opcode";
		gerrstring[2] = "Payload Length";
		break;

	case ACIT_EC_HWI_UE:
		gerrstring[0] = "Unrecoverable Error";
		gerrstring[1] = "UE_LOW";
		gerrstring[2] = "UE_HIGH";
		gerrstring[3] = "Firmware Line Number";
		break;

	default:
		break;
	}
}

ACIT_VOID 
be2iscsi_osm_error(
unsigned		uec_sub_module,
unsigned		uec_level,
unsigned		uec_phase,
unsigned		uec_additional_length,
unsigned		uec_code
)
{
	switch (uec_code) {

	case OSM_EC_HW_INIT_FAILED:

		switch (uec_level) {

		case UNI_CFG_LOG_LEVEL_INFO:
			gerrstring[0] = "OSM Hardware Initialization Failure";
			break;

 		case UNI_CFG_LOG_LEVEL_ERROR:
 			gerrstring[0] = "Target Initialization Failure";
 			break;

		case UNI_CFG_LOG_LEVEL_CRITICAL:
			if (uec_phase == UNI_PHASE_INIT) {
				gerrstring[0] = "Initialization Failure";
			} else if (uec_phase == UNI_PHASE_POWER_MGMT) {
				gerrstring[0] = "Initialization failure during "
					"Power Management Bootup";
			}
			break;
		}

		break;

	case OSM_EC_GET_SGLIST_FAILED:
		/* NOT USED */
		break;

	case OSM_EC_OSM_API_TABLE_INITIALIZE_FAILED:
		/* NOT USED */
		break;

	case OSM_EC_ACIT_API_FAILED:
		gerrstring[0] = "Internal API failed during Initialization";
		break;

	case OSM_EC_UNSUPPORTED_HARDWARE:
		gerrstring[0] = "PCI ID mismatch on Board";
		break;

	case OSM_EC_RESOURCE_ALLOCATION_FAILURE:

		switch (uec_level) {

		case UNI_CFG_LOG_LEVEL_INFO:
			gerrstring[0] = "OSM Resource Allocation Failure";
			break;

		case UNI_CFG_LOG_LEVEL_CRITICAL:
			if (uec_phase == UNI_PHASE_INIT) {
				gerrstring[0] = "Memory Allocation failure during init";
			} else {
				gerrstring[0] = "Memory Allocation failure ";
			}

			break;
		}
		break;

	case OSM_EC_CONFIG_PARAMETER_OUTSIDE_RANGE:
		gerrstring[0] = "Invalid Parameter specified during insmod, "
			"using default";
		break;

	case OSM_EC_CONFIG_PARAMETER_NOT_ADJUSTABLE:
		/* NOT USED */
		break;

	case OSM_EC_CONFIG_PARAMETER_INVALID:
		gerrstring[0] = "Ignored a invalid parameter";
		break;

	case OSM_EC_MAP_REGISTERS_FAILED:
		gerrstring[0] = "Failed to map one or more PCI Base Address "
			"Register";
		break;

	}
}

ACIT_VOID 
be2iscsi_log_error(
	PACIT_CONTROLLER_KEY pcontroller_key,
	ACIT_UEXACT32 error_code)
{
	se_hba		*phba;
	unsigned	uec_module;
	unsigned	uec_sub_module;
	unsigned	uec_level;
	unsigned	uec_phase;
	unsigned	uec_additional_length;
	unsigned	uec_code;

	if (pcontroller_key != NULL) {
		phba = pcontroller_key->osm_key;
	} else {
		phba = NULL;
	}

	gerrvar[gerrcnt] = error_code;

	uec_module = (gerrvar[0] & UNI_MODULE_BIT_MASK) >> 
		UNI_MODULE_BIT_SHIFT;

	uec_sub_module = (gerrvar[0] & UNI_SUBMODULE_BIT_MASK) >> 
		UNI_SUBMODULE_BIT_SHIFT;

	uec_level = (gerrvar[0] & UNI_ERROR_LEVEL_BIT_MASK) >> 
		UNI_ERROR_LEVEL_BIT_SHIFT;

	uec_phase = (gerrvar[0] & UNI_PHASE_BIT_MASK) >> 
		UNI_PHASE_BIT_SHIFT;

	uec_additional_length  = (gerrvar[0] & UNI_EXTENSION_BIT_MASK) >> 
		UNI_EXTENSION_BIT_SHIFT;

	uec_code  = (error_code & UNI_ERROR_CODE_BIT_MASK) >> 
		UNI_ERROR_CODE_BIT_SHIFT;

	if (gerrcnt == 0)
	{
		gerrmaxcnt = uec_additional_length;

		gerrstring[0] = NULL;
		gerrstring[1] = NULL;
		gerrstring[2] = NULL;
		gerrstring[3] = NULL;
		gerrignore = 0;

		switch (uec_module) {
		case UNI_MODULE_ACIT:
			be2iscsi_acit_error(uec_sub_module, uec_level,
				uec_phase, uec_additional_length, uec_code);
			break;

		case UNI_MODULE_OSM:
			be2iscsi_osm_error(uec_sub_module, uec_level, uec_phase,
				uec_additional_length, uec_code);
			break;

		case UNI_MODULE_FIRMWARE:
		case UNI_MODULE_MILI:
		case UNI_MODULE_CLI:
		case UNI_MODULE_GUI:
		default:
			gerrstring[0] = NULL;
			break;
		}
	}

	if (gerrcnt != gerrmaxcnt) {
		/*
		 * Gather up all of the parts of the error.
		 */
		gerrcnt++;
		return;
	} else {
		gerrcnt = 0;
	}

	if (gerrignore) {
		/*
		 * We have been told to not print out this error.
		 */
		return;
	}

	if ((phba != NULL) && (phba->host != NULL)) {
		printk("scsi%d: ", phba->host->host_no);
	}

	printk("BladeEngine 2 iSCSI Driver: ");

	if (gerrstring[0] != NULL) {
		printk("%s\n", gerrstring[0]);
	} else {
		printk("module 0x%x | sub-module 0x%x | level 0x%x | "
			"phase 0x%x | code 0x%x\n",
			uec_module, uec_sub_module, uec_level,
			uec_phase, uec_code);
	}

	if (gerrstring[1] != NULL) {
		if ((uec_module == UNI_MODULE_ACIT) && 
		    ((uec_code == ACIT_EC_OFFLOAD_ERROR) ||
		    (uec_code == ACIT_EC_MCC_SGE_EXCEEDED))) {

			printk("%s = 0x%x / 0x%x\n", gerrstring[1],
				((gerrvar[1] & 0xffff0000) >> 16),
				 (gerrvar[1] & 0x0000ffff));
		} else {
			printk("%s = 0x%x\n", gerrstring[1], gerrvar[1]);
		}
	}

	if (gerrstring[2] != NULL) {
		printk("%s = 0x%x\n", gerrstring[2], gerrvar[2]);
	}

	if (gerrstring[3] != NULL) {
		printk("%s = 0x%x\n", gerrstring[3], gerrvar[3]);
	}

	printk("\n");	
}

/*++
Function:
	be2iscsi_get_config_value

Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	pconfig_param:   [in]   pointer to ACIT_CONFIG_PARAM structure
				that describes the config parameter.

Return:
	ACIT_STATUS_SUCCESS - no errors
	ACIT_STATUS_ERROR - config parameter is not found

 Context:
	Init

Description:
	This routine is used by ACIT to read configuration parameter
	that is set in the host operating system environment.
	
--*/

ACIT_STATUS
be2iscsi_get_config_value(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	PACIT_CONFIG_PARAM      pconfig_param)
{
	int temp;
	BEGIN;
	PDEBUG(DBG_INIT, "getting value for %s\n",
			pconfig_param->osm_config_param);
	for (temp = 0; temp<PARAMETERS; temp++){
			/* check if its a config parameter */
		if (strcmp(se_param[temp].osm_config_param,
			pconfig_param->osm_config_param) == 0){
			/* check if its a valid numerical param */
			if (pconfig_param->acit_pflag == 
				ACIT_PFLAG_PARAMETER_NUMERIC){
			
				if (VALID_PARAM(temp,pconfig_param)){
					pconfig_param->numeric_value = 
						se_param[temp].val;
				}
				else{
					printk("WARNING: %s VALUE = %d OUT OF RANGE . LIMIT IS %d TO %d. TAKING THE DEFAULT VALUE AS %d \n",se_param[temp].command_line_param, se_param[temp].val,pconfig_param->min_value,pconfig_param->max_value,pconfig_param->numeric_value);
				/* check if its a config parameter */
					se_param[temp].val = 
					       pconfig_param->numeric_value; 
				}		
				return ACIT_STATUS_SUCCESS;
			} else if (pconfig_param->acit_pflag ==
				ACIT_PFLAG_PARAMETER_STRING){
				strncpy(pconfig_param->string_value,
					se_param[temp].str_val,
					ACIT_MAX_CONFIG_PARAMETER_LENGTH);
				return ACIT_STATUS_SUCCESS;
			}
		}
	}
	pconfig_param->osm_pflag = ACIT_OSM_PFLAG_DEFAULT_VALUE;

	/* return a sane value for unknown config numeric param */
	/* TODO: remove this later */
//
//	if (pconfig_param->acit_pflag == ACIT_PFLAG_PARAMETER_NUMERIC){
//		PDEBUG(DBG_AIFACE, "returning  config val\n");
//		pconfig_param->numeric_value = pconfig_param->max_value;
//		return ACIT_STATUS_SUCCESS;
//	}

	return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32 be2iscsi_complete_ioctl(
			PACIT_CONTROLLER_KEY controller_key,
			PACIT_IOCTL_REQUEST pioctl_request)
{
	se_hba	*phba;
	phba = pioctl_request->posm_reference;
	/* Wake up the sleeping process */
	complete(&phba->ioctl_completion);
	return ACIT_STATUS_SUCCESS;
}


/*++
Function:
	be2iscsi_complete_tmf

Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	ptmf_request     [in]   pointer to ACIT_IO_REQUEST that is completed

Return:
	ACIT_STATUS_SUCCESS - no errors
	otherwise, complete was not successful.

 Context:
	IO/Timer/Interrupt

Description:
	This routine is used by ACIT to complete an TMF request that was
	submitted by the OSM.

--*/

ACIT_STATUS
be2iscsi_complete_tmf(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	PACIT_TMF_REQUEST       ptmf_request)
{
	ACIT_STATUS     acit_status;
	se_hba	*phba;
	se_target     *ptarget;
	se_device     *pdevice;
	se_tmf	*posm_request;
	se_req	*preq;
	BEGIN;
	
	OSM_ASSERT(pcontroller_key);
	OSM_ASSERT(ptmf_request);

	PDEBUG(DBG_AIFACE, "pcontroller_key %p - ptmf_request %p\n",
		pcontroller_key, ptmf_request);
	
	phba = ptmf_request->pcontroller_key->osm_key;
	ptarget = ptmf_request->ptarget_key->osm_key;
	pdevice = ptmf_request->plun_key->osm_key;
	preq = ptmf_request->ptmf_key->osm_key;
	posm_request = &preq->u.tmf;
	
	acit_status = se_complete_tmf_request(
				phba,
				posm_request);
			

	/* put the free io request back on the free list */
	se_return_cmd(phba, (se_req *)preq);
	
	return ACIT_STATUS_SUCCESS;
}
/*++
Function:
	be2iscsi_complete_io

Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	pio_request:     [in]   pointer to ACIT_IO_REQUEST that is completed

Return:
	ACIT_STATUS_SUCCESS - no errors
	otherwise, complete was not successful.

 Context:
	IO/Timer/Interrupt

Description:
	This routine is used by ACIT to complete an IO request that was
	submitted by the OSM.

--*/

ACIT_STATUS
be2iscsi_complete_io(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	PACIT_IO_REQUEST	pio_request)
{
	ACIT_STATUS     acit_status;
	se_hba	*phba;
	se_target     *ptarget;
	se_device     *pdevice;
	se_req	*posm_request;

	BEGIN;
	
	OSM_ASSERT(pcontroller_key);
	OSM_ASSERT(pio_request);
	
	phba = pio_request->pcontroller_key->osm_key;
	ptarget = pio_request->ptarget_key->osm_key;
	pdevice = pio_request->plun_key->osm_key;
	posm_request = pio_request->pio_key->osm_key;
	
	acit_status = se_complete_io_request(
				phba,
				ptarget,
				pdevice,
				&posm_request->u.io,
				pio_request);
			
	/* put the free io request back on the free list */

	posm_request->u.io.old_ser_num= posm_request->u.io.curr_ser_num;
	se_return_cmd(phba, (se_req *)posm_request);
	
	return ACIT_STATUS_SUCCESS;
}

/*++
Function:
	be2iscsi_complete_abort_io

Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	pio_request:     [in]   pointer to ACIT_IO_REQUEST that is completed

Return:
	ACIT_STATUS_SUCCESS - no errors
	otherwise, complete was not successful.

 Context:
	IO/Timer/Interrupt

Description:
	This routine is used by ACIT to complete an abort request that was
	submitted by the OSM.

--*/

ACIT_UEXACT32
be2iscsi_complete_abort_io(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	PACIT_IO_REQUEST	pio_request)
{
	BEGIN;
	return ACIT_STATUS_SUCCESS;
}

/*++
Function:
	be2iscsi_complete_lun_reset

Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	pio_request:     [in]   pointer to ACIT_IO_REQUEST that is completed

Return:
	ACIT_STATUS_SUCCESS - no errors
	otherwise, complete was not successful.

 Context:
	IO/Timer/Interrupt

Description:
	This routine is used by ACIT to complete an LUN reset request submitted
	by OSM.

--*/

ACIT_UEXACT32
be2iscsi_complete_lun_reset(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	PACIT_IO_REQUEST	pio_request)
{
	BEGIN;
	return ACIT_STATUS_SUCCESS;
}

/*++
Function:
	be2iscsi_complete_target_reset

Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	pio_request:     [in]   pointer to ACIT_IO_REQUEST that is completed

Return:
	ACIT_STATUS_SUCCESS - no errors
	otherwise, complete was not successful.

 Context:
	IO/Timer/Interrupt

Description:
	This routine is used by ACIT to complete a target reset request 
	submitted by OSM.

--*/

ACIT_UEXACT32
be2iscsi_complete_target_reset(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	PACIT_IO_REQUEST	pio_request)
{
	BEGIN;
	return ACIT_STATUS_SUCCESS;
}

/*++
Function:
	be2iscsi_async_event

Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	event_class:     [in]   event class identifying the kind of event
				that occured.
	event_code:      [in]   code to identify the event within a particular
				kind of event.
	info:	    [in]   information pertaining to the event.

Return:
	ACIT_STATUS_SUCCESS - no errors
	otherwise, complete was not successful.

 Context:
	IO/Timer/Interrupt

Description:
	This routine is used by ACIT to notify OSMs about some events like
	controller busy, target add/remove etc.

--*/

ACIT_STATUS
be2iscsi_async_event(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	ACIT_UEXACT32	   event_class,
	ACIT_UEXACT32	   event_code,
	ACIT_UEXACT32	   info)
{
	se_hba		*phba;
	BEGIN;

	OSM_ASSERT(pcontroller_key);

	phba = (se_hba *)(pcontroller_key->osm_key);
	switch(event_class) {

		case ACIT_EVENT_CLASS_CONTROLLER:
			se_handle_controller_event(phba, event_code, info);
			break;

		case ACIT_EVENT_CLASS_TARGET:
			se_handle_target_event(phba, event_code, info);
			break;

		case ACIT_EVENT_CLASS_LUN:

			break;

	}
       	END; 
	return ACIT_STATUS_SUCCESS;
}

/*++
Function:
	be2iscsi_read8

Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	register_handle: [in]   handle to the register to read from
	offset:	  [in]   offset within the register base to read

Return:
	A byte value that was read from the specified register.

 Context:
	Any

Description:
	This routine is used by ACIT to read a byte value off a register.

--*/
ACIT_UEXACT8
be2iscsi_read8(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	ACIT_REGISTER_HANDLE    register_handle,
	ACIT_UEXACT32	   offset)
{
	se_pci_resource *presource = (se_pci_resource *)register_handle;
	BEGIN;
	if (presource->mmap) 
		return readb((void *)(presource->address+offset));
	else
		return inb(presource->address+offset);
}

/*++
Function:
	be2iscsi_read16

Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	register_handle: [in]   handle to the register to read from
	offset:	  [in]   offset within the register base to read

Return:
	A short value that was read from the specified register.

 Context:
	Any

Description:
	This routine is used by ACIT to read a word value off a register.

--*/

ACIT_UEXACT16
be2iscsi_read16(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	ACIT_REGISTER_HANDLE    register_handle,
	ACIT_UEXACT32	   offset)
{
	se_pci_resource *presource = (se_pci_resource *)register_handle;
	BEGIN;
	PDEBUG(DBG_AIFACE, "register handle = 0x%lx, offset = %d\n",
			(unsigned long)register_handle, offset)
	if (presource->mmap) 
		return readw((void *)(presource->address+offset));
	else
		return inw(presource->address+offset);
}

/*++
Function:
	be2iscsi_read32

Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	register_handle: [in]   handle to the register to read from
	offset:	  [in]   offset within the register base to read

Return:
	A dword value that was read from the specified register.

 Context:
	Any

Description:
	This routine is used by ACIT to read a dword value off a register.

--*/
ACIT_UEXACT32
be2iscsi_read32(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	ACIT_REGISTER_HANDLE    register_handle,
	ACIT_UEXACT32	   offset)
{
	se_pci_resource *presource = (se_pci_resource *)register_handle;
	BEGIN;
	PDEBUG(DBG_AIFACE, "register handle = 0x%lx, offset = 0x%x\n",
			(unsigned long)register_handle, offset)
	if (presource->mmap) 
		return readl((void *)(presource->address+offset));
	else
		return inl(presource->address+offset);
}

/*++
Function:
	be2iscsi_read_buffer

Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	register_handle: [in]   handle to the register to read from
	offset:	  [in]   offset within the register base to read
	pbuffer:	 [in]   pointer to the buffer to read from
	length:	  [in]   length of buffer to read.	

Return:
	ACIT_STATUS_SUCCESS - no errors

 Context:
	Any

Description:
	This routine is used by ACIT to read a given number of bytes off the
	specified register.

--*/

ACIT_STATUS
be2iscsi_read_buffer(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	ACIT_REGISTER_HANDLE    register_handle,
	ACIT_UEXACT32	   offset,
	PACIT_VOID	      pbuffer,
	ACIT_UEXACT32	   length)
{
	se_pci_resource *presource = (se_pci_resource *)register_handle;
	BEGIN;
	if (presource->mmap) 
       		memcpy_fromio(pbuffer,((char *)presource->address+offset),length);
	else
		insb(presource->address+offset, pbuffer, length);

	return ACIT_STATUS_SUCCESS;
}

/*++
Function:
	be2iscsi_write8

Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	register_handle: [in]   handle to the register to write to
	offset:	  [in]   offset within the register base to write
	value:	   [in]   char to write to the register

Return:
	None

 Context:
	Any

Description:
	This routine is used by ACIT to write a byte value to a register

--*/

ACIT_VOID
be2iscsi_write8(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	ACIT_REGISTER_HANDLE    register_handle,
	ACIT_UEXACT32	   offset,
	ACIT_UEXACT8	    value)
{
	se_pci_resource *presource = (se_pci_resource *)register_handle;
	BEGIN;
	if (presource->mmap) 
		writeb(value, (void *)(presource->address+offset));
	else
		outb(value, presource->address+offset);
}

/*++
Function:
	be2iscsi_write16

Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	register_handle: [in]   handle to the register to write to
	offset:	  [in]   offset within the register base to write
	value:	   [in]   word to write to the register

Return:
	None

 Context:
	Any

Description:
	This routine is used by ACIT to write a word value to a register

--*/
ACIT_VOID
be2iscsi_write16(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	ACIT_REGISTER_HANDLE    register_handle,
	ACIT_UEXACT32	   offset,
	ACIT_UEXACT16	   value)
{
	se_pci_resource *presource = (se_pci_resource *)register_handle;
	BEGIN;
	PDEBUG(DBG_AIFACE, "register handle = 0x%lx, offset = %d\n",
			(unsigned long)register_handle, offset)
	if (presource->mmap) 
		writew(value, (void *)(presource->address+offset));
	else
		outw(value, presource->address+offset);
}

/*++
Function:
	be2iscsi_write32

Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	register_handle: [in]   handle to the register to write to
	offset:	  [in]   offset within the register base to write
	value:	   [in]   dword to write to the register

Return:
	None

 Context:
	Any

Description:
	This routine is used by ACIT to write a dword value to a register

--*/

ACIT_VOID
be2iscsi_write32(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	ACIT_REGISTER_HANDLE    register_handle,
	ACIT_UEXACT32	   offset,
	ACIT_UEXACT32	   value)
{
	se_pci_resource *presource = (se_pci_resource *)register_handle;
	BEGIN;
	PDEBUG(DBG_AIFACE, "register handle = 0x%lx, offset = %x, val = %x\n",
			(unsigned long)register_handle, offset, value)
	if (presource->mmap) 
		writel(value, (void *)(presource->address+offset));
	else
		outl(value, presource->address+offset);
}

/*++
Function:
	be2iscsi_write_buffer

Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	register_handle: [in]   handle to the register to write to
	offset:	  [in]   offset within the register base to write
	pbuffer:	 [in]   pointer to the buffer to write
	length:	  [in]   length of buffer to write    

Return:
	ACIT_STATUS_SUCCESS - no errors

 Context:
	Any

Description:
	This routine is used by ACIT to write a given number of bytes to the
	specified register.

--*/

ACIT_STATUS
be2iscsi_write_buffer(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	ACIT_REGISTER_HANDLE    register_handle,
	ACIT_UEXACT32	   offset,
	PACIT_VOID	      pbuffer,
	ACIT_UEXACT32	   length)
{
	se_pci_resource *presource = (se_pci_resource *)register_handle;
	BEGIN;
	PDEBUG(DBG_AIFACE, "register_handle = %lx, offset = %d, length = %d\n",
		(unsigned long)register_handle, offset, length);
	if (presource->mmap) 
		memcpy_toio(((char *)presource->address+offset),pbuffer,length);
	else
		outsb(presource->address+offset, pbuffer, length);

	return ACIT_STATUS_SUCCESS;
}

/*++
Function:
	be2iscsi_read_config8
	
Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	offset:	  [in]   offset within the config space
	value:	   [in]   pointer to data to read to

Return:
	ACIT_STATUS_SUCCESS - no errors
	ACIT_STATUS_ERROR otherwise

 Context:
	Any

Description:
	This routine is used by the ACIT to read a byte from the PCI
	configuration space.

--*/

ACIT_STATUS
be2iscsi_read_config8(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	ACIT_UEXACT32	   offset,
	PACIT_UEXACT8	   pbuffer)
{
	int status;
	struct pci_dev *pcidev = ((se_hba *)(pcontroller_key->osm_key))->pcidev;
	BEGIN;
	status = pci_read_config_byte(pcidev, offset, pbuffer);
	if (!status) 
		return ACIT_STATUS_SUCCESS;
	return ACIT_STATUS_ERROR;

}

/*++
Function:
	be2iscsi_read_config16
	
Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	offset:	  [in]   offset within the config space
	value:	   [in]   pointer to data to read to

Return:
	ACIT_STATUS_SUCCESS - no errors
	ACIT_STATUS_ERROR otherwise

 Context:
	Any

Description:
	This routine is used by the ACIT to read a word from the PCI
	configuration space.

--*/

ACIT_STATUS
be2iscsi_read_config16(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	ACIT_UEXACT32	   offset,
	PACIT_UEXACT16	  pbuffer)
{
	int status;
	struct pci_dev *pcidev = ((se_hba *)(pcontroller_key->osm_key))->pcidev;
	BEGIN;
	status = pci_read_config_word(pcidev, offset, pbuffer);
	if (!status) 
		return ACIT_STATUS_SUCCESS;
	return ACIT_STATUS_ERROR;
}

/*++
Function:
	be2iscsi_read_config32
	
Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	offset:	  [in]   offset within the config space
	value:	   [in]   pointer to data to read to

Return:
	ACIT_STATUS_SUCCESS - no errors
	ACIT_STATUS_ERROR otherwise

 Context:
	Any

Description:
	This routine is used by the ACIT to read a dword from the PCI
	configuration space.

--*/

ACIT_STATUS
be2iscsi_read_config32(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	ACIT_UEXACT32	   offset,
	PACIT_UEXACT32	  pbuffer)
{
	int status;
	struct pci_dev *pcidev = ((se_hba *)(pcontroller_key->osm_key))->pcidev;
	BEGIN;
	status = pci_read_config_dword(pcidev, offset, pbuffer);
	if (!status) 
		return ACIT_STATUS_SUCCESS;
	return ACIT_STATUS_ERROR;


}

/*++
Function:
	be2iscsi_read_config_buffer
	
Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	??

Return:
	ACIT_STATUS_SUCCESS - no errors
	ACIT_STATUS_ERROR otherwise

 Context:
	Any

Description:
	This routine is used by the ACIT to read a buffer from the PCI
	configuration space.

--*/

ACIT_STATUS
be2iscsi_read_config_buffer(PACIT_CONTROLLER_KEY pcontroller_key)
{
	BEGIN;
	return ACIT_STATUS_SUCCESS;
}

/*++
Function:
	be2iscsi_write_config8
	
Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	offset:	  [in]   offset in PCI config space
	value:	   [in]   value to write to the config space

Return:
	ACIT_STATUS_SUCCESS - no errors
	ACIT_STATUS_ERROR otherwise

 Context:
	Any

Description:
	This routine is used by the ACIT to write a byte to the PCI
	configuration space.

--*/

ACIT_STATUS
be2iscsi_write_config8(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	ACIT_UEXACT32	   offset,
	ACIT_UEXACT8	    value)
{
	int status;
	struct pci_dev *pcidev = ((se_hba *)(pcontroller_key->osm_key))->pcidev;
	BEGIN;
	status = pci_write_config_byte(pcidev, offset, value);
	if (!status) 
		return ACIT_STATUS_SUCCESS;
	return ACIT_STATUS_ERROR;
}

/*++
Function:
	be2iscsi_write_config16
	
Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	offset:	  [in]   offset in PCI config space
	value:	   [in]   value to write to the config space

Return:
	ACIT_STATUS_SUCCESS - no errors
	ACIT_STATUS_ERROR otherwise

 Context:
	Any

Description:
	This routine is used by the ACIT to write a word to the PCI
	configuration space.

--*/

ACIT_STATUS
be2iscsi_write_config16(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	ACIT_UEXACT32	   offset,
	ACIT_UEXACT16	   value)
{
	int status;
	struct pci_dev *pcidev = ((se_hba *)(pcontroller_key->osm_key))->pcidev;
	BEGIN;
	status = pci_write_config_word(pcidev, offset, value);
	if (!status) 
		return ACIT_STATUS_SUCCESS;
	return ACIT_STATUS_ERROR;
}

/*++
Function:
	be2iscsi_write_config32
	
Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	offset:	  [in]   offset in PCI config space
	value:	   [in]   value to write to the config space

Return:
	ACIT_STATUS_SUCCESS - no errors
	ACIT_STATUS_ERROR otherwise

 Context:
	Any

Description:
	This routine is used by the ACIT to write a dword to the PCI
	configuration space.

--*/

ACIT_STATUS
be2iscsi_write_config32(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	ACIT_UEXACT32	   offset,
	ACIT_UEXACT32	   value)
{
	int status;
	struct pci_dev *pcidev = ((se_hba *)(pcontroller_key->osm_key))->pcidev;
	BEGIN;
	status = pci_write_config_dword(pcidev, offset, value);
	if (!status) 
		return ACIT_STATUS_SUCCESS;
	return ACIT_STATUS_ERROR;
}


/*++
Function:
	be2iscsi_write_config_buffer
	
Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	??

Return:
	ACIT_STATUS_SUCCESS - no errors
	ACIT_STATUS_ERROR otherwise

 Context:
	Any

Description:
	This routine writes a buffer to the PCI config space

--*/

ACIT_UEXACT32
be2iscsi_write_config_buffer(PACIT_CONTROLLER_KEY pcontroller_key)
{
	BEGIN;
	return ACIT_STATUS_SUCCESS;
}

/*++
Function:
	be2iscsi_get_bus_address
	
Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	pvirtual_address:[in]   virtual address of the buffer

Return:
	bus address for the virtual address

 Context:
	Any

Description:
	This routine is used by the ACIT to get the bus address for a
	virtual address.

--*/

ACIT_BUS_ADDRESS
be2iscsi_get_bus_address(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	PACIT_VOID		posm_reference,
	PACIT_VOID	      pvirtual_address,
	PACIT_UEXACT32		plength)
{


	ACIT_BUS_ADDRESS	bus_address;
//	void 			*data_buf;
//	unsigned int 		data_len, page_aligned_add;
//	se_hba 		*phba = (se_hba *)(pcontroller_key->osm_key);
	
	BEGIN;
	memset(&bus_address, 0, sizeof(bus_address));
#if defined(__VMKERNEL_MODULE__)
	bus_address.u.a64.address = (pvirtual_address == NULL) ? 0 : 
	virt_to_phys(pvirtual_address);
#else
	bus_address.u.a64.address = virt_to_bus(pvirtual_address);
#endif
	return bus_address;
}

/*++
Function:
	be2iscsi_delay_timer
	
Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	timer:	   [in]   number of microsecs to delay

Return:
	ACIT_STATUS_SUCCESS

 Context:
	Any

Description:
	This routine is used by the ACIT to delay the execution by a
	specified number of microsecs.

--*/
ACIT_STATUS
be2iscsi_delay_timer(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	ACIT_UEXACT32	   time)
{
	se_hba *phba;
	/* 
	 * prabal - need more thought into this. Ideally this should
	 * be routine where OSMs can decide if its OK to sleep or
	 * should it put in a tight delay loop. For now this should do.
	 */
	
	BEGIN;
	phba = (se_hba *)(pcontroller_key->osm_key);
	PDEBUG(DBG_AIFACE,"Inside be2iscsi_delay_timer\n");
	if ((0 == gudel_jiffies) || (gudel_jiffies > jiffies))
		gudel_jiffies = jiffies;
	
	if (phba->state & SE_STATE_SHUTDOWN){
		if ((jiffies-gudel_jiffies)/HZ >= 1){
			gudel_jiffies = jiffies; 
			schedule();
		}
	}

#if !defined(__VMKERNEL_MODULE__) && !defined(__VMKLNX__) && !defined(LINUX_2_4)
	/*
	 * we have problem with init discover_target
	 * checking for link status took 10secs
	 * due to the long wait with discovery
	 * will remove in be3...
	 */
	if (!(phba->state & SE_STATE_HBA_INITIALIZED) ||
			(phba->state & SE_STATE_SHUTDOWN)) {
		touch_softlockup_watchdog();
		touch_nmi_watchdog();
	}
#endif

	udelay(time);
	PDEBUG(DBG_AIFACE,"exiting be2iscsi_delay_timer\n");
	return ACIT_STATUS_SUCCESS;
}

/*++
Function:
	be2iscsi_map_io_handle
	
Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	range_index:     [in]   index of PCI base address register
	length:	  [in]   number of bytes to be mapped
	attributes:      [in]   properties specified for the mapping.
				Big-endian/little-endian
	pbus_address:    [out]  pointer where OSM returns the physical address 
				of the I/O handle given by OS.
	preg_handle:     [out]  pointer to handle mapped by the OSM

Return:
	ACIT_STATUS_SUCCESS if the mapping was successful
	ACIT_STATUS_ERROR otherwise

 Context:
	Init

Description:
	This routine is used by the ACIT to map an address specified
	at a particular BAR to the host memory.

--*/

ACIT_STATUS
be2iscsi_map_io_handle(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	ACIT_UEXACT32	   range_index,
	ACIT_UEXACT32	   length,
	ACIT_UEXACT32	   attributes,
	ACIT_BUS_ADDRESS *      pbus_address,
	PACIT_VOID *	    pvirtual_address,
	PACIT_REGISTER_HANDLE   pregister_handle)
{
	se_hba * phba = (se_hba *)(pcontroller_key->osm_key);
	se_pci_resource *presource = &phba->resourcemap[range_index];
	BEGIN;
	*pregister_handle = presource;
	pbus_address->u.a64.address = presource->paddr;
	*pvirtual_address = (void *)presource->address;

	 PDEBUG(DBG_INIT, " In be2iscsi_map_io_handle  range_index = %u len = %u\n",range_index, length);

	if (presource->address)
		return ACIT_STATUS_SUCCESS;

	return ACIT_STATUS_ERROR;
}

/*++
Function:
	be2iscsi_release_io_handle
	
Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	register_handle: [in]   handle to memory to unmap

Return:
	ACIT_STATUS_SUCCESS - no errors

 Context:
	Any

Description:
	This routine is used by the ACIT to unmap a memory range mapped 
	earlier via be2iscsi_osm_map_io_handle call.

--*/

ACIT_STATUS
be2iscsi_release_io_handle(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	ACIT_REGISTER_HANDLE    register_handle)
{
	BEGIN;
	return ACIT_STATUS_SUCCESS;
}

/*++
Function:
	be2iscsi_synchronize_range
	
Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	??

Return:
	ACIT_STATUS_SUCCESS - success

 Context:
	??

Description:
	??

--*/
ACIT_STATUS
be2iscsi_synchronize_range(PACIT_CONTROLLER_KEY    pcontroller_key)
{
	BEGIN;
	return ACIT_STATUS_SUCCESS;
}

/*++
Function:
	be2iscsi_poll_register
	
Parameter:
	pcontroller_key: [in]   pointer to controller key identifying
				the controller
	??

Return:
	ACIT_STATUS_SUCCESS - no errors

 Context:
	??

Description:
	??

--*/
ACIT_STATUS
be2iscsi_poll_register(PACIT_CONTROLLER_KEY    pcontroller_key)
{
	BEGIN;
	return ACIT_STATUS_SUCCESS;
}

