/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

        be2iscsi_common.h

    Purpose:

        This file contains declarations for structs that is used in different
        .c files.


--*/
#ifndef _BE2ISCSICOMMON_H
#define _BE2ISCSICOMMON_H
#if 0
#if !LINUX_2_4 
#error "This driver only support kernel 2.4.18 and greater"
#endif
#endif

#define BE2ISCSI_MAX_HBA		32
#ifdef __VMKLNX__
#define MAX_TARGETS_SUPPORTED   256	/*In ESX we need more target numbers to handle the case
					   where target destroy is pending and new targets got added(Bug #26660) */
#else
#define MAX_TARGETS_SUPPORTED   128
#endif

#define MAX_CHANNEL     	8	/* Maximum Channel # Supported */
#define MAX_ID        		128	/* Maximum Target ID Supported */
#ifndef BUS_FIRST_MAPPING
#define MAX_ID_PER_CHANNEL	256	/* 10 Targets per Channel */
#endif
#define	MAX_LUN			255	/* Maximum LUNs Supported */
#define MAX_SENSE_LEN           256

/* 
 * HBA state flags 
  */
#define SE_STATE_INT_ENABLED	        (1U << 0)
#define SE_STATE_IOCTL                (1U << 1)
#define SE_STATE_BUS_RESET            (1U << 2)
#define SE_STATE_SHUTDOWN             (1U << 3)
#define SE_STATE_LINK_DOWN            (1U << 4)
#define SE_STATE_LINK_UP              (1U << 5)
#define SE_STATE_RESCAN_IN_PROGRESS   (1U << 6)
#define SE_STATE_LCT_GET              (1U << 7)
#define SE_STATE_LOGIN_PENDING        (1U << 8)
#define SE_STATE_HBA_INITIALIZED      (1U << 9)
#define SE_STATE_DELETE_HBA	      (1U << 10)
#define SE_STATE_REMOVE_LUNS	      (1U << 11)
#define SE_STATE_SUSPEND	      (1U << 12)
/*
 * SCSI Status as per SAM2
 */
#define SCSI_STATUS_GOOD                        0x00
#define SCSI_STATUS_ACA_ACTIVE                  0x30        

/*
 * constant PAGE_SIZE
 */
#ifndef PAGE_SIZE_4K
#define PAGE_SHIFT_4K           12
#define PAGE_SIZE_4K            (1 << PAGE_SHIFT_4K)
#endif

/* Structure to store commandline parameters */
#define PARAMETERS 42 

#define MAX_PCI_RESOURCES 6
#define BE_NUM_MSIX_ENTRIES (1)
struct se_param_struct {
	char osm_config_param[ACIT_MAX_CONFIG_NAME_LENGTH];
	char command_line_param[ACIT_MAX_CONFIG_NAME_LENGTH];
	/* following might be a union */
	unsigned int val;
	char str_val[ACIT_MAX_CONFIG_PARAMETER_LENGTH];
};
typedef struct kthread_struct
{

        /* queue thread is waiting on. Gets initialized by
           init_kthread, can be used by thread itself.
        */
        wait_queue_head_t queue;
        struct completion work_queued;
        /* flag to tell thread whether to die or not.
           When the thread receives a signal, it must check
           the value of terminate and call exit_kthread and terminate
           if set.
        */
        int terminate;
	/* completion to wait when shutting down the thread */
	struct completion thread_completion;
} kthread_t;

/* prototypes */

/* start new kthread (called by creator) */
void start_kthread(void (*func)(kthread_t *), kthread_t *kthread);

/* stop a running thread (called by "killer") */
void stop_kthread(kthread_t *kthread);

/* setup thread environment (called by new thread) */
void init_kthread(kthread_t *kthread, char *name);

/* cleanup thread environment (called by thread upon receiving termination signal) */
void exit_kthread(kthread_t *kthread);

#define MAX_STATIC_LUN_LIMIT   1 

#define ADDRESSING_MODE(byte0)  (byte0>>6 & 0x3)
#define PA_BUS_ID(byte0)        (byte0 & 0x3f)
#define LU_TARGET(byte0)        (byte0 & 0x3f)
#define LU_BUS(byte1)           (byte1>>5 & 0x7)
#define LU_LUN(byte1)           (byte1 & 0x1f)
#define FA_LUN(word)            (word>>14 & 0x3fff)
#define EA_LEN(byte)            (byte>>4 & 0x2)
#define EA_METHOD(byte)         (byte & 0xf)
                
typedef struct _se_lun_peripheral_address {
        ACIT_UEXACT8    byte0;
        ACIT_UEXACT8    target_lun;
        ACIT_UEXACT16   level2;
        ACIT_UEXACT16   level3;
        ACIT_UEXACT16   level4;
}se_lun_peripheral_address, *pse_lun_peripheral_address;

typedef struct _se_lun_logical_unit_address {
        ACIT_UEXACT8    byte0;
        ACIT_UEXACT8    byte1;
        ACIT_UEXACT16   level2;
        ACIT_UEXACT16   level3;
        ACIT_UEXACT16   level4;
}se_lun_logical_unit_address, *pse_lun_logical_unit_address;

typedef struct _se_lun_flatspace_address {
        ACIT_UEXACT16   word0;
        ACIT_UEXACT16   level2;
        ACIT_UEXACT16   level3;
        ACIT_UEXACT16   level4;
}se_lun_flatspace_address, *pse_lun_flatspace_address;

typedef struct _se_lun_extended_address {
        ACIT_UEXACT8    byte0;
        ACIT_UEXACT8    address1;
        ACIT_UEXACT8    byte2;
        ACIT_UEXACT8    address2;
        ACIT_UEXACT8    byte4;
        ACIT_UEXACT8    address3;
        ACIT_UEXACT8    byte6;
        ACIT_UEXACT8    address4;
}se_lun_extended_address, *pse_lun_extended_address;

typedef struct _se_lun_byte_address {
        ACIT_UEXACT8    byte0;
        ACIT_UEXACT8    byte1;
        ACIT_UEXACT8    byte2;
        ACIT_UEXACT8    byte3;
        ACIT_UEXACT8    byte4;
        ACIT_UEXACT8    byte5;
        ACIT_UEXACT8    byte6;
        ACIT_UEXACT8    byte7;
}se_lun_byte_address, *pse_lun_byte_address;

typedef struct _se_lun_address {
        union {
                se_lun_peripheral_address     lun_pa;
                se_lun_logical_unit_address   lun_la;
                se_lun_flatspace_address      lun_fa;
                se_lun_extended_address       lun_ea;
                se_lun_byte_address           lun_bytes;
                ACIT_UEXACT64                   lun_complete;
        }u;
}se_lun_address, *pse_lun_address;



typedef struct _se_report_lun_header {
        ACIT_UEXACT32   lunlist_len;
        ACIT_UEXACT32   reserved;
} se_report_lun_header, *pse_report_lun_header;

typedef struct _se_report_lun {
        se_report_lun_header  header;
        se_lun_address  lun[MAX_STATIC_LUN_LIMIT];
}se_report_lun, *pse_report_lun;

typedef struct _se_session_data {
	UEXACT32 	login_state;
	EXACT8 		target_name[224];
}SE_SESSION_DATA, *PSE_SESSION_DATA;

typedef struct _se_disp_lg_tiqn {
	SE_SESSION_DATA   session_list[1];
}DISP_LG_TIQN, *PDISP_LG_TIQN;


typedef struct _acit_io_request_h {
        PACIT_IO_KEY            pio_key;
        PACIT_LUN_KEY           plun_key;
        PACIT_TARGET_KEY        ptarget_key;
        PACIT_CONTROLLER_KEY    pcontroller_key;
}acit_io_request_h, *pacit_io_request_h;

typedef struct _se_command_request {
	struct _se_command_request 	*prev;
	struct _se_command_request 	*next;
	ACIT_UEXACT32 			*msg;
       	ACIT_SEXACT32 			len;
       	ACIT_SEXACT32 			timeout;
	atomic_t			fw_started_processing;
	atomic_t			fw_completed_processing;
	atomic_t			fw_timed_out;
	wait_queue_head_t		thread_processing_queue;
	struct timer_list		cmd_timeout_timer;
}se_command_request,*pse_command_request;

struct se_queue_info {
       	struct	scsi_cmnd		*cmd;
	void			(*scsi_done) (struct scsi_cmnd *);
	struct timer_list	timerlist;
};

/*
 * The following structure is the same as ACIT representation
 * of SG, except that we define it with max possible elements
 */
#if defined(__VMKERNEL_MODULE__)
#define SG_ALL 0xff
#endif

typedef struct _se_sg {

        ACIT_UEXACT32   number_elements;
        ACIT_UEXACT32   reserved;
        ACIT_SG_ELEMENT sg[SG_ALL];
}se_sg, *pse_sg;
        
#define READ    0
#define WRITE   1
#define NONE    2

typedef struct _se_tmf {
        ACIT_TMF_KEY            acit_tmf_key;
       struct  scsi_cmnd               *command;
        ACIT_TMF_REQUEST        tmf_request;
        struct completion       complete_tmf;
        ACIT_UEXACT8            tmf_ws[0] __attribute__((aligned(8)));
} se_tmf;

typedef struct _se_io {
	ACIT_IO_KEY		acit_io_key;
	struct scsi_cmnd		*command;
	ACIT_IO_REQUEST		io_request;
        /* SG place holder */
        se_sg                 sg_list;
#ifdef __VMKLNX__
	/*sense data for ESX*/
	ACIT_UEXACT8            sense_data[SCSI_SENSE_BUFFERSIZE];
#endif
	/* internal io requests might need this. For example reportlun */
	struct completion	complete_io;
    	ACIT_S32_64	        curr_ser_num;
    	ACIT_S32_64	        old_ser_num;
	ACIT_UEXACT8    	io_ws[0] __attribute__((aligned(8)));
} se_io;
typedef struct _se_req{
        struct list_head        queue_link;
        union {
                se_io io;
                se_tmf tmf;
        }u;
} se_req;

typedef struct _se_device {
	/* LUN key */
	ACIT_LUN_KEY		lun_key;

        struct list_head        sibling_list;

	/* Back pointer to parent target */
	struct _se_target	*pparent_target;

        /* 8-byte LUN */
        se_lun_address        lun; 

	ACIT_UEXACT32		capacity;
	ACIT_UEXACT8		scsi_channel;
	ACIT_UEXACT8		scsi_id;
	ACIT_UEXACT8 		scsi_lun;
	ACIT_UEXACT8		state;
	struct scsi_device	*pScsi_dev;

        /* Sense data buffer */
        ACIT_UEXACT8            sense_data[MAX_SENSE_LEN];
        ACIT_UEXACT32           actual_sense_len;

	/* Pointer to ACIT workspace for LUN */
	PACIT_VOID		pacit_lun_ws;

#ifdef BE2ISCSI_DEBUG
	unsigned		num_ios_outstanding;
	unsigned		num_errors;
#endif
#if defined(__VMKERNEL_MODULE__)
	ACIT_UEXACT32		use_count;
#endif
} se_device;

typedef struct _se_target {
	/* Target Key */
	PACIT_TARGET_KEY		posm_target_key;
        
        struct list_head        lun_list;
        struct list_head        sibling_list;
#if defined(__VMKERNEL_MODULE__)
	se_device		*pdevice_lun0;
#endif
        unsigned int            bus, target;
	/* Back pointer to the parent HBA */
	struct _se_hba	*pparent_hba;

        /* Pointer to Report Lun data */
        struct se_internal_io  *pinternal_io;

        /* Length of current report lun data */
        ACIT_UEXACT32           internal_io_len;

        /* Bus address for the above report lun buffer */
        dma_addr_t              internal_io_bus_addr;

	ACIT_UEXACT32		state;

	/* Pointer to ACIT workspace for target */
        PACIT_VOID              pacit_target_ws;

        /* This is the total number of luns allocated */
        ACIT_U32                num_luns;
	
	/* This will hold the pointer to persistancy data 
		received from firmware */
	PMGMT_BTL_MAPPING	pbtl_mapping;	
	int 			extended_timeout;

#ifdef REGISTER_ISCSI
#ifdef __VMKLNX__
	struct iscsi_cls_session	*sess;
	struct iscsi_cls_conn		*conn;
	char buf_name[255];
	char buf_isid[255];
	char buf_port[255];
	char buf_ip[255];
	unsigned	portal_group_tag;
#endif
#endif
} se_target;

typedef struct _se_pci_resource {
        ACIT_U32_64   address;
        ACIT_U32_64   paddr;
        ACIT_UEXACT8  mmap;
        ACIT_UEXACT8  pci_bar_index;
}se_pci_resource, *pse_pci_resource;

struct be_msix_entry {
	int msix_enabled;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,32)
	uint32_t msix_vector;
#else
	uint16_t msix_vector;
#endif
	uint16_t msix_entry;
};


typedef struct _se_hba {
        /* Controller key identifier for this HBA */
        ACIT_CONTROLLER_KEY     controller_key;

        /* ACIT DEVICE */
        ACIT_DEVICE             acit_device;

        /* Pointer to ACIT Controller WorkSpace */
        PACIT_VOID              pcontroller_ws;

        /* Linked list of se_hbas */
        struct list_head        hba_queue;
        
        /* PCI representation of our HBA */
	struct 	pci_dev 	*pcidev;
        
        /* Spinlock to lock ISRs, IOs, Timer */
	spinlock_t 		controller_lock;

        /* Spinlock to call into ACIT */
        spinlock_t              acitapi_lock;

        struct list_head        target_list;
        /* Max iSCSI targets supported - operating level dependent */
        ACIT_U32                max_iscsi_targets;

#define OSM_TARGET_CACHE_NAME		"se_be2_osmtarget"
#define ACIT_TARGET_WS_CACHE_NAME	"se_be2_acittarget"
#define OSM_LUN_CACHE_NAME		"se_be2_osmlun"
#define ACIT_LUN_WS_CACHE_NAME		"se_be2_acitlun"

	char				osmtarget_name[20];	
	char				acittarget_name[20];	
	char				osmlun_name[20];	
	char				acitlun_name[20];	

	/* Cache memory to hold OSM target pointers */
	struct kmem_cache		*posm_target_cache;

	/* Cache memory to hold ACIT target ws */
	struct kmem_cache		*pacit_target_ws_cache;

	/* Capturing negotiated operating memory level */
	ACIT_UEXACT32 		operating_mem_level;
	
	/* Max LUNs per target - operating level dependent */
	ACIT_U32		max_luns_per_target;

	/* Cache memory to hold OSM LUN pointers */
	struct kmem_cache		*posm_lun_cache;

	/* Cache memory to hold ACIT LUN ws */
	struct kmem_cache		*pacit_lun_ws_cache;

	/* Max IOs per controller - operating level dependent */
        ACIT_U32                max_ios_per_controller;

	/* ACIT tells OSMs about targets with through this pointer */
	PACIT_TARGET_TABLE	pacit_target_table;

        /* Size of channel. Includes space for all possible target pointers */
        ACIT_U32                osm_channel_mem_size;

        /* Size of a target entry. Includes space for LUN pointers */
        ACIT_U32                osm_target_mem_size;

        /* ACIT WS sizes */
        ACIT_U32                acit_controller_ws_size;
        ACIT_U32                acit_target_ws_size;
        ACIT_U32                acit_lun_ws_size;
        ACIT_U32                acit_io_ws_size;
        ACIT_U32                acit_tmf_ws_size;
        
        /* Pointer to Scsi_Host structure, given upon scsi_register */
	struct 	Scsi_Host 	*host;

        /* unique id for the adapter */
        ACIT_U32                hba_unique_id;
       
	/* Free IO list for preallocated IO requests */
	spinlock_t		free_list_lock;
	struct list_head        free_req;

	/* A list of jobs to be processed in timer context */
	spinlock_t		job_queue_lock;
	struct list_head        job_queue;

	/* IOCTLS and IOs will be processed with the following lock held */
	spinlock_t		io_ioctl_lock;
	
	/* This semaphore serializes the ioctls and puts the process to
	   sleep if it tries to send an ioctl when one is in progress 
	   already */
	struct semaphore	ioctl_lock;

	ACIT_UEXACT32 		state;

        /* PCI Resource map for HBA */
	/* TODO - get value from a macro */
        se_pci_resource       resourcemap[MAX_PCI_RESOURCES];

	/* IOCTL completion. This can be per controller because we can 
	   process only one ioctl at a time */
	struct completion	ioctl_completion;
        /* We have to keep login completion pending until os enumeration
           of the target is finished. Until that login ioctl will wait
           on this completion */
	struct completion	login_completion;
	ACIT_STATUS		login_status;

	/* ACIT_IOCTL_REQUEST for internal ioctls generated by OSM */
	ACIT_IOCTL_REQUEST	internal_ioctl;
	/* data buffer pointers for intenal ioctl requests */
	ACIT_MEM_DESCRIPTOR	ioctl_buf;
        /* persisten bt mappings are stored here */
	PQUERY_OS_BUS_TARGET_INUSE_PARAMS_OUT   pbt_in_use;

	ACIT_UEXACT32		active_cmds;
	ACIT_UEXACT32		aborted_cmd_count;
	ACIT_UEXACT8 		in_use;
	ACIT_UEXACT8  		top_scsi_channel;
	ACIT_UEXACT8  		top_scsi_id;
	ACIT_UEXACT8  		top_scsi_lun;
        ACIT_SEXACT8            name[32];
	ACIT_SEXACT8 		detail[55];
	kthread_t		worker_thread;
	struct semaphore	kthread_sema;
#if defined(__VMKERNEL_MODULE__)
	spinlock_t 		worker_thread_lock;
#endif
	unsigned int 		major;
        struct timer_list hba_timer;
        ACIT_UEXACT8            timer_shutdown;
	struct be_msix_entry 	msix_entries[BE_NUM_MSIX_ENTRIES];
        /* lock to sync internal ios */
	struct semaphore	internal_io_lock;
       	unsigned int 	        boot_tgt_added;    
        unsigned char           bt_map[MAX_TARGETS_SUPPORTED/8];
	char  	buf_ip[255];
	char  	buf_name[255];
} se_hba;

struct se_target_job_data {
	PACIT_TARGET_ENTRY	ptarget_entry;
        se_target             *ptarget;
	int			command;
};

struct se_lunchange_info {
        unsigned int            bus_number;
        unsigned int            target_number;
        unsigned int            lun_number;
        int                     operation;
};

#if defined(__VMKERNEL_MODULE__)
struct se_process_reportlun_job {
        se_target		*ptarget;
	unsigned char		*rl_buffer;
        struct scsi_cmnd	*pscsi_cmnd;
};

#endif

struct se_free_dma_buf {
	void * virt;
	ACIT_SG_ELEMENT psg;
};

struct se_timer_queue {
	struct list_head        job_queue;
	void 		        (*fn)(se_hba *, void *);
	union {
		struct se_target_job_data tj_data;
                struct se_lunchange_info lc_data;
		struct se_free_dma_buf fdb;
#if defined(__VMKERNEL_MODULE__)
		struct se_process_reportlun_job	rl_data;
#endif
	} data;
};

struct se_internal_io {
        union {
                se_report_lun         rlun;
            /*    se_read_capacity      rcap;
                se_inquiry            inq;*/
        } u;
};

ACIT_STATUS
se_get_btl_info(se_hba *phba,
		PACIT_TARGET_ENTRY ptarget_entry, 
		PMGMT_BTL_MAPPING pbtl_mapping);

ACIT_STATUS se_get_cntlr_info(se_hba *phba, 
			PMGMT_HBA_ATTRIBUTES pattr_buf);

ACIT_STATUS se_get_boot_target_info(se_hba *phba,
			PGET_BOOT_TARGET_OUT	pattr_buf);

ACIT_STATUS se_get_default_gateway(se_hba *phba,
			PGET_DEFAULT_ROUTE_PARAMS	pattr_buf);

ACIT_STATUS se_get_all_if_id( se_hba	*phba,
			PGET_ALL_IF_ID_PARAMS	pattr_buf);

ACIT_STATUS se_get_nw_info(se_hba *phba,
			PGET_IF_INFO_PARAMS	pattr_buf);

ACIT_STATUS se_get_nic_info(se_hba *phba,
			PNIC_CONFIG_PARAMS pattr_buf);

ACIT_STATUS se_get_iqn_info(se_hba *phba,
			PISCSI_HBA_NAME_PARAMS pattr_buf);

ACIT_STATUS se_get_allsess_info(se_hba *phba,
			PGET_ALL_SESSION_ID_PARAMS pattr_buf,
			unsigned int data_length);

ACIT_STATUS se_get_asession_info(se_hba *phba,
			PGET_A_SESSION_PARAMS pattr_buf,
			UEXACT32 session_handle);

ACIT_STATUS
se_get_port_name(
	se_hba			*phba,
	uint8_t			*pport_name);


ACIT_STATUS
se_return_to_bt_slot(se_hba *phba,
		UEXACT16  channel,
		UEXACT16 target);

ACIT_VOID
se_return_cmd(se_hba *phba,
		se_req *prequest);

unsigned int
se_get_lun_mapping(se_target *ptarget,
		se_lun_address *plun);


ACIT_SEXACT32
se_complete_tmf_request(se_hba *phba, 
                        se_tmf *ptmf);

ACIT_STATUS
se_complete_io_request(se_hba                *phba,
                	se_target             *ptarget,
                	se_device             *pdevice,
                	se_io                 *posm_request,
                	PACIT_IO_REQUEST        pacit_io_request);

void
se_dma_unmap(se_hba *phba, se_io *posm_request,
		struct scsi_cmnd *pscsi_cmnd);

void 
se_schedule_job(se_hba *phba,
		struct se_timer_queue *job);

unsigned int 
se_set_session_conf(se_hba *phba, se_target *ptarget,
        PACIT_TARGET_ENTRY   ptarget_entry);

ACIT_STATUS
se_get_session_eto(
        se_hba                  *phba,
        PACIT_TARGET_ENTRY      ptarget_entry,
        se_target       *ptarget);


unsigned int
se_get_bt_inuse(
	se_hba		*phba,
	ACIT_UEXACT8	domain); 

void
se_handle_target_event(
                se_hba                *phba,
                ACIT_UEXACT32           event_code,
                ACIT_UEXACT32           info);

void
se_handle_controller_event(
                se_hba                *phba,
                ACIT_UEXACT32           event_code,
                ACIT_UEXACT32           info);

se_io*
se_build_internal_io_request(
                se_hba        *phba,
                se_target     *ptarget,
                se_device     *pdevice,
                ACIT_UEXACT8    *pcdb,
                ACIT_UEXACT32   cdb_len,
                ACIT_UEXACT32   direction,
                ACIT_UEXACT32   timeout);
ACIT_UEXACT32 
se_alloc_intenal_cmd_buffer(
                se_hba     *phba,
                se_target  *ptarget,
                int          len);
void 
se_free_internal_cmd_buffer(
                se_hba     *phba, 
                se_target  *ptarget);

int
se_get_new_hba_unique_id(struct pci_dev	*pcidev);


#endif /*_BE2ISCSICOMMON_H*/
