/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

#include <linux/version.h>
#include <linux/delay.h>
#ifdef LINUX_2_4
#include <linux/blk.h>
#include "sd.h"
#if defined(MODVERSIONS)
#include <linux/modversions.h>
#endif
#else
#include <linux/interrupt.h>
#include <linux/blkdev.h>
#include <linux/pci.h>
#endif
#if LINUX_VERSION_CODE != KERNEL_VERSION(2,6,5)
#include "scsi/scsi.h"
#else
#include "../drivers/scsi/scsi.h"
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include "scsi/scsi_cmnd.h"
#include "scsi/scsi_device.h"
#endif
#include "acitcfg.h"
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/reboot.h>
#include "acitdefs.h"
#include "acitscsi.h"
#include "uerr.h"
#include "ioctl_top_bmap.h"
#include "ioctl_top_enum.h"
#include "acitapi.h"
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/list.h>
#include <asm/unistd.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
#include <asm/semaphore.h>
#else
#include <linux/semaphore.h>
#endif
#include <asm/delay.h>
#include <linux/genhd.h>
#include <linux/spinlock.h>
#if defined(LINUX_2_4)
#include <linux/tqueue.h>
#include "hosts.h"
#else
#include <scsi/scsi_host.h>
#endif
#include "be2iscsi_common.h"
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
#endif
#include "version.h"
#include "be2iscsi_osm.h"
#include "be2iscsi_debug.h"
#include "acit_iface.h"
#if defined(__VMKERNEL_MODULE__) && defined(LINUX_2_4)
#include "vmklinux_dist.h"
#include "vmk_basic_types.h"
#endif //__VMKERNEL_MODULE__
#include <linux/slab.h>

#ifdef BE2ISCSI_DEBUG
unsigned int alloc_count = 0;
unsigned int free_count = 0;

extern struct list_head hba_chain;

#if 0
#undef kmalloc
#undef kfree

void *
dmalloc(unsigned int size, unsigned int flags, char *fn)
{
	void *ptr = kmalloc(size, flags);

	alloc_count++;

	return ptr;	
}

void
dfree(void *ptr, char *fn)
{
	free_count++;

	kfree(ptr);
}
#endif

void
clear_device_stats(void)
{
	struct list_head	*hba_entry;
	se_hba			*phba;
	se_target		*ptarget;
	se_device		*pdevice;
	unsigned		hba_index;
	unsigned		target_index;
	unsigned		lun_index;

	hba_index = 0;
	target_index = 0;
	lun_index = 0;

	list_for_each(hba_entry, &hba_chain) {

		phba = list_entry(hba_entry, se_hba, hba_queue);

		hba_index++;

		list_for_each_entry(ptarget, &phba->target_list, sibling_list) {

			target_index++;

			list_for_each_entry(pdevice, &ptarget->lun_list,
					sibling_list) {

				pdevice->num_errors = 0;

				lun_index++;
			}
		}
	}
}
#endif

#if defined(__VMKERNEL_MODULE__) || defined(LINUX_2_4)
ACIT_SEXACT32 		
se_proc_info(
char	*buffer, 
char	**start, 
off_t	offset,
int	length, 
int	inode, 
int	inout
)
{
#ifdef BE2ISCSI_DEBUG
	int			len;
	unsigned		i;
	u64			avg;
	struct list_head	*hba_entry;
	struct list_head	*target_entry;
	se_hba			*phba;
	se_target		*ptarget;
	se_device		*pdevice;
	unsigned		hba_index;
	unsigned		target_index;
	unsigned		lun_index;
	PACIT_TARGET_ENTRY	ptarget_entry; 

	len = 0;

	if (inout) {
		if (strncmp(buffer, "clear", 5) == 0) {
			printk("clearing\n");

			hba_index = 0;
			target_index = 0;
			lun_index = 0;

			clear_device_stats();

			return length;
		}
		if (strncmp(buffer, "help", 4) == 0) {
			printk("clear             - clear statistics\n");
		}
		printk("unknown: ");
		for (i = 0 ; i < length ; i++)
		{
			printk("%c", *(buffer + i));
		}
		printk("\n");
		
		return -EINVAL;
	}

	hba_index = 0;

        list_for_each(hba_entry, &hba_chain) {

		target_index = 0;

                phba = list_entry(hba_entry, se_hba, hba_queue);

		hba_index++;

		len += sprintf(buffer + len, "HBA%d - ", hba_index);

		list_for_each_entry(ptarget, &phba->target_list, sibling_list) {

			lun_index = 0;

			target_index++;

			list_for_each_entry(pdevice, &ptarget->lun_list,
				sibling_list) {

#if 0
				len += sprintf(buffer + len,
					"T%02dL%02d%c|%4d|%8d ",
						target_index, lun_index,
						(ptarget->state & SE_TARGET_BUSY?'B':' '),
						pdevice->num_ios_outstanding,
						pdevice->num_errors);
#else
				len += sprintf(buffer + len,
					"%2d|%c ",
						pdevice->num_ios_outstanding,
						(ptarget->state & SE_TARGET_BUSY?'B':' '));
#endif
				lun_index++;
			}
		}

		len += sprintf(buffer + len, "\n");

		ptarget_entry = &phba->pacit_target_table->target_entry[0];

		for (i = 0; i < phba->pacit_target_table->total_entries; i++)
		{
			if ((ptarget_entry[i].acit_tflags & ACIT_TFLAG_IN_USE)
				== 0)
			{
				continue;
			}

			len += sprintf(buffer + len, "  %d: phba=0x%x | "
				"ptarget_key=0x%x | ",
				i, phba, ptarget_entry[i].ptarget_key);

			if (ptarget_entry[i].ptarget_key != NULL)
			{
				len += sprintf(buffer + len, "osm_key %x\n",
					ptarget_entry[i].ptarget_key->osm_key);
			}
			else
			{
				len += sprintf(buffer + len, "osm_key NULL\n");
			}
		}
	}

	if (offset > len)
		return 0;

	*start = buffer + offset;
	len -= offset;

	if (len > length)
		len = length;

	return len;
#else
	return 0;
#endif
}
#endif
