/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */


/*++

    Filename:

        be2iscsi_debug.h

    Purpose:

        This file contains macro definitions for supporting debug


--*/

/*
 *
 *  Common Debug code for driver
 *
 */
#ifndef _BE2ISCSIDEBUG_H_
#define _BE2ISCSIDEBUG_H_
/*  Assertions and KDB codes  */
#if defined(BE2ISCSI_DEBUG) && !defined(_IA64_)
/*#define KDBENTER() __asm__ volatile ("int $3");*/
/*#define ASSERT() KDBENTER() */
#define KDBENTER() panic()
#define ASSERT(x) if(!(x)) __asm__ volatile ("int $3");
#else
#define KDBENTER() panic()
#define ASSERT(x)
#endif

/*
 * Debugging Macros.
 */ 
#ifdef BE2ISCSI_DEBUG

extern ACIT_U32	debug_level;
#define PDEBUG(debug_mask, fmt, args...){\
	if (debug_mask & debug_level)\
	{\
		printk("(%s():%d):0x%x:", __FUNCTION__,__LINE__, debug_level);\
		printk(fmt, ##args);\
	}\
}

#define PCRIT(debug_mask, fmt, args...){\
	if (debug_mask & debug_level)\
	{\
		printk("<c>");\
		printk(fmt, ##args);\
		printk("</c>");\
	}\
}

#define PERROR(fmt, args...) printk(fmt, ##args)
#define PINFO(fmt, args...) printk(fmt, ##args)
#define PWARN(fmt, args...) printk(fmt, ##args)
	
#else

#define PDEBUG(debug_mask, fmt, args...)	/**/
#define PCRIT(debug_mask, fmt, args...)	/**/
#define PERROR(fmt, args...) printk(fmt, ##args)
#define PINFO(fmt, args...) 
#define PWARN(fmt, args...) 

#endif /* ifdef BE2ISCSI_DEBUG */

#ifdef BE2ISCSI_DBG_FUNC

#define BEGIN PDEBUG(DBG_TRACE, "Entering\n");
#define END PDEBUG(DBG_TRACE, "Leaving\n");
#define ERROR_END	printk("<!- %s (%d)>", __FUNCTION__, __LINE__)

#else

#define BEGIN		/**/
#define END		/**/
#define ERROR_END	/**/

#endif /* ifdef BE2ISCSI_DBG_FUNC */

/*
 * Debugging switch level.
 */ 
#define DBG_ALL		0xFFFF		/* All DBG msg turned on, except ISR */
#define DBG_NONE	0x0000

#define DBG_INIT	0x0001		/* Driver initialization messages */
#define DBG_UNLOAD	0x0002		/* Driver unload/shutdown messages */
#define DBG_SCSI	0x0004		/* SCSI command related messages */
#define DBG_TRACE	0x0008		/* Function trace */
#define DBG_ERR_HNDL	0x0010		/* Error recovery routine messages */
#define DBG_TIMEOUT	0x0020		/* Messages timeout */
#define DBG_DISCOVER	0x0040		/* Discovery related messages */
#define DBG_AIFACE	0x0080		/* ACIT interface debug mesg */
#define DBG_POST	0x0100		/* Error message of posting I20 msg
					 * to IOP */
#define DBG_DEVICE	0x0200		/* I2O devices error messages */
#define DBG_IOCTL	0x0400		/* IOCTL related messages */
#define DBG_EVENT	0x0800		/* Private Event related message */
#define DBG_ISR		0x1000		/* Interrupt Service Routine message */
#define DBG_MISC	0x2000		/* Misc. Informational messages */

#define CRIT_Q	        0x4000		/* Critical Queue messages */

#define DBG_VOLUME_0    DBG_NONE        /* Define runtime volume level */
#define DBG_VOLUME_1    DBG_ERR_HNDL
#define DBG_VOLUME_2    DBG_VOLUME_1 | DBG_TIMEOUT
#define DBG_VOLUME_3    DBG_VOLUME_2 | DBG_DISCOVER
#define DBG_VOLUME_4    DBG_VOLUME_3 | DBG_INIT | DBG_UNLOAD
#define DBG_VOLUME_5    DBG_VOLUME_4 | DBG_ISR
#define DBG_VOLUME_6    DBG_VOLUME_5 | CRIT_Q 
#define DBG_VOLUME_7    DBG_VOLUME_6 | DBG_EVENT 
#define DBG_VOLUME_8    DBG_ALL 
#define YES 1
#define NO  0

#define BE2ISCSI_STABLE_CODE NO

#if BE2ISCSI_STABLE_CODE
 #define RETRY_CODE NO
#else
 #define RETRY_CODE YES 
#endif

#if 0
#ifdef BE2ISCSI_DEBUG
#undef kmalloc
#undef kfree

void *dmalloc(unsigned int size, unsigned int flags, char *fn);
void dfree(void *ptr, char *fn);
#define kmalloc(size, flag) dmalloc(size, flag, __FUNCTION__)
#define kfree(ptr) dfree(ptr, __FUNCTION__)
extern unsigned int alloc_count, free_count;
#endif
#endif
#endif /* _BE2ISCSIDEBUG_H_ */
