/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */


/*++

    Filename:

        be2iscsi_osm.h

    Purpose:

        This file contains the declarations of functions needed by
        be2iscsi_osm.c

--*/
#ifndef _BE2ISCSI_OSM_H
#define _BE2ISCSI_OSM_H

extern char	*be2iscsi_driver_version;

/* Helper Macros to declare configurable params */

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,0)
#define scsi_host_template	SHT
#endif

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,6)
                                                                                
#define DECLARE_CONF_PARAM_NUM(name, def_val) \
unsigned int name = def_val;    \
MODULE_PARM(name, "i");
#define DECLARE_CONF_PARAM_STR(name, def_val) \
unsigned char *name = def_val;          \
MODULE_PARM(name, "s");
                                                                                
#else /* LINUX_2_4 */
                                                                                
#define DECLARE_CONF_PARAM_NUM(name, def_val) \
unsigned int name = def_val;    \
module_param(name, int, S_IRUGO);
#define DECLARE_CONF_PARAM_STR(name, def_val) \
char *name = def_val;           \
module_param(name, charp, S_IRUGO);
                                                                                
#endif /* LINUX_2_4 */

#ifndef DMA_32BIT_MASK
#define DMA_32BIT_MASK 0xFFFFFFFF
#endif
#ifndef DMA_64BIT_MASK
#define DMA_64BIT_MASK 0xFFFFFFFFFFFFFFFF
#endif

#define INIT_NUM_CONF_PARAM(index, acit_name, cmdline_name) \
do { \
	strncpy(se_param[index].osm_config_param, #acit_name, ACIT_MAX_CONFIG_NAME_LENGTH); \
	strncpy(se_param[index].command_line_param, #cmdline_name, ACIT_MAX_CONFIG_NAME_LENGTH); \
	se_param[index].val = cmdline_name;\
} while(0);

#define INIT_STR_CONF_PARAM(index, name) \
do { \
        strncpy(se_param[index].osm_config_param, #name, ACIT_MAX_CONFIG_NAME_LENGTH); \
        strncpy(se_param[index].str_val, name, \
                                ACIT_MAX_CONFIG_PARAMETER_LENGTH);\
} while(0);
#define REPORT_LUN_HEADER_LENGTH 8
#define GET_ALLOCATED_REPORT_LUN_SIZE(ptarget) (ptarget->internal_io_len - 8)


#ifdef LINUX_2_4
#define RELEASE_IOR_LOCK() spin_unlock_irq(&io_request_lock);
#define ACQUIRE_IOR_LOCK() spin_lock_irq(&io_request_lock);
#define RELEASE_IOR_LOCK_RESTORE() spin_unlock_irqrestore(&io_request_lock, io_flags);
#define ACQUIRE_IOR_LOCK_SAVE() spin_lock_irqsave(&io_request_lock, io_flags);
#else
#define RELEASE_IOR_LOCK()
#define ACQUIRE_IOR_LOCK()
#endif


#define ADD_LUN         1
#define REMOVE_LUN      2
/*
 * Prototypes of entry points
 */
#ifndef LINUX_2_4
struct be_default_msix_entry {
	uint16_t entry;
	uint16_t index;
	const char *name;
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,18)
	irqreturn_t (*handler)(int, void *, struct pt_regs *);
#else
	irqreturn_t (*handler)(int, void *);
#endif
};
#endif

ACIT_SEXACT32 		
se_proc_info(
	ACIT_SEXACT8 *buffer, 
	ACIT_SEXACT8 **start, 
	off_t offset,
	ACIT_SEXACT32 length, 
	ACIT_SEXACT32 inode, 
	ACIT_SEXACT32 inout);

ACIT_SEXACT32
se_detect(
	struct scsi_host_template *sht);

ACIT_SEXACT32
se_queue(
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)) || defined(NEW_QUEUECMD)
	struct Scsi_Host *phost,
	struct scsi_cmnd *pscsi_cmnd);
#else
	struct scsi_cmnd *cmd, 
	void (*scsi_done) (struct scsi_cmnd *));
#endif

ACIT_SEXACT32
se_abort(
	struct scsi_cmnd *cmd);

ACIT_SEXACT32
se_reset(
	struct scsi_cmnd *cmd);

ACIT_SEXACT32
se_release(
	struct Scsi_Host *host);

void 
se_init_params(
        void );

void 
se_process_deferred_jobs(
        se_hba *phba);

#ifdef LINUX_2_4
ACIT_SEXACT32
se_bios_param(Disk *disk,
                kdev_t dev,
                ACIT_SEXACT32 geom[]);
#else
                                                                                
ACIT_SEXACT32
se_bios_param(struct scsi_device *sdev,
                struct block_device *bdev,
                sector_t capacity,
                int geom[]);
#endif

ACIT_SEXACT32
se_bus_reset(
	struct scsi_cmnd *cmd);

ACIT_SEXACT32
se_device_reset(
	struct scsi_cmnd *cmd);

const ACIT_SEXACT8*
se_info(
	struct Scsi_Host *host);

/*
 * File operation prototypes
 */

ACIT_SEXACT32
se_open(
	struct inode *inode, 
	struct file *file);

ACIT_SEXACT32
se_close(
	struct inode *inode, 
	struct file *file);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0)
long
se_ioctl(
	struct file *file, 
	ACIT_UEXACT32 cmd, 
	unsigned long arg);
#else
ACIT_SEXACT32
se_ioctl(
	struct inode *inode, 
	struct file *file, 
	ACIT_UEXACT32 cmd, 
	unsigned long arg);
#endif
/*
 * Shutdown event handler
 */

ACIT_SEXACT32
se_shutdown_event(
	struct notifier_block *nb, 
	u_long code,
	void *unused);

/*
 * PCI function handlers
 */

ACIT_SEXACT32
se_pci_dev_probe(
	struct pci_dev *pDev,
	const struct pci_device_id *id);

#ifdef POWER_MANAGEMENT

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16) 
int
se_pci_dev_suspend(struct pci_dev *pci,  
        pm_message_t state);
#else   /*  ChangeLog-2.6.11:22267:  [PATCH] pm: introduce pm_message_t  */
int
se_pci_dev_suspend(struct pci_dev *pci, 
        u32 state);
#endif

int
se_pci_dev_resume(
	struct pci_dev *pDev);
#endif

void
se_pci_dev_remove(
	struct pci_dev *pDev);


/*
 * struct scsi_host_template (see hosts.h) - some fields
 * to do with card config are filled in after the card is detected.
 */
#define	BE2ISCSI_DRIVER_NAME	"be2iscsi"

/* Temperory patch to support microcode */
/*#define SG_LIST_ELEMENTS	SG_ALL      */
#define SG_LIST_ELEMENTS 	30
#define BE2ISCSI_MAX_Q_DEPTH  1024 

#define BE2ISCSI_MAX_DEV_Q_DEPTH BE2ISCSI_MAX_Q_DEPTH
#define BE2ISCSI_MAX_INTERNAL_CMD BE2ISCSI_MAX_Q_DEPTH*2

#define BE2ISCSI_CMD_PER_LUN	255
#ifdef LINUX_2_4
#if  defined(BE2ISCSI_VARYIO_SHT_EXT) || defined(BE2ISCSI_VARYIO_SHT)
#define BE2ISCSI_SHT { 							  \
	next: NULL,							  \
	module: NULL,							  \
	proc_dir: NULL,							  \
	proc_info: se_proc_info,					  \
	name: BE2ISCSI_DRIVER_NAME,					  \
	detect: se_detect,						  \
	release: se_release,						  \
/*	info: se_info,						*/\
	queuecommand: se_queue,					  \
	eh_strategy_handler: NULL,					  \
	eh_abort_handler: se_abort,					\
	eh_device_reset_handler: se_device_reset,			  \
	eh_bus_reset_handler: se_bus_reset,				  \
	eh_host_reset_handler: NULL,					  \
	abort: NULL,							  \
	reset: NULL,							  \
	slave_attach: NULL,						  \
	bios_param: se_bios_param,					  \
	can_queue: BE2ISCSI_MAX_Q_DEPTH,	/* max simultaneous cmds */  	  \
	this_id: -1,			/* scsi id of host adapter */     \
	sg_tablesize: SG_LIST_ELEMENTS,	/* max scatter-gather cmds */     \
	cmd_per_lun: BE2ISCSI_CMD_PER_LUN,/* cmds per lun (linked cmds) */ \
	unchecked_isa_dma: 0,		/* no restrictions on DMA space */\
	use_clustering: ENABLE_CLUSTERING,	  			  \
	use_new_eh_code: 1,						  \
	can_do_varyio: 1,						  \
	proc_name: "be2iscsi"						  \
}
#else

#define BE2ISCSI_SHT { 							  \
	next: NULL,							  \
	module: NULL,							  \
	proc_dir: NULL,							  \
	proc_info: se_proc_info,					\
	name: BE2ISCSI_DRIVER_NAME,					  \
	detect: se_detect,						  \
	release: se_release,						  \
/*	info: se_info,						*/\
	queuecommand: se_queue,					  \
	eh_strategy_handler: NULL,					  \
	eh_abort_handler: se_abort,				          \
	eh_device_reset_handler: se_device_reset,			  \
	eh_bus_reset_handler: se_bus_reset,				  \
	eh_host_reset_handler: NULL,					  \
	abort: NULL,							  \
	reset: NULL,							  \
	slave_attach: NULL,						  \
	bios_param: se_bios_param,					  \
	can_queue: BE2ISCSI_MAX_Q_DEPTH,	/* max simultaneous cmds */  	  \
	this_id: -1,			/* scsi id of host adapter */     \
	sg_tablesize: SG_LIST_ELEMENTS,	/* max scatter-gather cmds */     \
	cmd_per_lun: BE2ISCSI_CMD_PER_LUN,/* cmds per lun (linked cmds) */ \
	unchecked_isa_dma: 0,		/* no restrictions on DMA space */\
	use_clustering: ENABLE_CLUSTERING,	  			  \
	use_new_eh_code: 1,						  \
	proc_name: "be2iscsi"						  \
}
#endif
#else /* linux 2.4 */

#if LINUX_VERSION_CODE != KERNEL_VERSION(2,6,9) 
#if defined(__VMKERNEL__)
#if defined BE2ISCSI_DEBUG
#define BE2ISCSI_SHT  {						\
        .module                 = THIS_MODULE,			\
        .name                   = "be2iscsi",			\
        /*.release                = se_release,*/ 	        \
	/*.detect                 = se_detect,*/        	\
        .proc_info              = se_proc_info,			\
        /*.info                   = se_info, */			\
	.proc_name 		= "be2iscsi",			\
        .queuecommand           = se_queue,			\
        .eh_abort_handler       = se_abort,			\
        .eh_device_reset_handler = se_device_reset,		\
        .eh_bus_reset_handler   = se_bus_reset,			\
        .bios_param             = se_bios_param,		\
	.sg_tablesize		= SG_LIST_ELEMENTS, 		\
        .can_queue              = BE2ISCSI_MAX_Q_DEPTH,	\
        .this_id                = -1,				\
        .cmd_per_lun            = BE2ISCSI_CMD_PER_LUN,	\
        .use_clustering         = ENABLE_CLUSTERING,		\
        .slave_alloc            = se_slave_alloc,		\
	.target_alloc           = se_target_alloc,		\
	.target_destroy         = se_target_destroy,		\
        .slave_destroy          = se_slave_destroy		\
};
#else
#define BE2ISCSI_SHT  {                                         \
        .module                 = THIS_MODULE,                  \
        .name                   = "be2iscsi",                   \
        /*.release                = se_release,*/               \
        /*.detect                 = se_detect,*/                \
        /*.proc_info              = se_proc_info,*/             \
        /*.info                   = se_info, */                 \
        /*.proc_name              = "be2iscsi",*/               \
        .queuecommand           = se_queue,                     \
        .eh_abort_handler       = se_abort,                     \
        .eh_device_reset_handler = se_device_reset,             \
        .eh_bus_reset_handler   = se_bus_reset,                 \
        .bios_param             = se_bios_param,                \
        .sg_tablesize           = SG_LIST_ELEMENTS,             \
        .can_queue              = BE2ISCSI_MAX_Q_DEPTH, \
        .this_id                = -1,                           \
        .cmd_per_lun            = BE2ISCSI_CMD_PER_LUN, \
        .use_clustering         = ENABLE_CLUSTERING,            \
        .slave_alloc            = se_slave_alloc,               \
        .target_alloc           = se_target_alloc,              \
        .target_destroy         = se_target_destroy,            \
        .slave_destroy          = se_slave_destroy              \
};
#endif

#else
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
#define BE2ISCSI_SHT  {						\
        .module                 = THIS_MODULE,			\
        .name                   = "be2iscsi",			\
        /*.release                = se_release,*/ 	        \
	/*.detect                 = se_detect,*/        	\
        /* .proc_info              = se_proc_info,*/		\
        /*.info                   = se_info, */			\
	.proc_name 		= "be2iscsi",			\
        .queuecommand           = se_queue,			\
        .eh_abort_handler       = se_abort,			\
        .eh_device_reset_handler = se_device_reset,		\
        .eh_bus_reset_handler   = se_bus_reset,			\
        .bios_param             = se_bios_param,		\
	.sg_tablesize		= SG_LIST_ELEMENTS, 		\
        .can_queue              = BE2ISCSI_MAX_Q_DEPTH,	\
        .this_id                = -1,				\
        .cmd_per_lun            = BE2ISCSI_CMD_PER_LUN,	\
        .use_clustering         = ENABLE_CLUSTERING,		\
        .slave_alloc            = se_slave_alloc,		\
        .shost_attrs            = be2iscsi_attrs,		\
        .slave_configure        = se_slave_configure,		\
        .slave_destroy          = se_slave_destroy		\
}
#else

#define BE2ISCSI_SHT  {						\
        .module                 = THIS_MODULE,			\
        .name                   = "be2iscsi",			\
        /*.release                = se_release,*/ 	        \
	/*.detect                 = se_detect,*/        	\
        /* .proc_info              = se_proc_info,*/		\
        /*.info                   = se_info, */			\
	.proc_name 		= "be2iscsi",			\
        .queuecommand           = se_queue,			\
        .eh_abort_handler       = se_abort,			\
        .eh_device_reset_handler = se_device_reset,		\
        .eh_bus_reset_handler   = se_bus_reset,			\
        .bios_param             = se_bios_param,		\
	.sg_tablesize		= SG_LIST_ELEMENTS, 		\
        .can_queue              = BE2ISCSI_MAX_Q_DEPTH,	\
        .this_id                = -1,				\
        .cmd_per_lun            = BE2ISCSI_CMD_PER_LUN,	\
        .use_clustering         = ENABLE_CLUSTERING,		\
        .slave_alloc            = se_slave_alloc,		\
        .shost_attrs            = be2iscsi_attrs,		\
        .slave_destroy          = se_slave_destroy		\
}

#endif /* kernel 2.6.25 */

#endif
#else
#define BE2ISCSI_SHT  {						\
        .module                 = THIS_MODULE,			\
        .name                   = "be2iscsi",			\
        /*.release                = se_release,*/ 	        \
	/*.detect                 = se_detect,*/        	\
        .proc_info              = se_proc_info,			\
        /*.info                   = se_info, */			\
	.proc_name 		= "be2iscsi",			\
        .queuecommand           = se_queue,			\
        .eh_abort_handler       = se_abort,			\
        .eh_device_reset_handler = se_device_reset,		\
        .eh_bus_reset_handler   = se_bus_reset,			\
        .bios_param             = se_bios_param,		\
	.sg_tablesize		= SG_LIST_ELEMENTS, 		\
        .can_queue              = BE2ISCSI_MAX_Q_DEPTH,	\
        .this_id                = -1,				\
        .cmd_per_lun            = BE2ISCSI_CMD_PER_LUN,	\
        .use_clustering         = ENABLE_CLUSTERING,		\
        .shost_attrs            = be2iscsi_attrs,		\
        .slave_alloc            = se_slave_alloc,		\
        .slave_destroy          = se_slave_destroy,		\
	.dump_sanity_check 	= be2iscsi_sanity_check,		\
	.dump_poll 		= be2iscsi_dump_poll		\
};

#endif
#endif

#define BE2ISCSI_I2O_VERSION	BUILDNUM

#define BE2ISCSI_REVISION	'1'
#define BE2ISCSI_SUBREVISION	'2'
#define BE2ISCSI_BETA		""
#define BE2ISCSI_MONTH		01 
#define BE2ISCSI_DAY		03
#define BE2ISCSI_YEAR		(1980-2004)

#define BE2ISCSI_DRIVER		"be2iscsi"
#define BE2ISCSI_MAJOR	152

#define BE2ISCSI_PCI_REVISION_ID   0x0

/* Internal Timer interval in sec */
#define INTERNAL_TICK 1

/* Internal scsi status, good for kernel 2.4.x */
#define DID_BUS_LINKDOWN 0xd

#define TMOUT_LINKDOWN  		(20)
#define TMOUT_SESS_RECOVERY		(120)

#define CMD_REQ_ACTIVE	0x01
#define CMD_REQ_DONE	0x02
#define CMD_REQ_ABORT	0x04
#define CMD_REQ_EXTENDED_COMPLETION 0x20/* due to linkdown timeout is extended*/

#define DRV_SUCCESS	0x2002
#define DRV_FAILED	0x2003

#ifndef TRUE
#define TRUE		1
#define FALSE		0
#endif

/* Device state flags */
#define SE_DEV_OFFLINE   		(1<<0)  // 0x01
#define SE_DEV_RL_PRESENT             	(1<<1)  // 0x02
#define SE_DEV_RESET   			(1<<2)  // 0x04
#define SE_DEV_DESTROY  		(1<<3)  // 0x08

/* Target state flags */
#define SE_TARGET_BUSY			(1<<0)  // 0x01
#define SE_TARGET_DESTROY		(1<<1)  // 0x02
#define SE_TARGET_OFFLINE		(1<<2)  // 0x04

/* 
 * If its been decided that a LUN map can be reported
 * to the OS. Should be set for simple LUN map scheme
 * and in custom map after the app has given the map
 * to the driver. Once a map is given by the app, driver
 * makes it persistent, and from then on even this can
 * cause the flag to set.
 */
  
#define BE2ISCSI_CAN_REPORT_LUN_MAP	0x40

/* delay timings */
#define	ONE_SECOND		1000
#define	TEN_SECONDS		10 * ONE_SECOND

/*
 * Function Prototypes
 */
#ifndef __VMKERNEL_MODULE__
void be2iscsi_alloc_sysfs_attr(se_hba *ha);
void be2iscsi_free_sysfs_attr(se_hba *ha);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
ssize_t be2iscsi_show_fw_version(struct class_device *class_dev, char *buf);

ssize_t be2iscsi_show_driver_version(
					struct class_device *class_dev, 
					char *buf);

ssize_t be2iscsi_show_ip_address    (
					struct class_device *class_dev, 
					char *buf);

ssize_t be2iscsi_show_gateway       (
					struct class_device *class_dev, 
					char *buf);
ssize_t be2iscsi_show_netmask       (
					struct class_device *class_dev, 
					char *buf);

ssize_t be2iscsi_show_mac_address   (
					struct class_device *class_dev, 
					char *buf);

ssize_t be2iscsi_show_adapter_iqn   (
					struct class_device *class_dev, 
					char *buf);

ssize_t be2iscsi_show_adapter_state (
					struct class_device *class_dev, 
					char *buf);

ssize_t be2iscsi_show_adapter_info (
					struct class_device *class_dev, 
					char *buf);

ssize_t be2iscsi_read_login_state(struct kobject *kobj, char *buf, loff_t off,
    size_t count);

ssize_t be2iscsi_read_target_iqn(struct kobject *kobj, char *buf, loff_t off,
    size_t count);

extern struct class_device_attribute *be2iscsi_attrs[];
#else 
extern struct device_attribute *be2iscsi_attrs[];
#endif
#endif

/*
 * Initialize and uninitialize
 */

ACIT_STATUS 
se_get_cntlr_info(
		se_hba *phba,
		PMGMT_HBA_ATTRIBUTES pattr_buf);


void
se_i2o_delete_hba(
	se_hba *pHba);

ACIT_SEXACT32
se_scsi_register(
	se_hba* pHba,
	struct scsi_host_template * sht);

void
se_delete_all_hba(
	void);

void se_delete_single_hba(se_hba *pHba);
/*
 * ISR
 */

#ifdef LINUX_2_4
void
#else
irqreturn_t
#endif
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,18)
se_isr(
	ACIT_SEXACT32 irq, 
	void *dev_id, 
	struct pt_regs *regs);
#else
se_isr(
	ACIT_SEXACT32 irq, 
	void *dev_id);
#endif


#ifdef LINUX_2_4
void
#else
irqreturn_t
#endif
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,18)
se_isr_msix(
	ACIT_SEXACT32 irq, 
	void *dev_id, 
	struct pt_regs *regs);
#else
se_isr_msix(
	ACIT_SEXACT32 irq, 
	void *dev_id);
#endif

int
se_setup_msix(
se_hba		*phba,
struct pci_dev	*pcidev, 
PACIT_DEVICE	pacit_device
);


/*
 * Timer
 */
#define TIMER_INTERVAL HZ/2
void 
se_timer (
	ACIT_U32_64 arg);

/*
 * Command completion
 */

se_req *
se_get_free_cmd(
                se_hba        *phba);

ACIT_VOID
se_return_cmd(
                se_hba        *phba,
                se_req        *prequest);


ACIT_STATUS
se_complete_io_request(
                se_hba                *phba,
                se_target             *ptarget,
                se_device             *pdevice,
                se_io                 *posm_request,
                PACIT_IO_REQUEST        pacit_io_request
                );


/* Target modify (add/remove) */
   
ACIT_STATUS
be2iscsi_osm_process_target_table(
                se_hba        *phba, 
                ACIT_U32        target_index,
                ACIT_U32        flags);

ACIT_STATUS
se_issue_discover_targets(
                se_hba *phba, 
                ACIT_U32 flag);

/*
 * Error handling
 */

/*
 * Thread functions
 */

ACIT_SEXACT32 
se_start_worker_thread(
	se_hba *pHba,
	void (*thread_fun)(kthread_t *),
	kthread_t *thread_id);

void
se_io_thread(
	kthread_t *kthread);

ACIT_SEXACT32
se_stop_worker_thread(
	se_hba *pHba,
	kthread_t *thread_id);

/*
 * IOCTL processing
 */


/*
 * Helper routines
 */

ACIT_UEXACT32 
se_init_free_req_list(
	se_hba *phba, 
	ACIT_UEXACT32 req_wse);

ACIT_VOID
se_delete_free_req_list(se_hba *phba);


ACIT_UEXACT32
se_init_workspace_for_io(
                se_hba        *phba,
                se_target     *ptarget,
                se_device     *pdevice,
                se_io         *pio_request,
                PACIT_IO_REQUEST pacit_io_request);


int se_worker_thread(void *data);
void se_free_mem_desc(se_hba  *phba);
void se_release_io_handles(se_hba *phba);
int se_alloc_io_handles(se_hba *phba);

void 
se_initialize_target(
                se_hba                        *phba, 
                void     			*entry_data);

void 
se_delete_hba(se_hba *phba,ACIT_UEXACT32 stop_type);

void
se_change_luns(
                se_hba                        *phba, 
                void                             *entry_data);

void
se_free_dma_buffer(
                se_hba                        *phba, 
                void                             *entry_data);


void
se_delete_target_job(
                se_hba                        *phba, 
                void     			*entry_data);

void
se_delete_target(
                se_hba                        *phba, 
		se_target			*ptarget);
#ifdef __VMKERNEL_MODULE__
void
se_deferred_process_reportlun(
se_hba		*phba, 
void		*entry_data
);

void
se_target_destroy(
struct scsi_target	*starget
);

int
se_target_alloc(
struct scsi_target	*starget
);
#endif
void
se_free_osm_res(
		se_hba *phba);

ACIT_STATUS
se_initialize_lun(
                se_target     *ptarget,
                se_device     **ppdevice,
                ACIT_UEXACT32   channel,
                ACIT_UEXACT32   target,
                ACIT_UEXACT32   lun,
                PACIT_LUN       pacit_lun);
void 
se_destroy_lun(
		se_hba *phba,
		se_device *pdevice);

void se_verify_parameters (void); 

#ifdef LINUX_2_4

int se_scan_target_for_luns(
		se_hba *phba,
                se_target     *ptarget);

void 
se_report_lunchange(
		se_hba *phba,
		void *plcdata);

inline ACIT_VOID 
se_delete_lun(
		se_hba *phba, 
		se_device *pdev);

ACIT_STATUS
se_internal_io_completion(
                PACIT_CONTROLLER_KEY    pcontroller_key,
                PACIT_IO_REQUEST        pio_request);
ACIT_STATUS
se_get_reportlun_data(
		se_hba	*phba,
                se_device     *pdevice);

se_device * 
find_lun_at (
		se_hba *phba, 
		ACIT_U32 channel, 
		ACIT_U32 target, 
		ACIT_U32 lun);


ACIT_UEXACT32
se_alloc_internal_cmd_buffer(
                se_hba     *phba,
                se_target  *ptarget,
                int          len);

void
se_free_internal_cmd_buffer(
                se_hba     *phba,
                se_target  *ptarget);

int 
find_lun(
		se_lun_address *plun, 
		se_target *ptarget);

ACIT_STATUS
se_process_reportlun_data(
                se_hba                    *phba,
                se_target                 *ptarget,
		char			  *buffer,
		char			  *cdb);
inline
ACIT_UEXACT32
se_init_workspace_for_io(
                se_hba        *phba,
                se_target     *ptarget,
                se_device     *pdevice,
                se_io         *pio_request,
                PACIT_IO_REQUEST pacit_io_request);

void 
se_select_queue_depths(
                struct Scsi_Host *host, 
                struct scsi_device *devicelist);

ACIT_SEXACT32 
se_detect(
		struct scsi_host_template *sht);

ACIT_SEXACT32
se_release(
		struct Scsi_Host *host);

#else           /* For Linux 2.6 Kernel */
int 
se_slave_alloc(
		struct scsi_device *sdev);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
int 
se_slave_configure(
		struct scsi_device *sdev);
#endif

void 
se_slave_destroy(
		struct scsi_device *sdev);
void
be2iscsi_dump_poll(struct scsi_device *device);

int
be2iscsi_sanity_check(struct scsi_device *device);

int 
se_call_scan_single_target(
		se_hba *phba,
                se_target     *ptarget);

#ifdef POWER_MANAGEMENT
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16) 
int
se_pci_dev_suspend(
		struct pci_dev *pci,
		pm_message_t state);
#else   /*  ChangeLog-2.6.11:22267:  [PATCH] pm: introduce pm_message_t  */

int
se_pci_dev_suspend(
		struct pci_dev *pci, 
		u32 state);
#endif
int
se_pci_dev_resume(
		struct pci_dev *pDev);
#endif

#endif


void LnxLogOSMError(
    UEXACT32                   log_level,
    UEXACT32                   phase,
    UEXACT32                   sub_module,
    UEXACT32                   error_code);

#endif /* _BE2ISCSI_OSM_H */
