/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

        be2iscsiosm24.c

    Purpose:

        This file implements the different entry points that are needed
        2.4 Kernel specific.


--*/
#ifdef LINUX_2_4

#include <linux/version.h>
#include <linux/reboot.h>
#include <linux/kmod.h> /* for call_usermodehelper */
#include <linux/fs.h> /* for open/write/close to proc calls */
#include <linux/blk.h>
#include "sd.h"
#if defined(MODVERSIONS)
#include <linux/modversions.h>
#endif
#include "scsi.h"
#include "acitcfg.h"
#include <linux/string.h>
#include <linux/kernel.h>
#include "acitdefs.h"
#include "acitscsi.h"
#include "ioctl_top_bmap.h"
#include "ioctl_top_enum.h"
#include "acitapi.h"
#include <linux/config.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/list.h>

#include <asm/unistd.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
#include <asm/semaphore.h>
#else
#include <linux/semaphore.h>
#endif
#include <linux/genhd.h>
#include <linux/blkdev.h>
#include <linux/spinlock.h>
#if defined(LINUX_2_4)
#include <linux/tqueue.h>
#include "hosts.h"
#else
#include <scsi/scsi_host.h>
#endif
#include "be2iscsi_common.h"
#include "version.h"
#include "be2iscsi_osm.h"
#include "be2iscsi_debug.h"
#include<linux/pci.h>
#include "acit_iface.h"
#ifdef __VMKERNEL_MODULE__
#include "vmk_basic_types.h"
#endif

/* External variable from OSM file */
extern struct semaphore se_configuration_lock;
extern struct pci_driver be2iscsi_pci_driver;
extern ACIT_API_TABLE acitapi_table;
extern ACIT_VERSION acit_api_version;

extern struct file_operations se_fops;
extern struct notifier_block se_shutdown_notifier;
extern ACIT_SEXACT8 hba_count;          /* Total count of hbas we have   */
extern unsigned int ios_per_lun;

#if 0
#ifndef __VMKERNEL_MODULE__
extern spinlock_t	io_request_lock;
#endif
#endif

se_device * 
find_lun_at(se_hba *phba, ACIT_U32 channel, ACIT_U32 target, ACIT_U32 lun)
{
        se_target     *ptarget = NULL, *ptemp = NULL;
	se_device *pdevice = NULL;
	ACIT_U32_64     flags = 0;

        BEGIN;

	list_for_each_entry(ptemp, &phba->target_list, sibling_list) { 
                if (ptemp->bus == channel && ptemp->target == target) {
                        ptarget = ptemp;
                        break;
                }
        }

       	if (ptarget == NULL) {

	        //PDEBUG(DBG_DISCOVER, "*********** !!! BAD !!! No targets "
		//	"in the list c=%d,t=%d,l=%d **********\n",
		//channel, target, lun);

		return NULL;
	} 

	PDEBUG(DBG_DISCOVER, "ptarget 0x%x FOUND \n", (unsigned int) ptarget);

	list_for_each_entry(pdevice, &ptarget->lun_list, sibling_list) { 
		if (pdevice->scsi_lun == lun) {
			PDEBUG(DBG_DISCOVER, "pdevice 0x%x FOUND \n", 
				(unsigned int) pdevice);

			return pdevice;
		}
	}

	PDEBUG(DBG_DISCOVER, "*********** !!! BAD !!! No devices in the "
		"list c=%d,t=%d,l=%d **********\n", channel, target, lun);

	return NULL;
}


ACIT_UEXACT32
se_alloc_internal_cmd_buffer(
                se_hba     *phba,
                se_target  *ptarget,
                int          len)
{
        BEGIN;
        if (ptarget->pinternal_io)
                return ACIT_STATUS_ERROR;
                                                                                
        ptarget->internal_io_len = len;
        ptarget->pinternal_io = pci_alloc_consistent(phba->pcidev,
                                        ptarget->internal_io_len,
                                        &ptarget->internal_io_bus_addr);
        if (!ptarget->pinternal_io)
                return ACIT_STATUS_INSUFFICIENT_RESOURCE;
                                                                                
        memset(ptarget->pinternal_io, 0 , ptarget->internal_io_len);
        return ACIT_STATUS_SUCCESS;
}


void
se_free_internal_cmd_buffer(
                se_hba     *phba,
                se_target  *ptarget)
{
        BEGIN;
        if (ptarget->pinternal_io)
                pci_free_consistent(phba->pcidev,
                        ptarget->internal_io_len,
                        ptarget->pinternal_io,
                        ptarget->internal_io_bus_addr);
        ptarget->internal_io_bus_addr = 0;
        ptarget->pinternal_io = NULL;
        ptarget->internal_io_len = 0;
        return;
}

se_device *
se_check_lun_exists(se_lun_address *plun, se_target *ptarget)
{
        se_device	*pdevice;
	se_hba		*phba;
	ACIT_U32_64     flags = 0;

        pdevice = NULL;
	phba = ptarget->pparent_hba;

	PDEBUG(DBG_DISCOVER, "Looking for LUN %llx\n", plun->u.lun_complete);

	list_for_each_entry(pdevice, &ptarget->lun_list, sibling_list) { 

		PDEBUG(DBG_DISCOVER, "checking BTL:%d:%d:%d | "
			"lun_complete = %llx\n",
			pdevice->scsi_channel,
			pdevice->scsi_id,
			pdevice->scsi_lun,
			pdevice->lun.u.lun_complete);

		if (memcmp((char *)&pdevice->lun.u.lun_complete,
			(char *)&plun->u.lun_complete, 8) != 0) {
			continue;
		}

		PDEBUG(DBG_DISCOVER, "match found\n");

#if defined(__VMKERNEL_MODULE__) && !defined(__VMKLNX__)
		se_device_hold(pdevice);
#endif

		return pdevice;
	}


	PDEBUG(DBG_DISCOVER, "no match\n");

        return NULL;
}

/*++
Function:
        se_process_reportlun_data

Parameter:
        phba:      [in]    pointer to hba
        ptarget:   [in]    pointer to target

Return:

Context:
        This is called from defferred execution thread context.

Description:

--*/

ACIT_STATUS
se_process_reportlun_data(
se_hba			*phba,
se_target		*ptarget,
char			*buffer,
char			*cdb
)
{
        ACIT_U32		num_luns;
        se_lun_address		*plun;
        ACIT_U32		index;
        ACIT_STATUS		acit_status;
        se_report_lun		*preport_lun;
        ACIT_U32		lunbytes;
	se_device		*pdevice;
	se_device		*tmp_pdevice;
	unsigned		alloc_length;
	struct Scsi_Host	*host;
	unsigned int		lun_number;
	char			*lun_array;
	ACIT_U32_64     	flags = 0;

        BEGIN;

        lunbytes = 0;

        preport_lun = (se_report_lun *)buffer;

        PDEBUG(DBG_DISCOVER,"Inside se_process_reportlun_data\n");

        lunbytes = ACIT_SWAP32(preport_lun->header.lunlist_len);

	if (cdb != NULL)
	{
		alloc_length = ACIT_SWAP32(*((unsigned *)&cdb[6]));

		/*
		 * The CDB has the length that we are requesting, and
		 * the response will have the actual amount of data
		 * present.  If we requested less than the amount of
		 * data that is available, then the lun list will be
		 * incomplete and we can't use it to determine if LUNs
		 * have been added or deleted.
		 */
		if (lunbytes + sizeof(se_report_lun_header) > alloc_length)
		{
			PDEBUG(DBG_DISCOVER, "All LUN info not present\n");
			PDEBUG(DBG_DISCOVER, "available = %x "
				"allocated = %x\n", 
				lunbytes + sizeof(se_report_lun_header),
				alloc_length);

			return ACIT_STATUS_SUCCESS;
		}
	}

        num_luns = lunbytes / 8;
        
        PDEBUG(DBG_DISCOVER, "luns returned by REPORT_LUNS: num_luns=%x\n",
		num_luns);

        if (num_luns > phba->max_luns_per_target) {

                /* We support only max_luns_per_target */
                PDEBUG(DBG_ERR_HNDL, "REPORT_LUNS has %d LUNs, supporting"
                                "only %d\n", num_luns, 
                                phba->max_luns_per_target);

                num_luns = phba->max_luns_per_target;

        }

        PDEBUG(DBG_DISCOVER, "num_luns = %d\n", ptarget->num_luns);

        for (index = 0; index < num_luns; index++) {

                plun = &preport_lun->lun[index];

		pdevice = se_check_lun_exists(plun, ptarget);

                /* 
                 * If this plun entry is not present in our BTL map, 
                 * add it now
                 */
                if (pdevice != NULL) {
			/*
			 * This LUN is in the Report LUN list and 
			 * the driver already knows about it.
			 */
			pdevice->state |= SE_DEV_RL_PRESENT;

			continue;
		}

		host = phba->host;

		PDEBUG(DBG_DISCOVER,"New lun, initializing...\n");
#ifdef __VMKERNEL_MODULE__
		lun_array = (char *)&plun->u.lun_complete;

		/*
		 * 2 SCSI byte-ordered shorts
		 */
		lun_number = (lun_array[0]) << 8;
		lun_number = lun_number | (lun_array[1]) << 0;
		lun_number = lun_number | (lun_array[2]) << 24;
		lun_number = lun_number | (lun_array[3]) << 16;
#else
		lun_number = se_get_lun_mapping(ptarget, plun); 

		if (lun_number > phba->max_luns_per_target)
			return ACIT_STATUS_ERROR;
#endif
	
		acit_status = se_initialize_lun(ptarget,
				&pdevice,
#if defined(__VMKERNEL_MODULE__)
				0,
#else
				ptarget->pbtl_mapping->os_bus,
#endif
				ptarget->pbtl_mapping->os_target,
				lun_number,
				(PACIT_LUN)plun
				);
		if (acit_status != ACIT_STATUS_SUCCESS) {
			PDEBUG(DBG_DISCOVER, "Couldn't initialize LUN"); 
			return ACIT_STATUS_INSUFFICIENT_RESOURCE;
		}

#if defined(__VMKERNEL_MODULE__)
		PDEBUG(DBG_DISCOVER,
			"LUN initialized at BTL:%d:%d:%d\n",
			0,
			ptarget->pbtl_mapping->os_target,
			lun_number);
#else
		PDEBUG(DBG_DISCOVER,
			"LUN initialized at BTL:%d:%d:%d\n",
			ptarget->pbtl_mapping->os_bus,
			ptarget->pbtl_mapping->os_target,
			lun_number);
#endif

		/*
		 * if this is called asynchronously do an upcall to 
		 * report this lun to the OS
		 */
		if (phba->state & SE_STATE_HBA_INITIALIZED){
			struct se_lunchange_info lcinfo;
#ifndef __VMKERNEL_MODULE__
			host->max_lun = phba->top_scsi_lun + 1;
#endif
			lcinfo.lun_number = lun_number;
#if defined(__VMKERNEL_MODULE__)
			lcinfo.bus_number = 0;
#else
			lcinfo.bus_number = ptarget->pbtl_mapping->os_bus;
#endif
			lcinfo.target_number = ptarget->pbtl_mapping->os_target;
			lcinfo.operation = ADD_LUN;
			se_report_lunchange(phba, &lcinfo);
		}
		pdevice->state |= SE_DEV_RL_PRESENT;
        }

	list_for_each_entry_safe(pdevice, tmp_pdevice, &ptarget->lun_list,
		sibling_list) { 

		if (pdevice->state & SE_DEV_RL_PRESENT) {
			/*
			 * This device is either brand new or already know
			 * by the driver.  All we need to do is release
			 * it and it is ready to go.
			 */
			pdevice->state &= ~SE_DEV_RL_PRESENT;

		} else {
			struct se_lunchange_info lcinfo;

 			/*
			 * After search the Report LUNS, this driver was
			 * known before, but it has gone missing.  We
			 * need to delete it.
			 */
			PDEBUG(DBG_DISCOVER, "device was removed\n");
                        PDEBUG(DBG_DISCOVER, "Removing LUN at BTL:%d:%d:%d\n",
						pdevice->scsi_channel,
						pdevice->scsi_id,
						pdevice->scsi_lun);

#if defined(__VMKERNEL_MODULE__) && !defined(__VMKLNX__)
			se_device_hold(pdevice);

			pdevice->state |= SE_DEV_DESTROY;
#else
			se_delete_lun(phba, pdevice);
#endif
                }
#if defined(__VMKERNEL_MODULE__) && !defined(__VMKLNX__)
		se_device_release(pdevice);
#endif
        }

        return ACIT_STATUS_SUCCESS;
}

/*++
Function:
        se_init_workspace_for_io

Parameter:
        phba:             [in] pointer to se_hba structure
        ptarget:          [in] pointer to target
        pdevice:          [in] pointer to lun
        pio_request:      [in] pointer to se_io
        pacit_io_request: [in] pointer to acit io request

Return:
        ACIT_STATUS_SUCCESS - no errors
        errors from acit_init_io_ws if any

Context:
        IO

Description:
        This routine sets up workspace information for controller,
        target, lun and io in acit_io_request structure.

--*/

inline
ACIT_UEXACT32
se_init_workspace_for_io(
                se_hba        *phba,
                se_target     *ptarget,
                se_device     *pdevice,
                se_io         *pio_request,
                PACIT_IO_REQUEST pacit_io_request)
{
        ACIT_STATUS acit_status;
        
        pacit_io_request->pio_key->osm_key = pio_request;
        pacit_io_request->pio_key->acit_key = 0;

        pacit_io_request->plun_key = &pdevice->lun_key;
        pacit_io_request->ptarget_key = ptarget->posm_target_key;
        pacit_io_request->pcontroller_key = &phba->controller_key;

        /* Initialize IO workspace */
        /* 
         * We dont handle NULL IO ws. That case will arise when
         * ACIT reported 0 for sizeof IO WS during init
         */
        PDEBUG(DBG_SCSI,"calling init io ws with CKey %lx %lx IKey %lx %lx\n",
                (ACIT_U32_64)phba->controller_key.osm_key,
                (ACIT_U32_64)phba->controller_key.acit_key,
                (ACIT_U32_64)pacit_io_request->pio_key->osm_key,
                (ACIT_U32_64)pacit_io_request->pio_key->acit_key);

        if (phba->acit_io_ws_size) {

                acit_status = acitapi_init_io_ws(phba, &phba->controller_key,
                                pacit_io_request,
                                pacit_io_request->pio_key,
                                ptarget,
                                &pio_request->io_ws);

                if (acit_status != ACIT_STATUS_SUCCESS) {

                        PDEBUG(DBG_ERR_HNDL,"could not initialize io ws %d\n", 
                                acit_status);

                        return acit_status;
                }
        }

        PDEBUG(DBG_SCSI,"Actual Key values" 
                "CK - acit %lx osm %lx\n"
                "TK - acit %lx osm %lx\n"
                "LK - acit %lx osm %lx\n"
                "IK - acit %lx osm %lx\n",
                (ACIT_U32_64)phba->controller_key.acit_key, 
                (ACIT_U32_64)phba->controller_key.osm_key,
                (ACIT_U32_64)ptarget->posm_target_key->acit_key, 
                (ACIT_U32_64)ptarget->posm_target_key->osm_key,
                (ACIT_U32_64)pdevice->lun_key.acit_key, 
                (ACIT_U32_64)pdevice->lun_key.osm_key,
                (ACIT_U32_64)pacit_io_request->pio_key->acit_key, 
                (ACIT_U32_64)pacit_io_request->pio_key->osm_key);

        return ACIT_STATUS_SUCCESS;

}

/*++
Function:
        se_select_queue_depths

Parameter:
        host:         [in] pointer to Scsi_Host structure
        devicelist:   [in] pointer to Scsi_Device belonging to the HBA

Return:
        None

Context:
        Init

Description:
        This is the entry point that is used to set queue depth for devices
        that support tagged queueing. This entry point is called immediately
        after the SCSI Midlayer completes the device discovery process.

--*/

void 
se_select_queue_depths(
                struct Scsi_Host *host, 
                Scsi_Device *devicelist)
{
        Scsi_Device     *device;   /* scsi layer per device information */
#ifndef LINUX_2_4
        struct list_head        *list_entry;
#endif
        se_hba        *phba;

        phba = (se_hba *) host->hostdata[0];
#ifdef LINUX_2_4
        for (device = devicelist; device != NULL; device = device->next) {
#else
        list_for_each(list_entry, &devicelist->siblings) {
                device = (Scsi_Device *)list_entry;
#endif
                if (device->host != host) {
                        continue;
                }

                if ((host->can_queue) && (device->tagged_supported)) {
                        device->queue_depth = ios_per_lun;
#ifdef LINUX_2_4
                        device->tagged_queue = 1;
#endif
                } else {
                        device->queue_depth = 2;
                }
        }
}


#if 0 /* A royal hack to allow sparse luns */

struct dev_info {
        const char *vendor;
        const char *model;
        const char *revision;        /* Latest revision known to be bad.  Not used yet */
        unsigned flags;
};
#define BLIST_SPARSELUN         0x040
#endif

/*++
Function:
        se_detect

Parameter:
        sht:    [inout]   pointer to Scsi Host Template
                          structure

Return:
        Number of hosts detected

Context:
        Entry point

Description:
        This is the first entry point called by the SCSI Midlayer. It is called
        when the midlayer is trying to add our driver into the system.
        Its responsibility is to detect the supported HBAs and report the 
        number of cards found.
        The following are done in this function
        - registration with PCI
        - registration with SCSI
        - configuration params setup
        - allocating all resources needed
        - initialization of the HBAs
        - update the max count for BTL so that OS enumerates them
--*/

ACIT_SEXACT32 
se_detect(Scsi_Host_Template *sht)
{
        se_hba        *phba = NULL;
        struct list_head *lh, *lh_sf;
        LIST_HEAD(hba_chain);

        BEGIN;

#ifdef __VMKERNEL_MODULE__
	if (!vmk_set_module_version("%s", be2iscsi_driver_version)) {
		return 0;
	}
        /* In the vmkernel, we do not hold the io_request lock during init,
         * so leave it unlocked and do not lock it before returning. */
#else
         RELEASE_IOR_LOCK();
#endif

        PWARN("Detecting Emulex BE2ISCSI  iSCSI Host Bus Adapters !\n");
#if 0 /* A royal hack to allow sparse luns */
        struct dev_info *dev_list = (struct dev_info *)0xc03ddc80;

        dev_list[0].vendor = "SBS";
        dev_list[0].model = "iSCSI Targed";
        dev_list[0].flags = BLIST_SPARSELUN;
        
        printk("the device name is %s\n", dev_list[0].vendor);
#endif
#if 0
#ifdef CONFIG_SYSFS
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13) 
        be2iscsi_class = class_simple_create(THIS_MODULE, "be2iscsi");
#else   /* ChangeLog-2.6.13:52566: [PATCH] class: remove class_simple code, as no one in the tree is using it anymore.*/
        be2iscsi_class = class_create(THIS_MODULE, "be2iscsi");
#endif        

        if (IS_ERR(be2iscsi_class)) {
                printk("Cannot create sysfs class\n");i
#ifndef __VMKERNEL_MODULE__
                 ACQUIRE_IOR_LOCK();
#endif
                return 0;
        }
#endif
#endif
#ifdef LINUX_2_4 
        if (pci_present() == 0) {
#ifndef __VMKERNEL_MODULE__
           ACQUIRE_IOR_LOCK();
#endif
                return (0);
        }
#endif
        /* initialize se_param with commandline params */
        se_init_params();

        /* setup procedure table so that ACIT can operate */
	acit_initialize(ACIT_OP_MODE_NORMAL, &acitapi_table, sizeof(ACIT_API_TABLE));
        acitapi_get_api_version(&acit_api_version);
        hba_count = 0; 
        pci_register_driver(&be2iscsi_pci_driver);
        if (hba_count == 0) {
                pci_unregister_driver(&be2iscsi_pci_driver);
#ifndef __VMKERNEL_MODULE__
                ACQUIRE_IOR_LOCK();
#endif
                return (0);
        }

#if 0
        /*
         * Register our control device node
         * nodes will need to be created in /dev to access this
         * the nodes can not be created from within the driver
         */ 
        if (hba_count && register_chrdev(BE2ISCSI_MAJOR, 
                                BE2ISCSI_DRIVER, &se_fops)) {

                PDEBUG(DBG_INIT,"Failed register_chrdev for 0x%lx\n",
                                        (unsigned long)phba);
                se_delete_all_hba();
#ifndef __VMKERNEL_MODULE__
                ACQUIRE_IOR_LOCK();
#endif
                return (0);

        }
#endif

        /*
         * Register our shutdown/reboot notifier to the kernel.
         */          
        register_reboot_notifier(&se_shutdown_notifier);

        /*
         * Register the driver with scsi core.
         */         
        down(&se_configuration_lock);
        list_for_each_safe(lh, lh_sf, &hba_chain) {
                phba = list_entry(lh, se_hba, hba_queue);  
                PDEBUG(DBG_INIT,"Calling register for phba = 0x%lx\n", 
                                (unsigned long)phba);
                if (se_scsi_register(phba, sht) < 0) {
                        PDEBUG(DBG_INIT,"Failed se_scsi_register for 0x%lx\n",
                                        (unsigned long)phba);
                        se_delete_all_hba();
#ifndef __VMKERNEL_MODULE__
                        ACQUIRE_IOR_LOCK();
#endif
                        up(&se_configuration_lock);
                        return -ENOMEM; 
                }
        }
        up(&se_configuration_lock);
	PDEBUG(DBG_INIT,"returning %d hosts\n", hba_count);
#ifndef __VMKERNEL_MODULE__
        ACQUIRE_IOR_LOCK();
#endif
        return hba_count;
}

/*++
Function:
        se_release
                                                                                
Parameter:
        Scsi_Host:    [in] pointer to scsi_host.
                                                                                
Return:
        0 - always
                                                                                
Context:
        entry point
                                                                                
Description:
        This routine is the entry point called by the SCSI Midlayer when
        it is attempting to remove the driver from the system. This will
        be called when SCSI module is being unloaded. This will also be
        called if scsi_register failed.
                                                                                
--*/
ACIT_SEXACT32
se_release(struct Scsi_Host *host)
{
        se_hba *phba;
        phba = (se_hba *) host->hostdata[0];
        /* clean up acit stuff */
        down(&se_configuration_lock);
        se_delete_hba(phba, ACIT_STOP_TYPE_NORMAL);
        up(&se_configuration_lock);
	PDEBUG(DBG_UNLOAD, "%d more HBAs remain\n",hba_count);
        if (!hba_count) {
          pci_unregister_driver(&be2iscsi_pci_driver);
          unregister_reboot_notifier(&se_shutdown_notifier);
#if 0
#ifdef CONFIG_SYSFS
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13) 
        class_simple_destroy(be2iscsi_class);
#else   /* ChangeLog-2.6.13:52566: [PATCH] class: remove class_simple code, as no one in the tree is using it anymore.*/
        class_destroy(be2iscsi_class);
#endif        
#endif
#endif
          unregister_chrdev(BE2ISCSI_MAJOR, BE2ISCSI_DRIVER);
          PDEBUG(DBG_UNLOAD, "Unregister device !\n");
        }

#if 0
#ifdef BE2ISCSI_DEBUG
        if (alloc_count != free_count)
		printk("alloc_count = %d, free_count = %d\n", alloc_count, free_count);
#endif
#endif
        return 0;
}

/*++
Function:
        se_get_reportlun_data
                                                                                
Parameter:
        phba:      [in]    pointer to hba
        ptarget:   [in]    pointer to target
                                                                                
Return:
        ACIT_STATUS_SUCCESS - no errors
        ACIT_STATUS_ERROR - failed to get reportlun data.
                                                                                
Context:
        Called from defferred execution thread context.
                                                                                
Description:
        This routine is called to submit a report lun command to LUN0
        of a target. This checks if the report lun data buffer length
        is sufficient to hold the entire data, and if not, it reallocates
        buffer needed for the command and submits the command to acit.
        This function will wait for the command to complete.
                                                                                
--*/
ACIT_STATUS
se_get_reportlun_data(
                se_hba        *phba,
                se_device     *pdevice)
{
        ACIT_STATUS     acit_status = 0;
        ACIT_UEXACT8    cdb[ACIT_SCSI_MAX_CDB_LENGTH] = {0};
        ACIT_UEXACT32   data_len;
        ACIT_UEXACT32   timeout = 20; 
        se_io         *posm_request;
        se_target     *ptarget = pdevice->pparent_target;
        se_report_lun *preport_lun = NULL;
        ACIT_U32 luns;
        ACIT_U32 lunbytes;

        /* 
         * allocate for report lun data for this target. Initially we
         * allocate for MAX_STATIC_LUN_LIMIT luns. So actual data needed is 
         * = header + sizeoflun * MAX_STATIC_LUN_LIMIT
         * Not sure if this scheme is of any help, because it could so happen
         * that we allocated 512 bytes of physically contiguous memory, and the
         * page is not given to anyone else.
         * We cannot go for cached memory, since we dont know the actual size.
         */
        PDEBUG(DBG_DISCOVER,"Allocating for Report LUN\n");

        if (se_alloc_internal_cmd_buffer(phba, ptarget, 
                REPORT_LUN_HEADER_LENGTH + (8*MAX_STATIC_LUN_LIMIT)) !=
                ACIT_STATUS_SUCCESS) {
                PDEBUG(DBG_ERR_HNDL, "Error allocating pci memory\n");
                return ACIT_STATUS_ERROR;
        }
        preport_lun = (se_report_lun *)ptarget->pinternal_io;
        


reissue:
      
        lunbytes =  (preport_lun->header.lunlist_len & 0xFF000000) >> 24;
        lunbytes |= (preport_lun->header.lunlist_len & 0x00FF0000) >> 8;
        lunbytes |= (preport_lun->header.lunlist_len & 0x0000FF00) << 8;
        lunbytes |= (preport_lun->header.lunlist_len & 0x000000FF) << 24;

	luns = lunbytes / 8;
        
        PDEBUG(DBG_DISCOVER,"issuing reportlun for %d luns\n",luns);
        if (GET_ALLOCATED_REPORT_LUN_SIZE(ptarget) < (luns * 8)) {

                /* Free earlier allocation and reallocate based on new length */
                se_free_internal_cmd_buffer(phba, ptarget);
                if (se_alloc_internal_cmd_buffer(phba, ptarget, 
                        (luns * 8)+REPORT_LUN_HEADER_LENGTH) != 
                        ACIT_STATUS_SUCCESS) {
                        PDEBUG(DBG_ERR_HNDL, "Error allocating pci memory\n");
                        return ACIT_STATUS_INSUFFICIENT_RESOURCE;

                }
                PDEBUG(DBG_DISCOVER,"get_reportlun_data: "
                          "report_lun_len %d for %d luns\n",
                           ptarget->internal_io_len, luns);
        }


        data_len = ptarget->internal_io_len;

        cdb[0] = ACIT_SCSI_CMD_REPORTLUNS;
        cdb[6] = (data_len >> 24) & 0xFF;
        cdb[7] = (data_len >> 16) & 0xFF;
        cdb[8] = (data_len >> 8) & 0xFF;
        cdb[9] = (data_len) & 0xFF;

        /* Put all the above into acit io request format from free list */
        posm_request = se_build_internal_io_request(
                                phba,
                                ptarget,
                                pdevice,
                                (ACIT_UEXACT8 *)&cdb,
                                12,
                                READ,
                                timeout);

        if (posm_request == NULL) {

                PDEBUG(DBG_DISCOVER,"get_reportlun_data - failed "
                                "build_internal_io_request failed\n");

                return ACIT_STATUS_ERROR;
        }
        /* Submit the request to acit */
        acit_status = acitapi_start_io(phba, &phba->controller_key,ptarget, 
                        &posm_request->io_request); 

        wait_for_completion(&posm_request->complete_io);

	switch (posm_request->io_request.scsi_status) {
		case SCSI_STATUS_GOOD:
		case SCSI_STATUS_CONDITION_MET:
		case SCSI_STATUS_INTERMEDIATE:
		case SCSI_STATUS_INTERMEDIATE_CONDITION_MET:
			break;
		case SCSI_STATUS_CHECK_CONDITION:
		default:
			PDEBUG(DBG_DISCOVER, "acitapi_start_io failed %d\n",
				posm_request->io_request.scsi_status);
			acit_status = posm_request->io_request.scsi_status;
			se_free_internal_cmd_buffer(phba, ptarget);

			/*
			 * We can get a check condition if there are no
			 * LUNs configured behind the target.
			 */
			//acit_status = ACIT_STATUS_SUCCESS;

			goto failed;
        }

        /* 
         * Check report lun data header to see if we got
         * all data that we needed
         */
        lunbytes =  (preport_lun->header.lunlist_len & 0xFF000000) >> 24;
        lunbytes |= (preport_lun->header.lunlist_len & 0x00FF0000) >> 8;
        lunbytes |= (preport_lun->header.lunlist_len & 0x0000FF00) << 8;
        lunbytes |= (preport_lun->header.lunlist_len & 0x000000FF) << 24;

	luns = lunbytes / 8;

        PDEBUG(DBG_DISCOVER,"received Rlun completion %x\n",
                        luns);
        
        if ((luns*8) > GET_ALLOCATED_REPORT_LUN_SIZE(ptarget)) {
                PDEBUG(DBG_DISCOVER, "reissuing RLun %d %d\n",
                                luns, ptarget->internal_io_len);
                se_return_cmd(phba, (se_req *)posm_request);
                /* Reissue the command */
                goto reissue;
        }

	if (luns > MAX_LUN) {
		luns = MAX_LUN;
	}

        /* adjust btl_mapping structure */

#if 0
        if (ptarget->num_luns < luns) {

                PMGMT_BTL_MAPPING ptarget_btl_mapping;
                ACIT_UEXACT32 data_length, old_data_length;

                PDEBUG(DBG_DISCOVER,"adjusting pbtl structure\n");

                old_data_length = sizeof(MGMT_BTL_MAPPING);

                if (ptarget->num_luns) {
                        old_data_length += sizeof(MGMT_BTL_MAPPING) * 
                                ptarget->num_luns;
		}

                data_length = sizeof(MGMT_BTL_MAPPING) +
                        (sizeof(MGMT_LUN_MAPPING) * (luns - 1));
       
                ptarget_btl_mapping = kmalloc(data_length, GFP_KERNEL);

                if (ptarget_btl_mapping == NULL)
                        return ACIT_STATUS_ERROR;

                memset(ptarget_btl_mapping, 0, data_length);
                memcpy(ptarget_btl_mapping, ptarget->pbtl_mapping,
			(data_length > old_data_length)?
				old_data_length:data_length);

                kfree(ptarget->pbtl_mapping);
                ptarget->pbtl_mapping = ptarget_btl_mapping;
        }
#endif

failed:
        se_return_cmd(phba, (se_req *)posm_request);

        return acit_status;
}

/*++
Function:
        se_build_internal_io_request

Parameter:
        phba:           [in]    pointer to hba
        ptarget:        [in]    pointer to target
        pdevice:        [in]    pointer to lun
        pcdb:           [in]    pointer to CDB
        cdb_len:        [in]    length of CDB
        pdata_buffer:   [inout] pointer to data buffer
        data_len:       [in]    length of data buffer
        dma_address:    [in]    physical address of data buffer
        direction:      [in]    READ/WRITE/NONE
        timeout:        [in]    how long ACIT should hold on to the request

Return:
        pointer to io request upon success.
        NULL on failure

Context:
        Timer, osm_async_event

Description:
        This routine is called to formulate an acit io request for an
        internal command that is needed by the OSM. This routine gets
        a free entry from the freeio_list, constructs a ACIT_IO_REQUEST
        for the command.
--*/

se_io*
se_build_internal_io_request(
                se_hba        *phba,
                se_target     *ptarget,
                se_device     *pdevice,
                ACIT_UEXACT8    *pcdb,
                ACIT_UEXACT32   cdb_len,
                ACIT_UEXACT32   direction,
                ACIT_UEXACT32   timeout)
{
        ACIT_STATUS             acit_status;
        PACIT_IO_REQUEST        pacit_io_request;
        ACIT_SG_ELEMENT         *psg;
        se_req                *prequest;
        se_io                 *pio_request;

        prequest = se_get_free_cmd(phba);

        if (!prequest) {

                PDEBUG(DBG_ERR_HNDL, "Could not get a free command\n");
                return NULL;
        }

        pio_request = &prequest->u.io;
        pio_request->command = NULL;
        init_completion(&pio_request->complete_io);
        pacit_io_request = &pio_request->io_request;
        memset(pacit_io_request, 0, sizeof(ACIT_IO_REQUEST));
        pacit_io_request->pio_key = &pio_request->acit_io_key;


        /* Command - CDB */

        memcpy((char *)pacit_io_request->cdb, (char *)pcdb, cdb_len);
        pacit_io_request->cdb_length = cdb_len;

        /* Data and SG */
        /*TODO: There is buffer pointer in command, is this what ACIT
         * is looking for?
         */
        pacit_io_request->pdata_buffer = ptarget->pinternal_io;
        
        /* construct SG list */
        pacit_io_request->psg_list = (PACIT_SG_LIST)&pio_request->sg_list;
        psg = &pio_request->sg_list.sg[0];

        psg[0].address.u.a64.address = ptarget->internal_io_bus_addr;
        psg[0].length = ptarget->internal_io_len;

        pacit_io_request->data_transfer_length = ptarget->internal_io_len;
        pio_request->sg_list.number_elements = 1;
        PDEBUG(DBG_SCSI,"pdata_buffer = %x, dma_address = %x\n", 
                        (unsigned int)ptarget->pinternal_io, 
                        (unsigned int)ptarget->internal_io_bus_addr);
        /* Sense */

#ifdef __VMKERNEL_MODULE__
        pacit_io_request->psense_buffer_physical.u.a64.address = 
		(&pdevice->sense_data == NULL) ? 0 : 
			virt_to_phys(&pdevice->sense_data);
#else

        pacit_io_request->psense_buffer_physical.u.a64.address = virt_to_phys(
                        &pdevice->sense_data);
#endif //__VMKERNEL_MODULE__
       
        pacit_io_request->psense_buffer_virtual = &pdevice->sense_data;
        pacit_io_request->sense_buffer_length = MAX_SENSE_LEN;

        pacit_io_request->ppost_routine = se_internal_io_completion;

        pacit_io_request->lun.u.lun_complete = pdevice->lun.u.lun_complete;

        if (direction == READ) 
                pacit_io_request->flags.inbound_data = 1;

        pacit_io_request->timeout = timeout;

        /* TODO - set queued option */
        /* For disk enable tagged queueing and for tapes disable it?? */
        pacit_io_request->task_attribute = 0;

        /* Workspace settings */

        acit_status = se_init_workspace_for_io(phba, ptarget, 
                                                pdevice, 
                                                pio_request,
                                                pacit_io_request);

        if (acit_status != ACIT_STATUS_SUCCESS) {

                PDEBUG(DBG_ERR_HNDL, "Could not initialize workspace %d\n",
                                        acit_status);
                se_return_cmd(phba, (se_req *)pio_request);
                return NULL;

        }

        return pio_request;
}

/*++
Function:
        se_internal_io_completion

Parameter:
        pcontroller_key:     [in]  pointer to controller key
        pio_request:         [in]  pointer to acit_io_request

Return:
        ACIT_STATUS_SUCCESS - no errors

Context:
        IO completion

Description:
        This routine is the completion routine for the commands that
        OSM internally issued to the ACIT.

--*/
ACIT_STATUS
se_internal_io_completion(
                PACIT_CONTROLLER_KEY    pcontroller_key,
                PACIT_IO_REQUEST        pio_request)
{
        se_hba                *phba;
        se_target             *ptarget;
        se_device             *pdevice;
        se_io                 *posm_request;
        BEGIN;
        OSM_ASSERT(pcontroller_key);
        OSM_ASSERT(pio_request);

        phba = pcontroller_key->osm_key;
        ptarget = pio_request->ptarget_key->osm_key;
        pdevice = pio_request->plun_key->osm_key;
        posm_request = pio_request->pio_key->osm_key;
        switch(pio_request->cdb[0]) {
                case ACIT_SCSI_CMD_REPORTLUNS:
                case 0x25:
                case ACIT_SCSI_CMD_INQUIRY:
                        complete(&posm_request->complete_io);
                        break;
        }
                        
        return ACIT_STATUS_SUCCESS;
}

/*++
Function:
        se_delete_lun

Parameter:
        phba:        [in]    pointer to hba
        pdev:        [in]    pointer to device

Return:
        None

Context:
        Any

Description:
        This routine does the cleanup for all operations done via
        se_initialize_lun. The device must not be accessed after
        this call.

--*/

inline ACIT_VOID 
se_delete_lun(se_hba *phba, se_device *pdev)
{
        se_target     *ptarget;
        struct se_timer_queue        *job;
        ptarget = pdev->pparent_target;

        job = kmalloc(sizeof(*job), GFP_ATOMIC);

        if (job){
                PDEBUG(DBG_DISCOVER, "will queue the job\n");
                job->fn = se_report_lunchange;
                job->data.lc_data.lun_number = pdev->scsi_lun;
#if defined(__VMKERNEL_MODULE__)
                job->data.lc_data.bus_number = 0;
#else
                job->data.lc_data.bus_number = ptarget->pbtl_mapping->os_bus;
#endif
                job->data.lc_data.target_number = ptarget->pbtl_mapping->os_target;
                job->data.lc_data.operation = REMOVE_LUN;
                se_schedule_job(phba, job);
                PDEBUG(DBG_DISCOVER, "queued the job %lx\n",
                                (unsigned long)job);
        }else{
                PDEBUG(DBG_DISCOVER, "cant queue the job\n");
        }

        se_destroy_lun(phba, pdev);
}        


void 
se_report_lunchange(se_hba *phba,
                      void *plcdata)
{
        char *argv[4], *envp[1];
        char cmd[120], swtch[3];
        struct se_lunchange_info *plcinfo = (struct se_lunchange_info *)plcdata;
        unsigned int lun_number = plcinfo->lun_number;
        unsigned int bus_number = plcinfo->bus_number;
        unsigned int target_number = plcinfo->target_number;
        int change = plcinfo->operation;
        int success = 0, status = 0;
        int retry;
        Scsi_Device *sdp;
        struct Scsi_Host *shp = phba->host;

        BEGIN;

#ifdef __VMKERNEL_MODULE__
	PDEBUG(DBG_DISCOVER, "Reporting change %s:bus=%d:target=%d:lun=%d\n",
		phba->host->adapter->name, bus_number, target_number,
		lun_number);

	vmk_scsi_state_change(phba->host);
#if 0
        if (change == ADD_LUN) {
		vmk_ScsiScanPaths(phba->host->adapter->name, bus_number, 
			target_number, lun_number);
	} else {
		vmk_ScsiScanPaths(phba->host->adapter->name, -1, -1, -1);
	}
#endif
#else
        PDEBUG(DBG_DISCOVER,"Inside se_lunchange_info\n");
        sprintf(swtch,"-c");
        if (change == ADD_LUN) {
                sprintf(cmd,
                   "\"\"/bin/echo scsi add-single-device %d %d %d %d >/proc/scsi/scsi\"\"",
                   phba->host->host_no,bus_number, target_number, lun_number);
        } else {
                sprintf(cmd,
                   "\"\"/bin/echo scsi remove-single-device %d %d %d %d >/proc/scsi/scsi\"\"",
                   phba->host->host_no, bus_number, target_number, lun_number);
        }
        argv[0] = "/bin/sh";
        argv[1] = swtch;
        argv[2] = cmd;
        argv[3] = 0;
        envp[0] = 0;

        status = call_usermodehelper (argv [0], argv, envp);
	PDEBUG(DBG_DISCOVER, "Status with /bin/echo %d\n",status);

	if (status) {

		PDEBUG( DBG_DISCOVER, "Calling /usr/bin/echo instead ... \n");
		if (change == ADD_LUN) {
			sprintf(cmd, "/usr/bin/echo scsi add-single-device %d %d %d %d >/proc/scsi/scsi",
					phba->host->host_no,bus_number, target_number, lun_number);
		} else {
			sprintf(cmd, "/usr/bin/echo scsi remove-single-device %d %d %d %d >/proc/scsi/scsi",
					phba->host->host_no, bus_number, target_number, lun_number);
		}
		status = call_usermodehelper (argv [0], argv, envp);
		PDEBUG(DBG_DISCOVER, "Call user mode status with /usr/bin/echo %d\n",status);

	}

        if (change == ADD_LUN) {
        for(retry = 0; retry <= 100; retry++) {
                for(sdp = shp->host_queue; sdp != NULL; sdp = sdp->next) {
                        if (sdp->id == target_number &&
                           sdp->channel == bus_number &&
                           sdp->lun == lun_number) {
                                success = 1;
                                break;
                        }
                }
                if (success)
                        break;

                interruptible_sleep_on_timeout(&phba->worker_thread.queue, HZ);
        }
        }
#endif //__VMKERNEL_MODULE__

}

int
se_scan_target_for_luns(
se_hba		*phba,
se_target	*ptarget
)
{
        unsigned int bus = ptarget->bus; 
        unsigned int target = ptarget->target;
        ACIT_STATUS  acit_status = 0;
#if defined(__VMKERNEL_MODULE__) && !defined(__VMKLNX__)
        se_lun_address lun0;
#endif
	ACIT_U32_64     flags = 0;


        ACIT_LUN        acit_lun;
	se_device     *pdevice = NULL;

        BEGIN;

#if defined(__VMKERNEL_MODULE__) && !defined(__VMKLNX__)
	memset(&lun0, 0, sizeof(lun0));

	acit_status = se_initialize_lun(ptarget,
			&ptarget->pdevice_lun0,
			0,
			ptarget->pbtl_mapping->os_target,
			0,
			(PACIT_LUN)&lun0
			);

        if (acit_status != ACIT_STATUS_SUCCESS) {

		PDEBUG(DBG_DISCOVER, "Couldn't initialize LUN"); 

                return acit_status;
        }

	/*
	 * Now take it off the sibling list and decrasing the numer of luns count.
	 */
	ptarget->num_luns--;
        list_del_init(&ptarget->pdevice_lun0->sibling_list);

	pdevice = ptarget->pdevice_lun0;

#else // __VMKERNEL_MODULE__
        acit_lun.u.lun_complete = 0;
        acit_status = se_initialize_lun(ptarget, &pdevice, 0, 0, 0, &acit_lun);
        if (acit_status != ACIT_STATUS_SUCCESS) {
                return acit_status;
        }
#endif // __VMKERNEL_MODULE__

        /*
	 * Here we will send reportlun scsi command
	 */
        acit_status = se_get_reportlun_data(phba, pdevice);

        if (acit_status != ACIT_STATUS_SUCCESS) {
#if defined(__VMKERNEL_MODULE__) && !defined(__VMKLNX__)
		se_device_release(pdevice);
#endif
                PDEBUG(DBG_DISCOVER, "get_reportlun_data failed %d\n",
			acit_status);

                return acit_status;
        }

#if defined(__VMKERNEL_MODULE__) && !defined(__VMKLNX__)
	se_device_release(pdevice);
#else
	/* free default lun */
        se_destroy_lun(phba, pdevice);
#endif

	/*
	 * If the se_get_reportlun_data failed then we will have already
	 * freed the buffer.
	 */
	if (ptarget->pinternal_io != NULL) {
		/*
		 * process the report lun data
		 */
		acit_status = se_process_reportlun_data(phba, ptarget,
			ptarget->pinternal_io, NULL);

		if (acit_status != ACIT_STATUS_SUCCESS) {
			PDEBUG(DBG_DISCOVER,
				"se_process_reportlun_data failed %d\n",
				acit_status);
		}

		se_free_internal_cmd_buffer(phba, ptarget);
	}

        return acit_status;
}

#endif
