/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

        be2iscsiosm26.c

    Purpose:

        This file implements the different entry points that are needed
        2.6 Kernel specific.

--*/
#ifndef LINUX_2_4

#include <linux/version.h>
#include <linux/kmod.h> /* for call_usermodehelper */
#include <linux/interrupt.h>
#include <linux/blkdev.h>
#include <linux/pci.h>
//#include "scsi.h"
#include "scsi/scsi_device.h"
#include "acitcfg.h"
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/reboot.h>
#include "acitdefs.h"
#include "acitscsi.h"
#include "uerr.h"
//#include "mgmtdefs.h"
//#include "hostioct.h"
#include "ioctl_top_bmap.h"
#include "ioctl_top_enum.h"
#include "acitapi.h"
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/list.h>
#include <asm/unistd.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
#include <asm/semaphore.h>
#else
#include <linux/semaphore.h>
#endif
#include <linux/genhd.h>
#include <linux/spinlock.h>
//#include "scsi.h"
#if defined(LINUX_2_4)
#include <linux/tqueue.h>
#include "hosts.h"
#else
#include <scsi/scsi_host.h>
#include "scsi/scsi_cmnd.h"
#endif
//#include "ioctl.h"
#include "be2iscsi_common.h"
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
//#include "be2iscsi_typedefs.h"
#endif
#include "version.h"
#include "be2iscsi_osm.h"
//#include "scsi.h"
//#include "ioctl.h"
#include "be2iscsi_debug.h"
#include<linux/pci.h>
#include "acit_iface.h"

#ifdef REGISTER_ISCSI
#ifdef __VMKLNX__
#include <scsi/scsi_transport.h>
#include <scsi/scsi_transport_iscsi.h>
#if defined(ESX_50)
#include <vmklinux_scsi.h>   // For Transport type
#include <asm/uaccess.h>
#else
#include <vmklinux26_scsi.h>   // For Transport type
#endif

extern struct iscsi_transport be2iscsi_iscsi_transport;
#endif
#endif

/* External variable from OSM file */
extern struct semaphore se_configuration_lock;
extern struct pci_driver be2iscsi_pci_driver;
extern ACIT_API_TABLE acitapi_table;
extern ACIT_VERSION acit_api_version;
extern struct notifier_block se_shutdown_notifier;
extern unsigned int  gcrashmode;
extern int eto;
extern int enable_msix;
extern struct be_default_msix_entry def_msix_entries[BE_NUM_MSIX_ENTRIES];

#ifdef CONFIG_SYSFS
struct class_simple *be2iscsi_class;
#endif

#ifdef POWER_MANAGEMENT
/*++
Function:
        se_pci_dev_suspend

Parameter:
        pci:    [in]    pointer to struct pci
        state:  [in]    Power State

Return:
        0 - for Success
        -1 - for Failure

Context:
        Entry point

Description:
        This routine is called as a result of change in Power state.
        This is the suspend function reported in struct pci_driver.         

--*/

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16) 
int
se_pci_dev_suspend(struct pci_dev *pcidev, pm_message_t state)
#else   /*  ChangeLog-2.6.11:22267:  [PATCH] pm: introduce pm_message_t  */
int
se_pci_dev_suspend(struct pci_dev *pcidev,  u32 state)
#endif
{
        ACIT_U32_64     flags = 0;
        se_hba *phba = (se_hba *)pci_get_drvdata(pcidev);
        if(!phba){
                PDEBUG(DBG_ERR_HNDL, "!!! NO pHBA !!!\n");
                return -EBUSY;
        }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16) 
        if(state.event < 1)
        {
               PDEBUG(DBG_ERR_HNDL, "Changing to power state: %d\n",state.event);
               return -EBUSY;
        }
#else
	
        if(state< 1)
        {
               PDEBUG(DBG_ERR_HNDL, "Changing to power state: %d\n",state);
               return -EBUSY;
        }
#endif 
        phba->timer_shutdown = 1;
        del_timer_sync(&phba->hba_timer);

        /* Disabling the IRQ ... */
    	PDEBUG(DBG_DISCOVER, "Disabling the IRQ for the HBA 0x%lx \n",(unsigned long) phba);
        acitapi_disable_irq(phba, &phba->controller_key);
	spin_lock_irqsave(&phba->acitapi_lock,flags);
	phba->state |= SE_STATE_SUSPEND;
	phba->state &= ~SE_STATE_INT_ENABLED;
        spin_unlock_irqrestore(&phba->acitapi_lock,flags);

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,5) 
	if (enable_msix) {
		int n;
		struct be_msix_entry  	*be_entry;
		for(n = 0; n < BE_NUM_MSIX_ENTRIES; n++) {
			be_entry = &phba->msix_entries[def_msix_entries[n].index];
			if (be_entry->msix_enabled)
				free_irq(be_entry->msix_vector, phba);
			}
			pci_disable_msix(phba->pcidev);
	} else 
#endif
	{
		if (phba->host->irq)
			free_irq(phba->host->irq, phba);
	}

        /* Stop the Controller in D3 state */
        PDEBUG(DBG_DISCOVER, "Calling Stop controller for HBA 0x%lx\n",(unsigned long) phba);
        acitapi_stop_controller(phba, &phba->controller_key, ACIT_STOP_TYPE_POWER_MGMT);
        phba->worker_thread.terminate = 1;
        complete(&phba->worker_thread.work_queued);
        PDEBUG(DBG_UNLOAD,"Running deferred jobs\n");
        
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16) 
        pci_save_state(pcidev);
#else
	pci_save_state(pcidev,&phba->state);		
#endif
        pci_disable_device(pcidev);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16) 
        PDEBUG(DBG_DISCOVER, "*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*\n");
        PDEBUG(DBG_DISCOVER, "Power transition state is %d\n",pci_choose_state(pcidev,state));
        PDEBUG(DBG_DISCOVER, "*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*\n");
        pci_set_power_state(pcidev, pci_choose_state(pcidev, state));

#else
	state = (state > 0) ? 3 : 0;
	pci_set_power_state(pcidev,state);
#endif
        return 0;
}



/*++
Function:
        se_pci_dev_resume

Parameter:
        pci:    [in]    pointer to struct pci

Return:
       0 - for Success
       -1 - for Failure

Context:
        Entry point

Description:
        This routine is called as a result of resume in Power state.
        This is the suspend function reported in struct pci_driver.         

--*/

int
se_pci_dev_resume(struct pci_dev *pcidev)
{

        ACIT_U32_64     flags = 0;
        ACIT_STATUS             acit_status = 0;
        ACIT_DISCOVER_PARAM        acit_discover_param;
	pid_t pid;
	unsigned		msix_initialized;

        se_hba *phba = (se_hba *)pci_get_drvdata(pcidev);
        if(!phba) {
                PDEBUG(DBG_ERR_HNDL, "!!! NO pHBA !!!\n");
                return -1;
        }

        PDEBUG(DBG_DISCOVER, "<><><><><><><><><><><><><><><><><><><><><><><><><><><><><><\n");
        PDEBUG(DBG_DISCOVER, "<><><><><><><><><><>RESUME CALLED><><><><><><><><><><><><><\n");
        PDEBUG(DBG_DISCOVER, "<><><><><><><><><><><><><><><><><><><><><><><><><><><><><><\n");

        // PCI_D0 isn't available on SLES9, so replacing with actual value to fix build break
        //pci_set_power_state(pcidev, PCI_D0);
        pci_set_power_state(pcidev, 0);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
        pci_restore_state(pcidev);
#else
        pci_restore_state(pcidev, &phba->state);
#endif
        acit_status = pci_enable_device(pcidev);
       
        PDEBUG(DBG_DISCOVER, "Calling Init Controller for HBA 0x%lx\n",(unsigned long) phba);
        /* Initialize the controller ... */
        acit_status = acitapi_init_controller(phba, &phba->controller_key,
                                          ACIT_INIT_TYPE_POWER_MGMT);

        if(acit_status != ACIT_STATUS_SUCCESS) {
                PDEBUG(DBG_ERR_HNDL,"acit_init_controller failed\n");
                goto init_failed;                
        }
        /*REQUEST IRQ .......*/
	msix_initialized = 0;
	if (enable_msix) {
		/*
		 * Try MSI-X first ...
		 */
		if (se_setup_msix(phba, pcidev, NULL) != 0) {
			/*
			 * It worked.
			 */
			msix_initialized = 1;
		} else {
			enable_msix = 0;
		}
	}

	if (msix_initialized == 0) {
		/*
		 * If we couldn't or didn't want to use MSI-X, then
		 * try regular interrupts.
		 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
        	if(request_irq(pcidev->irq, se_isr, SA_INTERRUPT | SA_SHIRQ,
#else 
		if (request_irq(pcidev->irq, se_isr,IRQF_DISABLED | IRQF_SHARED, 
#endif
                	"be2iscsi", phba) != 0) {
                	PDEBUG(DBG_INIT, "Cannot register irq\n");
                	goto rirq_failed;                
		}
                
        }

        /* Enable the IRQ ... */
        PDEBUG(DBG_DISCOVER, "Enabe IRQ for HBA 0x%lx\n",(unsigned long) phba);
        if(acitapi_enable_irq(phba, &phba->controller_key) != ACIT_TRUE) {
                PDEBUG(DBG_ERR_HNDL, "Cannot Enable irq\n");
                goto enable_failed;
        }

        PDEBUG(DBG_DISCOVER, "Enable IRQ successful\n");
	spin_lock_irqsave(&phba->acitapi_lock,flags);
	phba->state &= ~SE_STATE_SUSPEND;
        phba->state |= SE_STATE_INT_ENABLED;
        spin_unlock_irqrestore(&phba->acitapi_lock,flags);
        acit_discover_param.info = 0; /* undefined */
        acit_discover_param.pptarget_table = &phba->pacit_target_table;
        acit_discover_param.result = 0; /* undefined */
        acit_discover_param.reserved = 0;
        acit_discover_param.opcode = ACIT_DISCOVER_OPCODE_ASYNCHRONOUS;

        acit_status = acitapi_discover_targets(phba,&phba->controller_key,
                                                &acit_discover_param);

        if(acit_status != ACIT_STATUS_SUCCESS) {
                PDEBUG(DBG_ERR_HNDL, "Discover Targets failed!!!\n");
                goto discover_failed;
        }
        phba->worker_thread.terminate = 0;
        init_completion(&phba->worker_thread.thread_completion);
        init_waitqueue_head(&phba->worker_thread.queue);

	se_process_deferred_jobs(phba);

        init_timer(&phba->hba_timer);
        phba->timer_shutdown = 0;
        phba->hba_timer.function = se_timer;
        phba->hba_timer.data = (unsigned long)phba;
        phba->hba_timer.expires = jiffies + TIMER_INTERVAL;
        add_timer(&phba->hba_timer);
                                                                                

        pid = kernel_thread(se_worker_thread, phba, 0);
        if(pid < 0) {
                PDEBUG(DBG_ERR_HNDL,"kernel_thread failed with status %d\n",
                                acit_status);
                goto stop_thread;
        }

        return 0;

stop_thread:
        phba->worker_thread.terminate = 1;
        complete(&phba->worker_thread.work_queued);
        wait_for_completion(&phba->worker_thread.thread_completion);
init_failed:
rirq_failed:                
enable_failed:
discover_failed:

	LnxLogOSMError(UNI_CFG_LOG_LEVEL_CRITICAL,
		UNI_PHASE_POWER_MGMT,
		UNI_SUBMODULE_LNXOSM_GENERIC,
		OSM_EC_HW_INIT_FAILED);

        up(&se_configuration_lock);
        return -1;
}
#endif

#if 0
int se_call_scan_single_target(se_hba *phba,
                se_target     *ptarget)
{
        unsigned int bus = ptarget->bus; 
        unsigned int target = ptarget->target;
        ACIT_STATUS  acit_status = 0;

        char *argv[4], *envp[1];
        char cmd[120], swtch[3];

        BEGIN;

	if(phba->state & SE_STATE_HBA_INITIALIZED){
#ifdef scsi_scan_single_target 
        acit_status = scsi_scan_single_target(phba->host, bus, target);
#else

        PDEBUG(DBG_DISCOVER,"Calling usermodehelper to scan host\n");
        sprintf(swtch,"-c");
        sprintf(cmd, 
                "\"\"/bin/echo %d %d - >/sys/class/scsi_host/host%d/scan\"\"",
                bus, target, phba->host->host_no);
        argv[0] = "/bin/sh";
        argv[1] = swtch;
        argv[2] = cmd;
        argv[3] = 0;
        envp[0] = 0;
	PDEBUG(DBG_DISCOVER,"Calling usermode helper\n");

        acit_status = call_usermodehelper (argv [0], argv, envp, 1);
	PDEBUG(DBG_DISCOVER,"Done %d\n", acit_status);
        if(acit_status) {
		PDEBUG(DBG_DISCOVER,"/bin/echo failed using /usr/bin/echo\n");
		sprintf(cmd, 
		"/usr/bin/echo %d %d - >/sys/class/scsi_host/host%d/scan",
				bus, target, phba->host->host_no);

		acit_status = call_usermodehelper (argv [0], argv, envp, 1);
		PDEBUG(DBG_DISCOVER,"Done %d\n", acit_status);
	}
        if(acit_status) {
		PDEBUG(DBG_DISCOVER,"/usr/bin/echo failed using /lbin/echo\n");
		sprintf(cmd, 
		"/lbin/echo %d %d - >/sys/class/scsi_host/host%d/scan",
				bus, target, phba->host->host_no);

		acit_status = call_usermodehelper (argv [0], argv, envp, 1);
		PDEBUG(DBG_DISCOVER,"Done %d\n", acit_status);
	}
	if(acit_status) {
		scsi_scan_host(phba->host);
		acit_status =  ACIT_STATUS_SUCCESS;
	}
#endif   
    }
     return acit_status;
}
#endif 

int se_slave_alloc(struct scsi_device *sdev)
{
        unsigned int	id, lun, channel;
        se_target	*ptarget = NULL, *ptemp = NULL;
        ACIT_LUN	acit_lun;
        se_device	*pdevice = NULL;
        se_hba		*phba;
	struct		request_queue *q;
        ACIT_STATUS	status;
#ifdef REGISTER_ISCSI
	struct iscsi_cls_session *sess;
#endif

        phba = (se_hba *) sdev->host->hostdata[0];

	q = (struct request_queue *) sdev->request_queue;

        id      = sdev->id;
        lun     = sdev->lun;
        channel = sdev->channel;

        memset(&acit_lun,0,sizeof(ACIT_LUN));

        PDEBUG(DBG_DISCOVER,"in slave_alloc for %d %d %d\n", channel, id, lun);

        list_for_each_entry(ptemp, &phba->target_list, sibling_list){ 
                if(ptemp->bus == channel && ptemp->target == id){
                        ptarget = ptemp;
                        break;
                }
        }

        if (ptarget == NULL) {
		PDEBUG(DBG_DISCOVER, "Couldn't find ptarget\n");

		return -ENXIO;
        }

        acit_lun.u.lun_leveled[0] = 
                ((lun & 0x0000ff00)) >> 8 | ((lun & 0x000000ff) << 8);

        acit_lun.u.lun_leveled[1] = 
                ((lun & 0xff000000)) >> 24 | ((lun & 0x00ff0000) >> 8);

        status = se_initialize_lun(ptarget, &pdevice, channel, id, lun,
		&acit_lun);

       	if (status == ACIT_STATUS_SUCCESS) { 
		sdev->hostdata     = (void *)pdevice;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27))
		blk_queue_rq_timeout(sdev->request_queue, (ptarget->extended_timeout + 30) * HZ);
#else
		sdev->timeout = (ptarget->extended_timeout + 30) * HZ;
#endif

		pdevice->pScsi_dev = sdev;

#ifdef REGISTER_ISCSI
#ifndef __VMKLNX__
		sess = starget_to_session(sdev->sdev_target);
#else
		sess = iscsi_sdevice_to_session(sdev);
#endif
		if (sess != NULL) {
			sdev->sdev_target->hostdata = (void *)ptarget;
		} else {
			PDEBUG(DBG_DISCOVER, "sess is NULL\n");
		}
#endif

		if (gcrashmode) {
			scsi_adjust_queue_depth(sdev,
                                (sdev->tagged_supported ?  0x20 : 0), 1);
#if defined(NEW_BLK_API) | (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,34))
	     	        blk_queue_max_segments(q, 1);
#else
	     	        blk_queue_max_hw_segments(q, 1);
                        blk_queue_max_phys_segments(q, 1);
#endif
		}
	}

        return status;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
int se_slave_configure(struct scsi_device *sdev)
{
        unsigned int	id, lun, channel;
        se_target	*ptarget = NULL, *ptemp = NULL;
        se_device	*pdevice = NULL;
        se_hba		*phba;
        ACIT_STATUS	status;
	ACIT_LUN_INFO acit_lun_info;

        phba = (se_hba *) sdev->host->hostdata[0];

	pdevice = (se_device *)sdev->hostdata;

	if (pdevice == NULL) {
		PDEBUG(DBG_DISCOVER, "No device to configure\n");

		return -ENXIO;
	}

        memset(&acit_lun_info,0,sizeof(ACIT_LUN_INFO));

        id      = sdev->id;
        lun     = sdev->lun;
        channel = sdev->channel;

        PDEBUG(DBG_DISCOVER,"in slave_configure for %d %d %d\n", channel, id, lun);

        list_for_each_entry(ptemp, &phba->target_list, sibling_list){ 
                if(ptemp->bus == channel && ptemp->target == id){
                        ptarget = ptemp;
                        break;
                }
        }

        if (ptarget == NULL) {
		PDEBUG(DBG_DISCOVER, "Couldn't find ptarget\n");

		return -ENXIO;
        }

	acit_lun_info.device_type = sdev->type;
	status = acitapi_update_lun_info( phba,
					   &phba->controller_key,
					   ptarget->posm_target_key,
					   &pdevice->lun_key,
					   &acit_lun_info);

        return status;
}
#endif

void se_slave_destroy(struct scsi_device *sdev)
{
	se_device *pdevice;
	se_hba    *phba;

	pdevice = (se_device *)sdev->hostdata;

	if (pdevice == NULL) {
		PDEBUG(DBG_DISCOVER, "No device to destroy\n");

		return;
	}

	PDEBUG(DBG_DISCOVER, "in slave_destroy (%d,%d,%d)\n",
		pdevice->scsi_channel, pdevice->scsi_id, pdevice->scsi_lun);

	phba = (se_hba *) sdev->host->hostdata[0];

	se_destroy_lun(phba, pdevice);

	sdev->hostdata = NULL;
}

#ifdef __VMKERNEL__
int se_target_alloc(
struct scsi_target	*starget
)
{
	return 0;
}

void se_target_destroy(
struct scsi_target	*starget
)
{
#ifdef REGISTER_ISCSI
	se_target     *ptarget = NULL;
	se_hba         *phba = NULL;

        BEGIN;

	PDEBUG(DBG_DISCOVER, "Destroying target\n");

	ptarget = (se_target *)starget->hostdata;
	if(ptarget != NULL){
                phba = ptarget->pparent_hba;
                if(phba == NULL){
                        PDEBUG(DBG_DISCOVER,
				"IN se_target_destroy PHBA NULLL \n");
                        return;
                }
        }else{
		PDEBUG(DBG_DISCOVER, "ptarget NULL\n");
		return;
	}

	se_return_to_bt_slot(phba, ptarget->bus, ptarget->target);
	list_del_init(&ptarget->sibling_list);
	kfree(ptarget->pbtl_mapping);


	if ((ptarget != NULL) && (ptarget->sess != NULL)) {
		iscsi_free_session(ptarget->sess);
		ptarget->sess = NULL;
		kmem_cache_free(phba->posm_target_cache, ptarget);
	}

	END;
#endif
}
#endif

#ifdef REGISTER_ISCSI
#ifdef __VMKLNX__
extern struct scsi_transport_template *be2iscsi_scsi_transport;
#endif
#endif


/* This function will be called at module initialization time
   This will do all the one_time_things for this module and then
   register with the pci subsystem*/
static int __init be2iscsi_init(void)
{

        /*
         * Register our control device node
         * nodes will need to be created in /dev to access this
         * the nodes can not be created from within the driver
         */ 
#if 0
#ifdef CONFIG_SYSFS
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13) 
        be2iscsi_class = class_simple_create(THIS_MODULE, "be2iscsi");
#else   /* ChangeLog-2.6.13:52566: [PATCH] class: remove class_simple code, as no one in the tree is using it anymore.*/
        be2iscsi_class = class_create(THIS_MODULE, "be2iscsi");
#endif        
        if(IS_ERR(be2iscsi_class)){
                PDEBUG(DBG_ERR_HNDL, "Cannot create sysfs class\n");
                goto no_class;
        }
#endif
#endif
        /*
         * Register our shutdown/reboot notifier to the kernel.
         */          
        register_reboot_notifier(&se_shutdown_notifier);

#ifdef REGISTER_ISCSI
#ifdef __VMKLNX__
	be2iscsi_scsi_transport = iscsi_register_transport(
		&be2iscsi_iscsi_transport);

	if (be2iscsi_scsi_transport == NULL) {
		return 1;
	}
#endif
#endif

        return pci_register_driver(&be2iscsi_pci_driver);
#if 0
#ifdef CONFIG_SYSFS
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,13) 
        class_simple_destroy(be2iscsi_class);
#else   /* ChangeLog-2.6.13:52566: [PATCH] class: remove class_simple code, as no one in the tree is using it anymore.*/
        class_destroy(be2iscsi_class);
#endif        
#endif
no_class:
#endif

        return 1;

}

static void __exit be2iscsi_exit(void)
{
        pci_unregister_driver(&be2iscsi_pci_driver);
        unregister_reboot_notifier(&se_shutdown_notifier);
        unregister_chrdev(BE2ISCSI_MAJOR, BE2ISCSI_DRIVER);
#if 0
#ifdef CONFIG_SYSFS
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,13) 
        class_simple_destroy(be2iscsi_class);
#else   /* ChangeLog-2.6.13:52566: [PATCH] class: remove class_simple code, as no one in the tree is using it anymore.*/
        class_destroy(be2iscsi_class);
#endif        
#endif 
#endif       
#if 0
#ifdef BE2ISCSI_DEBUG
        if(alloc_count != free_count)
		printk("alloc_count = %d, free_count = %d\n", alloc_count, free_count);
#endif
#endif
}
module_init(be2iscsi_init);
module_exit(be2iscsi_exit);
#endif

