/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   beclib_private.h

@brief
   
CVS Header Information:
   $Header: /proj/be/archive/dsn/sw/host/drivers/beclib/beclib_private_ll.h,v 1.39 2006/04/20 23:49:03 menz Exp $
*/
#ifndef __beclib_private_h__
#define __beclib_private_h__

#define BE_BAD_ID ((u32)(0xFFFFFFFF))

#define  UT_PAGE_SIZE         SA_PAGE_SIZE      

#if defined(BE_EMULATION) || defined(BE_SELECTIVE_EMULATION)
    typedef union _CRA_UNION {
     
        u32                 dw;

        CRA            uc;  // unified cache view of address
        UC_CRA         uc_tag;
        PTC_CRA        ptc_tag;

        ISCSI_FRAG_CRA frag;
        ISCSI_SGL_CRA  sgl;
        RING_CRA       ring;  
        JE_CRA         je;
        TE_CRA         te;

    } CRA_UNION, *PCRA_UNION;

    
    STATIC INLINE 
    CRA_UNION be_cra_ring_create (u32 type, u32 qid) { 
        CRA_UNION cra;
        cra.dw = 0; // clear all bits.
        cra.ring.constant = UC_CRA_RINGS;
        cra.ring.queueid  = qid;
        cra.ring.type     = type; 
        return cra;
    }

    STATIC INLINE 
    CRA_UNION be_cra_frag_create (u32 page_offset) { 
        CRA_UNION cra;
        cra.dw = 0; // clear all bits.

        cra.sgl.constant  = UC_CRA_ISCSI;
        cra.sgl.sgl_index = 0;
        cra.sgl.frag_num  = 0;
        cra.sgl.offset    = 0;

        cra.dw += SA_PAGE_SIZE * page_offset;

        ASSERT (cra.sgl.constant == UC_CRA_ISCSI);

        return cra;
    }

    STATIC INLINE 
    CRA_UNION be_cra_jell_create (u32 page_offset) { 
        CRA_UNION cra;
        cra.dw = 0; // clear all bits.

        cra.je.constant = CRA_QUAD_JELL;
        cra.je.je = 0;

        cra.dw += SA_PAGE_SIZE * page_offset;

        ASSERT (cra.je.constant == CRA_QUAD_JELL);

        return cra;
    }

    /*!
    @brief
        Returns the a UT Virtual Address aligned on a page boundary
    @param    
        Va - UT Virtual Address
    @return   
        A UT Virtual Address aligned on a page boundary
    */
    STATIC INLINE
    u64
    be_ut_page_align ( 
        IN      u64 cra)
    {
        cra = (cra & ~(UT_PAGE_SIZE - 1));
        return cra;
    }


#endif

//
// Turn byte packing on cuz we'll have fairly large arrays of this structure
//                         
#include "system/sa_packon.h"
typedef struct _MEM_CLUSTER_ENTRY {

    PSA_SGL Sgl;
    u16 RefCount;

} SA_PACK MEM_CLUSTER_ENTRY, *PMEM_CLUSTER_ENTRY;
#include "system/sa_packoff.h"


//
// This guy manages the array of cluster entries
//
typedef struct _MEM_CLUSTER {

    u32                MaxMemory;
    u32                ItemSize;
    u32                ClusterSize;
    u32                PoolTag;

    u32                ItemCount;
    u32                ItemsPerCluster;
    u32                ClusterCount;


    PMEM_CLUSTER_ENTRY Clusters;

    PSA_SGL            LastMdlMapped;
    PVOID              MappedVa;
    
} MEM_CLUSTER, *PMEM_CLUSTER;

//
// A Cluster manager manages allocations of chunks of memory e.g. pages allocated
// at a time
//
typedef struct _MEM_CLUSTER_MGR {

    SA_BIT_VECTOR AllocationMap;
    u32           ClusterSize;
    MEM_CLUSTER   Clusters;
    u32*          BitmapBuffer;
} MEM_CLUSTER_MGR, *PMEM_CLUSTER_MGR;


// ------------------------------------------------------
// Chip Object
// ------------------------------------------------------

void be_chip_lock (IN PBE_CHIP_OBJECT chip_object);
void be_chip_unlock (IN PBE_CHIP_OBJECT chip_object);

// ------------------------------------------------------
//  Function Object
// ------------------------------------------------------

// Init function object.
BESTATUS
be_function_create (
    SA_DEV*             sa_dev,           // previously created device
    u32                 function_type,    // e.g FUNCTION_TYPE_ISCSI
    PSA_SGL             mailbox_sgl,
    PBE_FUNCTION_OBJECT function_object); // object to initialize

// Creates one chip object.
BESTATUS
be_chip_create (
    OUT PBE_CHIP_OBJECT  chip);

// Inserts the function object into the chip object.  Both must be previously
// created.  This function will initialize the chip during emulation.
BESTATUS
be_chip_insert_function_object (
    IN      PBE_CHIP_OBJECT         chip_object, 
    IN      PBE_FUNCTION_OBJECT     function_object);

// Cleans up the chip object.
// This function is called as part of be_function_cleanup.
void
be_chip_destroy (
    PBE_CHIP_OBJECT chip);

// Destroys the function object.  This assumes that all resources for
// this function have already been destroy, and it asserts if that
// is not true.
// This function is called as part of be_function_cleanup.
BESTATUS
be_function_destroy (
    PBE_FUNCTION_OBJECT function_object);


// Removes the function object from the chip.
// This function is called as part of be_function_cleanup.
void
be_chip_remove_function_object (
    IN      PBE_CHIP_OBJECT         chip_object, 
    IN      PBE_FUNCTION_OBJECT     function_object);

void _be_function_lock   (PBE_FUNCTION_OBJECT fo);
void _be_function_unlock (PBE_FUNCTION_OBJECT fo);

void _be_function_add_eq    (PBE_FUNCTION_OBJECT fo, PBE_EQ_OBJECT eq_object);
void _be_function_remove_eq (PBE_FUNCTION_OBJECT fo, PBE_EQ_OBJECT eq_object);

void _be_function_add_cq    (PBE_FUNCTION_OBJECT fo, PBE_CQ_OBJECT cq_object);
void _be_function_remove_cq (PBE_FUNCTION_OBJECT fo, PBE_CQ_OBJECT cq_object);

void _be_function_add_eth_sq    (PBE_FUNCTION_OBJECT fo, PBE_ETHSQ_OBJECT eth_sq);
void _be_function_remove_eth_sq (PBE_FUNCTION_OBJECT fo, PBE_ETHSQ_OBJECT eth_sq);

void _be_function_add_eth_rq    (PBE_FUNCTION_OBJECT fo, PBE_ETHRQ_OBJECT eth_rq);
void _be_function_remove_eth_rq (PBE_FUNCTION_OBJECT fo, PBE_ETHRQ_OBJECT eth_rq);

void _be_function_add_lro_rq    (PBE_FUNCTION_OBJECT fo, PBE_LRORQ_OBJECT lro_rq);
void _be_function_remove_lro_rq (PBE_FUNCTION_OBJECT fo, PBE_LRORQ_OBJECT lro_rq);

void _be_function_add_mcc       (PBE_FUNCTION_OBJECT fo, PBE_MCC_OBJECT mcc);
void _be_function_remove_mcc    (PBE_FUNCTION_OBJECT fo, PBE_MCC_OBJECT mcc);

void _be_function_add_iscsi_connection    (PBE_FUNCTION_OBJECT fo, PBE_ISCSI_CONNECTION_OBJECT connection_object);
void _be_function_remove_iscsi_connection (PBE_FUNCTION_OBJECT fo, PBE_ISCSI_CONNECTION_OBJECT connection_object);
PBE_ISCSI_CONNECTION_OBJECT _be_function_find_iscsi_connection (PBE_FUNCTION_OBJECT fo, u32 cid);

void _be_function_add_fcoe_interface      (PBE_FUNCTION_OBJECT fo, PBE_FCOE_INTERFACE_OBJECT interface_object);
void _be_function_remove_fcoe_interface   (PBE_FUNCTION_OBJECT fo, PBE_FCOE_INTERFACE_OBJECT interface_object);

void _be_function_add_toe_sq    (PBE_FUNCTION_OBJECT funcobj, PBE_TOE_SQ_OBJECT sq);
void _be_function_remove_toe_sq (PBE_FUNCTION_OBJECT funcobj, PBE_TOE_SQ_OBJECT sq);

void _be_function_add_toe_rq    (PBE_FUNCTION_OBJECT funcobj, PBE_TOE_RQ_OBJECT rq);
void _be_function_remove_toe_rq (PBE_FUNCTION_OBJECT funcobj, PBE_TOE_RQ_OBJECT rq);

void _be_function_add_toe_dq    (PBE_FUNCTION_OBJECT fo, PBE_TOE_DQ_OBJECT toe_dq);
void _be_function_remove_toe_dq (PBE_FUNCTION_OBJECT fo, PBE_TOE_DQ_OBJECT toe_dq);

void _be_function_add_toe_connection    (PBE_FUNCTION_OBJECT fo, PBE_TOE_CONNECTION_OBJECT connection_object);
void _be_function_remove_toe_connection (PBE_FUNCTION_OBJECT fo, PBE_TOE_CONNECTION_OBJECT connection_object);

void _be_function_add_wrbq (PBE_FUNCTION_OBJECT funcobj, PBE_ISCSI_WRB_QUEUE_OBJECT wrbq);
void _be_function_remove_wrbq (PBE_FUNCTION_OBJECT funcobj, PBE_ISCSI_WRB_QUEUE_OBJECT wrbq);

void _be_function_add_fcoe_wrbq (PBE_FUNCTION_OBJECT funcobj, PBE_FCOE_WRB_QUEUE_OBJECT wrbq);
void _be_function_remove_fcoe_wrbq (PBE_FUNCTION_OBJECT funcobj, PBE_FCOE_WRB_QUEUE_OBJECT wrbq);

void _be_function_add_default_pdu_queue (PBE_FUNCTION_OBJECT func_obj, PBE_DEFAULT_PDU_QUEUE_OBJECT defq);
void _be_function_remove_default_pdu_queue (PBE_FUNCTION_OBJECT func_obj, PBE_DEFAULT_PDU_QUEUE_OBJECT defq);

void _be_function_add_rdma_qp (PBE_FUNCTION_OBJECT function_object, PBE_RDMA_QP_OBJECT qp_object);
void _be_function_remove_rdma_qp (PBE_FUNCTION_OBJECT function_object, PBE_RDMA_QP_OBJECT qp_object);

void _be_function_add_rxif (PBE_FUNCTION_OBJECT function_object, PBE_RXIF_OBJECT rxif_object);
void _be_function_remove_rxif (PBE_FUNCTION_OBJECT function_object, PBE_RXIF_OBJECT rxif_object);
u32 _be_function_default_rxif_id (PBE_FUNCTION_OBJECT function_object);

u32 _be_function_reference (PBE_FUNCTION_OBJECT fo);
u32 _be_function_dereference (PBE_FUNCTION_OBJECT fo);

BESTATUS
be_function_ring_destroy (
    PBE_FUNCTION_OBJECT function_object,
    u32                 id,
    u32                 ring_type);

BESTATUS
be_function_ring_destroy_async (
    PBE_FUNCTION_OBJECT  function_object,
    u32                  id,
    u32                  ring_type,
    MCC_WRB_CQE_CALLBACK callback,
    PVOID                callback_context,
    MCC_WRB_CQE_CALLBACK internal_callback,
    PVOID                internal_callback_context);

//BE_LINK_STATUS_CONTEXT*
//be_function_allocate_link_context (
//    PBE_FUNCTION_OBJECT function_object);
//
//void
//be_function_free_link_context (
//    PBE_FUNCTION_OBJECT function_object,
//    BE_LINK_STATUS_CONTEXT* context);


#define BE_CREATE_MCC_RESPONSE_COPY(_ioctl_type_, _field_, _va_)           \
    be_create_mcc_response_copy (SA_FIELD_OFFSET (_ioctl_type_, _field_),  \
                                 SA_SIZEOF_FIELD (_ioctl_type_, _field_),  \
                                 (_va_))                                   \

STATIC INLINE 
BE_MCC_WRB_RESPONSE_COPY
be_create_mcc_response_copy (
    u16   offset,
    u16   length,
    PVOID va)
{
    BE_MCC_WRB_RESPONSE_COPY rc;
    rc.length = length;
    rc.ioctl_offset = offset;
    rc.va = va;
    return rc;
}

BESTATUS
be_function_post_mcc_wrb_with_queue_context (
    IN PBE_FUNCTION_OBJECT       function_object,
    IN PMCC_WRB                  wrb,
    IN PBE_GENERIC_QUEUE_CONTEXT queue_context,
    IN MCC_WRB_CQE_CALLBACK      callback,
    IN PVOID                     callback_context,
    IN PVOID                     optional_ioctl_va);

BESTATUS
be_function_post_mcc_wrb_with_copy (
    IN PBE_FUNCTION_OBJECT       function_object,
    IN PMCC_WRB                  wrb,
    IN PBE_GENERIC_QUEUE_CONTEXT queue_context,
    IN MCC_WRB_CQE_CALLBACK      callback,
    IN PVOID                     callback_context,
    IN PVOID                     optional_ioctl_va,
    IN BE_MCC_WRB_RESPONSE_COPY  response_copy);

BESTATUS
be_function_post_mcc_wrb_with_internal_callback (
    IN PBE_FUNCTION_OBJECT       function_object,
    IN PMCC_WRB                  wrb,
    IN PBE_GENERIC_QUEUE_CONTEXT queue_context,
    IN MCC_WRB_CQE_CALLBACK      callback,    
    IN PVOID                     callback_context,
    IN MCC_WRB_CQE_CALLBACK      internal_callback,
    IN PVOID                     internal_callback_context,
    IN PVOID                     optional_ioctl_va
    );

BESTATUS
be_function_post_mcc_wrb_complete (
    IN PBE_FUNCTION_OBJECT       function_object,
    IN PMCC_WRB                  wrb,
    IN PBE_GENERIC_QUEUE_CONTEXT queue_context,
    IN MCC_WRB_CQE_CALLBACK      callback,
    IN PVOID                     callback_context,
    IN MCC_WRB_CQE_CALLBACK      internal_callback,
    IN PVOID                     internal_callback_context,
    IN PVOID                     optional_ioctl_va,
    IN BE_MCC_WRB_RESPONSE_COPY  response_copy);

// Returns the protection domain number. 0 is host.
u32 
be_function_get_pd_number (
    PBE_FUNCTION_OBJECT function_object);

boolean
be_function_is_vm (
    PBE_FUNCTION_OBJECT function_object);

BESTATUS
be_function_post_page_table_ioctl (
    PBE_FUNCTION_OBJECT function_object,
    PSA_SGL sgl,   // optional, required if num_pages > 0
    u32 num_pages, // num_pages == 0 used internally to query number of free page tables
    u32 *num_free);

BESTATUS
be_function_query_num_free_page_tables (
    PBE_FUNCTION_OBJECT function_object,
    u32 *num_free);

BESTATUS
be_function_queue_mcc_wrb (
    PBE_FUNCTION_OBJECT       function_object,
    PBE_GENERIC_QUEUE_CONTEXT queue_context);

BESTATUS
be_function_internal_query_fw_cfg (
   IN PBE_FUNCTION_OBJECT  function_object,
   OUT BE_FW_CFG* config
);

// ------------------------------------------------------
//  Event Queue
// ------------------------------------------------------

void _be_eq_add_cq    (PBE_EQ_OBJECT eq_object, PBE_CQ_OBJECT cq_object);
void _be_eq_remove_cq (PBE_EQ_OBJECT eq_object, PBE_CQ_OBJECT cq_object);

u32 be_eq_reference (PBE_EQ_OBJECT cq_object);
u32 be_eq_dereference (PBE_EQ_OBJECT cq_object);

// Deprecate....
EQ_CALLBACK 
be_eq_set_callback (
    IN PBE_EQ_OBJECT eq_object,
    IN EQ_CALLBACK   callback,
    IN PVOID         context);

// Deprecate...
void
be_eq_delegate_processing (
    IN PBE_FUNCTION_OBJECT function_object,
    IN u32                 eq_id);

// ------------------------------------------------------
//  Completion Queue Object
// ------------------------------------------------------

u32
be_cq_object_reference (
    PBE_CQ_OBJECT cq_object);

u32
be_cq_object_dereference (
    PBE_CQ_OBJECT cq_object);

// Deprecate....
CQ_CALLBACK 
be_cq_object_set_callback (
    IN PBE_CQ_OBJECT   cq_object, 
    IN CQ_CALLBACK     callback, 
    IN PVOID           context);

// Deprecate....
void
be_cq_object_delegate_processing (
    PBE_FUNCTION_OBJECT     function_object, 
    u32                     cq_id);

// ------------------------------------------------------
//  MCC QUEUE
// ------------------------------------------------------

BESTATUS
be_mpu_init_mailbox (
    IN PBE_FUNCTION_OBJECT function_object,
    IN PSA_SGL             mailbox);

BESTATUS
be_mpu_uninit_mailbox (    
    IN      PBE_FUNCTION_OBJECT  function_object);

PMCC_WRB
_be_mpu_peek_ring_wrb (
    IN PBE_MCC_OBJECT mcc,
    IN boolean driving_queue);


PBE_MCC_WRB_CONTEXT
_be_mcc_allocate_wrb_context (
    PBE_FUNCTION_OBJECT function_object);

void
_be_mcc_free_wrb_context (
    PBE_FUNCTION_OBJECT function_object,
    PBE_MCC_WRB_CONTEXT context);

BESTATUS
_be_mpu_post_wrb_mailbox (
    IN PBE_FUNCTION_OBJECT   function_object,
    IN PMCC_WRB         wrb,
    IN PBE_MCC_WRB_CONTEXT   wrb_context);

BESTATUS
_be_mpu_post_wrb_ring (
    IN PBE_MCC_OBJECT        mcc,
    IN PMCC_WRB         wrb,
    IN PBE_MCC_WRB_CONTEXT   wrb_context);

void
be_mcc_process_cqe (
    IN PBE_FUNCTION_OBJECT function_object,
    IN PMCC_CQ_ENTRY cqe);

u32
be_mcc_get_id (PBE_MCC_OBJECT mcc);

boolean
be_drive_mcc_wrb_queue ( 
    IN      PBE_MCC_OBJECT mcc);

// ------------------------------------------------------
//  Mailbox Object
// ------------------------------------------------------

// ------------------------------------------------------
//  Default PDU Queue
// ------------------------------------------------------

// ------------------------------------------------------
//  iSCSI WRB ring
// ------------------------------------------------------

// ------------------------------------------------------
//  iSCSI connection
// ------------------------------------------------------

BESTATUS
be_iscsi_cxn_create (
    PBE_FUNCTION_OBJECT function_object,
    IN  PBE_ISCSI_WRB_QUEUE_OBJECT         wrbq,
    IN  PBE_DEFAULT_PDU_QUEUE_OBJECT       defq,
    IN  PBE_CQ_OBJECT                      cq,
    OUT PBE_ISCSI_CONNECTION_OBJECT        cxn);

void
be_iscsi_cxn_destroy (
    PBE_ISCSI_CONNECTION_OBJECT cxn);


// ------------------------------------------------------
//  FCoE Stuff 
// ------------------------------------------------------

BESTATUS
be_fcoe_interface_create (
    IN PBE_FUNCTION_OBJECT func_obj,
    IN PBE_FCOE_WRB_QUEUE_OBJECT wrbq,
    IN PBE_CQ_OBJECT             cq,
    IN PBE_FCOE_INTERFACE_OBJECT intf);


// ------------------------------------------------------
//  Ethernet Send Queue
// ------------------------------------------------------


// ------------------------------------------------------
//  Ethernet Receive Queue
// ------------------------------------------------------

// ------------------------------------------------------
//  LRO Queue
// ------------------------------------------------------

// ------------------------------------------------------
//  TOE
// ------------------------------------------------------

BESTATUS
be_toe_cxn_create (
    IN  PBE_FUNCTION_OBJECT       function_object,
    IN  PBE_TOE_DQ_OBJECT         dq,
    IN  PBE_TOE_SQ_OBJECT         sq,
    IN  PBE_TOE_RQ_OBJECT         rq,
    OUT PBE_TOE_CONNECTION_OBJECT cxn);

BESTATUS 
be_toe_cxn_destroy (
    IN  PBE_TOE_CONNECTION_OBJECT cxn);

u32 
be_toe_cxn_get_cid (
    IN  PBE_TOE_CONNECTION_OBJECT cxn);

u32
be_toe_sq_get_ring_size_encoding (
    IN PBE_TOE_SQ_OBJECT sq);

// TODO - Determine if we need!
//
//// Queries the current tcp_state.  This may not be an accurate snapshot of the
//// connection if it is currently enabled.
//BESTATUS
//be_toe_offload_query (
//    IN  PBE_TOE_CONNECTION_OBJECT cxn,
//    OUT PTCP_CONTEXT              tcp_state,
//    IN  MCC_WRB_CQE_CALLBACK      callback,
//    IN  PVOID                     callback_context);
//
//// Enables or disables the connection.  The template header and context 
//// remains untouched.
//BESTATUS
//be_toe_offload_enable (
//    IN PBE_TOE_CONNECTION_OBJECT cxn,
//    IN boolean                   enable,
//    IN MCC_WRB_CQE_CALLBACK      callback,
//    IN PVOID                     callback_context);

// ------------------------------------------------------
//  RDMA/iWARP
// ------------------------------------------------------



// ------------------------------------------------------
//  Offload
// ------------------------------------------------------

// ------------------------------------------------------
//  Ring Sizes
// ------------------------------------------------------
STATIC INLINE 
u32 be_ring_encoding_to_length (u32 encoding, u32 object_size) {

    ASSERT (encoding != 1); // 1 is rsvd
    ASSERT (encoding < 16);
    ASSERT (object_size > 0);

    if (encoding == 0) // 32k deep
        encoding = 16; 

    return (1 << (encoding - 1)) * object_size;
}

STATIC INLINE
u32 be_ring_length_to_encoding (u32 length_in_bytes, u32 object_size) {

    u32 count, encoding;

    ASSERT (object_size > 0);
    ASSERT (length_in_bytes % object_size == 0);

    count = length_in_bytes / object_size;
    
    ASSERT (count > 1);
    ASSERT (count <= 32*1024);
    ASSERT (sa_is_power2 (count));
    ASSERT (length_in_bytes <= 8 * SA_PAGE_SIZE); // max ring size in UT
    
    encoding = sa_log2 (count) + 1;

    if (encoding == 16)
        encoding = 0;   // 32k deep

    return encoding;
}




// ------------------------------------------------------
//  Locking
// ------------------------------------------------------

// Init, acquire, or release a lock at any IRQL
STATIC INLINE void be_lock_init (BE_LOCK* lock) {
    sa_init_spinlock (&lock->lock);
}

STATIC INLINE void be_lock_acquire (IN BE_LOCK* lock) {
    sa_acquire_spinlock (&lock->lock, &lock->irql);
}

STATIC INLINE void be_lock_release (IN BE_LOCK* lock) {
    sa_release_spinlock (&lock->lock, &lock->irql);
}

// Init, acquire, or release a blocking lock at IRQL<DISPATH_LEVEL. 
// TODO - These may need to be implemented with spinlocks on some platforms, hence this
// extra layer of abstraction.
STATIC INLINE void be_blocking_lock_init (IN BE_BLOCKING_LOCK *lock) {
    sa_initialize_fast_mutex (lock);
}

STATIC INLINE void be_blocking_lock_acquire (IN BE_BLOCKING_LOCK *lock) {
    SA_NOT_USED (lock);
    //SA_ASSERT_BLOCKING ();
    //sa_acquire_fast_mutex (lock);
}

STATIC INLINE void be_blocking_lock_release (IN BE_BLOCKING_LOCK *lock) {
    SA_NOT_USED (lock);
    //sa_release_fast_mutex (lock);
}



STATIC INLINE void be_lock_wrb_post (PBE_FUNCTION_OBJECT function_object)
{
    FUNCTION_ASSERT (function_object);
    //ASSERT (sa_atomic_increment (&function_object->post_lock) == 1);    
    sa_acquire_spinlock (&function_object->post_lock, &function_object->post_irq);    
}


STATIC INLINE void be_unlock_wrb_post_no_redrive (PBE_FUNCTION_OBJECT function_object)
{
    FUNCTION_ASSERT (function_object);
    // ASSERT (sa_atomic_decrement (&function_object->post_lock) == 0);       
    sa_release_spinlock (&function_object->post_lock, &function_object->post_irq);
}

STATIC INLINE void be_unlock_wrb_post (PBE_FUNCTION_OBJECT function_object)
{
    FUNCTION_ASSERT (function_object);
    // ASSERT (sa_atomic_decrement (&function_object->post_lock) == 0);       
    sa_release_spinlock (&function_object->post_lock, &function_object->post_irq);

    if (function_object->pend_queue_driving && function_object->links.mcc) {        
        function_object->stats.pended_queue_driving++;
        be_drive_mcc_wrb_queue (function_object->links.mcc);
    }
}

STATIC INLINE void be_lock_cq_process (PBE_FUNCTION_OBJECT function_object)
{
    FUNCTION_ASSERT (function_object);
    // ASSERT (sa_atomic_increment (&function_object->cq_lock) == 1);
    sa_acquire_spinlock (&function_object->cq_lock, &function_object->cq_irq);
}

STATIC INLINE void be_unlock_cq_process (PBE_FUNCTION_OBJECT function_object)
{
    FUNCTION_ASSERT (function_object);
    // ASSERT (sa_atomic_decrement (&function_object->cq_lock) == 0);
    sa_release_spinlock (&function_object->cq_lock, &function_object->cq_irq);
}

#if 0

    STATIC INLINE void
    be_unlock_hw (PBE_FUNCTION_OBJECT function)
    {
        ASSERT (function->own_semaphore == 1);
        function->own_semaphore = 0;

        // Clear the semaphore
        PCICFG0_WRITE_CONST (function, semaphore[0], 0);

        // Lower irql
        KeLowerIrql (function->current_irql);
    }

    STATIC INLINE void
    be_lock_hw (PBE_FUNCTION_OBJECT function)
    {
        u32 semaphore = 1;
        u32 wait = 0;

        // Don't allow nested lock acquisition.
        ASSERT (function->own_semaphore == 0);

        // If iSCSI accesses hw from interrupt, we need to sync to DIRQL.
        ASSERT (KeGetCurrentIrql () <= DISPATCH_LEVEL);
        
        do {
            // Raise IRQL to avoid preemption with the spinlock held.
            KeRaiseIrql (DISPATCH_LEVEL, &function->current_irql); // SA_IRQ current_irql

            semaphore = PCICFG_READ (function, semaphore[0]);

            if (semaphore != 0) {

                if ((wait++ & 0x3FFFF) == 0)  {
                    TRACE (DL_WARN, "Waiting for semaphore.");
                }

                // We did not acquire it.  Lower the irql.
                KeLowerIrql (function->current_irql);
            }

        } while (semaphore != 0);
        
        TRACE (DL_INFO, "Locked semaphore.");
        function->own_semaphore = 1;
    }

#else
    #define be_lock_hw(a_)
    #define be_unlock_hw(a_)
#endif

void
be_sgl_to_pa_list (
    PSA_SGL sgl,
    PHYS_ADDR* pa_list,
    u32 max_num);


#if defined(SA_DEBUG)

STATIC INLINE const c8* be_query_opcode_name (u32 subsystem, u32 opcode) {
    switch (subsystem) {
        case ENUM_IOCTL_SUBSYSTEM_RSVD:         return "rsvd_subsystem";
        case ENUM_IOCTL_SUBSYSTEM_COMMON:       return common_subsystem_opcodes_get_name (opcode);
        case ENUM_IOCTL_SUBSYSTEM_COMMON_ISCSI: return common_iscsi_subsystem_opcodes_get_name (opcode);
        case ENUM_IOCTL_SUBSYSTEM_ETH:          return eth_subsystem_opcodes_get_name (opcode);
        case ENUM_IOCTL_SUBSYSTEM_TOE:          return toe_subsystem_opcodes_get_name (opcode);
        case ENUM_IOCTL_SUBSYSTEM_ISCSI_INI:    return iscsi_ini_subsystem_opcodes_get_name (opcode);
        case ENUM_IOCTL_SUBSYSTEM_ISCSI_TGT:    return iscsi_tgt_subsystem_opcodes_get_name (opcode);
        case ENUM_IOCTL_SUBSYSTEM_RDMA:         return rdma_subsystem_opcodes_get_name (opcode);
        case ENUM_IOCTL_SUBSYSTEM_LOWLEVEL:     return lowlevel_subsystem_opcodes_get_name (opcode);
        default:
            return "invalid subsystem";
    }
}

#endif



#endif // __beclib_private_h__


