/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!

@file
    beregister.h

@brief
    Provides 'smart' wrappers for BladeEngine registers.  These wrappers are C++
    usable only.   

@author jflores

@note
    jflores   11/30/2004   initial
    $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/drivers/beclib/beregister.h,v 1.7 2006/02/21 03:20:07 menz Exp $
 
*/

#if !defined( __BEREGISTER_H__ )
#define __BEREGISTER_H__

/*!

@brief
    This macro zeros out a data structure whose contents happen to be
    exactly 32-bits. Many controller registers have bit map structures
    defined to represent them and this routine allows for easy and correct
    initialization of the register to 0.

@param    
    reg         - Pointer to the 32-bit structure to zero out.
    
@return
   
@note
    IRQL: any

*/


// 
// Macros for reading/writing a protection domain or CSR registers in BladeEngine.
//
#define PD_READ( _fo_, _field_ )                BePdRead ((_fo_)->sa_dev, SA_FIELD_OFFSET( PROTECTION_DOMAIN_DBMAP, _field_ ), SA_DBG_CSTR (#_field_) )
/*#define PD_WRITE(_fo_, _field_, _value_ )       BePdWrite((_fo_)->sa_dev, SA_FIELD_OFFSET( PROTECTION_DOMAIN_DBMAP, _field_ ), (_value_).dw, SA_DBG_CSTR (#_field_) )*/
#define PD_WRITE(_fo_, _field_, _value_ )       BePdWrite((_fo_)->sa_dev, SA_FIELD_OFFSET( PROTECTION_DOMAIN_DBMAP, _field_ ), (*(u32 *)&(_value_)), SA_DBG_CSTR (#_field_) )
#define PD_WRITE_CONST(_fo_, _field_, _value_ ) BePdWrite((_fo_)->sa_dev, SA_FIELD_OFFSET( PROTECTION_DOMAIN_DBMAP, _field_ ), (_value_), SA_DBG_CSTR (#_field_) )

// Returns virtual address for given doorbell register.
#define PD_ADDRESS( _fo_, _field_ )  (PVOID)(                                      \
    (u8*)sa_dev_get_bar_offset ((_fo_)->sa_dev, sa_get_pd_bar ((_fo_)->sa_dev)) +  \
    SA_FIELD_OFFSET( PROTECTION_DOMAIN_DBMAP, _field_)                             \
    )                                                                              \

#define CSR_READ( _fo_, _field_ )                BeCsrRead( (_fo_)->sa_dev, SA_FIELD_OFFSET( BLADE_ENGINE_CSRMAP, _field_ ), SA_DBG_CSTR (#_field_) )
/*#define CSR_WRITE(_fo_, _field_, _value_ )       BeCsrWrite((_fo_)->sa_dev, SA_FIELD_OFFSET( BLADE_ENGINE_CSRMAP, _field_ ), (_value_).dw, SA_DBG_CSTR (#_field_) )*/
#define CSR_WRITE(_fo_, _field_, _value_ )       BeCsrWrite((_fo_)->sa_dev, SA_FIELD_OFFSET( BLADE_ENGINE_CSRMAP, _field_ ), (*(u32 *)&(_value_)), SA_DBG_CSTR (#_field_) )
#define CSR_WRITE_CONST(_fo_, _field_, _value_ ) BeCsrWrite((_fo_)->sa_dev, SA_FIELD_OFFSET( BLADE_ENGINE_CSRMAP, _field_ ), (_value_), SA_DBG_CSTR (#_field_) )

// Returns virtual address for given doorbell register.
#define CSR_ADDRESS( _fo_, _field_ )  (PVOID)(                                      \
    (u8*)sa_dev_get_bar_offset ((_fo_)->sa_dev, sa_get_csr_bar ((_fo_)->sa_dev)) +  \
    SA_FIELD_OFFSET( BLADE_ENGINE_CSRMAP, _field_)                                  \
    )                                                                               \


// 
// Macros for reading/writing a protection domain or CSR registers in BladeEngine.
//
#define PCICFG_READ( _fo_, _field_ )                 BePciCfgRead( (_fo_)->sa_dev, SA_FIELD_OFFSET( PCICFG_PF0_CSRMAP, _field_ ), SA_DBG_CSTR (#_field_) )
/*#define PCICFG_WRITE(_fo_, _field_, _value_ )        BePciCfgWrite((_fo_)->sa_dev, SA_FIELD_OFFSET( PCICFG_PF0_CSRMAP, _field_ ), (_value_).dw, SA_DBG_CSTR (#_field_) )*/
#define PCICFG_WRITE(_fo_, _field_, _value_ )        BePciCfgWrite((_fo_)->sa_dev, SA_FIELD_OFFSET( PCICFG_PF0_CSRMAP, _field_ ), (*(u32 *)&(_value_)), SA_DBG_CSTR (#_field_) )
#define PCICFG_WRITE_CONST(_fo_, _field_, _value_ )  BePciCfgWrite((_fo_)->sa_dev, SA_FIELD_OFFSET( PCICFG_PF0_CSRMAP, _field_ ), (_value_), SA_DBG_CSTR (#_field_) )
#define VF_PCICFG_READ( _fo_, _field_ )              BePciCfgRead( (_fo_)->sa_dev, SA_FIELD_OFFSET( VF_PCICFG_CSRMAP , _field_ ), SA_DBG_CSTR (#_field_) )

// These names are preserved for legacy reasons, but the functionality is the same.
#define PCICFG0_READ( _fo_, _field_ )                PCICFG_READ(_fo_, _field_)
#define PCICFG0_WRITE(_fo_, _field_, _value_ )       PCICFG_WRITE(_fo_, _field_, _value_)
#define PCICFG0_WRITE_CONST(_fo_, _field_, _value_ ) PCICFG_WRITE_CONST(_fo_, _field_, _value_ )

#define PCICFG1_READ( _fo_, _field_ )                PCICFG_READ(_fo_, _field_)
#define PCICFG1_WRITE(_fo_, _field_, _value_ )       PCICFG_WRITE(_fo_, _field_, _value_)
#define PCICFG1_WRITE_CONST(_fo_, _field_, _value_ ) PCICFG_WRITE_CONST(_fo_, _field_, _value_ )


#define RXULP_PD_WRITE(_function_object_, _ulp_number_, _value_)   \
    switch (_ulp_number_) {                                        \
        case 0:                                                    \
            PD_WRITE ((_function_object_), rxulp0_db, (_value_));  \
            break;                                                 \
        case 1:                                                    \
            PD_WRITE ((_function_object_), rxulp1_db, (_value_));  \
            break;                                                 \
        case 2:                                                    \
            PD_WRITE ((_function_object_), rxulp2_db, (_value_));  \
            break;                                                 \
        default:                                                   \
            BREAKPOINT ();                                         \
    }                                                              \


#define TXULP_PD_WRITE(_function_object_, _ulp_number_, _value_)   \
    switch (_ulp_number_) {                                        \
        case 0:                                                    \
            PD_WRITE ((_function_object_), txulp0_db, (_value_));  \
            break;                                                 \
        case 1:                                                    \
            PD_WRITE ((_function_object_), txulp1_db, (_value_));  \
            break;                                                 \
        case 2:                                                    \
            PD_WRITE ((_function_object_), txulp2_db, (_value_));  \
            break;                                                 \
        default:                                                   \
            BREAKPOINT ();                                         \
    }                                                              \


/*!

@brief
    This routine reads from a register located within the protection domain
    doorbell space for a given function object.

@param    
    FuncObj     - Pointer to the function object to read from.
    
@param    
    Offset      - The Offset (in bytes) to read from within the function's PD space.

@return
    The value read. 

@note
    IRQL: any

*/
STATIC INLINE
u32  
BePdRead( 
    IN      PSA_DEV     sa_dev, 
    IN      u32         offset,
    IN      const c8*   dbg_name)
{
    u32 v;
    ASSERT( offset < sizeof (PROTECTION_DOMAIN_DBMAP) );
    v = sa_dev_read_u32 (sa_dev, sa_get_pd_bar (sa_dev), offset);
    TRACE (DL_HW, "PD Read.  offset:0x%04x value:0x%08x (%s)", offset, v, dbg_name);
    return v;
}

STATIC INLINE
u32
BeCsrBarAvailable(
    IN      PSA_DEV     sa_dev)
{
    return (sa_get_csr_bar (sa_dev) != ~(u32)0);
}


/*!

@brief
    This routine writes to a register located within the protection domain
    doorbell space for a given function object.

@param    
    FuncObj     - Pointer to the function object to read from.
    
@param    
    Offset      - The Offset (in bytes) to write to within the function's PD space.

@param    
    Value       - The value to write to the register.

@return

@note
    IRQL: any

*/
STATIC INLINE
void  
BePdWrite( 
    IN      PSA_DEV        sa_dev, 
    IN      u32            offset, 
    IN      u32            value,
    IN      const c8*      dbg_name)
{
    ASSERT( offset < sizeof(PROTECTION_DOMAIN_DBMAP) );    

    TRACE (DL_HW, "PD Write. offset:0x%04x value:0x%08x (%s)", offset, value, dbg_name);
    sa_dev_write_u32 (sa_dev, sa_get_pd_bar (sa_dev), offset, value);
}


/*!

@brief
    This routine reads from a register located within the CSR space for a 
    given function object.

@param    
    sa_dev - Context for this set of register space
    
@param    
    Offset          - The Offset (in bytes) to read from within the function's CSR space.

@return
    The value read.    

@note
    IRQL: any

*/
STATIC INLINE
u32  
BeCsrRead( 
    IN      PSA_DEV        sa_dev, 
    IN      u32            offset,
    IN      const c8*      dbg_name)
{
    u32 value;
    u32 bar;

    ASSERT( offset < sizeof(BLADE_ENGINE_CSRMAP) );

    bar = sa_get_csr_bar (sa_dev);
    value = sa_dev_read_u32 (sa_dev, bar, offset);

    TRACE (DL_HW, "CSR Read.  offset:0x%04x value:0x%08x (%s)", offset, value, dbg_name);

    return  value;
}

/*!

@brief
    This routine writes to a register located within the CSR
    space for a given function object.

@param    
    sa_dev - Context for this set of register space
    
@param    
    Offset          - The Offset (in bytes) to write to within the function's CSR space.

@param    
    Value           - The value to write to the register.

@return

@note
    IRQL: any

*/
STATIC INLINE
void  
BeCsrWrite( 
    IN      PSA_DEV        sa_dev,  
    IN      u32            offset, 
    IN      u32            value,
    IN      const c8*      dbg_name)
{
    ASSERT( offset < sizeof(BLADE_ENGINE_CSRMAP) );

    TRACE (DL_HW, "CSR Write. offset:0x%04x value:0x%08x (%s)", offset, value, dbg_name);

    sa_dev_write_u32 (sa_dev, sa_get_csr_bar (sa_dev), offset, value);
}

/*!

@brief
    This routine reads from a register located within the pci cfg
    space for a given function object.

@param    
    sa_dev - Context for this set of register space
    
@param    
    Offset          - The Offset (in bytes) to read from within the function's pci cfg space.

@return
    The value read. 

@note
    IRQL: any

*/
STATIC INLINE
u32  
BePciCfgRead( 
    IN      PSA_DEV        sa_dev, 
    IN      u32            offset,
    IN      const c8*      dbg_name)
{
    u32 value;

    ASSERT( (offset < sizeof(PCICFG_PF0_CSRMAP)));

    value = sa_dev_read_u32 (sa_dev, sa_get_pci_bar (sa_dev), offset);

    TRACE (DL_HW, "PCICFG Read.  offset:0x%04x value:0x%08x (%s)", offset, value, dbg_name);

    return  value;
}


/*!

@brief
    This routine writes to a register located within the pci cfg
    space for a given function object.

@param    
    sa_dev - Context for this set of register space
    
@param    
    Offset          - The Offset (in bytes) to write to within the function's pci cfg space.

@param    
    Value           - The value to write to the register.

@return

@note
    IRQL: any

*/
STATIC INLINE
void  
BePciCfgWrite (
    IN PSA_DEV sa_dev,
    IN u32       offset,
    IN u32       value,
    IN const c8* dbg_name)
{

    ASSERT( (offset < sizeof(PCICFG_PF0_CSRMAP)));

    TRACE (DL_HW, "PCICFG Write. offset:0x%04x value:0x%08x (%s)", offset, value, dbg_name);

    sa_dev_write_u32 (sa_dev, sa_get_pci_bar (sa_dev), offset, value);
    
}



STATIC INLINE
void 
BeSetDword( 
    IN 		PVOID		field,
    IN		u32			value)
{
    
    (*( u32*) field) = value;

    return ;
}



#endif  // __BEREGISTER_H__
