/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

///*++
//
//    Filename:
//
//        hwiwrb.c
//
//    Purpose:
//
//        This file implements functionality to interact with the WRB data
//        structures. These structures are used to interface with the BE ASIC.
//        
//
//--*/
//
//
#include "acitpch.h"

ACIT_VOID
hwi_unitt_init_sgl_handle(
    PACIT_CONTROLLER_WS pcontroller);

extern INLINE
ACIT_VOID
hwi_complete_cmd(
    PHWI_CONTROLLER_WS      phwi_controller,
    ACIT_UEXACT32           message_id,
    PI_SOL_CQE              psol);

/*
 * The following routines are actually part of HWI API, but are
 * defined by HWIWRB for efficiency
 */

ACIT_UEXACT32
hwi_fill_io_request(
    PHWI_CONTROLLER_WS          phwi_controller, 
    PACIT_IO_HANDLE             pio_handle)
{
    PWRB_HANDLE                 pwrb_handle;
    PISCSI_WRB                  pwrb;
    PACIT_IO_REQUEST            pio_request = pio_handle->pio->pio_request;

    pwrb_handle = (PWRB_HANDLE)pio_handle->pwrb_handle;
    pwrb = (PISCSI_WRB)pwrb_handle->pwrb;

    OSM_ASSERT (pwrb);
    OSM_MEMSET(pwrb, 0, sizeof(ISCSI_WRB));

    // TODO - handle multi-level LUNs here

    OSM_ASSERT(!(pio_request->lun.u.lun_leveled[1] |
                 pio_request->lun.u.lun_leveled[2] |
                 pio_request->lun.u.lun_leveled[3]));

    // 14 bits of LUN info, from the first level in bigendian format
    ACIT_PUT_BE16(pwrb->lun, pio_request->lun.u.lun_leveled[0]);

    if(pio_request->flags.outbound_data) 
    {
        pwrb->type    = INI_WR_CMD;
		pwrb->dsp     = 1;

        // for writes, use r2t_buf_offset field to communicate number of sg
        // elements in order to turn on the EP firmware's fast path
        //pwrb->r2t_buf_offset = pio_request->psg_list->number_elements;
    }
    else
    {
        pwrb->type = INI_RD_CMD;
    }
#if ACIT_CFG_INTERRUPT_REDIRECTION
    // Merely by compile flag alone, we assume we can stuff in the cq_id into
    // the ISCSI_WRB. The BE2 firmware has no such field however. But that
    // field on BE2 is a reserved member, so we go ahead and fill it anyway. 
    // Saves a couple of checks
    pwrb->cq_id             = pio_handle->cq_id;
#endif
    pwrb->wrb_idx           = pwrb_handle->wrb_index;
    pwrb->ptr2NextWrb       = pwrb->wrb_idx;
    ACIT_PUT_LE32(pwrb->r2t_exp_dtl, pio_handle->exp_xfer_length);
    ACIT_PUT_LE32(pwrb->cmdsn_itt, pio_handle->cmd_sn);

    // Fill the SGs.
    // split logic is not required upto 64KB.
    if(pio_handle->exp_xfer_length <= BE_MAX_BYTES_PER_FRAG)
    {
        hwi_write_sgl(pwrb,
                      &pio_handle->pio->bhs_pa,
                      pio_handle->pio->bhs_len,
                      pio_handle->psgl_handle, 
                      pio_request->psg_list);
    }
    else
    {
        hwi_write_large_sgl(pwrb,
                            &pio_handle->pio->bhs_pa,
                            pio_handle->pio->bhs_len,
                            pio_handle->psgl_handle,
                            pio_request->psg_list);
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,
            ("IO WRB: 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x cri 0x%x\n", 
            pwrb->dw[0], pwrb->dw[1], pwrb->dw[2], pwrb->dw[3], pwrb->dw[4], pwrb->dw[5], pwrb->dw[6], pwrb->dw[7], pio_handle->cri));

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_fill_logout_request(
    PHWI_CONTROLLER_WS      phwi_controller, 
    PACIT_LOGOUT_HANDLE     plogout_handle)
{
    PWRB_HANDLE             pwrb_handle;
    PISCSI_WRB              pwrb;

    pwrb_handle = plogout_handle->pwrb_handle;
    pwrb = (PISCSI_WRB)pwrb_handle->pwrb;

    OSM_ASSERT (pwrb);
    OSM_MEMSET (pwrb, 0, sizeof (*pwrb));

    // Logout is sent with INI_WR_CMD type and no dsp bit is
    // set since there's no data to transmit. 
    pwrb->type = INI_WR_CMD;
    pwrb->wrb_idx    = pwrb_handle->wrb_index;
    pwrb->ptr2NextWrb = pwrb->wrb_idx;
    ACIT_PUT_LE32(pwrb->cmdsn_itt, plogout_handle->cmd_sn);
    // put the logout completion on the first CQ
#if ACIT_CFG_INTERRUPT_REDIRECTION
    pwrb->cq_id = HWI_GET_CQ_ID(phwi_controller->pacit_ws, 0);
#endif

    /*
     * Fill the SGs.
     */
    hwi_write_sgl(pwrb,
                  &plogout_handle->bhs_pa,
                  plogout_handle->bhs_len,
                  plogout_handle->psgl_handle, 
                  ACIT_NULL);

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,
            ("Logout WRB: 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x cri 0x%x\n", 
            pwrb->dw[0], pwrb->dw[1], pwrb->dw[2], pwrb->dw[3], pwrb->dw[4], pwrb->dw[5], pwrb->dw[6], pwrb->dw[7], plogout_handle->cri));

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_fill_tmf_request(
    PHWI_CONTROLLER_WS      phwi_controller, 
    PACIT_TMF_HANDLE        ptmf_handle)
{
    PWRB_HANDLE             pwrb_handle;
    PISCSI_WRB              pwrb;

    pwrb_handle = ptmf_handle->pwrb_handle;
    pwrb = (PISCSI_WRB)pwrb_handle->pwrb;

    OSM_ASSERT (pwrb);
    OSM_MEMSET (pwrb, 0, sizeof (*pwrb));

    // TMF is sent with INI_WR_CMD type and no dsp bit is
    // set since there's no data to transmit
    pwrb->type = INI_WR_CMD;
    pwrb->wrb_idx    = pwrb_handle->wrb_index;
    pwrb->ptr2NextWrb = pwrb->wrb_idx;
    ACIT_PUT_LE32(pwrb->cmdsn_itt, ptmf_handle->cmd_sn);

    if(ptmf_handle->tmf_function == ACIT_TMF_ABORT_TASK)
    {
        pwrb->ref_sgl_icd_idx = (ptmf_handle->pio_handle->itt & 0xfff);
    }
    // put the tmf completion on the first CQ
#if ACIT_CFG_INTERRUPT_REDIRECTION
    pwrb->cq_id = HWI_GET_CQ_ID(phwi_controller->pacit_ws, 0);
#endif

    /*
     * Fill the SGs.
     */
    hwi_write_sgl(pwrb,
                  &ptmf_handle->bhs_pa,
                  ptmf_handle->bhs_len,
                  ptmf_handle->psgl_handle, 
                  ACIT_NULL);

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,
            ("TMF WRB: 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x cri 0x%x\n", 
            pwrb->dw[0], pwrb->dw[1], pwrb->dw[2], pwrb->dw[3], pwrb->dw[4], pwrb->dw[5], pwrb->dw[6], pwrb->dw[7], ptmf_handle->cri));

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_fill_nop_request(
    PHWI_CONTROLLER_WS      phwi_controller, 
    PACIT_NOP_HANDLE        pnop_handle,
    PACIT_NOP_HANDLE        poptional_handle)
{
    PWRB_HANDLE             pwrb_handle;
    PISCSI_WRB              pwrb;

    pwrb_handle = pnop_handle->pwrb_handle;
    pwrb = (PISCSI_WRB)pwrb_handle->pwrb;

    OSM_ASSERT (pwrb);
    OSM_MEMSET (pwrb, 0, sizeof (*pwrb));

    // NOPOut is sent with INI_WR_CMD type and no dsp bit is
    // set since there's no data to transmit. 
    // Note!! If there is a payload that needs to go out, then 
    // dsp bit should be set
    pwrb->type = INI_WR_CMD;
    pwrb->wrb_idx     = pwrb_handle->wrb_index;
    pwrb->ptr2NextWrb = pwrb->wrb_idx;
    pwrb->dmsg        = ACIT_TRUE;  // ack based completion
    ACIT_PUT_LE32(pwrb->cmdsn_itt, pnop_handle->cmd_sn);

    if(poptional_handle)
    {
        /*
         * Fill the SGs.
         */
        hwi_write_sgl(pwrb,
                      &pnop_handle->bhs_pa,
                      pnop_handle->bhs_len,
                      pnop_handle->psgl_handle, 
                      ACIT_NULL);
    }
    else
    {
        hwi_write_wrb_sgl(pwrb,
                      &pnop_handle->bhs_pa,
                      pnop_handle->bhs_len,
                      ACIT_NULL);

        // Fill in the SGL ICD we reserved earlier
        ACIT_PUT_LE32(pwrb->sgl_icd_idx, phwi_controller->nop_icd);
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,
            ("NOP WRB: 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x cri 0x%x\n", 
            pwrb->dw[0], pwrb->dw[1], pwrb->dw[2], pwrb->dw[3], pwrb->dw[4], pwrb->dw[5], pwrb->dw[6], pwrb->dw[7], pnop_handle->cri));


    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_submit_request(
    PACIT_CONTROLLER_WS     pcontroller,
    PWRB_HANDLE             pwrb_handle,
    ACIT_UEXACT32           cri)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    WRB_POST_DB             wrb_post_db; 
    ACIT_UEXACT32           wrb_index;
    PHWI_WRB_CONTEXT        pwrb_context;

    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);
    wrb_index = ((PWRB_HANDLE)(pwrb_handle))->wrb_index;

    //OSM_ASSERT(!(((PWRB_HANDLE)(pwrb_handle))->count));
    //((PWRB_HANDLE)(pwrb_handle))->count++;

    //
    // WRBs are outstanding so we need to update the tail pointer.  There
    // is a chance that the tail WRB is:
    //   1. Actually the tail and requires linking.
    //   2. Already freed and sitting in the free list.
    //   3. The same as the current WRB that we are posting.
    // In case 2 & 3 the linking is not required but it will not hurt
    // anything.  It is done in all cases to avoid synchronizing the
    // tailWrb variable between the completion and posting routines. In
    // a full duplex storport driver, they are not synchronized by default.
    //

    pwrb_context = &phwi_controller->wrb_context[cri];

    if(pwrb_context->plast_wrb)
    {
        pwrb_context->plast_wrb->ptr2NextWrb = wrb_index;
        pwrb_context->plast_wrb = ((PWRB_HANDLE)(pwrb_handle))->pwrb;
    }
    else
    {
        pwrb_context->plast_wrb = ((PWRB_HANDLE)(pwrb_handle))->pwrb;
    }

    //
    // Format doorbell. This will post the next command.
    //
	memset(&wrb_post_db,0,sizeof(wrb_post_db));
    wrb_post_db.wrb_cid = pwrb_context->cid;
    wrb_post_db.wrb_index = wrb_index;
    wrb_post_db.numberPosted = 1;

    //TRACE (DL_HW, "WRB Post Doorbell:0x%x", doorbell);

#if (BE_CONFIG==4)
    if(pwrb_context->cid & 1)
        ACIT_PD_WRITE (pcontroller, txulp1_db, wrb_post_db.dw);
    else
        ACIT_PD_WRITE (pcontroller, txulp0_db, wrb_post_db.dw);
#else
    ACIT_PD_WRITE (pcontroller, txulp0_db, wrb_post_db.dw);
#endif /* (BE_CONFIG==4) */

    ACIT_UPDATE_TXULP_DB_COUNT(pcontroller);

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_submit_drvr_request(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_VOID              pwrb_handle,
    ACIT_UEXACT32           cri)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    PHWI_WRB_CONTEXT        pwrb_context;
    PWRB_HANDLE             pwrb_h = (PWRB_HANDLE)pwrb_handle;
    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);

    pwrb_context = &phwi_controller->wrb_context[cri];

    LIST_ADDTO_TAIL(&pwrb_h->link,
                    &pwrb_context->wrb_handle_drvr_list);

    ACIT_UPDATE_DRVR_MSG_SENT_COUNT(pcontroller);

    hwi_submit_request(pcontroller,
                       pwrb_handle,
                       cri);

    return ACIT_STATUS_SUCCESS;
}

INLINE
ACIT_VOID
hwi_complete_cmd(
    PHWI_CONTROLLER_WS      phwi_controller,
    ACIT_UEXACT32           message_id,
    PI_SOL_CQE              psol)
{
    PHWI_WRB_CONTEXT        pwrb_context;
    PWRB_HANDLE             pwrb_handle;
    PWRB_HANDLE             pwrb_handle_base;

    // Find the command object.
    ASSERT (psol->wrb_index < phwi_controller->pacit_ws->params.num_cmd_per_cxn);

    pwrb_context = &phwi_controller->wrb_context[CRI_FROM_CID(phwi_controller->pacit_ws, psol->cid)];

    pwrb_handle_base = pwrb_context->pwrb_handle_base;
    pwrb_handle = (PWRB_HANDLE)pwrb_handle_base + psol->wrb_index;

    switch(pwrb_handle->type)
    {
    case HWH_TYPE_IO: {
        ACIT_BIT_SET(pwrb_handle->pio_handle->flags, FLAG_IOH_IO_COMPLETION);

#if ACIT_CFG_INTERRUPT_REDIRECTION
        OSM_ASSERT(pwrb_handle->pio_handle->message_id == message_id);
#endif

        phwi_controller->pacit_ws->partition[message_id].dic.ios_per_sec++;

        ap_complete_io(phwi_controller->pacit_ws, 
                       pwrb_handle->pio_handle, 
                       psol);

        }
        break;

    case HWH_TYPE_LOGOUT:

        ap_complete_logout(phwi_controller->pacit_ws, 
                           pwrb_handle->plogout_handle,
                           (ACIT_UEXACT8)psol->i_resp,
                           0);  // no timeout

        break;

    case HWH_TYPE_TMF:
        
        ap_complete_tmf(phwi_controller->pacit_ws, 
                        pwrb_handle->ptmf_handle,
                        (ACIT_UEXACT8)psol->i_resp,
                        (ACIT_UEXACT8)psol->i_cmd_wnd,
                        psol->i_exp_cmd_sn);

        break;

    case HWH_TYPE_NOP:
        /* This is only in case initiator sent out a NOPOut and gets
         * back a NOPIn response. Async NOPIns end up in default PDU
         * queue
         */
        ap_complete_nop_in(phwi_controller->pacit_ws, pwrb_handle->pnop_handle);
        break;

    default:
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Invalid command completion WRB Index 0x%x CID 0x%x\n",
                                            psol->wrb_index, psol->cid));

        OSM_ASSERT(0);
        phwi_controller->pacit_ws->partition[message_id].interrupt_stats.unknown_cmds++;
        break;
    }
}

ACIT_UEXACT32
hwi_complete_drvr_msgs(
    PHWI_CONTROLLER_WS      phwi_controller,
    ACIT_UEXACT32           cid,
    ACIT_UEXACT32           wrb_index)
{
    PHWI_WRB_CONTEXT        pwrb_context;
    PWRB_HANDLE             pwrb_handle;
    PWRB_HANDLE             pwrb_handle_base = ACIT_NULL;
    PACIT_DC_LIST           plink;
    ACIT_UEXACT32           this_wrb_index, count = 0;

    /*
     * Walk the WRB handle list until we've found the WRB
     * whose index matches with wrb_index passed
     */
    pwrb_context = &phwi_controller->wrb_context[CRI_FROM_CID(phwi_controller->pacit_ws, cid)];
    pwrb_handle_base = pwrb_context->pwrb_handle_base;

    do
    {
        OSM_ASSERT(IS_LIST_NOT_EMPTY(&pwrb_context->wrb_handle_drvr_list));

        LIST_REMOVEFROM_HEAD (&plink, 
                              &pwrb_context->wrb_handle_drvr_list);

        pwrb_handle = STRUCT_BASE(WRB_HANDLE,
                                  link, 
                                  plink);

        OSM_ASSERT(pwrb_handle);
        OSM_ASSERT(pwrb_handle->type == HWH_TYPE_NOP);

        this_wrb_index = pwrb_handle->wrb_index;

        ap_free_nop_resources(phwi_controller->pacit_ws, pwrb_handle->pnop_handle);
        count++;
    }
    while(this_wrb_index != wrb_index);

    ACIT_UPDATE_DRVR_MSG_RECVD_COUNT(phwi_controller->pacit_ws, count);

    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
hwi_complete_cmd_error(
    PHWI_CONTROLLER_WS      phwi_controller,
    ACIT_UEXACT32           message_id,
    ACIT_UEXACT16           cid,
    ACIT_UEXACT16           wrb_index,
    ACIT_UEXACT16           code)
{
    PACIT_CONTROLLER_WS     pcontroller;
    PHWI_WRB_CONTEXT        pwrb_context;
    PWRB_HANDLE             pwrb_handle;
    PWRB_HANDLE             pwrb_handle_base;

    pcontroller = phwi_controller->pacit_ws;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,(("CQ Error 0x%x, CMD kill received for WRB 0x%x on CID 0x%x...\n"), 
                                        code, wrb_index, cid));

    // Find the command object.
    ASSERT (wrb_index < pcontroller->params.num_cmd_per_cxn);

    pwrb_context = &phwi_controller->wrb_context[CRI_FROM_CID(phwi_controller->pacit_ws, cid)];

    pwrb_handle_base = pwrb_context->pwrb_handle_base;
    pwrb_handle = (PWRB_HANDLE)pwrb_handle_base + wrb_index;

    if(pwrb_handle->type != HWH_TYPE_INVALID)
    {
        OSM_ASSERT(pwrb_handle->pio_handle);

        if(ACIT_IS_BIT_CLR(pwrb_handle->pio_handle->flags, FLAG_IOH_INVALIDATE_PENDING) &&
            ACIT_IS_BIT_CLR(pwrb_handle->pio_handle->flags, FLAG_IOH_INVALIDATE_COMPLETE))
        {
            switch(pwrb_handle->type)
            {
            case HWH_TYPE_IO:

                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("CEV error on IOH 0x%p in target 0x%p error, will trigger abort after invalidation...\n"),
                    pwrb_handle->pio_handle, pwrb_handle->pio_handle->ptarget));

                ap_complete_io_error(pcontroller, 
                                     pwrb_handle->pio_handle,
                                     FLAG_IOH_CQ_ERROR);

                break;

            case HWH_TYPE_LOGOUT:

                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("CEV error on LOH 0x%p in target 0x%p, invalidating the connection...\n"),
                    pwrb_handle->pio_handle, pwrb_handle->pio_handle->ptarget));

                ap_complete_logout(pcontroller, 
                                pwrb_handle->plogout_handle,
                                ISCSI_PDU_LOGOUT_RESPONSE_FAIL, // not used if timeout
                                ACIT_TRUE);  // timeout = Yes
                break;

            case HWH_TYPE_TMF:

                ACIT_SET_TARGET_SUBSTATE(pwrb_handle->ptmf_handle->ptarget, 
                                        TARGET_SUB_STATE_TMF_ERROR);

                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("CEV error on TMFH 0x%p in target 0x%p, will trigger session recovery...\n"),
                    pwrb_handle->ptmf_handle, pwrb_handle->ptmf_handle->ptarget));

                iscsi_recover_session(pcontroller, 
                                    pwrb_handle->ptmf_handle->ptarget,
                                    ACIT_NULL);

                break;

            case HWH_TYPE_NOP:

                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("CEV error on NOPH 0x%p in target 0x%p timeout, invalidating...\n"),
                    pwrb_handle->pnop_handle, pwrb_handle->pnop_handle->ptarget));

                //
                // We don't send out NOPOuts on our own, so we don't expect valid
                // ITT values for NOP handles..
                //
                OSM_ASSERT(0);
                phwi_controller->pacit_ws->partition[message_id].interrupt_stats.unknown_cmd_errors++;

                break;

            default:

                OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("CQ Error - unknown WRB 0x%x on CID 0x%x...\n", wrb_index, cid));
                OSM_ASSERT(0);
                phwi_controller->pacit_ws->partition[message_id].interrupt_stats.unknown_cmd_errors++;
                break;
            }
        }
        else
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Ignoring CQ Error on already invalidated cmd\n"));
        }
    }
    else
    {
        PACIT_TARGET_WS ptarget;

        pcontroller->eh_stats.cq_errors_invalid++;

        ptarget = iscsi_get_target_from_cid(pcontroller, cid);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("CQ CMD Error on invalid WRB 0x%x cid 0x%x! Forcing session recovery on target 0x%p\n",
                    wrb_index, cid, ptarget));

        ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_INVALID_CQ_ERR);

        iscsi_recover_session(pcontroller, 
                              ptarget,
                              ACIT_NULL);
    }
}

ACIT_VOID
hwi_mark_wrb_for_invalidate(
    PWRB_HANDLE             pwrb_handle)
{
    pwrb_handle->pwrb->invld = 1;
}

/* HWI API end */

PWRB_HANDLE
hwi_alloc_wrb_handle(
    PHWI_CONTROLLER_WS      phwi_controller,
    ACIT_UEXACT32           cri)
{
    PACIT_DC_LIST           plink;
    PWRB_HANDLE             pwrb_handle = ACIT_NULL;
    PHWI_WRB_CONTEXT        pwrb_context;

    pwrb_context = &phwi_controller->wrb_context[cri];

    OSM_ASSERT(IS_LIST_NOT_EMPTY(&pwrb_context->wrb_handle_list));

    LIST_REMOVEFROM_HEAD (&plink, 
                            &pwrb_context->wrb_handle_list);

    pwrb_handle = STRUCT_BASE(WRB_HANDLE,
                                link, 
                                plink);

    OSM_ASSERT(pwrb_handle);
    //pwrb_handle->count = 0;
    pwrb_context->wrb_handles_available--;


    if(!pwrb_context->wrb_handles_available)
    {
        /*
         * No more WRB handles available, this might be a case of
         * target going busy, let iscsi module handle this..
         */
        iscsi_flag_connection(phwi_controller->pacit_ws, 
                              cri, 
                              CONNECTION_BUSY_WRB_HANDLE,
                              ACIT_FALSE);  // ready = FALSE
    }

    return pwrb_handle;
}

ACIT_VOID
hwi_free_wrb_handle(
    PHWI_CONTROLLER_WS      phwi_controller,
    ACIT_UEXACT32           cri,
    PWRB_HANDLE             pwrb_handle)
{
    PHWI_WRB_CONTEXT        pwrb_context = &phwi_controller->wrb_context[cri];

    /* Add to free list */
#ifdef RECYCLE_WRB_INDICES
    LIST_ADDTO_HEAD(&pwrb_handle->link, &pwrb_context->wrb_handle_list);
#else
    LIST_ADDTO_TAIL(&pwrb_handle->link, &pwrb_context->wrb_handle_list);
#endif

    if(pwrb_context->wrb_handles_available == 0)
    {
        iscsi_flag_connection(phwi_controller->pacit_ws,
                              cri,
                              CONNECTION_BUSY_WRB_HANDLE,
                              ACIT_TRUE);   // ready = TRUE
    }

    pwrb_context->wrb_handles_available++;

}

PSGL_HANDLE
hwi_alloc_io_sgl_handle(
    PHWI_CONTROLLER_WS      phwi_controller)
{
    PACIT_DC_LIST           plink;
    PSGL_HANDLE             psgl_handle = ACIT_NULL;

    OSM_ASSERT(IS_LIST_NOT_EMPTY(&phwi_controller->io_sgl_list));

    LIST_REMOVEFROM_HEAD (&plink, 
                          &phwi_controller->io_sgl_list);

    psgl_handle = STRUCT_BASE(SGL_HANDLE,
                                link, 
                                plink);

    OSM_ASSERT(psgl_handle);

    phwi_controller->io_sgl_handles_available--;

   if(!phwi_controller->io_sgl_handles_available)
    {
        /*
         * No more IO WRB handles available, notify OSM that controller is busy
         */
        ap_flag_busy_controller(phwi_controller->pacit_ws, 
                                CONTROLLER_BUSY_IO_SGL_HANDLE);
    }

    return psgl_handle;
}

ACIT_VOID
hwi_free_io_sgl_handle(
    PHWI_CONTROLLER_WS      phwi_controller,
    PSGL_HANDLE             psgl_handle)
{
#ifdef RECYCLE_SGL_INDICES
    LIST_ADDTO_HEAD(&psgl_handle->link, &phwi_controller->io_sgl_list);
#else
    LIST_ADDTO_TAIL(&psgl_handle->link, &phwi_controller->io_sgl_list);
#endif

   if(!phwi_controller->io_sgl_handles_available)
    {
        /*
         * Mark controller as ready
         */
        ap_flag_ready_controller(phwi_controller->pacit_ws, 
                                CONTROLLER_BUSY_IO_SGL_HANDLE);
    }

    phwi_controller->io_sgl_handles_available++;
}

PSGL_HANDLE
hwi_alloc_eh_sgl_handle(
    PHWI_CONTROLLER_WS      phwi_controller)
{
    PACIT_DC_LIST           plink;
    PSGL_HANDLE             psgl_handle = ACIT_NULL;

    OSM_ASSERT(IS_LIST_NOT_EMPTY(&phwi_controller->eh_sgl_list));

    LIST_REMOVEFROM_HEAD (&plink, 
                          &phwi_controller->eh_sgl_list);

    psgl_handle = STRUCT_BASE(SGL_HANDLE,
                                link, 
                                plink);

    OSM_ASSERT(psgl_handle);

    phwi_controller->eh_sgl_handles_available--;

   if(!phwi_controller->eh_sgl_handles_available)
    {
        /*
         * No more IO WRB handles available, notify OSM that controller is busy
         */
        ap_flag_busy_controller(phwi_controller->pacit_ws, 
                                CONTROLLER_BUSY_EH_SGL_HANDLE);
    }

    return psgl_handle;
}

ACIT_VOID
hwi_free_eh_sgl_handle(
    PHWI_CONTROLLER_WS      phwi_controller,
    PSGL_HANDLE             psgl_handle)
{
#ifdef RECYCLE_SGL_INDICES
    LIST_ADDTO_HEAD(&psgl_handle->link, &phwi_controller->eh_sgl_list);
#else
    LIST_ADDTO_TAIL(&psgl_handle->link, &phwi_controller->eh_sgl_list);
#endif

   if(!phwi_controller->eh_sgl_handles_available)
    {
        /*
         * Mark controller as ready
         */
        ap_flag_ready_controller(phwi_controller->pacit_ws, 
                                CONTROLLER_BUSY_EH_SGL_HANDLE);
    }

    phwi_controller->eh_sgl_handles_available++;
}

ACIT_VOID
hwi_get_wrb_mem(
    PHWI_CONTROLLER_WS      phwi_controller,
    ACIT_UEXACT32           cid,
    PACIT_UEXACT32          pwrb_mem_index,
    PACIT_UEXACT32          poffset,
    PACIT_UEXACT32          psize)
{
    *pwrb_mem_index = cid / phwi_controller->max_cxns_per_descriptor;
    *poffset = (cid % phwi_controller->max_cxns_per_descriptor) * phwi_controller->wrb_bytes_per_cxn;
    *psize = phwi_controller->wrb_bytes_per_cxn;
    *pwrb_mem_index += phwi_controller->wrb_mem_index;
}

ACIT_UEXACT32
hwi_write_wrb_sgl(
    PISCSI_WRB              pwrb,
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           bhs_len,
    PACIT_SG_LIST           pdata_sg_list)
{
    OSM_ASSERT(pwrb);
    OSM_ASSERT(pbhs_pa);
    OSM_ASSERT(bhs_len);

    //
    // This routine only updates the WRB. No SGL is assumed present.
    // NOPOuts in response to NOPIns will use this routine
    //

    // Fill in BHS portion of WRB
    ACIT_PUT_LE32(pwrb->iscsi_bhs_addr_lo, pbhs_pa->u.a32.address_lo);
    ACIT_PUT_LE32(pwrb->iscsi_bhs_addr_hi, pbhs_pa->u.a32.address_hi);

    // Fill in the data SG1 and SG2
    if(pdata_sg_list)
    {
        if(pdata_sg_list->number_elements >= 1)
        {
            ACIT_PUT_LE32(pwrb->sge0_addr_lo, pdata_sg_list->sglist[0].address.u.a32.address_lo);
            ACIT_PUT_LE32(pwrb->sge0_addr_hi, pdata_sg_list->sglist[0].address.u.a32.address_hi);
            ACIT_PUT_LE32(pwrb->sge0_len, pdata_sg_list->sglist[0].length);

            pwrb->sge0_last = 1;
        }

        if(pdata_sg_list->number_elements >= 2)
        {
            pwrb->sge0_last = 0;
            ACIT_PUT_LE32(pwrb->sge1_addr_lo, pdata_sg_list->sglist[1].address.u.a32.address_lo);
            ACIT_PUT_LE32(pwrb->sge1_addr_hi, pdata_sg_list->sglist[1].address.u.a32.address_hi);
            ACIT_PUT_LE32(pwrb->sge1_r2t_offset, pdata_sg_list->sglist[0].length);
            ACIT_PUT_LE32(pwrb->sge1_len, pdata_sg_list->sglist[1].length);
        }

        if(pdata_sg_list->number_elements == 2)
            pwrb->sge1_last = 1;
    }


    // Fill in sgl_icd_index
    pwrb->sgl_icd_idx = 0;

    return ACIT_STATUS_SUCCESS;
}


ACIT_UEXACT32
hwi_write_sgl(
    PISCSI_WRB              pwrb,
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           bhs_len,
    PSGL_HANDLE             psgl_handle,
    PACIT_SG_LIST           pdata_sg_list)
{
    ISCSI_SGE               *sgl;
    PACIT_SG_ELEMENT        acit_sgl;
    ACIT_UEXACT32           i, offset = 0;

    OSM_ASSERT(pwrb);
    OSM_ASSERT(psgl_handle);
    OSM_ASSERT(pbhs_pa);
    OSM_ASSERT(bhs_len);

    // Fill in BHS portion of WRB
    ACIT_PUT_LE32(pwrb->iscsi_bhs_addr_lo, pbhs_pa->u.a32.address_lo);
    ACIT_PUT_LE32(pwrb->iscsi_bhs_addr_hi, pbhs_pa->u.a32.address_hi);

    // Fill in the data SG1 and SG2
    if(pdata_sg_list)
    {
        if(pdata_sg_list->number_elements >= 1)
        {
            ACIT_PUT_LE32(pwrb->sge0_addr_lo, pdata_sg_list->sglist[0].address.u.a32.address_lo);
            ACIT_PUT_LE32(pwrb->sge0_addr_hi, pdata_sg_list->sglist[0].address.u.a32.address_hi);
            ACIT_PUT_LE32(pwrb->sge0_len, pdata_sg_list->sglist[0].length);

            pwrb->sge0_last = 1;
        }

        if(pdata_sg_list->number_elements >= 2)
        {
            pwrb->sge0_last = 0;
            ACIT_PUT_LE32(pwrb->sge1_addr_lo, pdata_sg_list->sglist[1].address.u.a32.address_lo);
            ACIT_PUT_LE32(pwrb->sge1_addr_hi, pdata_sg_list->sglist[1].address.u.a32.address_hi);
            ACIT_PUT_LE32(pwrb->sge1_r2t_offset, pdata_sg_list->sglist[0].length);
            ACIT_PUT_LE32(pwrb->sge1_len, pdata_sg_list->sglist[1].length);
        }

        if(pdata_sg_list->number_elements == 2)
            pwrb->sge1_last = 1;
    }


    // Fill in sgl_icd_index
    ACIT_PUT_LE32(pwrb->sgl_icd_idx, psgl_handle->sgl_index);

    //
    // The first FR must still point to the area where the
    // PDU templates are located
    //
    sgl = (PISCSI_SGE)psgl_handle->pfrag;
    ACIT_PUT_LE32(sgl->len, bhs_len);
    ACIT_PUT_LE32(sgl->addr_lo, pbhs_pa->u.a32.address_lo);
    ACIT_PUT_LE32(sgl->addr_hi, pbhs_pa->u.a32.address_hi);
       
    // Check if the command has additional SGE's
    if (pdata_sg_list && pdata_sg_list->number_elements) 
    {
        sgl++;

        //
        // Get pointer to the ACIT scatter gather list and fill
        // in rest of the FRs
        //

        acit_sgl = pdata_sg_list->sglist;

        for(i = 0; i < pdata_sg_list->number_elements; i++)
        { 
            //
            // do a increment of SGL before we start in order to skip over the first
            // fragment
            //
            ++sgl;

            OSM_ASSERT (acit_sgl->length > 0);
            OSM_ASSERT (acit_sgl->address.u.a64.address != 0);

            ACIT_PUT_LE32(sgl->len, acit_sgl->length);
            ACIT_PUT_LE32(sgl->addr_lo, acit_sgl->address.u.a32.address_lo);
            ACIT_PUT_LE32(sgl->addr_hi, acit_sgl->address.u.a32.address_hi);
            ACIT_PUT_LE32(sgl->sge_offset, offset);
            sgl->last_sge = 0;

            offset+=sgl->len;
            ++acit_sgl;
        }

        sgl->last_sge = 1;
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_write_large_sgl(
    PISCSI_WRB              pwrb,
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           bhs_len,
    PSGL_HANDLE             psgl_handle,
    PACIT_SG_LIST           pdata_sg_list)
{
    ISCSI_SGE               *sgl;
    PACIT_SG_ELEMENT        acit_sgl;
    ACIT_UEXACT32           i, offset = 0;

    OSM_ASSERT(pwrb);
    OSM_ASSERT(psgl_handle);
    OSM_ASSERT(pbhs_pa);
    OSM_ASSERT(bhs_len);

    // Fill in BHS portion of WRB
    ACIT_PUT_LE32(pwrb->iscsi_bhs_addr_lo, pbhs_pa->u.a32.address_lo);
    ACIT_PUT_LE32(pwrb->iscsi_bhs_addr_hi, pbhs_pa->u.a32.address_hi);

    // Fill in data SG1 and SG2
    if(pdata_sg_list)
    {
        if(pdata_sg_list->number_elements >= 1)
        {
            ACIT_PUT_LE32(pwrb->sge0_addr_lo, pdata_sg_list->sglist[0].address.u.a32.address_lo);
            ACIT_PUT_LE32(pwrb->sge0_addr_hi, pdata_sg_list->sglist[0].address.u.a32.address_hi);

            if(pdata_sg_list->sglist[0].length > BE_MAX_BYTES_PER_FRAG)
            {
                ACIT_PUT_LE32(pwrb->sge0_len, BE_MAX_BYTES_PER_FRAG);

                ACIT_PUT_LE32(pwrb->sge1_addr_lo, pdata_sg_list->sglist[0].address.u.a32.address_lo +
                                                    BE_MAX_BYTES_PER_FRAG);
                ACIT_PUT_LE32(pwrb->sge1_addr_hi, pdata_sg_list->sglist[0].address.u.a32.address_hi);
                ACIT_PUT_LE32(pwrb->sge1_r2t_offset, BE_MAX_BYTES_PER_FRAG);

                if((pdata_sg_list->sglist[0].length - BE_MAX_BYTES_PER_FRAG) > BE_MAX_BYTES_PER_FRAG)
                {
                    ACIT_PUT_LE32(pwrb->sge1_len, BE_MAX_BYTES_PER_FRAG);
                }
                else
                {
                    ACIT_PUT_LE32(pwrb->sge1_len, pdata_sg_list->sglist[0].length - BE_MAX_BYTES_PER_FRAG);
                }

                goto fill_rest;
            }
            else
            {
                ACIT_PUT_LE32(pwrb->sge0_len, pdata_sg_list->sglist[0].length);
                pwrb->sge0_last = 1;
            }
        }

        if(pdata_sg_list->number_elements >= 2)
        {
            pwrb->sge0_last = 0;
            ACIT_PUT_LE32(pwrb->sge1_addr_lo, pdata_sg_list->sglist[1].address.u.a32.address_lo);
            ACIT_PUT_LE32(pwrb->sge1_addr_hi, pdata_sg_list->sglist[1].address.u.a32.address_hi);
            ACIT_PUT_LE32(pwrb->sge1_r2t_offset, pdata_sg_list->sglist[0].length);

            if(pdata_sg_list->sglist[1].length > BE_MAX_BYTES_PER_FRAG)
            {
                ACIT_PUT_LE32(pwrb->sge1_len, BE_MAX_BYTES_PER_FRAG);
            }
            else
            {
                ACIT_PUT_LE32(pwrb->sge1_len, pdata_sg_list->sglist[1].length);

                if(pdata_sg_list->number_elements == 2)
                    pwrb->sge1_last = 1;
            }
        }
    }

fill_rest:

    // Fill in sgl_icd_index
    ACIT_PUT_LE32(pwrb->sgl_icd_idx, psgl_handle->sgl_index);

    //
    // The first FR must still point to the area where the
    // PDU templates are located
    //
    sgl = (PISCSI_SGE)psgl_handle->pfrag;
    ACIT_PUT_LE32(sgl->len, bhs_len);
    ACIT_PUT_LE32(sgl->addr_lo, pbhs_pa->u.a32.address_lo);
    ACIT_PUT_LE32(sgl->addr_hi, pbhs_pa->u.a32.address_hi);
       
    // Check if the command has additional SGE's
    if (pdata_sg_list && pdata_sg_list->number_elements) 
    {
        sgl++;

        //
        // Get pointer to the ACIT scatter gather list and fill
        // in rest of the FRs
        //

        acit_sgl = pdata_sg_list->sglist;

        for(i = 0; i < pdata_sg_list->number_elements; i++)
        { 
            //
            // do a increment of SGL before we start in order to skip over the first
            // fragment
            //
            ++sgl;

            OSM_ASSERT (acit_sgl->length > 0);
            OSM_ASSERT (acit_sgl->address.u.a64.address != 0);

            ACIT_PUT_LE32(sgl->addr_lo, acit_sgl->address.u.a32.address_lo);
            ACIT_PUT_LE32(sgl->addr_hi, acit_sgl->address.u.a32.address_hi);
            ACIT_PUT_LE32(sgl->sge_offset, offset);
            sgl->last_sge = 0;

            //
            // The hardware has a limitation with DMA in that only 64KB or less
            // can be in a frag. Split anything that is more than this limit
            //
            if(acit_sgl->length > BE_MAX_BYTES_PER_FRAG)
            {
                ACIT_UEXACT32   bytes_remaining, len;
                ACIT_BUS_ADDRESS address;

                address.u.a32.address_lo = sgl->addr_lo;
                address.u.a32.address_hi = sgl->addr_hi;

                // limit the len to max allowed
                ACIT_PUT_LE32(sgl->len, BE_MAX_BYTES_PER_FRAG);

                bytes_remaining = acit_sgl->length - sgl->len;

                do
                {
                    offset+=sgl->len;
                    ++sgl;

                    if(bytes_remaining > BE_MAX_BYTES_PER_FRAG)
                        len = BE_MAX_BYTES_PER_FRAG;
                    else
                        len = bytes_remaining;

                    ACIT_PUT_LE32(sgl->len, len);
                    ACIT_PUT_LE32(sgl->addr_lo, address.u.a32.address_lo + 
                                                (acit_sgl->length - bytes_remaining));
                    ACIT_PUT_LE32(sgl->addr_hi, address.u.a32.address_hi);
                    ACIT_PUT_LE32(sgl->sge_offset, offset);
                    sgl->last_sge = 0;

                    bytes_remaining = bytes_remaining - len;

                } while(bytes_remaining > 0);
            }
            else
            {
                ACIT_PUT_LE32(sgl->len, acit_sgl->length);
            }

            offset+=sgl->len;
            ++acit_sgl;
        }

        sgl->last_sge = 1;
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
hwi_init_wrb_handle(
    PACIT_CONTROLLER_WS      pcontroller)
{
    PHWI_CONTROLLER_WS      phwi;
    PACIT_CONTROLLER_WS     pc = pcontroller;
    PACIT_PARAMETERS        p = &pc->params;
    ACIT_UEXACT32           wrbh_desc, wrb_desc;
    ACIT_UEXACT32           wrb_index;
    ACIT_UEXACT32           index, wrbh_total, wrb_total, cxns;
    PWRB_HANDLE             pwrb_handle = ACIT_NULL;
    PISCSI_WRB              pwrb = ACIT_NULL;
    PHWI_WRB_CONTEXT        pwrb_context;

    phwi = GET_HWI_CONTROLLER_WS(pcontroller);

    /*
     * Split the WRB handles across the pwrb_handle_base
     */
    wrbh_desc = phwi->wrbh_mem_index;
    OSM_ASSERT(wrbh_desc == pc->mem_descriptors[wrbh_desc].index);
    pwrb_handle = (PWRB_HANDLE)pc->mem_descriptors[wrbh_desc].virtual_address;

    wrb_desc = phwi->wrb_mem_index;
    OSM_ASSERT(wrb_desc == pc->mem_descriptors[wrb_desc].index);
    pwrb = (PISCSI_WRB)pc->mem_descriptors[wrb_desc].virtual_address;

    wrbh_total = 0;
    wrb_total = 0;
    cxns = 0;

    for(index = 0; index < p->num_cxn; index++)
    {
        pwrb_context = &phwi->wrb_context[index];

        LIST_INIT_HEADER(&pwrb_context->wrb_handle_list);
        LIST_INIT_HEADER(&pwrb_context->wrb_handle_drvr_list);
        pwrb_context->plast_wrb = ACIT_NULL;

        pwrb_context->pwrb_handle_base = pwrb_handle;

        // Initialize this to 0 since after a hibernation resume this is not 
        // guaranteed to be 0
        pwrb_context->wrb_handles_available = 0;

        for(wrb_index = 0; wrb_index < p->num_cmd_per_cxn; wrb_index++)
        {
            pwrb_handle->wrb_index = (ACIT_UEXACT16)wrb_index;
            pwrb_handle->pwrb = pwrb;

            LIST_ADDTO_TAIL(&pwrb_handle->link,
                            &pwrb_context->wrb_handle_list);

            pwrb_context->wrb_handles_available++;

            pwrb_handle ++;
            pwrb++;
            OSM_ASSERT(wrbh_total * sizeof(WRB_HANDLE) <= ACIT_CFG_MAX_CONTIGUOUS_MEM);
            OSM_ASSERT(wrb_total * sizeof(ISCSI_WRB) <= ACIT_CFG_MAX_CONTIGUOUS_MEM);
            wrbh_total ++;
            wrb_total ++;
        }

        cxns++;

        if(index < p->num_cxn - 1)
        {
            if(wrbh_total  >= phwi->max_wrbh_per_descriptor)
            {
                wrbh_desc++;
                pwrb_handle = (PWRB_HANDLE)pc->mem_descriptors[wrbh_desc].virtual_address;
                wrbh_total = 0;
            }

            if(cxns >= phwi->max_cxns_per_descriptor)
            {
                wrb_desc++;
                pwrb = (PISCSI_WRB)pc->mem_descriptors[wrb_desc].virtual_address;
                wrb_total = 0;
                cxns = 0;
            }
        }
    }
}

ACIT_VOID
hwi_init_sgl_handle(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PHWI_CONTROLLER_WS      phwi;
    PACIT_CONTROLLER_WS     pc = pcontroller;
    PACIT_PARAMETERS        p = &pc->params;
    ACIT_UEXACT16           index, sgl_total, sgl_desc, sglh_total;
    PISCSI_SGE              pfrag;
    PSGL_HANDLE             psgl_handle;

    phwi = GET_HWI_CONTROLLER_WS(pcontroller);

    psgl_handle = (PSGL_HANDLE)pc->mem_descriptors[HWI_MEM_SGLH].virtual_address;

    sgl_desc = (ACIT_UEXACT16)phwi->frag_mem_index;
    OSM_ASSERT(sgl_desc == pc->mem_descriptors[sgl_desc].index);
    pfrag = (PISCSI_SGE)pc->mem_descriptors[sgl_desc].virtual_address;

    LIST_INIT_HEADER(&phwi->io_sgl_list);
    LIST_INIT_HEADER(&phwi->eh_sgl_list);

    phwi->psgl_handle_base = psgl_handle;
    phwi->io_sgl_handles_available = 0;
    phwi->eh_sgl_handles_available = 0;

    sgl_total = 0;
    sglh_total = 0;

    for(index = 0; index < p->icds_per_controller; index++)
    {
        psgl_handle->sgl_index = pcontroller->fw_config.iscsi_icd_start + index;
        psgl_handle->pfrag = pfrag;

        if(index < p->ios_per_controller)
        {
            LIST_ADDTO_TAIL(&psgl_handle->link,
                            &phwi->io_sgl_list);

            phwi->io_sgl_handles_available++;
        }
        else
        {
            LIST_ADDTO_TAIL(&psgl_handle->link,
                            &phwi->eh_sgl_list);

            phwi->eh_sgl_handles_available++;
        }

        psgl_handle ++;
        pfrag = (PISCSI_SGE)((ACIT_PTR)pfrag + (p->num_sge * sizeof(ISCSI_SGE)));
        sgl_total ++;
        sglh_total ++;

        if(sgl_total >= phwi->max_sgl_per_descriptor)
        {
            if(index < p->icds_per_controller - 1)
            {
                sgl_desc++;
                pfrag = (PISCSI_SGE)pc->mem_descriptors[sgl_desc].virtual_address;
                sgl_total = 0;
            }
            else
            {
                break;
            }
        }
        OSM_ASSERT(sgl_total * sizeof(ISCSI_SGE) <= ACIT_CFG_MAX_CONTIGUOUS_MEM);
        OSM_ASSERT(sglh_total * sizeof(SGL_HANDLE) <= ACIT_CFG_MAX_CONTIGUOUS_MEM);
    }

    hwi_unitt_init_sgl_handle(pcontroller);
}

ACIT_UEXACT32
hwi_reserve_nop_sgl(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PHWI_CONTROLLER_WS      phwi;
    PSGL_HANDLE             psgl_handle;

    phwi = GET_HWI_CONTROLLER_WS(pcontroller);

    psgl_handle = hwi_alloc_eh_sgl_handle(phwi);

    if(psgl_handle)
    {
        phwi->nop_icd = psgl_handle->sgl_index;
        return ACIT_STATUS_SUCCESS;
    }

    return ACIT_STATUS_ERROR;

}

ACIT_VOID
hwi_unitt_init_sgl_handle(
    PACIT_CONTROLLER_WS pcontroller)
{
    PHWI_CONTROLLER_WS  phwi;
    PSGL_HANDLE         psgl_handle;
    PACIT_DC_LIST       flink;
    ACIT_UEXACT32       num_handles = 0;

    phwi = GET_HWI_CONTROLLER_WS(pcontroller);

    OSM_ASSERT(IS_LIST_NOT_EMPTY(&phwi->io_sgl_list));

    flink = phwi->io_sgl_list.next;

    while(flink != &phwi->io_sgl_list)
    {
        psgl_handle = STRUCT_BASE(SGL_HANDLE,
                                  link,
                                  flink);

        OSM_ASSERT(psgl_handle->sgl_index == num_handles +
            pcontroller->fw_config.iscsi_icd_start);
        
        num_handles++;
        flink = flink->next;
    }

    OSM_ASSERT(num_handles == phwi->io_sgl_handles_available);
}
