/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

    iscsiapi.c - iSCSI Layer Interface Routines

    Purpose:

        File for internal iSCSI implementation.
        

--*/

#include "acitpch.h"

/*
 * File for private iSCSI functions and structures
 */


ACIT_UEXACT32
iscsi_init_connection_information (
    PACIT_TARGET_WS         ptarget,
    PCONNECTION_DATA        pcxn_data,
    ACIT_UEXACT32           is_first_cxn)
{
    PACIT_CONTROLLER_WS     pcontroller;
    PISCSI_CONTROLLER_WS    piscsi_controller;
    PISCSI_TARGET_WS        piscsi_target;
    PISCSI_CONNECTION       pcxn = ACIT_NULL;
    ACIT_UEXACT32           i;

//#ifdef FPGA_ARM
//    ACIT_UEXACT32           cid = 0;
//#endif
        
    pcontroller = ptarget->pcontroller;
    piscsi_controller = GET_ISCSI_CONTROLLER_WS(pcontroller);
    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);

    for(i = 0; i < MAX_CXNS_PER_SESSION; i++)
    {
        if(is_first_cxn || ((piscsi_target->pcxn_list[i]) && 
          (piscsi_target->pcxn_list[i]->connection_handle == pcxn_data->connection_handle)))
        {
            pcxn = piscsi_target->pcxn_list[i];
            pcxn->piscsi_target_ws = piscsi_target;
            pcxn->status = ISCSI_CONN_STATUS_VALID;
            pcxn->connection_handle = pcxn_data->connection_handle;
            pcxn->iscsi_cid = (ACIT_UEXACT16)pcxn_data->conn_iscsi_cid;
            pcxn->login_options.digest.header = (ACIT_UEXACT16)pcxn_data->header_digest;
            pcxn->login_options.digest.data = (ACIT_UEXACT16)pcxn_data->data_digest;

            if(pcxn->login_options.digest.header ||
               pcxn->login_options.digest.data)
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Target 0x%p Cxn 0x%p Hdr Digest %d Data Digest %d\n",
                                    ptarget,
                                    pcxn,
                                    pcxn->login_options.digest.header,
                                    pcxn->login_options.digest.data));
            }

            OSM_ASSERT(pcxn_data->conn_status & MGMT_CONN_STATUS_OPENED);

            piscsi_target->last_cxn = (ACIT_UEXACT16)i;
            
            break;
        }
        else 
        {
            OSM_ASSERT(0);
        }

    }

    // get cid from ARM fw
//#ifdef FPGA_ARM

    // estimated_throughput is overloaded with CID from ARM fw
//    cid = pcxn_data->estimated_throughput;
//    pcxn->cid = (ACIT_UEXACT16)cid;
//    pcxn->cri = pcxn->cid;  // this is needed since the HWI is expecting

    // get cid from ARM fw end
//#endif

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Connection structure is 0x%p\n", pcxn));
    OSM_ASSERT(pcxn);

    if(pcxn->async_pending)
    {
        hwi_deliver_pending_async_pdu(pcontroller, pcxn->cri);
        
        pcxn->async_pending = 0;
        piscsi_controller->async_pending--;
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Target 0x%p Session %d Cid 0x%x\n", 
                    ptarget, piscsi_target->session_handle, pcxn->cid));

    return ISCSI_STATUS_SUCCESS;
}


ACIT_UEXACT32
iscsi_reset_connection(
    PACIT_CONTROLLER_WS     pcontroller,
    PISCSI_CONNECTION       pcxn)
{
    ACIT_UEXACT32           status = 0;
    if(ACIT_IS_BIT_CLR(pcxn->status, ISCSI_CONN_STATUS_INVALIDATE_PENDING) &&
       ACIT_IS_BIT_CLR(pcxn->status, ISCSI_CONN_STATUS_INVALIDATED))
    {
        status = iscsi_invalidate_connection(pcontroller,
                                    pcxn->piscsi_target_ws,
                                    pcxn,
                                    ACIT_TRUE); // issue reset

        if(status == ACIT_STATUS_SUCCESS)
        {
            ACIT_BIT_SET(pcxn->status, ISCSI_CONN_STATUS_TCP_RESET_PENDING);
        }
    }
    else
    {
        // resetting an already reset connection?
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Attempting reset on already reset \
                   connection CID 0x%x session 0x%x target 0x%p\n",
                   pcxn->cid,
                   pcxn->piscsi_target_ws->session_handle,
                   pcxn->piscsi_target_ws->pacit_ws));
    }

    return status;
}

ACIT_UEXACT32
iscsi_delete_session_info(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget)
{
    PISCSI_CONTROLLER_WS    piscsi_controller;
    PISCSI_TARGET_WS        piscsi_target;
    ACIT_UEXACT32           i;
    PACIT_IOCTL_REQUEST     pioctl_request;

    /*
     * This API is called when MGMT posts completion for session cleanup
     */
    piscsi_controller = GET_ISCSI_CONTROLLER_WS(pcontroller);
    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);

    /* All connections are gone, delete the target */

    for(i = 0; i < piscsi_target->num_connections; i ++)
    {
        OSM_ASSERT(piscsi_target->pcxn_list[i]);
        piscsi_target->pcxn_list[i]->status = ISCSI_CONN_STATUS_NULL;
        piscsi_target->pcxn_list[i]->piscsi_target_ws = ACIT_NULL;
        iscsi_free_cxn (pcontroller, piscsi_target->pcxn_list[i]);
        piscsi_target->pcxn_list[i] = ACIT_NULL;
    }

    piscsi_target->num_connections = 0; 
    piscsi_target->login_options.max_connections = 0;

    pioctl_request = piscsi_target->mgmt_ioctl_request;

    ptarget->piscsi_ws = ACIT_NULL;
    ap_remove_target(pcontroller, ptarget);

    if (pioctl_request)
        mgmt_return_ioctl_request(pcontroller, 
                                  pioctl_request,
                                  ACIT_STATUS_SUCCESS);

    return ACIT_STATUS_SUCCESS;
}

INLINE
ACIT_UEXACT32
iscsi_update_sn(
    PACIT_CONTROLLER_WS     pcontroller,
    PISCSI_TARGET_WS        piscsi_target,
    ACIT_UEXACT32           max_cmd_sn,
    ACIT_UEXACT32           exp_cmd_sn)
{
    PACIT_TARGET_WS         ptarget = piscsi_target->pacit_ws;

    /* If CQ completion happens out of order in that case local
     * exp_cmd_sn and max_cmd_sn would have got updated in the 
     * previous completion in that do not update any local values
     * and do not make target ready.
     */
    if (SNA_FORMER_IS_LESSER(max_cmd_sn, piscsi_target->max_cmd_sn))
        return ISCSI_STATUS_SUCCESS;

    if (SNA_FORMER_IS_LESSER(exp_cmd_sn, piscsi_target->exp_cmd_sn))
        return ISCSI_STATUS_SUCCESS;

    /* Check if the max_cmd_sn + 1 > exp_cmd_sn */
    if (SNA_FORMER_IS_GREATER(SNA_ADD(max_cmd_sn, 1), exp_cmd_sn))
    {
        if (SNA_FORMER_IS_GREATER(exp_cmd_sn, piscsi_target->exp_cmd_sn))
            piscsi_target->exp_cmd_sn = exp_cmd_sn;
        if (SNA_FORMER_IS_GREATER(max_cmd_sn, piscsi_target->max_cmd_sn))
            piscsi_target->max_cmd_sn = max_cmd_sn;

        if(ACIT_IS_BIT_SET(ptarget->busy, TARGET_BUSY_ISCSI_WINDOW))
        {
            if(!SNA_FORMER_IS_GREATER(piscsi_target->cmd_sn,
                                       piscsi_target->max_cmd_sn))
            {
                ap_flag_ready_target(   pcontroller,
                                        ptarget, 
                                        TARGET_BUSY_ISCSI_WINDOW);

                // stop the command window timer
                iscsi_stop_cmd_window_timer(pcontroller, ptarget);

            }
        }
    }
    /* Mark target as busy */
    else
    {
        if(ACIT_IS_BIT_CLR(ptarget->busy, TARGET_BUSY_ISCSI_WINDOW))
        {
            ACIT_UEXACT32   ext_error[3];

            ext_error[0] = piscsi_target->session_handle;
            ext_error[1] = max_cmd_sn;
            ext_error[2] = exp_cmd_sn;

            /* If we came here means target window should have been closed in the 
            * start IO path itself. We should not close window in the completion
            * path.
            */

            // extended error will contain the session id of the target maxcmdsn 
            // and expcmdsn received on the wire.

            ap_log_osm_event(pcontroller, 
                             UNI_CFG_LOG_LEVEL_ERROR,
                             UNI_SUBMODULE_ACIT_ISCSI,
                             UNI_PHASE_RECOVERY,
                             3,
                             ACIT_EC_ISCSI_CMDSN_ERROR,
                             0, ACIT_NULL,
                             &ext_error[0]);

            //OSM_ASSERT(ACIT_IS_BIT_SET(ptarget->busy, TARGET_BUSY_ISCSI_WINDOW));
            ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_IO_ERROR);

            return (iscsi_recover_session(pcontroller, ptarget, ACIT_NULL));
        }
    }
   return ISCSI_STATUS_SUCCESS;
}


ACIT_UEXACT32
iscsi_update_sn_by_offset(
    PACIT_CONTROLLER_WS     pcontroller,
    PISCSI_TARGET_WS        piscsi_target,
    ACIT_UEXACT32           max_cmd_sn_offset,
    ACIT_UEXACT32           exp_cmd_sn)
{
    //ACIT_UEXACT32           max_cmd_sn = SNA_ADD(max_cmd_sn_offset, exp_cmd_sn);
    /* With new EP firmware change the max_cmd_sn_offset is not passed in the
     * CQ but it passes the window. Which is 8 bit value thus no need to do serial
     * arithmatic instead do simple arithmatic to determine the max_cmd_sn.
     */
    ACIT_UEXACT32           max_cmd_sn;

    max_cmd_sn = max_cmd_sn_offset + exp_cmd_sn - 1;
    iscsi_update_sn(pcontroller, piscsi_target, max_cmd_sn, exp_cmd_sn);

    return ISCSI_STATUS_SUCCESS;
}


ACIT_UEXACT16
iscsi_get_conn_index(
    PISCSI_TARGET_WS        piscsi_target,
    PISCSI_CONNECTION       pcxn)
{
    ACIT_UEXACT16   i;
    for (i = 0; i < piscsi_target->num_connections; i++)
    {
        OSM_ASSERT(piscsi_target->pcxn_list[i]); 
        if (piscsi_target->pcxn_list[i] == pcxn)
            return i;
    }

    OSM_ASSERT(0);
    return 0xFFFF;
}

ACIT_UEXACT16
iscsi_get_cfg_flag(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget)
{
    PISCSI_TARGET_WS        piscsi_target;
    ACIT_UEXACT16           flags = 0;
    
    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);
 
    if((ACIT_IS_BIT_SET(pcontroller->busy, CONTROLLER_BUSY_SHUTDOWN_IN_PROGRESS)) ||
       (ACIT_IS_TARGET_BUSY(ptarget, TARGET_BUSY_SESSION_RECOVERY)))
    {
        flags = IOCTL_ISCSI_SESSION_SAVE_CFG_ON_FLASH;
    }
    else if (ACIT_IS_TARGET_BUSY(ptarget, TARGET_BUSY_SESSION_LOGOUT))
    {
        flags = IOCTL_ISCSI_SESSION_DEL_CFG_FROM_FLASH;
    }
    else if(ACIT_IS_TARGET_BUSY(ptarget, TARGET_BUSY_REMOVE_CONNECTION))
    {
        flags = IOCTL_ISCSI_SESSION_DEL_CFG_FROM_FLASH;
    }
    else
    {
        OSM_ASSERT(0);
        // why are we here?
    }

    return flags;
}
