/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

        mgmtapi.c

    Purpose:

        source file for Management module
                

--*/

#include "acitpch.h"

ACIT_UEXACT32
mgmt_get_ws_sizes(
    PACIT_UEXACT32          pcontroller_ws_size,
    PACIT_UEXACT32          ptarget_ws_size,
    PACIT_UEXACT32          plun_ws_size,
    PACIT_UEXACT32          pio_ws_size,
    PACIT_UEXACT32          ptmf_ws_size)
{
    *pcontroller_ws_size = sizeof(MGMT_CONTROLLER_WS);
    *ptarget_ws_size = sizeof(MGMT_TARGET_WS);
    *plun_ws_size = 0;
    *pio_ws_size = 0;
    *ptmf_ws_size = 0;

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
mgmt_init_mem_descriptors(
    PACIT_CONTROLLER_WS         pcontroller)
{
    mgmt_init_const_mem_desc(pcontroller);

    return ACIT_STATUS_SUCCESS;
}


ACIT_UEXACT32
mgmt_init_controller_ws(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PMGMT_CONTROLLER_WS pmgmt_controller;

    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);
    OSM_MEMSET(pmgmt_controller , 0, sizeof(MGMT_CONTROLLER_WS));

    pmgmt_controller->pacit_ws = pcontroller;

    LIST_INIT_HEADER(&pmgmt_controller->ioctl_buffer_free_list);
    LIST_INIT_HEADER(&pmgmt_controller->ioctl_buffer_busy_list);
    LIST_INIT_HEADER(&pmgmt_controller->ioctl_handle_free_list);
    LIST_INIT_HEADER(&pmgmt_controller->ioctl_handle_busy_list);

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
mgmt_init_const_mem_desc(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PACIT_CONTROLLER_WS     pc = pcontroller;
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    PACIT_PARAMETERS        p = &pc->params;

    const ACIT_UEXACT32 a8 = ACIT_ALIGN_8BYTE;
    const ACIT_UEXACT32 non_cached = ACIT_MEM_CATEGORY_NON_CACHED;

    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pc);

    ap_set_mem_const (pc, (PUEXACT8)"ioctl_handle", MGMT_MEM_IOCTL_HANDLE, p->ioctls_per_ctrl *sizeof(MGMT_IOCTL_HANDLE),          a8, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"ioctl_buffer", MGMT_MEM_IOCTL_BUFFER, p->ioctls_per_ctrl * sizeof(MGMT_INTERNAL_IOCTL_BUFFER_HANDLE),a8, non_cached);

    return ACIT_STATUS_SUCCESS;
}


ACIT_UEXACT32
mgmt_init_controller(
    PACIT_CONTROLLER_WS     pcontroller)
{

    PMGMT_CONTROLLER_WS pmgmt_controller;
    ACIT_U32 i;
    PMGMT_IOCTL_HANDLE pioctl_handle;
    PMGMT_INTERNAL_IOCTL_BUFFER_HANDLE pioctl_buf_handle;

    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);

    pmgmt_controller->pioctl_handle_base = (PMGMT_IOCTL_HANDLE)
                                            pcontroller->mem_descriptors[MGMT_MEM_IOCTL_HANDLE].virtual_address;
    pmgmt_controller->pioctl_buffer_base = (PMGMT_INTERNAL_IOCTL_BUFFER_HANDLE)
                                            pcontroller->mem_descriptors[MGMT_MEM_IOCTL_BUFFER].virtual_address;

    /*
     * Initialize all the IOCTL Handles
     */
    LIST_INIT_HEADER(&pmgmt_controller->ioctl_buffer_free_list);
    LIST_INIT_HEADER(&pmgmt_controller->ioctl_buffer_busy_list);
    LIST_INIT_HEADER(&pmgmt_controller->ioctl_handle_free_list);
    LIST_INIT_HEADER(&pmgmt_controller->ioctl_handle_busy_list);

    pioctl_handle = (PMGMT_IOCTL_HANDLE) pmgmt_controller->pioctl_handle_base;
    pmgmt_controller->ioctl_handles_available = 0;

    for(i = 0; i < pcontroller->params.ioctls_per_ctrl; i++)
    {
        LIST_ADDTO_TAIL(&pioctl_handle->controller_link,
                        &pmgmt_controller->ioctl_handle_free_list);

        pmgmt_controller->ioctl_handles_available ++;
        pioctl_handle->index = i+1;
        pioctl_handle->flags = MGMT_IOCTLH_FREE;
        pioctl_handle ++;
    }

    /*
     * Initialize all the IOCTL Buffers
     */
    pioctl_buf_handle = (PMGMT_INTERNAL_IOCTL_BUFFER_HANDLE) pmgmt_controller->pioctl_buffer_base;
    pmgmt_controller->ioctl_buffers_available = 0;

    for(i = 0; i < pcontroller->params.ioctls_per_ctrl; i++)
    {
        LIST_ADDTO_TAIL(&pioctl_buf_handle->controller_link,
                        &pmgmt_controller->ioctl_buffer_free_list);

        pmgmt_controller->ioctl_buffers_available ++;
        pioctl_buf_handle ++;
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
mgmt_init_target_ws(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           session_handle,
    ACIT_UEXACT16           connection_handle,
    ACIT_UEXACT16           cid,
    ACIT_UEXACT16           cq_id,
    ACIT_UEXACT16           def_hdr_id,
    ACIT_UEXACT16           def_data_id,
    PPHYS_ADDR              ptemplate_address,
    ACIT_UEXACT32           sync)
{
    PMGMT_TARGET_WS         pmgmt_target;
    MGMT_STATUS             status;
    OFFLOAD_SESSION_IN      offload_session_in = {{{0,0,0,0,0,{{{0,0}}}}}};

    pmgmt_target = GET_MGMT_TARGET_WS(ptarget);
    OSM_MEMSET(pmgmt_target , 0, sizeof(MGMT_TARGET_WS));

    offload_session_in.session_handle = session_handle;
    offload_session_in.connection_handle = connection_handle;
    offload_session_in.cid = cid;
    offload_session_in.cq_id = cq_id;
    offload_session_in.defq_id = def_hdr_id;
    offload_session_in.hdr_ring_id = def_hdr_id;
    offload_session_in.data_ring_id = def_data_id;
    offload_session_in.dataout_template_pa.lo = ptemplate_address->lo;
    offload_session_in.dataout_template_pa.hi = ptemplate_address->hi;

    status = mgmt_offload_session(pcontroller, 
                                  ptarget, 
                                  &offload_session_in, 
                                  sync);

    if(status != MGMT_STATUS_SUCCESS)
    {
        if(MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED == status)
            return ACIT_STATUS_INSUFFICIENT_RESOURCE;

        return ACIT_STATUS_ERROR;
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
mgmt_init_lun_ws(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun)
{
    return ACIT_STATUS_SUCCESS;
}


ACIT_UEXACT32
mgmt_init_io_ws(
    PACIT_CONTROLLER_WS     pcontroller,    
    PACIT_IO_WS             pio)
{
    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
mgmt_init_tmf_ws(
    PACIT_CONTROLLER_WS     pcontroller,    
    PACIT_TMF_WS            ptmf)
{
    return ACIT_STATUS_SUCCESS;
}


ACIT_UEXACT32
mgmt_stop_controller(
    PACIT_CONTROLLER_WS     pcontroller, 
    ACIT_UEXACT32           stop_type)
{
    // Sava the non persistent target if coming from hibernation shutdown.
    if (stop_type == ACIT_STOP_TYPE_POWER_MGMT)
        mgmt_save_nonpersistent_targets(pcontroller);

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
mgmt_discover_targets(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           is_polled)
{
    MGMT_STATUS             mgmt_status;

    mgmt_status = mgmt_get_all_session_id(pcontroller, is_polled);

    if(mgmt_status != ACIT_STATUS_SUCCESS)
    {
        return ACIT_STATUS_ERROR;
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
mgmt_discover_boot_target(
    PACIT_CONTROLLER_WS     pcontroller)
{
    MGMT_STATUS             mgmt_status;

    mgmt_status = mgmt_get_boot_session(pcontroller, ACIT_TRUE);

    if(mgmt_status != ACIT_STATUS_SUCCESS)
    {
        return ACIT_STATUS_ERROR;
    }

    return ACIT_STATUS_SUCCESS;
}

MGMT_STATUS
mgmt_get_link_status(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    PMGMT_IOCTL_HANDLE      pioctl_handle;
    MGMT_STATUS             mgmt_status;
    ACIT_UEXACT32           flags, expected_data_size;
    PBE_LINK_STATUS         plink_status = ACIT_NULL;
    PBE_LINK_STATUS_V1      plink_status_v1 = ACIT_NULL;

    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);

    flags = MGMT_SET_ALLOCATION_FLAGS(MGMT_IOCTL_POLLED,
                                      MGMT_IOCTL_INTERNAL,
                                      MGMT_IOCTL_RESPONSE_BUFFER);

    expected_data_size = (sizeof(IOCTL_COMMON_NTWK_LINK_STATUS_QUERY) -
                                               sizeof(IOCTL_HEADER));

    pioctl_handle = mgmt_alloc_ioctl_resources(pcontroller,
                                               ACIT_NULL,
                                               expected_data_size,
                                               flags);

    if (pioctl_handle == ACIT_NULL)
    {
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }

    pioctl_handle->timeout = MGMT_INTERNAL_IOCTL_DEFAULT_TIMEOUT;

    mgmt_fill_ioctl_hdr(pioctl_handle,
                        IOCTL_SUBSYSTEM_COMMON,
                        OPCODE_COMMON_NTWK_LINK_STATUS_QUERY,
                        0, // domain 0
                        expected_data_size,
                        pioctl_handle->timeout);

    mgmt_status = mgmt_fill_ioctl_request(pcontroller, pioctl_handle);

    if(pcontroller->state.asic_generation >= ASIC_GEN_BE3)
    {
        mgmt_fill_ioctl_hdr_version(pioctl_handle, 1);
    }
    
    if(mgmt_status == MGMT_STATUS_SUCCESS)
    {
        mgmt_status = mgmt_submit_ioctl(pcontroller, pioctl_handle);
    }

    if(pcontroller->state.asic_generation < ASIC_GEN_BE3)
    {
        plink_status = (PBE_LINK_STATUS)pioctl_handle->pmgmt_buffer;
        
        if(PHY_LINK_SPEED_ZERO == plink_status->mac_speed)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("Physical Link down, Port: %c\n"),
                (plink_status->physical_port == NTWK_PORT_A ? 'A' : 'B')));
            pcontroller->link_down[0] = ACIT_TRUE;

            ap_log_osm_event(pcontroller,
                             UNI_CFG_LOG_LEVEL_CRITICAL,
                             UNI_SUBMODULE_ACIT_GENERIC,
                             UNI_PHASE_DONT_CARE,
                             0,
                             ACIT_EC_LINK_DOWN,
                             0, ACIT_NULL,
                             ACIT_NULL);
        }
        else
        {
            // Anything other than 0-speed is a good link
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("Physical Link up, port: %c\n"),
                (plink_status->physical_port == NTWK_PORT_A ? 'A' : 'B')));
            pcontroller->link_down[0] = ACIT_FALSE;
        }
    }
    else
    {
        // check logical link status
        plink_status_v1 = (PBE_LINK_STATUS_V1)pioctl_handle->pmgmt_buffer;

        if(plink_status_v1->logical_link_status == 0)
        {
            pcontroller->link_down[0] = ACIT_TRUE;

            // Link is down, check if its due to physical link or logical link
            if(plink_status_v1->physical_port_fault == PHY_LINK_FAULT_NONE)
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("Physical Link Up, Logical Link down, Port: %c\n"),
                    (plink_status_v1->port_number == NTWK_PORT_A ? 'A' : 'B')));

                // logical alone
                ap_log_osm_event(pcontroller,
                                 UNI_CFG_LOG_LEVEL_CRITICAL,
                                 UNI_SUBMODULE_ACIT_GENERIC,
                                 UNI_PHASE_DONT_CARE,
                                 0,
                                 ACIT_EC_LOGICAL_LINK_DOWN,
                                 0, ACIT_NULL,
                                 ACIT_NULL);
            }
            else
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("Physical Link down, Port: %c\n"),
                    (plink_status_v1->port_number == NTWK_PORT_A ? 'A' : 'B')));

                // physical port fault
                ap_log_osm_event(pcontroller,
                                 UNI_CFG_LOG_LEVEL_CRITICAL,
                                 UNI_SUBMODULE_ACIT_GENERIC,
                                 UNI_PHASE_DONT_CARE,
                                 0,
                                 ACIT_EC_LINK_DOWN,
                                 0, ACIT_NULL,
                                 ACIT_NULL);
            }
        }
        else
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("Logical Link up, port: %c\n"),
                (plink_status_v1->port_number == NTWK_PORT_A ? 'A' : 'B')));
            pcontroller->link_down[0] = ACIT_FALSE;
        }
    }

    mgmt_free_ioctl_resources(pcontroller, pioctl_handle);

    return MGMT_STATUS_SUCCESS;
}

MGMT_STATUS
mgmt_process_ioctl_tick(
    PACIT_CONTROLLER_WS     pcontroller)
{
#if ACIT_CFG_DEBUG_BUILD
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    PMGMT_IOCTL_HANDLE      pioctl_handle;
    ACIT_DC_LIST            *tmp, *pioch_link;
    ACIT_UEXACT32           wait_time;

    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);

    for(pioch_link = pmgmt_controller->ioctl_handle_busy_list.next, tmp = pioch_link->next;
        pioch_link != &pmgmt_controller->ioctl_handle_busy_list;
        pioch_link = tmp, tmp = pioch_link->next) 
    {
        pioctl_handle = STRUCT_BASE(MGMT_IOCTL_HANDLE,
                                    controller_link,
                                    pioch_link);

        wait_time = (pcontroller->counters.current_tick_count - pioctl_handle->ticks) *
                        pcontroller->params.ticks_per_second;

        if(wait_time > pioctl_handle->timeout)
        {
            if((wait_time % 5) == 0)
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("IOCTL handle 0x%p pending beyond timeout value\n", pioctl_handle));
            }
        }
    }
#endif
    return MGMT_STATUS_SUCCESS;
}

MGMT_STATUS
mgmt_generic_ioctl_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    MGMT_STATUS             mgmt_status;
    
    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);
    mgmt_status = presponse_hdr->status;
    mgmt_free_ioctl_resources(pcontroller, pioctl_handle);

    return mgmt_status;
}

MGMT_STATUS
mgmt_issue_function_reset(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    PMGMT_IOCTL_HANDLE      pioctl_handle;
    MGMT_STATUS             mgmt_status;
    ACIT_UEXACT32           flags;

    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);

    flags = MGMT_SET_ALLOCATION_FLAGS(MGMT_IOCTL_POLLED,
                                      MGMT_IOCTL_INTERNAL,
                                      MGMT_IOCTL_NO_RESPONSE_BUFFER);

    pioctl_handle = mgmt_alloc_ioctl_resources(pcontroller,
                                               ACIT_NULL,
                                               0,
                                               flags);

    if (pioctl_handle == ACIT_NULL)
    {
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }

    pioctl_handle->timeout = MGMT_INTERNAL_IOCTL_DEFAULT_TIMEOUT;

    mgmt_fill_ioctl_hdr(pioctl_handle,
                        SUBSYSTEM_COMMON_FUNCTION_RESET,
                        OPCODE_COMMON_FUNCTION_RESET,
                        0, // domain 0
                        0,
                        pioctl_handle->timeout);

    pioctl_handle->pmgmt_post_routine = ACIT_NULL;

    mgmt_status = mgmt_fill_ioctl_request(pcontroller, pioctl_handle);
    
    if(mgmt_status == MGMT_STATUS_SUCCESS)
    {
        mgmt_status = mgmt_submit_ioctl(pcontroller, pioctl_handle);
    }

    mgmt_free_ioctl_resources(pcontroller, pioctl_handle);

    return mgmt_status;
}


MGMT_STATUS
mgmt_check_supported_fw(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    PMGMT_IOCTL_HANDLE      pioctl_handle;
    MGMT_STATUS             mgmt_status;
    ACIT_UEXACT32           flags;
    PMGMT_CONTROLLER_ATTRIBUTES pmgmt_controller_attrib = ACIT_NULL;

#if (ACIT_CFG_DEBUG_BUILD)	
    ACIT_UEXACT8 ioctl_version[MGMT_MAX_FLASHROM_VERSION_STRING] = MGMT_FLASHROM_PARAM_VERSION;
#endif

#if (ENABLE_VERSION_CHECK)
    ACIT_UEXACT32           i = 0;
#if !(ACIT_CFG_DEBUG_BUILD)
    ACIT_UEXACT8 ioctl_version[MGMT_MAX_FLASHROM_VERSION_STRING] = MGMT_FLASHROM_PARAM_VERSION;
#endif
#endif

    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);

    flags = MGMT_SET_ALLOCATION_FLAGS(MGMT_IOCTL_POLLED,
                                      MGMT_IOCTL_INTERNAL,
                                      MGMT_IOCTL_RESPONSE_BUFFER);

    pioctl_handle = mgmt_alloc_ioctl_resources(pcontroller,
                                               ACIT_NULL,
                                               sizeof(IOCTL_COMMON_GET_CNTL_ATTRIBUTES),
                                               flags);

    if (pioctl_handle == ACIT_NULL)
    {
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }

    pioctl_handle->timeout = MGMT_INTERNAL_IOCTL_DEFAULT_TIMEOUT;

    mgmt_fill_ioctl_hdr(pioctl_handle,
                        IOCTL_SUBSYSTEM_COMMON,
                        OPCODE_COMMON_GET_CNTL_ATTRIBUTES,
                        0, // domain 0
                        sizeof(IOCTL_COMMON_GET_CNTL_ATTRIBUTES),
                        pioctl_handle->timeout);

    pioctl_handle->pmgmt_post_routine = ACIT_NULL;

    mgmt_status = mgmt_fill_ioctl_request(pcontroller, pioctl_handle);
    
    if(mgmt_status == MGMT_STATUS_SUCCESS)
    {
        mgmt_status = mgmt_submit_ioctl(pcontroller, pioctl_handle);
    }

    if(mgmt_status == MGMT_STATUS_SUCCESS)
    {
        pmgmt_controller_attrib = (PMGMT_CONTROLLER_ATTRIBUTES)pioctl_handle->pmgmt_buffer;
        pcontroller->state.asic_revision = pmgmt_controller_attrib->hba_attribs.asic_revision;
        pcontroller->state.asic_generation = pmgmt_controller_attrib->hba_attribs.asic_generation;
        pcontroller->state.iscsi_features = pmgmt_controller_attrib->hba_attribs.iSCSI_features;
    }

#if (ENABLE_VERSION_CHECK)

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Driver version of IOCTL  : %s\n", ioctl_version));
    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Firmware version of IOCTL: %s\n", pmgmt_controller_attrib->hba_attribs.flashrom_version_string));
    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Release Build, performing version check...\n"));

    for(i = 0; i < MGMT_MAX_FLASHROM_VERSION_STRING; i++)
    {
        if(pmgmt_controller_attrib->hba_attribs.flashrom_version_string[i] !=
            ioctl_version[i])
        {
            mgmt_status = MGMT_STATUS_FAILED;
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Version mismatch!! Failing driver load...\n"));
            break;
        }
    }

    if(mgmt_status != MGMT_STATUS_FAILED)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Version matched, continuing driver load...\n"));
    }

#else

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Driver version of IOCTL  : %s\n", ioctl_version));
    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Firmware version of IOCTL: %s\n", pmgmt_controller_attrib->hba_attribs.flashrom_version_string));
    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Developer Build, not performing version check...\n"));

#endif

    mgmt_free_ioctl_resources(pcontroller, pioctl_handle);

    if(mgmt_status != MGMT_STATUS_SUCCESS)
    {
        ap_log_osm_event(pcontroller,
                         UNI_CFG_LOG_LEVEL_CRITICAL,
                         UNI_SUBMODULE_ACIT_GENERIC,
                         UNI_PHASE_INIT,
                         0,
                         ACIT_EC_FIRMWARE_MISMATCH,
                         0, ACIT_NULL,
                         ACIT_NULL);
    }

    return mgmt_status;
}

MGMT_STATUS
mgmt_get_fw_config(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    PMGMT_IOCTL_HANDLE      pioctl_handle;
    MGMT_STATUS             mgmt_status;
    ACIT_UEXACT32           flags;
    PBE_FW_CFG              pfw_cfg = ACIT_NULL;

    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);

    flags = MGMT_SET_ALLOCATION_FLAGS(MGMT_IOCTL_POLLED,
                                      MGMT_IOCTL_INTERNAL,
                                      MGMT_IOCTL_RESPONSE_BUFFER);

    pioctl_handle = mgmt_alloc_ioctl_resources(pcontroller,
                                               ACIT_NULL,
                                               sizeof(IOCTL_COMMON_QUERY_FIRMWARE_CONFIG),
                                               flags);

    if (pioctl_handle == ACIT_NULL)
    {
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }

    pioctl_handle->timeout = MGMT_INTERNAL_IOCTL_DEFAULT_TIMEOUT;

    mgmt_fill_ioctl_hdr(pioctl_handle,
                        SUBSYSTEM_COMMON_QUERY_FIRMWARE_CONFIG,
                        OPCODE_COMMON_QUERY_FIRMWARE_CONFIG,
                        0, // domain 0
                        sizeof(IOCTL_COMMON_QUERY_FIRMWARE_CONFIG),
                        pioctl_handle->timeout);

    pioctl_handle->pmgmt_post_routine = ACIT_NULL;

    mgmt_status = mgmt_fill_ioctl_request(pcontroller, pioctl_handle);
    
    if(mgmt_status == MGMT_STATUS_SUCCESS)
    {
        mgmt_status = mgmt_submit_ioctl(pcontroller, pioctl_handle);
    }

    if(mgmt_status == MGMT_STATUS_SUCCESS)
    {
        ACIT_UEXACT32 status;

        pfw_cfg = (PBE_FW_CFG)pioctl_handle->pmgmt_buffer;
        pcontroller->fw_config.be_config_number = pfw_cfg->be_config_number;
        pcontroller->fw_config.phys_port = pfw_cfg->phys_port;
        pcontroller->fw_config.function_mode = pfw_cfg->function_mode;
        pcontroller->fw_config.iscsi_icd_start = pfw_cfg->ulp[0].icd_base;
        pcontroller->fw_config.iscsi_icd_count = pfw_cfg->ulp[0].icd_count;
        pcontroller->fw_config.iscsi_cid_start = pfw_cfg->ulp[0].sq_base;
        pcontroller->fw_config.iscsi_cid_count = pfw_cfg->ulp[0].sq_count;

        status = ap_init_fw_config(pcontroller);

        if(status != ACIT_STATUS_SUCCESS)
            mgmt_status = MGMT_STATUS_FAILED;
    }

    mgmt_free_ioctl_resources(pcontroller, pioctl_handle);

    if(mgmt_status != MGMT_STATUS_SUCCESS)
    {
        ap_log_osm_event(pcontroller,
                         UNI_CFG_LOG_LEVEL_CRITICAL,
                         UNI_SUBMODULE_ACIT_GENERIC,
                         UNI_PHASE_INIT,
                         0,
                         ACIT_EC_FW_CONFIG_FAILED,
                         0, ACIT_NULL,
                         ACIT_NULL);
    }

    return mgmt_status;
}

MGMT_STATUS
mgmt_set_fragnum_bits(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           num_bits)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    PMGMT_IOCTL_HANDLE      pioctl_handle;
    MGMT_STATUS             mgmt_status;
    ACIT_UEXACT32           flags;
    PSET_FRAGNUM_BITS_FOR_SGL_CRA_IN pset_fragnum;

    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);

    flags = MGMT_SET_ALLOCATION_FLAGS(MGMT_IOCTL_POLLED,
                                      MGMT_IOCTL_INTERNAL,
                                      MGMT_IOCTL_NO_RESPONSE_BUFFER);

    pioctl_handle = mgmt_alloc_ioctl_resources(pcontroller,
                                               ACIT_NULL,
                                               sizeof(SET_FRAGNUM_BITS_FOR_SGL_CRA_PARAMS),
                                               flags);

    if (pioctl_handle == ACIT_NULL)
    {
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }

    pioctl_handle->timeout = MGMT_INTERNAL_IOCTL_DEFAULT_TIMEOUT;

    mgmt_fill_ioctl_hdr(pioctl_handle,
                        IOCTL_SUBSYSTEM_COMMON_ISCSI,
                        OPCODE_COMMON_ISCSI_SET_FRAGNUM_BITS_FOR_SGL_CRA,
                        0, // domain 0
                        sizeof(SET_FRAGNUM_BITS_FOR_SGL_CRA_PARAMS),
                        pioctl_handle->timeout);

    pioctl_handle->pmgmt_post_routine = ACIT_NULL;
    pset_fragnum = (PSET_FRAGNUM_BITS_FOR_SGL_CRA_IN)pioctl_handle->pmgmt_buffer;
    pset_fragnum->num_bits = num_bits;

    mgmt_status = mgmt_fill_ioctl_request(pcontroller, pioctl_handle);
    
    if(mgmt_status == MGMT_STATUS_SUCCESS)
    {
        mgmt_status = mgmt_submit_ioctl(pcontroller, pioctl_handle);
    }

    mgmt_free_ioctl_resources(pcontroller, pioctl_handle);
    return mgmt_status;
}

MGMT_STATUS
mgmt_epfw_cleanup(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT16           chute)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    PMGMT_IOCTL_HANDLE      pioctl_handle;
    MGMT_STATUS             mgmt_status;
    ACIT_UEXACT32           flags;
    PISCSI_CLEANUP_PARAMS_IN pcleanup;

    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);

    flags = MGMT_SET_ALLOCATION_FLAGS(MGMT_IOCTL_POLLED,
                                      MGMT_IOCTL_INTERNAL,
                                      MGMT_IOCTL_NO_RESPONSE_BUFFER);

    pioctl_handle = mgmt_alloc_ioctl_resources(pcontroller,
                                               ACIT_NULL,
                                               sizeof(ISCSI_CLEANUP_PARAMS),
                                               flags);

    if (pioctl_handle == ACIT_NULL)
    {
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }

    pioctl_handle->timeout = MGMT_INTERNAL_IOCTL_DEFAULT_TIMEOUT;

    mgmt_fill_ioctl_hdr(pioctl_handle,
                        IOCTL_SUBSYSTEM_COMMON_ISCSI,
                        OPCODE_COMMON_ISCSI_CLEANUP,
                        0, // domain 0
                        sizeof(ISCSI_CLEANUP_PARAMS),
                        pioctl_handle->timeout);

    pioctl_handle->pmgmt_post_routine = mgmt_generic_ioctl_callback;

    pcleanup = (PISCSI_CLEANUP_PARAMS_IN)pioctl_handle->pmgmt_buffer;
    pcleanup->chute = chute;
    pcleanup->hdr_ring_id = (ACIT_UEXACT8)HWI_GET_DEF_HDRQ_ID(pcontroller);
    pcleanup->data_ring_id = (ACIT_UEXACT8)HWI_GET_DEF_BUFQ_ID(pcontroller);

    mgmt_status = mgmt_fill_ioctl_request(pcontroller, pioctl_handle);
    
    if(mgmt_status == MGMT_STATUS_SUCCESS)
    {
        mgmt_status = mgmt_submit_ioctl(pcontroller, pioctl_handle);
    }
        
    return MGMT_STATUS_SUCCESS;
}

MGMT_STATUS
mgmt_modify_eq_delay(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           pid,
    ACIT_UEXACT32           eq_id,
    ACIT_UEXACT32           new_delay)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    PMGMT_IOCTL_HANDLE      pioctl_handle;
    MGMT_STATUS             mgmt_status;
    ACIT_UEXACT32           flags;
    PMODIFY_EQ_DELAY_PARAMS pmodify_delay = NULL;

    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);

    flags = MGMT_SET_ALLOCATION_FLAGS(MGMT_IOCTL_NON_POLLED,
                                      MGMT_IOCTL_INTERNAL,
                                      MGMT_IOCTL_NO_RESPONSE_BUFFER);

    pioctl_handle = mgmt_alloc_ioctl_resources(pcontroller,
                                               ACIT_NULL,
                                               sizeof(MODIFY_EQ_DELAY_PARAMS),
                                               flags);

    if (pioctl_handle == ACIT_NULL)
    {
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }

    pioctl_handle->timeout = MGMT_INTERNAL_IOCTL_DEFAULT_TIMEOUT;
    pioctl_handle->ptarget = (PACIT_VOID)((ACIT_U32_64)pid);

    mgmt_fill_ioctl_hdr(pioctl_handle,
                        IOCTL_SUBSYSTEM_COMMON,
                        OPCODE_COMMON_MODIFY_EQ_DELAY,
                        0, // domain 0
                        sizeof(MODIFY_EQ_DELAY_PARAMS),
                        pioctl_handle->timeout);

    pioctl_handle->pmgmt_post_routine = mgmt_modify_eq_delay_callback;

    pmodify_delay = (PMODIFY_EQ_DELAY_PARAMS)pioctl_handle->pmgmt_buffer;

    pmodify_delay->request.num_eq = 1;
    pmodify_delay->request.delay[0].eq_id = eq_id;
    pmodify_delay->request.delay[0].phase = 0;
    pmodify_delay->request.delay[0].delay_multiplier = new_delay;

    pcontroller->partition[pid].dic.reconfigure_in_progress ++;

    mgmt_status = mgmt_fill_ioctl_request(pcontroller, pioctl_handle);
    
    if(mgmt_status == MGMT_STATUS_SUCCESS)
    {
        mgmt_status = mgmt_submit_ioctl(pcontroller, pioctl_handle);
    }
        
    return MGMT_STATUS_SUCCESS;
}

MGMT_STATUS
mgmt_modify_eq_delay_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    MGMT_STATUS             mgmt_status;
    ACIT_UEXACT32           pid;
    
    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);
    mgmt_status = presponse_hdr->status;

    pid = (ACIT_UEXACT32)((ACIT_U32_64)pioctl_handle->ptarget);
    mgmt_free_ioctl_resources(pcontroller, pioctl_handle);
    OSM_ASSERT(pcontroller->partition[pid].dic.reconfigure_in_progress);
    pcontroller->partition[pid].dic.reconfigure_in_progress --;
    return mgmt_status;
}

MGMT_STATUS
mgmt_get_addi_attribs(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    PMGMT_IOCTL_HANDLE      pioctl_handle;
    MGMT_STATUS             mgmt_status;
    ACIT_UEXACT32           flags;
    PGET_CNTL_ADDITIONAL_ATTRIBUTES_PARAMS_OUT paddi_attribs = ACIT_NULL;

    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);

    flags = MGMT_SET_ALLOCATION_FLAGS(MGMT_IOCTL_POLLED,
                                      MGMT_IOCTL_INTERNAL,
                                      MGMT_IOCTL_RESPONSE_BUFFER);

    pioctl_handle = mgmt_alloc_ioctl_resources(pcontroller,
                                               ACIT_NULL,
                                               sizeof(IOCTL_COMMON_GET_CNTL_ADDITIONAL_ATTRIBUTES),
                                               flags);

    if (pioctl_handle == ACIT_NULL)
    {
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }

    pioctl_handle->timeout = MGMT_INTERNAL_IOCTL_DEFAULT_TIMEOUT;

    mgmt_fill_ioctl_hdr(pioctl_handle,
                        SUBSYSTEM_COMMON_GET_CNTL_ADDITIONAL_ATTRIBUTES,
                        OPCODE_COMMON_GET_CNTL_ADDITIONAL_ATTRIBUTES,
                        0, // domain 0
                        sizeof(IOCTL_COMMON_GET_CNTL_ADDITIONAL_ATTRIBUTES),
                        pioctl_handle->timeout);

    pioctl_handle->pmgmt_post_routine = ACIT_NULL;

    mgmt_status = mgmt_fill_ioctl_request(pcontroller, pioctl_handle);

    if(mgmt_status == MGMT_STATUS_SUCCESS)
    {
        mgmt_status = mgmt_submit_ioctl(pcontroller, pioctl_handle);
    }

    if(mgmt_status == MGMT_STATUS_SUCCESS)
    {
        paddi_attribs = (PGET_CNTL_ADDITIONAL_ATTRIBUTES_PARAMS_OUT)pioctl_handle->pmgmt_buffer;
        pcontroller->addi_attribs.ipl_file_number = paddi_attribs->ipl_file_number;
        pcontroller->addi_attribs.ipl_file_version = paddi_attribs->ipl_file_version;
        pcontroller->addi_attribs.interrupt_redirection = (ACIT_UEXACT8)(paddi_attribs->driver_advanced_features_supported & DRIVER_ADV_FEATURE_IFD);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("IPLFileNum 0x%x IPLFileVer 0x%x Redirection 0x%x\n", 
                                pcontroller->addi_attribs.ipl_file_number,
                                pcontroller->addi_attribs.ipl_file_version,
                                pcontroller->addi_attribs.interrupt_redirection));
    }

    mgmt_free_ioctl_resources(pcontroller, pioctl_handle);

    return mgmt_status;
}


MGMT_STATUS
mgmt_process_ioctl_login(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    PLOGIN_TARGET_PARAMS_IN plogin_in;
    MGMT_STATUS             mgmt_status;

    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);
    plogin_in        = (PLOGIN_TARGET_PARAMS_IN)pioctl_request->pdata_buffer;

    /*
     * Check if the number of targets are more than configured limit
     */
    if(plogin_in->session_flags & MGMT_SESSION_FLAG_DATA_SESSION)
    {
        if(pcontroller->ptarget_table->occupied_entries >= 
           pcontroller->params.targets_per_controller)
        {
            pcontroller->eh_stats.login_failure_driver++;
            return MGMT_STATUS_TGTTBL_FULL;
        }

        if(plogin_in->tgt_login_options.sess_login_opts.max_connections == 0)
        {
            PIOCTL_RESPONSE_HEADER   presponse_hdr;
            pcontroller->eh_stats.login_failure_driver++;
            presponse_hdr = (PIOCTL_RESPONSE_HEADER)&(pioctl_request->ioctl_hdr);
            presponse_hdr->additional_status = MGMT_ADDI_STATUS_INVALID_MAXCONNECTION_PARAM;
            
            return MGMT_STATUS_ILLEGAL_FIELD;
        }

        if(plogin_in->extended_timeout != ACIT_U32_MINUS_ONE)
        {   
            if(!(ACIT_IS_ETO_VALID(plogin_in->extended_timeout)))
            {
                PIOCTL_RESPONSE_HEADER   presponse_hdr;
                pcontroller->eh_stats.login_failure_driver++;
                presponse_hdr = (PIOCTL_RESPONSE_HEADER)&(pioctl_request->ioctl_hdr);
                presponse_hdr->additional_status = MGMT_ADDI_STATUS_INVALID_EXTENDED_TIMEOUT;
                OSM_ASSERT(0);
                
                return MGMT_STATUS_ILLEGAL_FIELD;
            }
        }

        pcontroller->stats.data_session_logins++;
    }

    mgmt_status = mgmt_start_external_ioctl(pcontroller,
                                            pioctl_request,
                                            NULL,
                                            mgmt_process_ioctl_login_post);

    return MGMT_STATUS_SUCCESS;
}

ACIT_UEXACT8
mgmt_process_ioctl_login_post(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer)
{
    PLOGIN_TARGET_PARAMS_OUT    plogin_out;
    MGMT_STATUS                 mgmt_status;

    plogin_out = (PLOGIN_TARGET_PARAMS_OUT)(pdata_buffer);

    if(presponse_hdr->status != MGMT_STATUS_SUCCESS)
    {
        goto login_done;
    }

    if(plogin_out->session_flags & MGMT_SESSION_FLAG_DATA_SESSION)
    {
        /*
         * Add the target to the table only if this is a data session.  
         * Adding the target to the table causes OSM to expose all the 
         * luns belongs to this session to the host operating system.
         */
     
        /* All the other session type bits should be off */
        OSM_ASSERT(!(plogin_out->session_flags 
            & (MGMT_SESSION_FLAG_DISCOVERY | 
               MGMT_SESSION_FLAG_INFORMATIONAL)));

        if(ap_check_duplicate_target(pcontroller, 
                                     plogin_out->session_handle,
                                     ACIT_NULL) == ACIT_STATUS_DUPLICATE_ENTRY)
        {
            /* ASSERT here. Target table is not expected to have this target*/
            OSM_ASSERT(0);
            //presponse_hdr->status = MGMT_STATUS_ASSERT_FAILED;
            presponse_hdr->status = MGMT_STATUS_INVALID_SESSION;
            goto login_done;
        }

        ap_add_target(pcontroller, 
                      plogin_out->session_handle, 
                      presponse_hdr->domain,
                      MGMT_SESSION_STATUS_OPENED,
                      1 //is_login_pending
                      );
    }

login_done:
    mgmt_status = presponse_hdr->status;

    if(plogin_out->session_flags & MGMT_SESSION_FLAG_DATA_SESSION)
    {
        if(mgmt_status != MGMT_STATUS_SUCCESS)
        {
            pcontroller->eh_stats.login_failure_arm_fw++;
        }
    }
    
    mgmt_complete_external_ioctl(pcontroller, 
                                 pioctl_handle, 
                                 presponse_hdr, 
                                 pdata_buffer);
    return mgmt_status;
}

MGMT_STATUS
mgmt_process_ioctl_logout(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller; 
    PIOCTL_RESPONSE_HEADER  presponse_hdr;
    PLOGOUT_TARGET_PARAMS_IN    plogout_in;
    PLOGOUT_TARGET_PARAMS_OUT   plogout_out;
    PACIT_TARGET_WS         ptarget;
    ACIT_UEXACT32           session_handle, status, target_index = 0;
    MGMT_STATUS             mgmt_status;
    ACIT_U32_64             target_param = 0;

    pmgmt_controller= GET_MGMT_CONTROLLER_WS(pcontroller);
    presponse_hdr    = (PIOCTL_RESPONSE_HEADER)&(pioctl_request->ioctl_hdr);
    plogout_in      = (PLOGOUT_TARGET_PARAMS_IN)pioctl_request->pdata_buffer;
    plogout_out     = (PLOGOUT_TARGET_PARAMS_OUT)pioctl_request->pdata_buffer;

    session_handle = plogout_in->session_handle;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE, ("SessionId: 0x%x\n", session_handle));

    //
    // If the target entry is present in ACIT library 
    // then logout by sending logout PDU, else forward it to firmware
    //
    ptarget = ap_get_target_ws_from_session_handle(pcontroller, 
                                                   session_handle);

    if(ptarget && ACIT_IS_TARGET_INITIALIZED(ptarget))
    {
        if(IS_LIST_NOT_EMPTY(&ptarget->io_handle_busy_list))
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING, (("target 0x%p has pending IOs, failing logout\n"),
                                    ptarget));

            presponse_hdr->additional_status = MGMT_ADDI_STATUS_LOGOUT_FAILED_IO_BUSY;
            pcontroller->eh_stats.logout_failure_io_busy++;
            return MGMT_STATUS_TARGET_BUSY;
        }
        else if (ACIT_IS_TARGET_BUSY(ptarget, TARGET_BUSY_SESSION_RECOVERY))
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING, (("target 0x%p is in recovery, failing logout\n"),
                                    ptarget));

            presponse_hdr->additional_status = MGMT_ADDI_STATUS_LOGOUT_FAILED_TARGET_IN_RECOVERY;
            pcontroller->eh_stats.logout_failure_recovery++;
            return MGMT_STATUS_TARGET_BUSY;
        }

        pcontroller->stats.data_session_drvr_logouts++;

        //
        // Initiate logout process with iSCSI since this session still
        // is in offload
        //
        plogout_out->session_status = MGMT_SESSION_STATUS_CLOSING;
        iscsi_logout_session(pcontroller, ptarget, pioctl_request);
    }
    else 
    {
        status = ap_check_duplicate_target(pcontroller, 
                                           session_handle,
                                           &target_index);

        if(status == ACIT_STATUS_DUPLICATE_ENTRY)
        {
            //
            // Bug 20826 : A logout and a offload during recovery crossed each other.
            // When offload eventually completed, the iSCSI workspace was gone and
            // caused a BSOD. So if ARM fw owns the session, allow logout only if ETO
            // is counting down
            //
            if((IS_CONTROLLER_LDTO_EXPIRED(pcontroller, 0)) ||
               (IS_TARGET_MISSING(ptarget)))
            {
                // logout of those targets that are in ACIT's table but not
                // offloaded
                ACIT_BIT_SET(pcontroller->ptarget_table->target_entry[target_index].acit_tflags,
                                ACIT_TFLAG_REMOVE_PENDING);

                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING, (("target 0x%p is in recovery/LDTO expired, passing logout to ARM fw\n"),
                                        ptarget));
                        
                target_param = (ACIT_U32_64)target_index;
                mgmt_status = mgmt_start_external_ioctl(pcontroller,
                                                        pioctl_request,
                                                        (PACIT_VOID)target_param,
                                                        mgmt_process_ioctl_logout_post);
            }
            else
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING, (("target 0x%p has not completed recovery, failing logout\n"),
                                        ptarget));

                presponse_hdr->additional_status = MGMT_ADDI_STATUS_LOGOUT_FAILED_TARGET_IN_RECOVERY;
                pcontroller->eh_stats.logout_failure_recovery++;
                return MGMT_STATUS_TARGET_BUSY;
            }
        }
        else
        {
            // logout of those targets that are owned by ARM fw (discovery and
            // non data session targets)
            mgmt_process_external_ioctl_default(pcontroller, pioctl_request);
        }
    }

    return MGMT_STATUS_SUCCESS;
}

ACIT_UEXACT8
mgmt_process_ioctl_logout_post(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer)
{
    PACIT_TARGET_WS         ptarget;
    ACIT_U32_64             target_index = 0;
    ACIT_UEXACT32           session_handle;
    MGMT_STATUS             mgmt_status;

    //
    // This routine is ONLY for logout of targets for
    // which the ARM fw still owns the session
    //
    if(presponse_hdr->status == MGMT_STATUS_SUCCESS)
    {
        pcontroller->stats.data_session_armfw_logouts++;

        // ptarget is the placeholder of target table index to delete
        target_index = (ACIT_U32_64)pioctl_handle->ptarget;

        session_handle = pcontroller->ptarget_table->target_entry[target_index].session_handle;

        ptarget = ap_get_target_ws_from_session_handle(pcontroller, 
                                                    session_handle);

        if(ptarget)
        {
            //
            // this is for those targets who have been initialized once earlier, but
            // have now been uploaded but not logged out
            //
            //OSM_ASSERT(ACIT_IS_BIT_SET(ptarget->flags, FLAG_TGT_UPLOAD_COMPLETE));

            //
            // Bug 18196 - 04/09/2011:
            // Call iscsi function to cleanup the session resources. It will call ap_remove_target
            // at the end
            //
            iscsi_delete_session_info(pcontroller, ptarget);
            //ap_remove_target(pcontroller, ptarget);
        }
        else
        {
            //
            // those targets that haven't been initialized. Just remove the
            // target entry from ACIT target table
            //
            ap_delete_target_entry(pcontroller, (ACIT_UEXACT32)target_index);
        }
    }
    else
    {
        pcontroller->eh_stats.logout_failure_armfw_session++;
        OSM_ASSERT(0);
    }

    mgmt_status = presponse_hdr->status;

    mgmt_complete_external_ioctl(pcontroller, 
                                 pioctl_handle, 
                                 presponse_hdr, 
                                 pdata_buffer);
    return mgmt_status;
}

MGMT_STATUS
mgmt_process_ioctl_modify_sessionloginkeys(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request)
{
    PACIT_TARGET_WS         ptarget;
    PMODIFY_SESSION_LOGINKEYS_PARAMS_IN pmodify_in;
    ACIT_UEXACT32           status;

    pmodify_in      = (PMODIFY_SESSION_LOGINKEYS_PARAMS_IN)(pioctl_request->pdata_buffer);

    ptarget = ap_get_target_ws_from_session_handle(pcontroller, 
                                                   pmodify_in->session_handle);

    /*
     * NOTE: We do not supporting modifying session parameters of informational
     * sessions. So its incorrect for the target to be not present in ACIT
     */
    OSM_ASSERT(ptarget);

    status = iscsi_recover_session(pcontroller, 
                                   ptarget, 
                                   pioctl_request);

    if(status == ACIT_STATUS_PENDING)
    {
        mgmt_return_ioctl(pcontroller, pioctl_request, MGMT_STATUS_TARGET_BUSY);
    }
    
    return MGMT_STATUS_SUCCESS;
}

ACIT_UEXACT8
mgmt_process_ioctl_modify_sessionloginkeys_post(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer)
{
    PACIT_TARGET_WS         ptarget;
    PMODIFY_SESSION_LOGINKEYS_PARAMS_OUT pmodify_out;
    MGMT_STATUS             mgmt_status;
    ACIT_UEXACT8            domain;

    pmodify_out = (PMODIFY_SESSION_LOGINKEYS_PARAMS_OUT)pdata_buffer;
    if(presponse_hdr->status != MGMT_STATUS_SUCCESS)
    {
        OSM_ASSERT(0);
    }
    ptarget = ap_get_target_ws_from_session_handle(pcontroller, 
                                                   pmodify_out->session_handle);

    OSM_ASSERT(ptarget);

    mgmt_status = presponse_hdr->status;
    domain = presponse_hdr->domain;

    mgmt_complete_external_ioctl(pcontroller, 
                                 pioctl_handle, 
                                 presponse_hdr, 
                                 pdata_buffer);

    mgmt_reopen_sessions(pcontroller, 
                         IOCTL_ISCSI_REOPEN_A_SESSION, 
                         pmodify_out->session_handle,
                         domain,
                         ACIT_FALSE);

    return mgmt_status;

}

MGMT_STATUS
mgmt_process_ioctl_modify_connloginkeys(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pacit_ioctl)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;

    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);

    /* 
     * TODO: Enable this call when ISCSI module provides the api 
     * in the future. Assert for now.
     */
    /*iscsi_remove_connection(pController, 
     *                        pParamsIn->session_id, pParams->connection_id);
     */
    OSM_ASSERT(0);
    
    return MGMT_STATUS_SUCCESS;
}

ACIT_UEXACT8
mgmt_process_ioctl_modify_connloginkeys_post(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer)
{
    PMODIFY_CONNECTION_LOGINKEYS_PARAMS_OUT pmodify_out;
    MGMT_STATUS             mgmt_status;

    if(presponse_hdr->status != MGMT_STATUS_SUCCESS)
        goto modify_done;

    pmodify_out = (PMODIFY_CONNECTION_LOGINKEYS_PARAMS_OUT)(pdata_buffer);

    /*
     * TODO: Enable this call when ISCSI module provides the 
     * api in the future.  Assert for now
     */
    /*
     * iscsi_save_connection_info(pController, 
     *                            pParamsOut->session_id, 
     *                            pParamsOut->reserved_conn_data);
     */
    OSM_ASSERT(0);

modify_done:

    mgmt_status = presponse_hdr->status;

    mgmt_complete_external_ioctl(pcontroller, 
                                 pioctl_handle, 
                                 presponse_hdr, 
                                 pdata_buffer);
    
    return mgmt_status;
}

// No MCS support
# if 0
MGMT_STATUS
mgmt_process_ioctl_addconn(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    PADD_CONNECTION_TO_SESSION_PARAMS_IN padd_cxn_in;
    MGMT_STATUS             mgmt_status;
    PACIT_TARGET_WS         ptarget;

    pmgmt_controller= GET_MGMT_CONTROLLER_WS(pcontroller);
    padd_cxn_in     = (PADD_CONNECTION_TO_SESSION_PARAMS_IN)pioctl_request->pdata_buffer;

    ptarget = ap_get_target_ws_from_session_handle(
                                            pcontroller, 
                                            padd_cxn_in->session_handle);

    padd_cxn_in->cmd_sn = ISCSI_GET_CMD_SN(ptarget);

    mgmt_status = mgmt_start_external_ioctl(pcontroller,
                                            pioctl_request,
                                            ptarget,
                                            mgmt_process_ioctl_addconn_post);

    // Mark target as busy and initiate a reopen a connection IOCTL
    ap_flag_busy_target(pcontroller, ptarget, TARGET_BUSY_NEW_CONNECTION_ADD);

    if(mgmt_status != MGMT_STATUS_SUCCESS)
    {
        /* Mark target status as ready irrespective */
        ap_flag_ready_target(pcontroller, ptarget, TARGET_BUSY_NEW_CONNECTION_ADD);
        return mgmt_status;
    }
   
    return MGMT_STATUS_SUCCESS;
}

MGMT_STATUS
mgmt_process_ioctl_addconn_post(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer)
{
    PADD_CONNECTION_TO_SESSION_PARAMS_OUT padd_cxn_out;
    PACIT_TARGET_WS                       ptarget;
    ACIT_UEXACT32                         connection_handle, session_handle;
    MGMT_STATUS                           mgmt_status;

    padd_cxn_out = (PADD_CONNECTION_TO_SESSION_PARAMS_OUT)pdata_buffer;

    ptarget = pioctl_handle->ptarget;   

    /*
     *   Update ISCSI module about the new connection
     */
    
    if(presponse_hdr->status == MGMT_STATUS_SUCCESS)
        ap_add_connection(pcontroller, 
                          padd_cxn_out->session_handle, 
                          padd_cxn_out->conn_data.connection_handle);
    
    /* Mark target status as ready irrespective */
    ap_flag_ready_target(pcontroller, ptarget, TARGET_BUSY_NEW_CONNECTION_ADD);
    connection_handle = padd_cxn_out->conn_data.connection_handle;
    session_handle = padd_cxn_out->conn_data.session_handle;
    
    mgmt_status = presponse_hdr->status;

    mgmt_complete_external_ioctl(pcontroller, 
                                 pioctl_handle, 
                                 presponse_hdr, 
                                 pdata_buffer);
    
    if(mgmt_status == MGMT_STATUS_SUCCESS)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING, ("iscsi_process_connection_event called for newly added connection\n"));
        iscsi_process_connection_event(pcontroller,
                                        ptarget,
                                        connection_handle,
                                        session_handle);
    }
    
    return mgmt_status;
}

MGMT_STATUS
mgmt_process_ioctl_closeconn(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    PIOCTL_RESPONSE_HEADER  presponse_hdr;
    PCLOSE_CONNECTION_PARAMS_IN pclose_cxn_in;
    MGMT_STATUS             status, mgmt_status = MGMT_STATUS_SUCCESS;
    PACIT_TARGET_WS         ptarget;
    ACIT_UEXACT32           cxn_cid;


    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);
    presponse_hdr    = (PIOCTL_RESPONSE_HEADER)&(pioctl_request->ioctl_hdr);
    pclose_cxn_in    = (PCLOSE_CONNECTION_PARAMS_IN)pioctl_request->pdata_buffer;

    ptarget = ap_get_target_ws_from_session_handle(
                                            pcontroller, 
                                            pclose_cxn_in->session_handle);

    if(!ptarget)
    {
        status = ap_check_duplicate_target(pcontroller, 
                                           pclose_cxn_in->session_handle,
                                           ACIT_NULL);

        if(status == ACIT_STATUS_DUPLICATE_ENTRY)
        {
            //
            // This is the case where target was logged in but has not been offloaded
            // yet. So wait until the login process is clean before we can logout
            //
            presponse_hdr->status = MGMT_STATUS_FAILED;
            presponse_hdr->status = MGMT_ADDI_STATUS_LOGIN_IN_PROGRESS;
            return presponse_hdr->status;
        }

        mgmt_status = mgmt_process_external_ioctl_default(pcontroller, pioctl_request);
    }
    else
    {
        /* If the connection is not part of the session then this connection may not
         * be offloaded. Pass it to the ARM f/w.
         */
        cxn_cid = iscsi_get_cid_from_connection_handle (pcontroller, 
                                                        ptarget, 
                                                        pclose_cxn_in->connection_handle);
        if (cxn_cid == 0xffffffff)
        {
            mgmt_status = mgmt_process_external_ioctl_default(pcontroller, pioctl_request);
        }
        else
        {
            /* Initiate logout process */

            OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE, ("Driver IOCTL\n"));

            mgmt_status = (ACIT_UEXACT8)iscsi_close_connection(pcontroller, ptarget, pioctl_request);
        }
    }
   
    return mgmt_status;
}
#endif

MGMT_STATUS
mgmt_process_set_tcpconfig(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;

    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);
    OSM_ASSERT(!pmgmt_controller->porig_ioctl_request);
    pmgmt_controller->porig_ioctl_request = pioctl_request;

    //
    // if we have a boot session configured, do not attempt to reopen
    // sessions. Simply forward the request to ARM fw. ARM fw will save
    // the address and commit it in the next boot
    //
    
    mgmt_get_boot_session(pcontroller, 
                          ACIT_FALSE); // discovery = FALSE
    
    return MGMT_STATUS_SUCCESS;
}

ACIT_UEXACT8
mgmt_process_set_nw_params_post(
    PACIT_CONTROLLER_WS     pcontroller, 
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer)
{
    MGMT_STATUS             mgmt_status;

    if(presponse_hdr->status != MGMT_STATUS_SUCCESS)
    {
        /*
         * sessions are in a inconsistent state. We closed the session,
         * but global nw params could'nt be set, so reopen the sessions anyways.
         * The older nw params (IP/iSCSIName/NIC) should hold good
         */

        OSM_ASSERT(0);
    }

    mgmt_status = presponse_hdr->status;

    mgmt_complete_external_ioctl(pcontroller, 
                                 pioctl_handle, 
                                 presponse_hdr, 
                                 pdata_buffer);

    if(pcontroller->ptarget_table->occupied_entries)
    {
        mgmt_reopen_sessions(pcontroller, 
                             IOCTL_ISCSI_REOPEN_ALL_SESSIONS, 
                             ACIT_U32_MINUS_ONE,
                             0, // domain 0
                             ACIT_FALSE);
    }

    return mgmt_status;
}

MGMT_STATUS
mgmt_process_set_sessionconfig(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request)
{
    PACIT_TARGET_WS         ptarget;
    MGMT_STATUS             mgmt_status;
    PSET_SESSION_CONFIGURATION_PARAMS_IN    psession_config;

    psession_config = (PSET_SESSION_CONFIGURATION_PARAMS_IN)pioctl_request->pdata_buffer;

    ptarget = ap_get_target_ws_from_session_handle(
                                            pcontroller, 
                                            psession_config->session_handle);

    if(ptarget)
    {
        if(psession_config->param_flags & SET_SESSCONFIG_PARAM_FLAG_ETO)
        {
            if(psession_config->extended_timeout != ACIT_U32_MINUS_ONE)
            {   
                if(ACIT_IS_ETO_VALID(psession_config->extended_timeout))
                {
                    ptarget->eto.original_ticks = psession_config->extended_timeout * pcontroller->params.ticks_per_second;
                    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Target 0x%p - Commiting new ETO value %d\n", ptarget, psession_config->extended_timeout));
                }
                else
                {
                    PIOCTL_RESPONSE_HEADER  presponse_hdr;

                    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Target 0x%p - Rejecting out of range ETO value Max %d Set %d\n", ptarget, ACIT_CFG_MAX_ETO, psession_config->extended_timeout));
                    presponse_hdr = (PIOCTL_RESPONSE_HEADER)&(pioctl_request->ioctl_hdr);
                    presponse_hdr->additional_status = MGMT_ADDI_STATUS_INVALID_EXTENDED_TIMEOUT;
                    OSM_ASSERT(0);
                    return MGMT_STATUS_ILLEGAL_FIELD;
                }
            }
            else
            {
                ptarget->eto.original_ticks = pcontroller->params.eto * pcontroller->params.ticks_per_second;
                OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Target 0x%p - Commiting new ETO value %d\n", ptarget, pcontroller->params.eto));
            }
        }
    }
    else
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Target workspace not found for session %d\n", 
                            psession_config->session_handle));
        OSM_ASSERT(0);
    }

    mgmt_status = mgmt_process_external_ioctl_default(pcontroller, pioctl_request);

    return mgmt_status;
}

MGMT_STATUS
mgmt_process_get_nic_statistics(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request)
{
    PACIT_TARGET_WS         ptarget;
    PACIT_TARGET_ENTRY      ptarget_entry = ACIT_NULL;
    ACIT_UEXACT32           index;
    PIOCTL_RESPONSE_HEADER  presponse_hdr;
    PMIB_NIC_STATISTICS     presp;
    PISCSI_TARGET_WS        piscsi_target;
    ACIT_UEXACT32           i;

    presponse_hdr = (PIOCTL_RESPONSE_HEADER)&(pioctl_request->ioctl_hdr);

    presp = (PMIB_NIC_STATISTICS)pioctl_request->pdata_buffer;
    ptarget_entry = &pcontroller->ptarget_table->target_entry[0];

    for(index = 0; 
        index < pcontroller->params.targets_per_controller;
        index ++)
    {
        if(ACIT_IS_BIT_SET(ptarget_entry[index].acit_tflags,
                            ACIT_TFLAG_IN_USE))
        {
            ptarget = GET_TARGET_WS_SAFE(ptarget_entry[index].ptarget_key);

            if(!ptarget)
                continue;
            
            piscsi_target = GET_ISCSI_TARGET_WS(ptarget);
            for (i = 0; i < piscsi_target->num_connections; i++)
            {
                if(piscsi_target->pcxn_list[i])
                {
                    presp->pdu_transmitted += piscsi_target->pcxn_list[i]->iscsi_cmd_pdus;
                    presp->pdu_received += piscsi_target->pcxn_list[i]->iscsi_rsp_pdus;
                    presp->byte_transmitted += piscsi_target->pcxn_list[i]->total_bytes_xferred;
                    presp->byte_received += piscsi_target->pcxn_list[i]->total_bytes_read;
                }
            }
        }
    }
    presponse_hdr->status = MGMT_STATUS_SUCCESS;
    presponse_hdr->response_length = sizeof(MIB_NIC_STATISTICS);
    presponse_hdr->actual_response_length = sizeof(MIB_NIC_STATISTICS);

    mgmt_return_ioctl(pcontroller, 
                      pioctl_request, 
                      (MGMT_STATUS)presponse_hdr->status);
        
    return MGMT_STATUS_SUCCESS;
}

MGMT_STATUS
mgmt_process_get_pcie_info_post(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer)
{
    PMGMT_PCIE_INFO         presp;

    // Restore the opcode and subsystem!
    presponse_hdr->subsystem = IOCTL_SUBSYSTEM_ISCSI_INI;
    presponse_hdr->opcode = OPCODE_ISCSI_INI_CFG_GET_PCIE_INFO;

    presp = (PMGMT_PCIE_INFO)pdata_buffer;

    OSM_MEMCPY(presp, &pcontroller->pcie_info, sizeof(MGMT_PCIE_INFO));

    presponse_hdr->status = MGMT_STATUS_SUCCESS;
    presponse_hdr->response_length = sizeof(MGMT_PCIE_INFO);
    presponse_hdr->actual_response_length = sizeof(MGMT_PCIE_INFO);

    mgmt_complete_external_ioctl(pcontroller, 
                                 pioctl_handle, 
                                 presponse_hdr, 
                                 pdata_buffer);

    return MGMT_STATUS_SUCCESS;

}

MGMT_STATUS
mgmt_nop_ioctl(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request,
    PMGMT_POST              post_routine)
{
    PIOCTL_REQUEST_HEADER   prequest_hdr;
    MGMT_STATUS             mgmt_status;

    prequest_hdr = (PIOCTL_REQUEST_HEADER)&(pioctl_request->ioctl_hdr);

    // Modify the IOCTL header and opcode to issue a NOP IOCTL.
    // It is the post routine's responsibility to change it
    // back to whatever the caller issued originally

    prequest_hdr->opcode = OPCODE_COMMON_NOP;
    prequest_hdr->subsystem = IOCTL_SUBSYSTEM_COMMON;

    mgmt_status = mgmt_start_external_ioctl(pcontroller,
                                            pioctl_request,
                                            NULL,
                                            post_routine);

    return mgmt_status;
}

ACIT_UEXACT8
mgmt_process_external_ioctl_default(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request)
{
    MGMT_STATUS             mgmt_status;

    /*
     *  The default case is for IOCTLs that don't require ACIT to look at 
     *  the IOCTL payload for special processing prior to delivering the
     *  IOCTLs to the firmware.
     */

    mgmt_status = mgmt_start_external_ioctl(pcontroller,
                                            pioctl_request,
                                            NULL,
                                            mgmt_complete_external_ioctl);

    return mgmt_status;
}

MGMT_STATUS
mgmt_start_external_ioctl(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request,
    PACIT_VOID              ptarget,
    PMGMT_POST              post_routine)
{
    PMGMT_IOCTL_HANDLE      pioctl_handle;
    ACIT_UEXACT32           poll = 0, flags;
    MGMT_STATUS             mgmt_status;

    if(pioctl_request->ioctl_flag & ACIT_IOCTL_FLAG_POLLED)
        poll = MGMT_IOCTL_POLLED;
    else
        poll = MGMT_IOCTL_NON_POLLED;

    flags = MGMT_SET_ALLOCATION_FLAGS(poll, MGMT_IOCTL_EXTERNAL, 0);

    pioctl_handle = mgmt_alloc_ioctl_resources(pcontroller,
                                               pioctl_request,
                                               pioctl_request->data_length,
                                               flags);

    if (pioctl_handle == ACIT_NULL)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("ioctl handle allocation failure..\n"));
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }

    pioctl_handle->pioctl_request = pioctl_request;
    pioctl_handle->timeout = pioctl_request->ioctl_hdr.request.timeout;
    pioctl_handle->ptarget = (PACIT_VOID)ptarget;
    pioctl_handle->pmgmt_post_routine = post_routine;

    mgmt_status = mgmt_fill_ioctl_request(pcontroller, pioctl_handle);

    if(mgmt_status != MGMT_STATUS_SUCCESS)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("mgmt_fill_ioctl_request failure - 0x%x\n", mgmt_status));
        OSM_ASSERT(0);
        mgmt_free_ioctl_resources(pcontroller, pioctl_handle);
        return mgmt_status;
    }

    mgmt_status = mgmt_submit_ioctl(pcontroller, pioctl_handle);

    if(mgmt_status == MGMT_STATUS_FAILED)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("mgmt_submit_ioctl failure - 0x%x\n", mgmt_status));
        OSM_ASSERT(0);
    }
    
    return mgmt_status;
}

ACIT_UEXACT32
mgmt_start_ioctl(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    PIOCTL_REQUEST_HEADER   prequest_hdr;
    MGMT_STATUS             mgmt_status;

    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);

    /* 
     * Process IOCTLs that require MGMT to handle before delivering it to ISCSI or HWI 
     */

    prequest_hdr = (PIOCTL_REQUEST_HEADER)&(pioctl_request->ioctl_hdr);
    OSM_ASSERT(prequest_hdr->subsystem);

    OSM_DEBUG_PRINT (ACIT_DBGMSG_TRACE, ("MGMT Subsystem 0x%x Opcode:0x%x\n", prequest_hdr->subsystem, prequest_hdr->opcode));

    switch(prequest_hdr->subsystem)
    {
    case IOCTL_SUBSYSTEM_ISCSI_INI:

        switch(prequest_hdr->opcode)
        {
            case OPCODE_ISCSI_INI_SESSION_LOGIN_TARGET:
                
                mgmt_status = mgmt_process_ioctl_login(pcontroller, pioctl_request);
                break;

            case OPCODE_ISCSI_INI_SESSION_LOGOUT_TARGET:

                mgmt_status = mgmt_process_ioctl_logout(pcontroller, pioctl_request);
                break;

            case OPCODE_ISCSI_INI_SESSION_MODIFY_SESSION_LOGINKEYS:

                mgmt_status = mgmt_process_ioctl_modify_sessionloginkeys(pcontroller, pioctl_request);
                break;

            case OPCODE_ISCSI_INI_SESSION_MODIFY_CONNECTION_LOGINKEYS:

                mgmt_status = mgmt_process_ioctl_modify_connloginkeys(pcontroller, pioctl_request);
                break;

                // No MCS support
//            case OPCODE_ISCSI_INI_SESSION_ADD_CONNECTION_TO_SESSION:
//
//                mgmt_status = mgmt_process_ioctl_addconn(pcontroller, pioctl_request);
//                break;
//
//            case OPCODE_ISCSI_INI_SESSION_CLOSE_CONNECTION:
//
//                mgmt_status = mgmt_process_ioctl_closeconn(pcontroller, pioctl_request);
//                break;

            case OPCODE_ISCSI_INI_SESSION_SET_CONFIGURATION:

                mgmt_status = mgmt_process_set_sessionconfig(pcontroller, pioctl_request);
                break;

            case OPCODE_ISCSI_INI_NTWK_SET_TCPIP_CONFIG:

                mgmt_status = mgmt_process_set_tcpconfig(pcontroller, pioctl_request);
                break;

            case OPCODE_ISCSI_INI_CFG_GET_PCIE_INFO:

                mgmt_status = mgmt_nop_ioctl(pcontroller, pioctl_request, mgmt_process_get_pcie_info_post);
                break;

            default:

                mgmt_status = mgmt_process_external_ioctl_default(pcontroller, 
                                                                  pioctl_request);
                break;
        }

        break;

    case IOCTL_SUBSYSTEM_COMMON_ISCSI:

        switch(prequest_hdr->opcode)
        {
        case OPCODE_COMMON_ISCSI_STATS_GET_NIC_STATISTICS:
            
            mgmt_status = mgmt_process_get_nic_statistics(pcontroller, pioctl_request);
            break;

        case OPCODE_COMMON_ISCSI_NTWK_MODIFY_IP_ADDR:

            mgmt_status = mgmt_process_set_tcpconfig(pcontroller, pioctl_request);
            break;

        default:

            mgmt_status = mgmt_process_external_ioctl_default(pcontroller, 
                                                                pioctl_request);
            break;

        }
        break;

    default:

            mgmt_status = mgmt_process_external_ioctl_default(pcontroller, 
                                                              pioctl_request);
        break;
    }

    if(mgmt_status != MGMT_STATUS_SUCCESS)
    {
        /*
         *  The IOCTL cannot be delivered to the ARM layer, return it back to OSM
         */
        mgmt_return_ioctl(pcontroller, pioctl_request, mgmt_status);
    }

    return ACIT_STATUS_SUCCESS;
}


MGMT_STATUS
mgmt_start_hiber_io_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    PACIT_IO_REQUEST        pio_request;
    PSCSI_READ_WRITE_PARAMS_OUT pscsi_cmd_out;
    PACIT_VOID              psense_buffer;
    ACIT_UEXACT32           status = 0;

    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);
    pio_request = (PACIT_IO_REQUEST)pioctl_handle->pioctl_request;
    pscsi_cmd_out = (PSCSI_READ_WRITE_PARAMS_OUT)pdata_buffer;

    OSM_ASSERT(presponse_hdr->opcode == OPCODE_ISCSI_INI_BOOT_READ_WRITE);

    if(presponse_hdr->status == MGMT_STATUS_SUCCESS)
    {
        pio_request->io_status = ACIT_IO_STATUS_SUCCESS;
        pio_request->scsi_status = pscsi_cmd_out->scsi_status;

        //
        // Copy sense data if necessary.
        //
        if (ACIT_SCSI_STATUS_CHECK_CONDITION == pscsi_cmd_out->scsi_status) {

            ACIT_UEXACT32 sense_length;

            psense_buffer = &pscsi_cmd_out->sense_buffer;

            // The start of the payload contains the sense data length as a UINT16.
            sense_length = MGMT_SENSE_BUFFER_LEN;
            OSM_DEBUG_PRINT (ACIT_DBGMSG_WARNING, ("SCSI check condition. Length:0x%x\n", sense_length));            

            // Hardware may truncate sense data to fit in the header fragment.
            if (sense_length > ACIT_SENSE_INFO_SIZE) {
                sense_length = ACIT_SENSE_INFO_SIZE;
                OSM_DEBUG_PRINT (ACIT_DBGMSG_WARNING, ("Sense data truncated by header frag size to 0x%x bytes\n", sense_length));
            }

            // The OS buffer may truncate the sense data as well.
            if (pio_request->sense_buffer_length < sense_length) {
                sense_length = pio_request->sense_buffer_length;
                OSM_DEBUG_PRINT (ACIT_DBGMSG_WARNING, ("Sense data truncated to 0x%x bytes\n", sense_length));
            }

            // Copy the sense data into the io request.
            OSM_MEMCPY (pio_request->psense_buffer_virtual, 
                        (PACIT_UEXACT8)psense_buffer,
                        sense_length);

            // update OSM request sense_len 
            pio_request->sense_buffer_length = sense_length;
        }

        mgmt_free_ioctl_resources(pcontroller, pioctl_handle);
    }
    else
    {
        // Firmware returns unexpected error status.
        OSM_DEBUG_PRINT (ACIT_DBGMSG_ERROR, ("ARM fw returned error %d\n", presponse_hdr->status));

        pio_request->io_status = ACIT_IO_STATUS_ERROR;
        pio_request->scsi_status = 0;

        //
        // bug 13183. For additional status MGMT_ADDI_STATUS_SESSION_ALREADY_CLOSED,
        // issue reopen session
        //
        if((presponse_hdr->status == MGMT_STATUS_INVALID_SESSION) &&
           (presponse_hdr->additional_status == MGMT_ADDI_STATUS_SESSION_ALREADY_CLOSED))
        {
            mgmt_free_ioctl_resources(pcontroller, pioctl_handle);

            OSM_DEBUG_PRINT (ACIT_DBGMSG_ERROR, ("Reopen boot session for status 0x%x addi 0x%x\n", 
                                                presponse_hdr->status,
                                                presponse_hdr->additional_status));

            status = mgmt_reopen_sessions(pcontroller,
                                            IOCTL_ISCSI_REOPEN_BOOT_SESSION,
                                            ACIT_U32_MINUS_ONE,   // not used
                                            0, // domain 0
                                            ACIT_TRUE);

            if(status != MGMT_STATUS_SUCCESS)
            {
                // end of story as far as crashdump/hibernation is concerned.
                OSM_DEBUG_PRINT (ACIT_DBGMSG_ERROR, ("Reopen boot session failed with status 0x%x\n", 
                                                    status));
            }
        }
        else
        {
            mgmt_free_ioctl_resources(pcontroller, pioctl_handle);
        }
    }

    ap_return_hiber_io(pcontroller, pio_request);
    return MGMT_STATUS_SUCCESS;    
}

MGMT_STATUS
mgmt_start_hiber_io(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_REQUEST        pio_request)
{
    PACIT_TARGET_WS             ptarget;
    PISCSI_TARGET_WS            piscsi_target;
    PSCSI_READ_WRITE_PARAMS_IN  pscsi_io_req;
    PMGMT_IOCTL_HANDLE          pioctl_handle;
    MGMT_STATUS                 mgmt_status;
    ACIT_UEXACT32               expected_data_size, payload_length, flags, i;

    ptarget = GET_TARGET_WS(pio_request->ptarget_key);
    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);

    expected_data_size = pio_request->data_transfer_length;

    /* modify the payload length since driver IO can have more than one sg element */
    payload_length = sizeof(SCSI_READ_WRITE_PARAMS_IN);
    if ( (pio_request->psg_list) && (pio_request->psg_list->number_elements > 1) )
        payload_length +=(sizeof(MGMT_SG_ELEMENT) * (pio_request->psg_list->number_elements-1));

    flags = MGMT_SET_ALLOCATION_FLAGS(MGMT_IOCTL_POLLED,
                                      MGMT_IOCTL_INTERNAL,
                                      0);
                                      //MGMT_IOCTL_RESPONSE_BUFFER);

    pioctl_handle = mgmt_alloc_ioctl_resources(pcontroller,
                                               ACIT_NULL,
                                               payload_length,
                                               flags);

    if(pioctl_handle == ACIT_NULL)
    {
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }

    /* Fill up the SCSI cmd IOCTL request */
    pioctl_handle->pmgmt_post_routine = mgmt_start_hiber_io_callback;
    pioctl_handle->timeout = MGMT_IOCTL_BOOT_READ_WRITE_TIMEOUT;
    /* Place holder to get the io request back from ioctl handle */
    pioctl_handle->pioctl_request = (PACIT_IOCTL_REQUEST)pio_request;
    
    pscsi_io_req = pioctl_handle->pmgmt_buffer;
    pscsi_io_req->session_handle = piscsi_target->session_handle;
    if (pio_request->flags.inbound_data) 
        pscsi_io_req->data_direction_flag = SCB_FLAG_XFER_FROM_DEVICE ;
    else if(pio_request->flags.outbound_data)
        pscsi_io_req->data_direction_flag = SCB_FLAG_XFER_TO_DEVICE;
    else
        pscsi_io_req->data_direction_flag = 0;

    OSM_MEMCPY(pscsi_io_req->cdb, pio_request->cdb, pio_request->cdb_length); 
    OSM_MEMCPY(pscsi_io_req->lun, pio_request->lun.u.lun_bytes, MGMT_LUN_SIZE);
    if (pscsi_io_req->data_direction_flag)
    {
        /* copy the sg list from iorequest to the IOCTL if data xfer is involved */
        pscsi_io_req->sgtable.sg_count = (pio_request->psg_list->number_elements << 24) ;
        for (i=0; i<pio_request->psg_list->number_elements; i++)
        {
            pscsi_io_req->sgtable.sg_list[i].address.u.a64.address = pio_request->psg_list->sglist[i].address.u.a64.address;
            pscsi_io_req->sgtable.sg_list[i].length = pio_request->psg_list->sglist[i].length;
            //OSM_MEMCPY(pscsi_io_req->sgtable.sg_list[i], pio_request->psg_list->sglist[i], sizeof(MGMT_SG_ELEMENT));
        }
        pscsi_io_req->sgtable.sg_count = ((pscsi_io_req->sgtable.sg_count & 0xFF000000) | (expected_data_size & 0x00FFFFFF)) ;
    }


    mgmt_fill_ioctl_hdr(pioctl_handle,
                        IOCTL_SUBSYSTEM_ISCSI_INI,
                        OPCODE_ISCSI_INI_BOOT_READ_WRITE,
                        0, // domain 0
                        sizeof(SCSI_READ_WRITE_PARAMS_OUT),//expected_data_size,
                        pioctl_handle->timeout);
    
    mgmt_status = mgmt_fill_ioctl_request(pcontroller, pioctl_handle);
    
    if(mgmt_status == MGMT_STATUS_SUCCESS)
    {
        mgmt_status = mgmt_submit_ioctl(pcontroller, pioctl_handle);
    }

    return mgmt_status;
}

MGMT_STATUS
mgmt_get_all_session_id(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           is_polled)
{
    PMGMT_IOCTL_HANDLE      pioctl_handle;
    MGMT_STATUS             mgmt_status;
    ACIT_UEXACT32           expected_data_size, flags, poll;
    PGET_ALL_SESSION_ID_PARAMS_IN psessionid_in;

    if(is_polled)
    {
        poll = MGMT_IOCTL_POLLED;
    }
    else
    {
        poll = MGMT_IOCTL_NON_POLLED;
    }

    expected_data_size = GET_ALL_SESSION_ID_PARAMS_MAXSIZE;

    flags = MGMT_SET_ALLOCATION_FLAGS(poll,
                                      MGMT_IOCTL_INTERNAL,
                                      MGMT_IOCTL_RESPONSE_BUFFER);

    pioctl_handle = mgmt_alloc_ioctl_resources(pcontroller,
                                               ACIT_NULL,
                                               expected_data_size,
                                               flags);

    if(pioctl_handle == ACIT_NULL)
    {
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }

    psessionid_in = (PGET_ALL_SESSION_ID_PARAMS_IN)pioctl_handle->pmgmt_buffer;
    psessionid_in->session_flags = MGMT_SESSION_FLAG_DATA_SESSION;

    pioctl_handle->pmgmt_post_routine = mgmt_get_all_session_id_callback;
    pioctl_handle->timeout = MGMT_IOCTL_GET_ALL_SESSION_ID_TIMEOUT;

    mgmt_fill_ioctl_hdr(pioctl_handle,
                        IOCTL_SUBSYSTEM_ISCSI_INI,
                        OPCODE_ISCSI_INI_SESSION_GET_ALL_SESSION_ID,
                        0, // domain 0
                        expected_data_size,
                        pioctl_handle->timeout);
    
    mgmt_status = mgmt_fill_ioctl_request(pcontroller, pioctl_handle);
    
    if(mgmt_status == MGMT_STATUS_SUCCESS)
    {
        mgmt_status = mgmt_submit_ioctl(pcontroller, pioctl_handle);
    }

    return mgmt_status;
}

MGMT_STATUS
mgmt_get_all_session_id_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    ACIT_U32                i;
    PGET_ALL_SESSION_ID_PARAMS_OUT psessions_out;
    
    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);

    OSM_ASSERT(presponse_hdr->opcode == OPCODE_ISCSI_INI_SESSION_GET_ALL_SESSION_ID);

    if(presponse_hdr->status == MGMT_STATUS_SUCCESS)
    {
        psessions_out = (PGET_ALL_SESSION_ID_PARAMS_OUT)pdata_buffer;

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Session Count = %d\n", psessions_out->session_id_count));

        for (i=0; i < psessions_out->session_id_count; i++)
        {
            if(ap_check_duplicate_target(pcontroller, 
                                         psessions_out->session_id_list[i].session_handle,
                                         ACIT_NULL) != 
                ACIT_STATUS_DUPLICATE_ENTRY)
            {
                ap_add_target(pcontroller, 
                              psessions_out->session_id_list[i].session_handle, 
                              0,    // domain is 0 for persistent targets
                              MGMT_SESSION_STATUS_OPENED,
                              0 // is_login_pending
                              );
            }
        }
    }
    else
    {
        // Firmware returns unexpected error status.
        OSM_ASSERT(0);
    }

    mgmt_free_ioctl_resources(pcontroller, pioctl_handle);
    ap_complete_discovery(pcontroller);
    return MGMT_STATUS_SUCCESS;
    
}

MGMT_STATUS
mgmt_offload_session(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_TARGET_WS         ptarget,
    POFFLOAD_SESSION_IN     poffload_session_in,
    ACIT_UEXACT32           is_polled)
{ 
    MGMT_STATUS             mgmt_status;
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    PMGMT_TARGET_WS         pmgmt_target;
    PMGMT_IOCTL_HANDLE      pioctl_handle;
    ACIT_UEXACT32           flags, poll_flag, request_size;
   
    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);
    pmgmt_target     = GET_MGMT_TARGET_WS(ptarget);

    poll_flag = (is_polled? MGMT_IOCTL_POLLED: MGMT_IOCTL_NON_POLLED);

    flags = MGMT_SET_ALLOCATION_FLAGS(poll_flag,
                                      MGMT_IOCTL_INTERNAL,
                                      0);

#if ACIT_CFG_INTERRUPT_REDIRECTION
    if(FW_SUPPORTS_INTERRUPT_REDIRECTION(pcontroller))
    {
        request_size = sizeof(OFFLOAD_SESSION_PARAMS_V1);
    }
    else
    {
        request_size = sizeof(OFFLOAD_SESSION_PARAMS);
    }
#else
    request_size = sizeof(OFFLOAD_SESSION_PARAMS);
#endif

    pioctl_handle = mgmt_alloc_ioctl_resources(pcontroller,
                                               ACIT_NULL,
                                               request_size,
                                               flags);

    if(pioctl_handle == ACIT_NULL)
    {
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }
   
    pioctl_handle->ptarget = ptarget;   
    pioctl_handle->pmgmt_post_routine = mgmt_offload_session_callback;
    pioctl_handle->timeout = MGMT_IOCTL_OFFLOAD_TIMEOUT;

    // copy only the _V0 size. If fw supports interrupt redirection it will be
    // appended later
    OSM_MEMCPY(pioctl_handle->pmgmt_buffer, poffload_session_in, sizeof(OFFLOAD_SESSION_IN));

    mgmt_fill_ioctl_hdr(pioctl_handle,
                        IOCTL_SUBSYSTEM_ISCSI_INI,
                        OPCODE_ISCSI_INI_DRIVER_OFFLOAD_SESSION,
                        GET_TARGET_DOMAIN(pcontroller, ptarget),
                        request_size,
                        pioctl_handle->timeout);

#if ACIT_CFG_INTERRUPT_REDIRECTION
    //
    // If interrupt redirection is enabled, then also setup version to 1 to
    // tell the firmware that we need interrupt redirection
    //
    if(FW_SUPPORTS_INTERRUPT_REDIRECTION(pcontroller))
    {
        POFFLOAD_SESSION_IN_V1 poffload_session_in_v1;

        poffload_session_in_v1 = (POFFLOAD_SESSION_IN_V1)pioctl_handle->pmgmt_buffer;

        if(DRVR_SUPPORTS_INTERRUPT_REDIRECTION(pcontroller))
        {
            mgmt_fill_ioctl_hdr_version(pioctl_handle, 1);
            poffload_session_in_v1->ifd_state = IFD_STATE_ENABLE_V1;
        }
        else
        {
            poffload_session_in_v1->ifd_state = IFD_STATE_DISABLE_V1;
        }
    }
#endif

    mgmt_status = mgmt_fill_ioctl_request(pcontroller, pioctl_handle);

    if(mgmt_status != MGMT_STATUS_SUCCESS)
    {
        mgmt_free_ioctl_resources(pcontroller, pioctl_handle);
    }
    else
    {
        mgmt_status = mgmt_submit_ioctl(pcontroller, pioctl_handle);
    }

    return mgmt_status;
}

MGMT_STATUS
mgmt_offload_session_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  pioctl_hdr,
    PACIT_VOID              pdata_buffer)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    POFFLOAD_SESSION_PARAMS psession_info;
    PACIT_TARGET_WS         ptarget;
    PMGMT_TARGET_WS         pmgmt_target;
    ACIT_UEXACT32           iscsi_status;
    MGMT_STATUS             mgmt_status;
    
    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);     

    OSM_ASSERT(pioctl_hdr->opcode == OPCODE_ISCSI_INI_DRIVER_OFFLOAD_SESSION);

    ptarget = pioctl_handle->ptarget;
    pmgmt_target = GET_MGMT_TARGET_WS(ptarget);
    mgmt_status = pioctl_hdr->status;
    psession_info = (POFFLOAD_SESSION_PARAMS)(pdata_buffer);

    if(mgmt_status == MGMT_STATUS_SUCCESS)
    {
        ptarget->eh_stats.offloads++;
        iscsi_status = iscsi_update_session_info(pcontroller, ptarget, &psession_info->response);
        if(iscsi_status != ACIT_STATUS_SUCCESS)
            mgmt_status = MGMT_STATUS_ERROR_ACITISCSI;
    }

    if(mgmt_status != MGMT_STATUS_SUCCESS)
    {
        ACIT_UEXACT32       extended_error = 0;

        pcontroller->eh_stats.offload_failures++;

        // the error log is also filled with mgmt_status, additional status and
        // the session id which saw the offload error
        extended_error = GET_SESSION_HANDLE(pcontroller, ptarget) << 16 |
                         ((ACIT_UEXACT32)mgmt_status) << 8 |
                         pioctl_hdr->additional_status;

        ap_log_osm_event(pcontroller,
                         UNI_CFG_LOG_LEVEL_CRITICAL,
                         UNI_SUBMODULE_ACIT_MGMT,
                         UNI_PHASE_IOCTL,
                         1,
                         ACIT_EC_OFFLOAD_ERROR,
                         0, ACIT_NULL,
                         &extended_error);

        // Firmware returns unexpected error status
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Offload failed for target 0x%p session %d - mgmt_status 0x%x\n", 
                        ptarget, psession_info->response.session_handle, mgmt_status));

        ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_OFFLOAD_ERR);

        //
        // Notify iSCSI that offload has failed. iSCSI will decide what
        // mgmt_status to return
        //
        mgmt_status = iscsi_process_ws_init_failure(pcontroller, 
                                                    ptarget, 
                                                    mgmt_status);
    }

    mgmt_free_ioctl_resources(pcontroller, pioctl_handle);

    return mgmt_status;
}

// No MCS support

#if 0
MGMT_STATUS
mgmt_offload_cxn(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           session_handle,
    ACIT_UEXACT32           connection_handle,
    ACIT_UEXACT16           cid,
    ACIT_UEXACT16           cq_id,
    ACIT_UEXACT16           def_hdr_id,
    ACIT_UEXACT16           def_data_id,
    PPHYS_ADDR              ptemplate_address,
    ACIT_UEXACT32           is_polled)
{ 
    MGMT_STATUS             mgmt_status;
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    PMGMT_TARGET_WS         pmgmt_target;
    PMGMT_IOCTL_HANDLE      pioctl_handle;
    ACIT_UEXACT32           flags, poll_flag, request_size;
    POFFLOAD_CONN_IN        poffload_conx_in;
    POFFLOAD_CONN_IN_V1     poffload_conx_in_v1;

   
    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);
    pmgmt_target     = GET_MGMT_TARGET_WS(ptarget);

    poll_flag = (is_polled? MGMT_IOCTL_POLLED: MGMT_IOCTL_NON_POLLED);

    flags = MGMT_SET_ALLOCATION_FLAGS(poll_flag,
                                      MGMT_IOCTL_INTERNAL,
                                      0);

#if ACIT_CFG_INTERRUPT_REDIRECTION
    if(FW_SUPPORTS_INTERRUPT_REDIRECTION(pcontroller))
    {
        request_size = sizeof(OFFLOAD_CONN_PARAMS_V1);
    }
    else
    {
        request_size = sizeof(OFFLOAD_CONN_PARAMS);
    }
#else
    request_size = sizeof(OFFLOAD_CONN_PARAMS);
#endif

    pioctl_handle = mgmt_alloc_ioctl_resources(pcontroller,
                                               ACIT_NULL,
                                               request_size,
                                               flags);

    if(pioctl_handle == ACIT_NULL)
    {
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }
   
    pioctl_handle->ptarget = ptarget;   
    pioctl_handle->pmgmt_post_routine = mgmt_offload_cxn_callback;
    pioctl_handle->timeout = MGMT_IOCTL_OFFLOAD_TIMEOUT;

    poffload_conx_in = pioctl_handle->pmgmt_buffer;
    poffload_conx_in->session_handle = session_handle;
    poffload_conx_in->connection_handle = (ACIT_UEXACT16)connection_handle;
    poffload_conx_in->cid = cid;
    poffload_conx_in->cq_id = cq_id;
    poffload_conx_in->defq_id = def_hdr_id;
    poffload_conx_in->hdr_ring_id = def_hdr_id;
    poffload_conx_in->data_ring_id = def_data_id;
    poffload_conx_in->dataout_template_pa.lo = ptemplate_address->lo;
    poffload_conx_in->dataout_template_pa.hi = ptemplate_address->hi;

    mgmt_fill_ioctl_hdr(pioctl_handle,
                        IOCTL_SUBSYSTEM_ISCSI_INI,
                        OPCODE_ISCSI_INI_DRIVER_OFFLOAD_CONN,
                        GET_TARGET_DOMAIN(pcontroller, ptarget),
                        request_size,
                        pioctl_handle->timeout);
#if ACIT_CFG_INTERRUPT_REDIRECTION
    //
    // If interrupt redirection is enabled, then also setup version to 1 and
    // ifd_enable flag to tell the firmware that we need interrupt redirection
    //
    if(FW_SUPPORTS_INTERRUPT_REDIRECTION(pcontroller))
    {
        poffload_conx_in_v1 = poffload_conx_in;
        if(DRVR_SUPPORTS_INTERRUPT_REDIRECTION(pcontroller))
        {
            mgmt_fill_ioctl_hdr_version(pioctl_handle, 1);
            poffload_conx_in_v1->ifd_state = IFD_STATE_ENABLE_V1;
        }
        else
        {
            poffload_conx_in_v1->ifd_state = IFD_STATE_DISABLE_V1;
        }
    }
#endif
    mgmt_status = mgmt_fill_ioctl_request(pcontroller, pioctl_handle);

    if(mgmt_status != MGMT_STATUS_SUCCESS)
    {
        mgmt_free_ioctl_resources(pcontroller, pioctl_handle);
    }
    else
    {
        mgmt_status = mgmt_submit_ioctl(pcontroller, pioctl_handle);
    }

    return mgmt_status;
}

MGMT_STATUS
mgmt_offload_cxn_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  pioctl_hdr,
    PACIT_VOID              pdata_buffer)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    POFFLOAD_CONN_PARAMS    pcxn_info;
    PACIT_TARGET_WS         ptarget;
    PMGMT_TARGET_WS         pmgmt_target;
    MGMT_STATUS             mgmt_status;
    
    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);     

    ptarget = pioctl_handle->ptarget;
    pmgmt_target = GET_MGMT_TARGET_WS(ptarget);

    pcxn_info = (POFFLOAD_CONN_PARAMS)pdata_buffer;

    if(pioctl_hdr->status == MGMT_STATUS_SUCCESS &&
       (pcxn_info->response.connection_info.conn_status & MGMT_CONN_STATUS_ISCSI_LOGIN_STATE_FULL))
    {
        iscsi_init_connection_information(ptarget, &pcxn_info->response.connection_info, ACIT_FALSE);
    }
    else
    {
        // Firmware returns unexpected error status.
        OSM_ASSERT(0);
        pioctl_hdr->status = MGMT_STATUS_FAILED;
    }

    mgmt_status = pioctl_hdr->status;

    mgmt_free_ioctl_resources(pcontroller, pioctl_handle);
    return mgmt_status;
}
#endif

MGMT_STATUS
mgmt_get_boot_session(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           discover)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    PMGMT_IOCTL_HANDLE      pioctl_handle;
    MGMT_STATUS             mgmt_status;
    ACIT_UEXACT32           flags, expected_data_size, poll;

    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);

    if(discover)
        poll = MGMT_IOCTL_POLLED;
    else
        poll = MGMT_IOCTL_NON_POLLED;

    flags = MGMT_SET_ALLOCATION_FLAGS(poll,
                                      MGMT_IOCTL_INTERNAL,
                                      MGMT_IOCTL_RESPONSE_BUFFER);

    expected_data_size = (sizeof(IOCTL_ISCSI_INI_BOOT_GET_BOOT_TARGET) -
                                               sizeof(IOCTL_HEADER));

    pioctl_handle = mgmt_alloc_ioctl_resources(pcontroller,
                                               ACIT_NULL,
                                               expected_data_size,
                                               flags);

    if (pioctl_handle == ACIT_NULL)
    {
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }

    pioctl_handle->timeout = MGMT_EH_IOCTL_TIMEOUT;

    if(discover)
    {
        pioctl_handle->pmgmt_post_routine = mgmt_discover_boot_target_callback;
    }
    else
    {
        pioctl_handle->pmgmt_post_routine = mgmt_get_boot_session_callback;
    }

    mgmt_fill_ioctl_hdr(pioctl_handle,
                        IOCTL_SUBSYSTEM_ISCSI_INI,
                        OPCODE_ISCSI_INI_BOOT_GET_BOOT_TARGET,
                        0, // domain 0
                        expected_data_size,
                        pioctl_handle->timeout);

    mgmt_status = mgmt_fill_ioctl_request(pcontroller, pioctl_handle);
    
    if(mgmt_status == MGMT_STATUS_SUCCESS)
    {
        mgmt_status = mgmt_submit_ioctl(pcontroller, pioctl_handle);
    }
        
    return mgmt_status;
}

MGMT_STATUS
mgmt_discover_boot_target_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    MGMT_STATUS             mgmt_status;
    PGET_BOOT_TARGET_OUT    pboot_target = ACIT_NULL;
    
    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);
    mgmt_status = presponse_hdr->status;

    pboot_target = (PGET_BOOT_TARGET_OUT)pioctl_handle->pmgmt_buffer;
    
    if(IS_BOOT_SESSION_OPEN(pboot_target))
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Reporting boot session, handle = %d\n", pboot_target->boot_session_handle));
        ap_add_target(pcontroller, pboot_target->boot_session_handle, 0, 0, 0);
    }
    else
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Boot session not found while discovery!!!\n"));
        OSM_ASSERT(0);
    }

    mgmt_free_ioctl_resources(pcontroller, pioctl_handle);
    ap_complete_discovery(pcontroller);

    return mgmt_status;
}


MGMT_STATUS
mgmt_get_boot_session_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    MGMT_STATUS             mgmt_status;
    PGET_BOOT_TARGET_OUT    pboot_target = ACIT_NULL;
    ACIT_UEXACT32           boot_session;
    
    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);
    mgmt_status = presponse_hdr->status;

    pboot_target = (PGET_BOOT_TARGET_OUT)pioctl_handle->pmgmt_buffer;
    
    if(IS_BOOT_SESSION_OPEN(pboot_target))
    {
        boot_session = ACIT_TRUE;
    }
    else
    {
        boot_session = ACIT_FALSE;
    }

    mgmt_free_ioctl_resources(pcontroller, pioctl_handle);

    //
    // If boot session is configured, forward the IOCTL to ARM fw, else
    // close all sessions, issue the set IP IOCTL and reopen sessions again
    //
    if(boot_session)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Boot session configured, will not reopen session!!\n"));

        mgmt_process_external_ioctl_default(pcontroller, 
                                            pmgmt_controller->porig_ioctl_request);
    }
    else
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Boot session not configured, reopen all sessions!!\n"));

        iscsi_recover_all_sessions(pcontroller, 
                                   pmgmt_controller->porig_ioctl_request);

    }

    pmgmt_controller->porig_ioctl_request = ACIT_NULL;

    return mgmt_status;
}


MGMT_STATUS
mgmt_delete_nonpersistent_targets(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PMGMT_IOCTL_HANDLE      pioctl_handle;
    MGMT_STATUS             mgmt_status;
    ACIT_UEXACT32           flags;

    flags = MGMT_SET_ALLOCATION_FLAGS(MGMT_IOCTL_POLLED,
                                      MGMT_IOCTL_INTERNAL,
                                      MGMT_IOCTL_NO_RESPONSE_BUFFER);

    pioctl_handle = mgmt_alloc_ioctl_resources(pcontroller,
                                               ACIT_NULL,
                                               0,
                                               flags);

    if (pioctl_handle == ACIT_NULL)
    {
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }
 
    pioctl_handle->timeout = MGMT_INTERNAL_IOCTL_DEFAULT_TIMEOUT;
    pioctl_handle->pmgmt_post_routine = ACIT_NULL;

    mgmt_fill_ioctl_hdr(pioctl_handle,
                        IOCTL_SUBSYSTEM_ISCSI_INI,
                        OPCODE_ISCSI_INI_DRIVER_DEL_ALL_NONPERSISTENT_SESSIONS,
                        0, // domain 0
                        0,
                        pioctl_handle->timeout);

    mgmt_status = mgmt_fill_ioctl_request(pcontroller, pioctl_handle);

    if(mgmt_status != MGMT_STATUS_SUCCESS)
    {
        mgmt_free_ioctl_resources(pcontroller, pioctl_handle);
        return mgmt_status;
    }

    mgmt_status = mgmt_submit_ioctl(pcontroller, pioctl_handle);

    mgmt_free_ioctl_resources(pcontroller, pioctl_handle);

    return mgmt_status;
}

MGMT_STATUS
mgmt_login_all_targets(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           is_polled)
{
    PMGMT_IOCTL_HANDLE      pioctl_handle;
    MGMT_STATUS             mgmt_status;
    ACIT_UEXACT32           flags, poll = MGMT_IOCTL_NON_POLLED;

    if(is_polled)
    {
        poll = MGMT_IOCTL_POLLED;
    }

    flags = MGMT_SET_ALLOCATION_FLAGS(poll,
                                      MGMT_IOCTL_INTERNAL,
                                      MGMT_IOCTL_NO_RESPONSE_BUFFER);

    pioctl_handle = mgmt_alloc_ioctl_resources(pcontroller,
                                               ACIT_NULL,
                                               0,
                                               flags);

    if (pioctl_handle == ACIT_NULL)
    {
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }

    //
    // 05/20 - This is a bit of a hack. The Windows OSM needs a huge timeout
    // to work in the case of resuming from standby where persistent target
    // logins will take sometime to complete. But Linux OSM doesn't need such
    // huge timeout and will infact panic (NMI watchdog) if ACIT polls for
    // more than 5 secs. So the solution is for the Linux OSM to use async
    // mode of discovery even during init. ACIT will also provide smaller
    // timeout in this case. This is to prevent any wait if an unload is attempted
    // immediately after load (stop_controller first waits for all IOCTLs to complete).
    //
    if(is_polled)
    {
        // Set Timeout to 90 sec for 64 targets
        pioctl_handle->timeout = MGMT_INTERNAL_IOCTL_DEFAULT_TIMEOUT * 3;
    }
    else
    {
        pioctl_handle->timeout = MGMT_EH_IOCTL_TIMEOUT;
    }

    pioctl_handle->pmgmt_post_routine = ACIT_NULL;

    mgmt_fill_ioctl_hdr(pioctl_handle,
                        IOCTL_SUBSYSTEM_ISCSI_INI,
                        OPCODE_ISCSI_INI_DRIVER_LOGIN_ALL_SESSIONS,
                        0, // domain 0
                        0,
                        pioctl_handle->timeout);

    mgmt_status = mgmt_fill_ioctl_request(pcontroller, pioctl_handle);

    if(mgmt_status != MGMT_STATUS_SUCCESS)
    {
        mgmt_free_ioctl_resources(pcontroller, pioctl_handle);
        return mgmt_status;
    }

    mgmt_status = mgmt_submit_ioctl(pcontroller, pioctl_handle);

    mgmt_free_ioctl_resources(pcontroller, pioctl_handle);

    return mgmt_status;

}

MGMT_STATUS
mgmt_invalidate_connection(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT16           cid,
    ACIT_UEXACT16           issue_reset,
    ACIT_UEXACT16           savecfg_flag)
{
    PMGMT_IOCTL_HANDLE      pioctl_handle;
    MGMT_STATUS             mgmt_status;
    ACIT_UEXACT32           flags, expected_data_size;
    PIOCTL_ISCSI_INVALIDATE_CONNECTION_PARAMS pinv_conn;

    expected_data_size = sizeof(IOCTL_ISCSI_INVALIDATE_CONNECTION_PARAMS);

    flags = MGMT_SET_ALLOCATION_FLAGS(MGMT_IOCTL_NON_POLLED,
                                      MGMT_IOCTL_INTERNAL,
                                      MGMT_IOCTL_NO_RESPONSE_BUFFER);

    pioctl_handle = mgmt_alloc_ioctl_resources(pcontroller,
                                               ACIT_NULL,
                                               expected_data_size,
                                               flags);

    if (pioctl_handle == ACIT_NULL)
    {
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }

    pioctl_handle->ptarget = (PACIT_VOID)(ACIT_U32_64)cid;
    pioctl_handle->pmgmt_post_routine = mgmt_invalidate_connection_callback;
    pioctl_handle->timeout = MGMT_EH_IOCTL_TIMEOUT;

    pinv_conn = (PIOCTL_ISCSI_INVALIDATE_CONNECTION_PARAMS)pioctl_handle->pmgmt_buffer;

    pinv_conn->request.session_handle = GET_SESSION_HANDLE(pcontroller, ptarget);
    pinv_conn->request.cid= cid;
    if(issue_reset)
        pinv_conn->request.cleanup_type = IOCTL_ISCSI_CONNECTION_ISSUE_TCP_RST;
    else
        pinv_conn->request.cleanup_type = IOCTL_ISCSI_CONNECTION_INVALIDATE;

    pinv_conn->request.save_cfg = savecfg_flag;

    mgmt_fill_ioctl_hdr(pioctl_handle,
                        IOCTL_SUBSYSTEM_ISCSI_INI,
                        OPCODE_ISCSI_INI_DRIVER_INVALIDATE_CONNECTION,
                        GET_TARGET_DOMAIN(pcontroller, ptarget),
                        expected_data_size,
                        pioctl_handle->timeout);

    mgmt_status = mgmt_fill_ioctl_request(pcontroller, pioctl_handle);
    
    if(mgmt_status == MGMT_STATUS_SUCCESS)
    {
        mgmt_status = mgmt_submit_ioctl(pcontroller, pioctl_handle);
    }
    else
    {
        mgmt_free_ioctl_resources(pcontroller, pioctl_handle);
    }

    return mgmt_status;
}

MGMT_STATUS
mgmt_invalidate_connection_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer)
{
    PIOCTL_ISCSI_INVALIDATE_CONNECTION_PARAMS pinv_conn;
    ACIT_UEXACT16           cid;

    OSM_ASSERT(presponse_hdr->status == MGMT_STATUS_SUCCESS);

    cid = (ACIT_UEXACT16)((ACIT_U32_64)pioctl_handle->ptarget);
    pinv_conn = (PIOCTL_ISCSI_INVALIDATE_CONNECTION_PARAMS)pdata_buffer;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING, 
                   ("Received invalidate connection completion for cid 0x%x\n", cid));

    mgmt_free_ioctl_resources(pcontroller, pioctl_handle);

    iscsi_connection_invalidate_notify(pcontroller, 
                                       (ACIT_UEXACT16)cid);

    return MGMT_STATUS_SUCCESS;
}

MGMT_STATUS
mgmt_upload_connection(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT16           cid,
    ACIT_UEXACT32           upload_flag)
{
    PMGMT_IOCTL_HANDLE      pioctl_handle;
    MGMT_STATUS             mgmt_status;
    PIOCTL_COMMON_TCP_UPLOAD_PARAMS pupload_cxn;
    ACIT_UEXACT32           flags;

    //
    // for reopen a session and upload IOCTL, check if the delayed IOCTL
    // count is greater that allowed. If it is, then do not
    // submit this IOCTL since it will backup in the ARM fw queue
    // and will waste a MCC WRB for the driver. It will instead
    // be queued up in the driver
    //

    if(pcontroller->counters.delayed_ioctls >= (pcontroller->params.ioctls_per_ctrl / 2))
    {
        // this return code will cause the caller to queue the request
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }

    flags = MGMT_SET_ALLOCATION_FLAGS(MGMT_IOCTL_NON_POLLED,
                                      MGMT_IOCTL_INTERNAL,
                                      MGMT_IOCTL_RESPONSE_BUFFER);

    pioctl_handle = mgmt_alloc_ioctl_resources(pcontroller,
                                               ACIT_NULL,
                                               sizeof(IOCTL_COMMON_TCP_UPLOAD),
                                               flags);

    if(pioctl_handle == ACIT_NULL)
    {
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }

    pupload_cxn = (PIOCTL_COMMON_TCP_UPLOAD_PARAMS)pioctl_handle->pmgmt_buffer;

    pupload_cxn->request.id = (ACIT_UEXACT16)cid;
    pupload_cxn->request.upload_type = (ACIT_UEXACT8)upload_flag;

    pioctl_handle->pmgmt_post_routine = mgmt_upload_connection_callback;
    pioctl_handle->timeout = MGMT_EH_IOCTL_TIMEOUT;
    pioctl_handle->ptarget = (PACIT_VOID)(ACIT_U32_64)cid;

    mgmt_fill_ioctl_hdr(pioctl_handle,
                        SUBSYSTEM_COMMON_TCP_UPLOAD,
                        OPCODE_COMMON_TCP_UPLOAD,
                        GET_TARGET_DOMAIN(pcontroller, ptarget),
                        sizeof(IOCTL_COMMON_TCP_UPLOAD),
                        pioctl_handle->timeout);
    
    mgmt_status = mgmt_fill_ioctl_request(pcontroller, pioctl_handle);
    
    if(mgmt_status != MGMT_STATUS_SUCCESS)
    {
        mgmt_free_ioctl_resources(pcontroller, pioctl_handle);
    }
    else
    {
        pcontroller->counters.delayed_ioctls++;
        mgmt_status = mgmt_submit_ioctl(pcontroller, pioctl_handle);
    }

    return mgmt_status;
}

MGMT_STATUS
mgmt_upload_connection_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer)
{
    PIOCTL_COMMON_TCP_UPLOAD_PARAMS pupload_cxn;
    ACIT_UEXACT32           cid;
    
    if(presponse_hdr->status != MGMT_STATUS_SUCCESS)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("IOCTL response header status = 0x%x\n", presponse_hdr->status));
        OSM_ASSERT(0);
    }

    pupload_cxn = (PIOCTL_COMMON_TCP_UPLOAD_PARAMS)pdata_buffer;
    cid = (ACIT_UEXACT16)((ACIT_U32_64)pioctl_handle->ptarget);

    if(presponse_hdr->status == MGMT_STATUS_SUCCESS)
    {
        OSM_ASSERT(presponse_hdr->response_length);
    }
    
    mgmt_free_ioctl_resources(pcontroller, pioctl_handle);
    pcontroller->counters.delayed_ioctls--;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING, 
                   ("Received upload connection completion cid 0x%x\n", cid));

    iscsi_connection_upload_notify(pcontroller, 
                                   cid);

    return MGMT_STATUS_SUCCESS;
}

MGMT_STATUS
mgmt_logout_fw_session(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget)
{
    PMGMT_IOCTL_HANDLE      pioctl_handle;
    MGMT_STATUS             mgmt_status;
    ACIT_UEXACT32           flags;
    PLOGOUT_TARGET_PARAMS_IN    plogout_in;

    flags = MGMT_SET_ALLOCATION_FLAGS(MGMT_IOCTL_POLLED,
                                      MGMT_IOCTL_INTERNAL,
                                      MGMT_IOCTL_RESPONSE_BUFFER);

    pioctl_handle = mgmt_alloc_ioctl_resources(pcontroller,
                                               ACIT_NULL,
                                               sizeof(LOGOUT_TARGET_PARAMS),
                                               flags);

    if (pioctl_handle == ACIT_NULL)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Failed allocating IOCTL handle!\n"));
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }

    pioctl_handle->timeout = MGMT_EH_IOCTL_TIMEOUT;
    pioctl_handle->pmgmt_post_routine = ACIT_NULL;

    plogout_in = (PLOGOUT_TARGET_PARAMS_IN)pioctl_handle->pmgmt_buffer;
    plogout_in->session_handle = GET_SESSION_HANDLE(pcontroller, ptarget);

    mgmt_fill_ioctl_hdr(pioctl_handle,
                        IOCTL_SUBSYSTEM_ISCSI_INI,
                        OPCODE_ISCSI_INI_SESSION_LOGOUT_TARGET,
                        GET_TARGET_DOMAIN(pcontroller, ptarget),
                        sizeof(PLOGOUT_TARGET_PARAMS),
                        pioctl_handle->timeout);

    mgmt_status = mgmt_fill_ioctl_request(pcontroller, pioctl_handle);

    if(mgmt_status != MGMT_STATUS_SUCCESS)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Failed filling IOCTL request!\n"));
        mgmt_free_ioctl_resources(pcontroller, pioctl_handle);
        return mgmt_status;
    }

    mgmt_status = mgmt_submit_ioctl(pcontroller, pioctl_handle);

    mgmt_free_ioctl_resources(pcontroller, pioctl_handle);

    if(mgmt_status != MGMT_STATUS_SUCCESS)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Failed logging out of ARM fw session!\n"));
        OSM_ASSERT(0);
    }

    pcontroller->stats.data_session_armfw_logouts++;

    return mgmt_status;
}

MGMT_STATUS
mgmt_peek_icd_inv_resources(
    PACIT_CONTROLLER_WS     pcontroller)
{
    return mgmt_peek_internal_resources(GET_MGMT_CONTROLLER_WS(pcontroller));
}

MGMT_STATUS
mgmt_invalidate_icds(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_INVALIDATE_HANDLE pinv_handle)
{
    PMGMT_IOCTL_HANDLE      pioctl_handle;
    MGMT_STATUS             mgmt_status;
    ACIT_UEXACT32           flags, expected_data_size;
    PACIT_DC_LIST           flink;
    ACIT_UEXACT16           icd = 0, cid = 0;
    PINVALIDATE_COMMANDS_PARAMS pinv_cmds;

    pinv_handle->submit_tick = pcontroller->counters.current_tick_count;

    expected_data_size = sizeof(INVALIDATE_COMMANDS_PARAMS);

    flags = MGMT_SET_ALLOCATION_FLAGS(MGMT_IOCTL_NON_POLLED,
                                      MGMT_IOCTL_INTERNAL,
                                      MGMT_IOCTL_RESPONSE_BUFFER);

    pioctl_handle = mgmt_alloc_ioctl_resources(pcontroller,
                                               ACIT_NULL,
                                               expected_data_size,
                                               flags);

    if (pioctl_handle == ACIT_NULL)
    {
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }

    pioctl_handle->pmgmt_post_routine = mgmt_invalidate_icds_callback;
    pioctl_handle->timeout = MGMT_INTERNAL_IOCTL_DEFAULT_TIMEOUT;

    mgmt_fill_ioctl_hdr(pioctl_handle,
                        IOCTL_SUBSYSTEM_COMMON_ISCSI,
                        OPCODE_COMMON_ISCSI_ERROR_RECOVERY_INVALIDATE_COMMANDS,
                        0, // domain 0
                        expected_data_size,
                        pioctl_handle->timeout);

    pinv_cmds = (PINVALIDATE_COMMANDS_PARAMS)pioctl_handle->pmgmt_buffer;

    /* Populate the ICDs and CIDs from the list */
    flink = pinv_handle->list_header.next;

    pinv_cmds->request.ref_handle = pinv_handle->index;
    pinv_cmds->request.cleanup_type = IOCTL_ISCSI_COMMAND_INVALIDATE;
    pinv_cmds->request.icd_count = 0;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("Invalidating num_invs %d ICDs of type %d invh 0x%p\n"), 
                                            pinv_handle->num_invs,
                                            pinv_handle->flags,
                                            pinv_handle));

    while(flink != &pinv_handle->list_header)
    {
        OSM_ASSERT(pinv_cmds->request.icd_count < MGMT_MAX_SGL_ICD_INVALIDATE);

        icd = ACIT_U16_MINUS_ONE; cid = ACIT_U16_MINUS_ONE;
        ap_get_inv_params(pcontroller, pinv_handle, flink, &icd, &cid);

        pinv_cmds->request.table[pinv_cmds->request.icd_count].icd = icd;
        pinv_cmds->request.table[pinv_cmds->request.icd_count].cid = cid;
        pinv_cmds->request.icd_count++;

        flink = flink->next;
    }
    
    mgmt_status = mgmt_fill_ioctl_request(pcontroller, pioctl_handle);
    
    if(mgmt_status != MGMT_STATUS_SUCCESS)
    {
        mgmt_free_ioctl_resources(pcontroller, pioctl_handle);
    }
    else
    {
        mgmt_status = mgmt_submit_ioctl(pcontroller, pioctl_handle);
    }

    OSM_ASSERT(mgmt_status == MGMT_STATUS_SUCCESS);
    return mgmt_status;

}

INLINE
ACIT_UEXACT32
mgmt_status_from_icd_status(
    ACIT_UEXACT32           icd_status)
{
    switch(icd_status & MGMT_SGL_ICD_STATUS_MASK)
    {
    case SGL_ICD_CLEANUP_SUCCESS:
        return ACIT_STATUS_SUCCESS;

    case SGL_ICD_NOT_FOUND:
        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Invalidation returned unknown ICD %d\n", icd_status));
        return ACIT_STATUS_INVALID_PARAMETER;

    case SGL_ICD_INVALIDATION_ERROR:
        OSM_ASSERT(0);
        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Invalidation returns a failure %d\n", icd_status));
        return ACIT_STATUS_ERROR;

    default:
        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Invalidation returned unknown status %d\n", icd_status));
        return ACIT_STATUS_ERROR;
    }

    return ACIT_STATUS_ERROR;
}

MGMT_STATUS
mgmt_invalidate_icds_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer)
{
    PACIT_DC_LIST           flink;
    PACIT_INVALIDATE_HANDLE pinv_handle;
    ACIT_UEXACT32           status, byte_count = 0, index = 0;
    PINVALIDATE_COMMANDS_PARAMS pinv_cmds;
    
    pinv_cmds = (PINVALIDATE_COMMANDS_PARAMS)pdata_buffer;
    pinv_handle = &pcontroller->pinv_handle_base[pinv_cmds->response.ref_handle];
    OSM_ASSERT(pinv_handle);

    pinv_handle->complete_tick = pcontroller->counters.current_tick_count;

    flink = pinv_handle->list_header.next;

    while(flink != &pinv_handle->list_header)
    {
        OSM_ASSERT(index < pinv_cmds->response.icd_count);

        if(presponse_hdr->status == MGMT_STATUS_SUCCESS)
        {
            status = mgmt_status_from_icd_status(pinv_cmds->response.icd_status[index]);
        }
        else
        {
            status = ACIT_STATUS_ERROR;

            if(presponse_hdr->status == MGMT_STATUS_POLL_IOCTL_TIMEOUT)
                pcontroller->eh_stats.inv_sts_ioctl_timeout++;
            else
                pcontroller->eh_stats.inv_sts_ioctl_err++;
        }

        switch (status)
        {
        case ACIT_STATUS_SUCCESS:
            pcontroller->eh_stats.inv_sts_success++;
            break;
        case ACIT_STATUS_INVALID_PARAMETER:
            pcontroller->eh_stats.inv_sts_not_found++;
            break;
        case ACIT_STATUS_ERROR:
            pcontroller->eh_stats.inv_sts_err++;
            break;
        }

        byte_count = (pinv_cmds->response.icd_status[index]) >> MGMT_SGL_ICD_BYTE_COUNT_SHIFT;

        ap_set_invalidate_icd_status(pcontroller, 
                                     pinv_handle, 
                                     flink, 
                                     status,
                                     byte_count);

        flink = flink->next;
        index++;
    }

    OSM_ASSERT(index == pinv_handle->num_invs);
    mgmt_free_ioctl_resources(pcontroller, pioctl_handle);
    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("Completing invalidations for %d ICDs of type %d invh 0x%p\n"), 
                                            pinv_handle->num_invs, pinv_handle->flags, pinv_handle));
    ap_complete_invalidate_icds(pcontroller, pinv_handle);

    return MGMT_STATUS_SUCCESS;
}

MGMT_STATUS
mgmt_target_upload_complete(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_TARGET_WS         ptarget, 
    ACIT_UEXACT32           session_handle,
    PACIT_IOCTL_REQUEST     pioctl_request)
{
    MGMT_STATUS             mgmt_status = 0;
    PIOCTL_RESPONSE_HEADER  presponse_hdr;

    /*
     * Before the session is reopened, check to see if there are any
     * IOCTLs that need to be sent
     */

    if(pioctl_request)
    {
        ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_REOPEN_IN_PROGRESS);

        mgmt_status = mgmt_start_external_ioctl(pcontroller,
                                                pioctl_request,
                                                NULL,
                                                mgmt_process_ioctl_modify_sessionloginkeys_post);
        if(mgmt_status != MGMT_STATUS_SUCCESS)
        {
            presponse_hdr = &pioctl_request->ioctl_hdr.response;
            presponse_hdr->status = mgmt_status;

            OSM_ASSERT(0);

            mgmt_return_ioctl(pcontroller, pioctl_request, mgmt_status);

            /*
             * Error submitting the IOCTL, atleast reopen the session that was
             * closed earlier
             */
        }
    }
    else
    {
        if(ACIT_IS_BIT_CLR(ptarget->flags, TGT_FLAG_OFFLOAD_RETRY_EXCEEDED))
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("issuing reopen_session IOCTL for session %d\n", session_handle));

            ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_REOPEN_IN_PROGRESS);

            mgmt_status = mgmt_reopen_sessions(pcontroller, 
                                                IOCTL_ISCSI_REOPEN_A_SESSION,
                                                session_handle,
                                                GET_TARGET_DOMAIN(pcontroller, ptarget),
                                                ACIT_FALSE);

            if(mgmt_status == MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED)
            {
                ap_set_target_event(pcontroller, ptarget, EVENT_TGT_REOPEN_PENDING);
            }
            else
            {
                OSM_ASSERT(mgmt_status == MGMT_STATUS_SUCCESS);
            }
        }
        else
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Not issuing reopen_session IOCTL for session %d because offload failure threshold exceeded\n", session_handle));
            ap_force_eto_expiry(pcontroller, ptarget);
        }
    }

    return mgmt_status;
}

MGMT_STATUS
mgmt_all_targets_close_notify(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request)
{
    MGMT_STATUS             mgmt_status;
    PIOCTL_RESPONSE_HEADER  presponse_hdr;

    OSM_ASSERT(pioctl_request);

    mgmt_status = mgmt_start_external_ioctl(pcontroller,
                                            pioctl_request,
                                            NULL,
                                            mgmt_process_set_nw_params_post);

    if(mgmt_status != MGMT_STATUS_SUCCESS)
    {
        presponse_hdr = &pioctl_request->ioctl_hdr.response;
        presponse_hdr->status = mgmt_status;

        OSM_ASSERT(0);

        /*
         * Error submitting the IOCTL, atleast reopen all sessions that were
         * closed earlier
         */
        if(pcontroller->ptarget_table->occupied_entries)
        {
            mgmt_reopen_sessions(pcontroller, 
                                 IOCTL_ISCSI_REOPEN_ALL_SESSIONS, 
                                 ACIT_U32_MINUS_ONE,
                                 0, // domain 0
                                 ACIT_FALSE);
        }
    }

    return MGMT_STATUS_SUCCESS;
}

// No MCS support

#if 0
MGMT_STATUS
mgmt_reopen_cxn(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_TARGET_WS         ptarget, 
    ACIT_UEXACT32           session_handle,
    ACIT_UEXACT32           connection_handle,
    ACIT_UEXACT32           cmd_sn)
{
    PMGMT_IOCTL_HANDLE      pioctl_handle;
    MGMT_STATUS             mgmt_status;
    ACIT_UEXACT32           flags;
    PIOCTL_ISCSI_REOPEN_A_CONN_IN p_req;

    ap_add_connection(pcontroller, 
                        session_handle, 
                        connection_handle);

    flags = MGMT_SET_ALLOCATION_FLAGS(MGMT_IOCTL_NON_POLLED,
                                      MGMT_IOCTL_INTERNAL,
                                      MGMT_IOCTL_NO_RESPONSE_BUFFER);

    pioctl_handle = mgmt_alloc_ioctl_resources(pcontroller,
                                               ACIT_NULL,
                                               0,
                                               flags);

    if (pioctl_handle == ACIT_NULL)
    {
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }

    pioctl_handle->ptarget = ptarget;   
    pioctl_handle->pmgmt_post_routine = mgmt_reopen_cxn_callback;
    pioctl_handle->timeout = MGMT_IOCTL_LOGIN_TIMEOUT;
    p_req = (PIOCTL_ISCSI_REOPEN_A_CONN_IN)pioctl_handle->pmgmt_buffer;
    p_req->session_handle = session_handle;
    p_req->connection_handle= connection_handle;
    p_req->cmd_sn = cmd_sn;

    mgmt_fill_ioctl_hdr(pioctl_handle,
                        IOCTL_SUBSYSTEM_ISCSI_INI,
                        OPCODE_ISCSI_INI_DRIVER_REOPEN_A_CONN,
                        GET_TARGET_DOMAIN(pcontroller, ptarget),
                        sizeof(IOCTL_ISCSI_REOPEN_A_CONN_PARAMS),
                        pioctl_handle->timeout);

    mgmt_status = mgmt_fill_ioctl_request(pcontroller, pioctl_handle);

    if(mgmt_status != MGMT_STATUS_SUCCESS)
    {
        mgmt_free_ioctl_resources(pcontroller, pioctl_handle);
    }
    else
    {
        mgmt_status = mgmt_submit_ioctl(pcontroller, pioctl_handle);
    }

    return mgmt_status;
}

MGMT_STATUS
mgmt_reopen_cxn_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    PACIT_TARGET_WS         ptarget;
    PMGMT_TARGET_WS         pmgmt_target;
    PIOCTL_ISCSI_REOPEN_A_CONN_PARAMS preopen_cxn;
    MGMT_STATUS             mgmt_status;

    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);
    ptarget = pioctl_handle->ptarget;
    preopen_cxn = (PIOCTL_ISCSI_REOPEN_A_CONN_PARAMS)pdata_buffer;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE, 
                   ("mgmt_reopen_cxn_callback: Received Reopen cxn completion from f/w\n"));

    mgmt_status = presponse_hdr->status;
    mgmt_free_ioctl_resources(pcontroller, pioctl_handle);

    pmgmt_target = GET_MGMT_TARGET_WS(ptarget);

    /* Mark target status as ready irrespective */
    ap_flag_ready_target(pcontroller, ptarget, TARGET_BUSY_NEW_CONNECTION_ADD);

    if(mgmt_status == MGMT_STATUS_SUCCESS)
    {
    }

    /*mgmt_offload_cxn(pcontroller, 
                     ptarget, 
                     preopen_cxn->response.session_handle,
                     preopen_cxn->response.connection_handle,
                     ACIT_FALSE);   // non polled
                     */

    return mgmt_status;
}

#endif

MGMT_STATUS
mgmt_reopen_sessions(
    PACIT_CONTROLLER_WS     pcontroller, 
    ACIT_UEXACT32           reopen_type,
    ACIT_UEXACT32           session_handle,
    ACIT_UEXACT8            domain,
    ACIT_UEXACT32           is_polled)
{
    PMGMT_IOCTL_HANDLE      pioctl_handle;
    MGMT_STATUS             mgmt_status;
    ACIT_UEXACT32           flags, poll;
    PIOCTL_ISCSI_REOPEN_ALL_SESSIONS_IN p_req;

    //
    // for reopen a session and upload IOCTL, check if the delayed IOCTL
    // count is greater that allowed. If it is, then do not
    // submit this IOCTL since it will backup in the ARM fw queue
    // and will waste a MCC WRB for the driver. It will instead
    // be queued up in the driver
    //

    if((reopen_type == IOCTL_ISCSI_REOPEN_A_SESSION) &&
       (pcontroller->counters.delayed_ioctls >= (pcontroller->params.ioctls_per_ctrl / 2)))
    {
        // this return code will cause the caller to queue the request
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }

    if(is_polled)
    {
        poll = MGMT_IOCTL_POLLED;
    }
    else
    {
        poll = MGMT_IOCTL_NON_POLLED;
    }

    flags = MGMT_SET_ALLOCATION_FLAGS(poll,
                                      MGMT_IOCTL_INTERNAL,
                                      MGMT_IOCTL_NO_RESPONSE_BUFFER);

    pioctl_handle = mgmt_alloc_ioctl_resources(pcontroller,
                                               ACIT_NULL,
                                               0,
                                               flags);
    if (pioctl_handle == ACIT_NULL)
    {
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }

    pioctl_handle->timeout = MGMT_REOPEN_SESSION_TIMEOUT;

    switch(reopen_type)
    {
    case IOCTL_ISCSI_REOPEN_ALL_SESSIONS:
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Issuing reopen all sessions...\n"));
        pcontroller->stats.reopen_all_sessions++;
        pioctl_handle->pmgmt_post_routine = mgmt_reopen_all_session_callback;
        break;

    case IOCTL_ISCSI_REOPEN_BOOT_SESSION:
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Issuing reopen boot session...\n"));
        pioctl_handle->pmgmt_post_routine = mgmt_generic_ioctl_callback;
        // For boot target alone, set a higher timeout value
        pioctl_handle->timeout = MGMT_INTERNAL_IOCTL_DEFAULT_TIMEOUT;
        break;

    case IOCTL_ISCSI_REOPEN_A_SESSION:
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Issuing reopen session - handle 0x%x...\n", session_handle));
        pioctl_handle->pmgmt_post_routine = mgmt_reopen_a_session_callback;
        // setup ptarget context in IOCTL handle. This will be used in the completion for validation
        pioctl_handle->ptarget = ap_get_target_ws_from_session_handle(pcontroller, 
                                                    session_handle);
        OSM_ASSERT(pioctl_handle->ptarget);
        pioctl_handle->ptarget->eh_stats.reopen_issued++;
        pcontroller->counters.delayed_ioctls++;
        break;

    default:

        OSM_ASSERT(0);
        break;
    }

    p_req = (PIOCTL_ISCSI_REOPEN_ALL_SESSIONS_IN)pioctl_handle->pmgmt_buffer;
    p_req->reopen_type = (ACIT_UEXACT16)reopen_type;
    p_req->session_handle = session_handle;

    mgmt_fill_ioctl_hdr(pioctl_handle,
                        IOCTL_SUBSYSTEM_ISCSI_INI,
                        OPCODE_ISCSI_INI_DRIVER_REOPEN_ALL_SESSIONS,
                        domain,
                        sizeof (IOCTL_ISCSI_REOPEN_ALL_SESSION_PARAMS),
                        pioctl_handle->timeout);

    mgmt_status = mgmt_fill_ioctl_request(pcontroller, pioctl_handle);

    if(mgmt_status != MGMT_STATUS_SUCCESS)
    {
        mgmt_free_ioctl_resources(pcontroller, pioctl_handle);

        if(reopen_type == IOCTL_ISCSI_REOPEN_A_SESSION)
        {
            pcontroller->counters.delayed_ioctls--;
        }
    }
    else
    {
        mgmt_status = mgmt_submit_ioctl(pcontroller, pioctl_handle);
    }

    return mgmt_status;
}

MGMT_STATUS
mgmt_reopen_a_session_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    PACIT_TARGET_WS         ptarget;
    ACIT_UEXACT32           session_handle;
    MGMT_STATUS             mgmt_status;
    PIOCTL_ISCSI_REOPEN_ALL_SESSIONS_OUT preopen;

    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);

    mgmt_status = presponse_hdr->status;

    //
    // bug 14131 : under some conditions ARM fw can return totally invalid value of 
    // session handle. So use the ptarget that was saved locally
    // Note!! If the ARM fw does not give a valid completion via async notification,
    // this target will be lost forever!!
    //
    ptarget = pioctl_handle->ptarget;
    OSM_ASSERT(ptarget);

    preopen = (PIOCTL_ISCSI_REOPEN_ALL_SESSIONS_OUT)pdata_buffer;
    session_handle = preopen->session_handle;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING, 
                   (("Received Reopen session completion for session_handle %d status 0x%x\n"),
                                        session_handle, mgmt_status));

    if(ptarget != ap_get_target_ws_from_session_handle(pcontroller, session_handle))
    {
#if ACIT_CFG_DEBUG_BUILD
        PACIT_TARGET_WS preceived_tgt = ap_get_target_ws_from_session_handle(pcontroller, 
                                                                             session_handle);
#endif
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,
                    (("Target mismatch - sent 0x%p received 0x%p\n"), 
                                        ptarget, preceived_tgt));
        pcontroller->eh_stats.reopen_mismatch++;
        ptarget->eh_stats.reopen_mismatch++;
        OSM_ASSERT(0);
    }

    mgmt_free_ioctl_resources(pcontroller, pioctl_handle);

    OSM_ASSERT(pcontroller->counters.delayed_ioctls);
    pcontroller->counters.delayed_ioctls--;

    if(ACIT_IS_BIT_SET(pcontroller->busy, CONTROLLER_BUSY_SHUTDOWN_IN_PROGRESS))
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING, ("Return from reopen_session_callback without completing recovery\n"));
        return MGMT_STATUS_SUCCESS;
    }

    //
    // ARM can report target notifications async BEFORE completing reopen session
    // IOCTL, so we ignore this response if that's the case
    //
    if(!(iscsi_is_session_open(ptarget)))
    {
        //        
        // If there were errors while reopening the session, then do not switch the session,
        // complete the recovery process with error status
        //

        if(mgmt_status != MGMT_STATUS_SUCCESS)
        {
            ptarget->eh_stats.reopen_error++;

            //
            // Bug 19481 : ETO should be started irrespective of link
            // status. Do not call ap_complete_target_recovery, since
            // that function checks the link. Directly call ap_start_eto
            // instead
            //
//            ap_complete_target_recovery(pcontroller, 
//                                        ptarget, 
//                                        ACIT_STATUS_ERROR);

            //
            // Report this target as missing. This could happen if either
            // the login or offload session has failed
            //
            ap_start_eto(pcontroller, ptarget);

            return mgmt_status;
        }

        ptarget->eh_stats.reopen_success++;

        //
        // Update session info. It may be a duplicate and connection_handle
        // could be incorrect. But this is required for those cases when ARM
        // fw has failed offload and a reopen IOCTL completes successfully
        //
        iscsi_update_session_info_from_async(pcontroller,
                                             ptarget,
                                             session_handle,
                                             0);
        //
        // Target is back from missing state
        //
        ap_reinit_target_ws(pcontroller, ptarget, session_handle);
    }
    else
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING, ("Ignorning status from reopen completion since async has already arrived!!\n"));
    }

    return MGMT_STATUS_SUCCESS;
}

MGMT_STATUS
mgmt_reopen_all_session_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer)
{
    mgmt_free_ioctl_resources(pcontroller, pioctl_handle);
    ap_complete_all_target_recovery(pcontroller);
    return MGMT_STATUS_SUCCESS;
}

MGMT_STATUS
mgmt_save_nonpersistent_targets(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PMGMT_IOCTL_HANDLE      pioctl_handle;
    MGMT_STATUS             mgmt_status;
    ACIT_UEXACT32           flags;

    flags = MGMT_SET_ALLOCATION_FLAGS(MGMT_IOCTL_POLLED,
                                      MGMT_IOCTL_INTERNAL,
                                      MGMT_IOCTL_NO_RESPONSE_BUFFER);

    pioctl_handle = mgmt_alloc_ioctl_resources(pcontroller,
                                               ACIT_NULL,
                                               0,
                                               flags);

    if (pioctl_handle == ACIT_NULL)
    {
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }

    pioctl_handle->pmgmt_post_routine = mgmt_generic_ioctl_callback;
    pioctl_handle->timeout = MGMT_IOCTL_LOGIN_TIMEOUT;

    mgmt_fill_ioctl_hdr(pioctl_handle,
                        IOCTL_SUBSYSTEM_ISCSI_INI,
                        OPCODE_ISCSI_INI_DRIVER_STORE_NONPERSISTENT_SESSIONS,
                        0, // domain 0
                        0,
                        pioctl_handle->timeout);

    mgmt_status = mgmt_fill_ioctl_request(pcontroller, pioctl_handle);
    
    if(mgmt_status != MGMT_STATUS_SUCCESS)
    {
        mgmt_free_ioctl_resources(pcontroller, pioctl_handle);
    }
    else
    {
        mgmt_status = mgmt_submit_ioctl(pcontroller, pioctl_handle);
    }

    return mgmt_status;
}

ACIT_UEXACT32
mgmt_return_pending_ioctls
(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget
)
{
    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
mgmt_return_ioctl_request(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request,
    ACIT_UEXACT32           status)
{
    MGMT_STATUS             mgmt_status;

    OSM_ASSERT(pioctl_request);

    if(status == ACIT_STATUS_SUCCESS)
        mgmt_status = MGMT_STATUS_SUCCESS;
    else
        mgmt_status = MGMT_STATUS_ERROR_ACITISCSI;

    mgmt_return_ioctl(pcontroller, pioctl_request, mgmt_status);

}

ACIT_VOID
mgmt_fill_ioctl_hdr_version(
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    ACIT_UEXACT8            version)
{
    PIOCTL_REQUEST_HEADER   pioctl_hdr;

    pioctl_hdr = (PIOCTL_REQUEST_HEADER)pioctl_handle->pmgmt_buffer_h;
    pioctl_hdr->version = version;
}

ACIT_VOID
mgmt_fill_ioctl_hdr(
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    ACIT_UEXACT8            subsystem,
    ACIT_UEXACT8            opcode,
    ACIT_UEXACT8            domain,
    ACIT_UEXACT32           transfer_length,
    ACIT_UEXACT32           timeout)
{
    PIOCTL_REQUEST_HEADER   pioctl_hdr;

    pioctl_hdr = (PIOCTL_REQUEST_HEADER)pioctl_handle->pmgmt_buffer_h;

    pioctl_hdr->subsystem = subsystem;
    pioctl_hdr->port_number = 0;
    pioctl_hdr->domain = domain;
    pioctl_hdr->opcode = opcode;
    ACIT_PUT_LE32(pioctl_hdr->request_length, transfer_length);
    ACIT_PUT_LE32(pioctl_hdr->timeout, timeout);
}

ACIT_VOID
mgmt_fill_drvr_request(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    ACIT_UEXACT8            opcode,
    PACIT_VOID              ppayload,
    ACIT_UEXACT32           payload_len)
{
    PIOCTL_REQUEST_HEADER   pioctl_hdr;

    pioctl_hdr = (PIOCTL_REQUEST_HEADER)pioctl_handle->pmgmt_buffer_h;

    pioctl_hdr->opcode = opcode;
    pioctl_hdr->timeout = MGMT_INTERNAL_IOCTL_DEFAULT_TIMEOUT;
    pioctl_hdr->request_length = payload_len;

    if(payload_len)
    {
        OSM_MEMCPY(pioctl_handle->pmgmt_buffer, ppayload, payload_len);
    }
    
    mgmt_fill_ioctl_request(pcontroller, pioctl_handle);
}


MGMT_STATUS
mgmt_complete_external_ioctl(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer)
{
    PACIT_IOCTL_REQUEST     pioctl_request;

    pioctl_request = pioctl_handle->pioctl_request;
    mgmt_free_ioctl_resources(pcontroller, pioctl_handle);

    mgmt_return_ioctl(pcontroller, 
                      pioctl_request, 
                      (MGMT_STATUS)presponse_hdr->status);
        
    return MGMT_STATUS_SUCCESS;
}

ACIT_UEXACT32
mgmt_return_ioctl(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request,
    MGMT_STATUS             mgmt_status)
{
    PIOCTL_RESPONSE_HEADER   presponse_hdr;

    presponse_hdr = (PIOCTL_RESPONSE_HEADER)&(pioctl_request->ioctl_hdr);

    OSM_ASSERT(presponse_hdr->subsystem);

    switch(presponse_hdr->subsystem)
    {
    case IOCTL_SUBSYSTEM_ISCSI_INI:

        switch(presponse_hdr->opcode)
        {
            case OPCODE_ISCSI_INI_STATS_GET_INSTANCE_STATISTICS:
                {
                    PMIB_INITIATOR_INSTANCE_STATISTICS  pintr_inst_stats;

                    if (mgmt_status == MGMT_STATUS_SUCCESS)
                    {
                        pintr_inst_stats = (PMIB_INITIATOR_INSTANCE_STATISTICS)pioctl_request->pdata_buffer;
                        pintr_inst_stats->iscsi_ctl_last_ssn_failure_type = pcontroller->eh_stats.last_ssn_failure_type;
                        pintr_inst_stats->iscsi_ctl_ssn_failure_count += pcontroller->eh_stats.ssn_failure_count;
                        pintr_inst_stats->iscsi_ctl_ssn_cxn_timeout_errors += pcontroller->eh_stats.ssn_cxn_timeout_errors;
                        pintr_inst_stats->iscsi_ctl_ssn_digest_errors += pcontroller->eh_stats.ssn_digest_errors;
                        pintr_inst_stats->iscsi_ctl_ssn_format_errors += pcontroller->eh_stats.ssn_format_errors;
                    }
                }
                break;
            
            case OPCODE_ISCSI_INI_STATS_GET_LOGIN_STATISTICS:
                {
                    PMIB_INTR_LOGIN_STATISTICS      pintr_login_stats;

                    if (mgmt_status == MGMT_STATUS_SUCCESS)
                    {
                        pintr_login_stats = (PMIB_INTR_LOGIN_STATISTICS)pioctl_request->pdata_buffer;
                        pintr_login_stats->iscsi_intr_logout_normals += pcontroller->eh_stats.ssn_logout_normals;
                        pintr_login_stats->iscsi_intr_logout_others += pcontroller->eh_stats.ssn_logout_others;
                    }
                }
                break;

            case OPCODE_ISCSI_INI_STATS_GET_SESSION_ATTRIBUTES:
                {
                    PGET_SESSION_ATTRIBUTES_PARAMS_OUT  presp;
                    PACIT_TARGET_WS         ptarget;
                    PMIB_SESSION_ATTRIBUTES psession_param;
                    PMGMT_CONN_INFO         pconnection_param;
                    UEXACT8                 *psession_ptr, *pconnection_ptr;
                    PISCSI_CONNECTION       pcxn;
                    ACIT_UEXACT32           i, j, is_leading_cxn = 0, conn_size = 0;

                    if (mgmt_status == MGMT_STATUS_SUCCESS)
                    {
                        presp = (PGET_SESSION_ATTRIBUTES_PARAMS_OUT)pioctl_request->pdata_buffer;
                        psession_param = &presp->session_attribs[0];
                        psession_ptr = (UEXACT8 *)psession_param;
                        for(i=0;i<presp->session_attribs_count;i++)
                        {
                            ptarget = ap_get_target_ws_from_session_handle(pcontroller,
                                                                            psession_param->session_handle);
                            if (ptarget)
                            {
                                psession_param->iscsi_ssn_statistics.iscsi_cxn_timeout_errors += 
                                                                            ptarget->eh_stats.ssn_cxn_timeout_errors;
                                psession_param->iscsi_ssn_statistics.iscsi_digest_errors += 
                                                                            ptarget->eh_stats.ssn_digest_errors;
                                psession_param->iscsi_ssn_statistics.iscsi_format_errors += 
                                                                            ptarget->eh_stats.ssn_format_errors;
                                pconnection_param = &psession_param->iscsi_ssn_info.conn_list[0];
                                pconnection_ptr = (UEXACT8 *)pconnection_param;
                                for (j=0;j<psession_param->iscsi_ssn_info.conn_count;j++)
                                {
                                    pcxn = iscsi_get_cxn_from_cxn_handle(pcontroller, 
                                                                ptarget, 
                                                                pconnection_param->connection_handle,
                                                                &is_leading_cxn);
									if (pcxn != ACIT_NULL) {
										psession_param->iscsi_ssn_statistics.iscsi_cmd_pdus += pcxn->iscsi_cmd_pdus;
										psession_param->iscsi_ssn_statistics.iscsi_rsp_pdus += pcxn->iscsi_rsp_pdus;
										psession_param->iscsi_ssn_statistics.iscsi_tx_data_octets += pcxn->total_bytes_xferred;
										psession_param->iscsi_ssn_statistics.iscsi_rx_data_octets += pcxn->total_bytes_read;
										pconnection_ptr += sizeof(MGMT_CONN_INFO);
										pconnection_param = (PMGMT_CONN_INFO)pconnection_ptr;
										if (j > 0)
											conn_size += sizeof(MGMT_CONN_INFO);
									}
									else {
										OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("mgmt_return_ioctl:: Invalid connection\n"));
										mgmt_status = MGMT_STATUS_FAILED;
										break;
									}
                                }
			                    if (mgmt_status == MGMT_STATUS_FAILED)
									break;
                                psession_ptr += (sizeof (MIB_SESSION_ATTRIBUTES) + conn_size);
                                psession_param = (PMIB_SESSION_ATTRIBUTES) psession_ptr;
                            }
                            else
							{
                                mgmt_status = MGMT_STATUS_FAILED;
								break;
							}
                        }
                    }

                }
                break;

            case OPCODE_ISCSI_INI_STATS_GET_A_SESSION_STATISTICS:
                {
                    PGET_A_SESSION_STATISTICS_PARAMS_OUT    presp;
                    PACIT_TARGET_WS         ptarget;
                    PISCSI_TARGET_WS        piscsi_target;
                    ACIT_UEXACT32           i;

                    if (mgmt_status == MGMT_STATUS_SUCCESS)
                    {
                        presp = (PGET_A_SESSION_STATISTICS_PARAMS_OUT)pioctl_request->pdata_buffer;
                        ptarget = ap_get_target_ws_from_session_handle(pcontroller, presp->session_stats.session_handle);
                        if (ptarget)
                        {
                            presp->session_stats.iscsi_ssn_stats.iscsi_cxn_timeout_errors += ptarget->eh_stats.ssn_cxn_timeout_errors;
                            presp->session_stats.iscsi_ssn_stats.iscsi_digest_errors += ptarget->eh_stats.ssn_digest_errors;
                            presp->session_stats.iscsi_ssn_stats.iscsi_format_errors += ptarget->eh_stats.ssn_format_errors;
                            piscsi_target = GET_ISCSI_TARGET_WS(ptarget);
                            for (i = 0; i < piscsi_target->num_connections; i++)
                            {
                                if(piscsi_target->pcxn_list[i])
                                {
                                    presp->session_stats.iscsi_ssn_stats.iscsi_cmd_pdus += 
                                            piscsi_target->pcxn_list[i]->iscsi_cmd_pdus;
                                    presp->session_stats.iscsi_ssn_stats.iscsi_rsp_pdus += 
                                            piscsi_target->pcxn_list[i]->iscsi_rsp_pdus;
                                    presp->session_stats.iscsi_ssn_stats.iscsi_tx_data_octets += 
                                            piscsi_target->pcxn_list[i]->total_bytes_xferred;
                                    presp->session_stats.iscsi_ssn_stats.iscsi_rx_data_octets += 
                                            piscsi_target->pcxn_list[i]->total_bytes_read;
                                }
                            }
                        }
                        else
                            mgmt_status = MGMT_STATUS_FAILED;
                    }
                }
                break;

            case OPCODE_ISCSI_INI_STATS_GET_CONN_STATISTICS_PER_SESSION:
                {
                    PGET_CONN_STATISTICS_PER_SESSION_OUT    presp;
                    PACIT_TARGET_WS         ptarget;
                    PISCSI_TARGET_WS        piscsi_target;
                    ACIT_UEXACT32           i, j;

                    if (mgmt_status == MGMT_STATUS_SUCCESS)
                    {
                        presp = (PGET_CONN_STATISTICS_PER_SESSION_OUT)pioctl_request->pdata_buffer;
                        ptarget = ap_get_target_ws_from_session_handle(pcontroller, presp->session_handle);
                        if (ptarget)
                        {
                            for (i=0;i<presp->conn_stats_count;i++)
                            {
                                piscsi_target = GET_ISCSI_TARGET_WS(ptarget);
                                for(j=0; j<MAX_CXNS_PER_SESSION; j++)
                                {
                                    if ((piscsi_target->pcxn_list[j]) &&
                                        (piscsi_target->pcxn_list[j]->iscsi_cid == presp->conn_stats[j].cid))
                                    {
                                        presp->conn_stats[i].iscsi_conn_stats.iscsi_cxn_timeout_errors += 
                                                piscsi_target->pcxn_list[j]->timeout_errors;
                                        presp->conn_stats[i].iscsi_conn_stats.iscsi_digest_errors += 
                                                piscsi_target->pcxn_list[j]->digest_errors;
                                        presp->conn_stats[i].iscsi_conn_stats.iscsi_format_errors += 
                                                piscsi_target->pcxn_list[j]->format_errors;
                                        presp->conn_stats[i].iscsi_conn_stats.iscsi_cmd_pdus += 
                                                piscsi_target->pcxn_list[j]->iscsi_cmd_pdus;
                                        presp->conn_stats[i].iscsi_conn_stats.iscsi_rsp_pdus += 
                                                piscsi_target->pcxn_list[j]->iscsi_rsp_pdus;
                                        presp->conn_stats[i].iscsi_conn_stats.iscsi_tx_data_octets += 
                                                piscsi_target->pcxn_list[j]->total_bytes_xferred;
                                        presp->conn_stats[i].iscsi_conn_stats.iscsi_rx_data_octets += 
                                                piscsi_target->pcxn_list[j]->total_bytes_read;
                                    }
                                }
                            }
                        }
                    }
                }
                break;

            case OPCODE_ISCSI_INI_SESSION_GET_CONFIGURATION:
                {
                    PGET_SESSION_CONFIGURATION_PARAMS psession_config;
                    PACIT_TARGET_WS ptarget;

                    if (mgmt_status == MGMT_STATUS_SUCCESS)
                    {
                        psession_config = (PGET_SESSION_CONFIGURATION_PARAMS)pioctl_request->pdata_buffer;
                        ptarget = ap_get_target_ws_from_session_handle(pcontroller, 
                                                                       psession_config->response.session_handle);
                        if (ptarget)
                        {
                            psession_config->response.extended_timeout = ptarget->eto.original_ticks /
                                pcontroller->params.ticks_per_second;

                            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("get session config: return ETO value %d\n", psession_config->response.extended_timeout));
                        }

                        // Bug 21257 : If ETO is un-initialized (as a result of BIOS configured target), then
                        // override the uninitialized value with what the controller defaults are
                        if(!ACIT_IS_ETO_VALID(psession_config->response.extended_timeout))
                        {
                            psession_config->response.extended_timeout = pcontroller->params.eto;
                        }
                    }
                }

                break;

            default:
                break;
        }
        break;

    case IOCTL_SUBSYSTEM_COMMON:

        switch(presponse_hdr->opcode)
        {
        case OPCODE_COMMON_GET_CNTL_ATTRIBUTES:
            {
                PMGMT_CONTROLLER_ATTRIBUTES pmgmt_controller_attrib;

                if (mgmt_status == MGMT_STATUS_SUCCESS)
                {
                    pmgmt_controller_attrib = (PMGMT_CONTROLLER_ATTRIBUTES)pioctl_request->pdata_buffer;

                    pmgmt_controller_attrib->hba_attribs.default_extended_timeout = pcontroller->params.eto;
                    pmgmt_controller_attrib->hba_attribs.default_link_down_timeout = (ACIT_UEXACT16)pcontroller->params.ldto;
                }
            }
            break;
        }

        break;

    default:
        break;
    }

    presponse_hdr->status = mgmt_status;

    if(pioctl_request->ppost_routine)
    {
        (pioctl_request->ppost_routine)(pcontroller->pcontroller_key, pioctl_request);   
    }
    else
    {
        OSM_COMPLETE_IOCTL(pcontroller->pcontroller_key, pioctl_request);
    }

    return ACIT_STATUS_SUCCESS;
}

PMGMT_IOCTL_HANDLE
mgmt_alloc_ioctl_handle(
   PMGMT_CONTROLLER_WS      pmgmt_controller)
{
    PACIT_DC_LIST           plink;
    PMGMT_IOCTL_HANDLE      pioctl_handle;

    pioctl_handle = ACIT_NULL;

    if(IS_LIST_NOT_EMPTY(&pmgmt_controller->ioctl_handle_free_list))
    {
        LIST_REMOVEFROM_HEAD(&plink, 
                            &pmgmt_controller->ioctl_handle_free_list);

        pioctl_handle = STRUCT_BASE(MGMT_IOCTL_HANDLE,
                                    controller_link,
                                    plink);

        OSM_ASSERT(pioctl_handle);

        LIST_ADDTO_TAIL(&pioctl_handle->controller_link, 
                        &pmgmt_controller->ioctl_handle_busy_list);
        
        pioctl_handle->flags &= ~MGMT_IOCTLH_FREE;
        pioctl_handle->pmgmt_buffer = ACIT_NULL;
        pioctl_handle->pmcc_wrb = ACIT_NULL;
        pioctl_handle->pmgmt_post_routine = ACIT_NULL;
        pioctl_handle->timeout = 0;
        pioctl_handle->ticks = ((PACIT_CONTROLLER_WS)(pmgmt_controller->pacit_ws))->counters.current_tick_count;
        pmgmt_controller->ioctl_handles_available--;
    }
    else
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Failed to obtain IOCTL Handle\n"));
        OSM_ENTER_DEBUGGER();
    }

    if(!pmgmt_controller->ioctl_handles_available)
    {
        mgmt_flag_busy_controller(pmgmt_controller, MGMT_CONTROLLER_BUSY_IOCTL_HANDLE);
    }
    return pioctl_handle;
}

ACIT_VOID
mgmt_free_ioctl_handle(
   PMGMT_CONTROLLER_WS      pmgmt_controller,
   PMGMT_IOCTL_HANDLE       pioctl_handle)
{
    OSM_ASSERT(!(pioctl_handle->flags & MGMT_IOCTLH_FREE));

    pioctl_handle->flags = MGMT_IOCTLH_FREE;

    LIST_REMOVE_ELEMENT(&pioctl_handle->controller_link);

    LIST_ADDTO_TAIL(&pioctl_handle->controller_link,
                    &pmgmt_controller->ioctl_handle_free_list);

    pioctl_handle->pioctl_request = NULL;
    pioctl_handle->pmcc_wrb = NULL;
    pioctl_handle->pmgmt_buf_handle = NULL;
    pioctl_handle->pmgmt_buffer = NULL;
    pioctl_handle->pmgmt_buffer_h = NULL;
    pioctl_handle->pmgmt_post_routine = NULL;
    pioctl_handle->ptarget = NULL;

    pmgmt_controller->ioctl_handles_available++;

    if(ACIT_IS_BIT_SET(pmgmt_controller->busy, MGMT_CONTROLLER_BUSY_IOCTL_HANDLE))
    {
        mgmt_flag_ready_controller( pmgmt_controller,
                                    MGMT_CONTROLLER_BUSY_IOCTL_HANDLE);
    }
}


MGMT_STATUS
mgmt_alloc_ioctl_buffer(
   PMGMT_CONTROLLER_WS      pmgmt_controller,
   PMGMT_IOCTL_HANDLE       pioctl_handle)
{
    PACIT_DC_LIST           plink;
    PMGMT_INTERNAL_IOCTL_BUFFER_HANDLE pmgmt_buf_handle;

    if(IS_LIST_NOT_EMPTY(&pmgmt_controller->ioctl_buffer_free_list))
    {
        LIST_REMOVEFROM_HEAD(&plink, 
                             &pmgmt_controller->ioctl_buffer_free_list);

        pmgmt_buf_handle = STRUCT_BASE(MGMT_INTERNAL_IOCTL_BUFFER_HANDLE,
                                         controller_link,
                                         plink);

        OSM_ASSERT(pmgmt_buf_handle);

        LIST_ADDTO_TAIL(&pmgmt_buf_handle->controller_link, 
                        &pmgmt_controller->ioctl_buffer_busy_list);
        
        pmgmt_buf_handle->flags &= ~MGMT_IOCTL_BUFFER_FLAG_FREE;

        pioctl_handle->pmgmt_buffer_h = (PACIT_UEXACT8)pmgmt_buf_handle + 
                                        STRUCT_FIELD_OFFSET(MGMT_INTERNAL_IOCTL_BUFFER_HANDLE, ioctl_header);

        pioctl_handle->pmgmt_buffer = (PACIT_UEXACT8)pmgmt_buf_handle + 
                                        STRUCT_FIELD_OFFSET(MGMT_INTERNAL_IOCTL_BUFFER_HANDLE, data_buffer);
        OSM_MEMSET(pioctl_handle->pmgmt_buffer_h, 0, sizeof(IOCTL_HEADER));
        OSM_MEMSET(pioctl_handle->pmgmt_buffer, 0, sizeof (pmgmt_buf_handle->data_buffer));

        pioctl_handle->flags |= MGMT_IOCTLH_EXPLICIT_BUFFER;
        pioctl_handle->pmgmt_buf_handle = pmgmt_buf_handle;

        pmgmt_controller->ioctl_buffers_available--;
    }
    else
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Failed to obtain IOCTL Buffer\n"));
        OSM_ENTER_DEBUGGER();
        return MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED;
    }

    if(pioctl_handle->payload_length > sizeof(pmgmt_buf_handle->data_buffer))
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("IOCTL buffer size too big\n"));
        OSM_ENTER_DEBUGGER();
        return MGMT_STATUS_INSUFFICIENT_BUFFER;
    }

    if(!pmgmt_controller->ioctl_buffers_available)
    {
        mgmt_flag_busy_controller(pmgmt_controller, MGMT_CONTROLLER_BUSY_IOCTL_BUFFER);
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
mgmt_free_ioctl_buffer(
   PMGMT_CONTROLLER_WS      pmgmt_controller,
   PMGMT_IOCTL_HANDLE       pioctl_handle)
{
    PMGMT_INTERNAL_IOCTL_BUFFER_HANDLE pmgmt_buf_handle;
    pmgmt_buf_handle = pioctl_handle->pmgmt_buf_handle;

    OSM_ASSERT(!(pmgmt_buf_handle->flags & MGMT_IOCTL_BUFFER_FLAG_FREE));
    pmgmt_buf_handle->flags = MGMT_IOCTL_BUFFER_FLAG_FREE;

    LIST_REMOVE_ELEMENT(&pmgmt_buf_handle->controller_link);
    LIST_ADDTO_TAIL(&pmgmt_buf_handle->controller_link,
                    &pmgmt_controller->ioctl_buffer_free_list);
    pmgmt_controller->ioctl_buffers_available++;

    if(ACIT_IS_BIT_SET(pmgmt_controller->busy, MGMT_CONTROLLER_BUSY_IOCTL_BUFFER))
    {
        mgmt_flag_ready_controller( pmgmt_controller,
                                    MGMT_CONTROLLER_BUSY_IOCTL_BUFFER);
    }
}

PMGMT_IOCTL_HANDLE
mgmt_alloc_ioctl_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request,
    ACIT_UEXACT32           transfer_length,
    ACIT_UEXACT32           flags)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;
    PMGMT_IOCTL_HANDLE      pioctl_handle = ACIT_NULL;
    ACIT_UEXACT32           status;

    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);

    pioctl_handle = mgmt_alloc_ioctl_handle(pmgmt_controller);

    if(pioctl_handle)
    {
        pioctl_handle->pcontroller = pcontroller;
        pioctl_handle->payload_length = transfer_length;
        pioctl_handle->flags = 0;

        if(!transfer_length)
        {
            OSM_ASSERT(!(flags & MGMT_IOCTL_RESPONSE_BUFFER));
        }

        if(flags & MGMT_IOCTL_POLLED)
            pioctl_handle->flags |= MGMT_IOCTLH_POLLED_MODE;
        
        if(flags & MGMT_IOCTL_INTERNAL)
            pioctl_handle->flags |= MGMT_IOCTLH_INTERNAL;

        if(flags & MGMT_IOCTL_RESPONSE_BUFFER)
            pioctl_handle->flags |= MGMT_IOCTLH_RESPONSE_BUFFER;

        /* external IOCTLs have buffer and header supplied in IOCTL request */
        if(!(pioctl_handle->flags & MGMT_IOCTLH_INTERNAL))
        {
            pioctl_handle->pmgmt_buffer_h = (PACIT_VOID)&pioctl_request->ioctl_hdr;
            pioctl_handle->pmgmt_buffer = pioctl_request->pdata_buffer;
        }

        status = mgmt_alloc_internal_resources(pmgmt_controller,
                                               pioctl_handle);

        if(status != MGMT_STATUS_SUCCESS)
        {
            mgmt_free_ioctl_handle(pmgmt_controller, pioctl_handle);
            return ACIT_NULL;
        }

        /*
         * Seperate buffers are needed only if one of the following is true
         * a. IOCTL is internal and transfer_length > TRANSFER_CONSTANT
         * b. flags has MGMT_IOCTL_RESPONSE_BUFFER bit set
         */
        if(((flags & MGMT_IOCTL_INTERNAL) && (transfer_length > BE_MCC_PAYLOAD_THRESHOLD)) ||
            (flags & MGMT_IOCTL_RESPONSE_BUFFER))
        {
            MGMT_STATUS mgmt_status;
            OSM_ASSERT(transfer_length <= sizeof(MGMT_INTERNAL_IOCTL_BUFFER_HANDLE));

            mgmt_status = mgmt_alloc_ioctl_buffer(pmgmt_controller, pioctl_handle);

            if(mgmt_status != MGMT_STATUS_SUCCESS)
            {
                mgmt_free_internal_resources(pmgmt_controller, pioctl_handle);
                mgmt_free_ioctl_handle(pmgmt_controller, pioctl_handle);
                pioctl_handle = ACIT_NULL;
            }
        }
    }

    if(pioctl_handle)
    {
        OSM_ASSERT(pioctl_handle->pmgmt_buffer);
        OSM_ASSERT(pioctl_handle->pmgmt_buffer_h);
        OSM_ASSERT(pioctl_handle->pmcc_wrb);

        OSM_MEMSET(pioctl_handle->pmcc_wrb, 0, sizeof(MCC_WRB));
    }

    return pioctl_handle;
}

ACIT_VOID
mgmt_free_ioctl_resources(
   PACIT_CONTROLLER_WS      pcontroller,
   PMGMT_IOCTL_HANDLE       pioctl_handle)
{
    PMGMT_CONTROLLER_WS     pmgmt_controller;

    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);

    OSM_ASSERT(pioctl_handle);

    if(pioctl_handle->flags & MGMT_IOCTLH_EXPLICIT_BUFFER)
        mgmt_free_ioctl_buffer(pmgmt_controller, pioctl_handle);

    mgmt_free_internal_resources(pmgmt_controller, pioctl_handle);
    mgmt_free_ioctl_handle(pmgmt_controller, pioctl_handle);
}

ACIT_UEXACT32
mgmt_flag_busy_controller(
   PMGMT_CONTROLLER_WS      pmgmt_controller,
   ACIT_UEXACT32            reason)
{
    ACIT_BIT_SET(pmgmt_controller->busy, reason);

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
mgmt_flag_ready_controller(
   PMGMT_CONTROLLER_WS      pmgmt_controller,
   ACIT_UEXACT32            reason)
{
    ACIT_BIT_CLR(pmgmt_controller->busy, reason);

    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
mgmt_unitt_init_controller(
    PACIT_CONTROLLER_WS     pcontroller)
{
#ifdef ACIT_UNIT_TEST
    ACIT_UEXACT32           flags, i;
    PMGMT_IOCTL_HANDLE      pioctl_h[128] = {ACIT_NULL};
    PMGMT_CONTROLLER_WS     pmgmt_controller;

    pmgmt_controller = GET_MGMT_CONTROLLER_WS(pcontroller);

    flags = MGMT_SET_ALLOCATION_FLAGS(MGMT_IOCTL_NON_POLLED,
                                        MGMT_IOCTL_INTERNAL,
                                        MGMT_IOCTL_RESPONSE_BUFFER);

    // test IOCTL handle allocation
    for(i = 0; i < pcontroller->params.ioctls_per_ctrl; i++)
    {
        pioctl_h[i] = mgmt_alloc_ioctl_resources(pcontroller,
                                                 ACIT_NULL,
                                                 64,
                                                 flags);

        OSM_ASSERT(pioctl_h);
    }

    // mgmt_controller must already be busy
    OSM_ASSERT(pmgmt_controller->busy);
    OSM_ASSERT(!pmgmt_controller->ioctl_handles_available);
    OSM_ASSERT(!pmgmt_controller->ioctl_buffers_available);

    // and this MUST fail now
    pioctl_h[127] = mgmt_alloc_ioctl_resources(pcontroller,
                                                    ACIT_NULL,
                                                    64,
                                                    flags);

    OSM_ASSERT(pioctl_h);

    for(i = 0; i < pcontroller->params.ioctls_per_ctrl; i++)
    {
        mgmt_free_ioctl_resources(pcontroller, pioctl_h[i]);
    }

    // mgmt_controller not busy anymore
    OSM_ASSERT(!pmgmt_controller->busy);
    OSM_ASSERT(pmgmt_controller->ioctl_handles_available == pcontroller->params.ioctls_per_ctrl);
    OSM_ASSERT(pmgmt_controller->ioctl_buffers_available == pcontroller->params.ioctls_per_ctrl);

#endif
}
