/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_chap.h

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/runtime/sa_chap.h,v 1.5 2006/03/10 19:16:12 menz Exp $
*/
#ifndef __sa_chap_h__
#define __sa_chap_h__

#include "system/sa_system.h"
SA_CPP_HEADER

#define SA_MAX_CHAP_CHALLENGE 1024
#define SA_MAX_CHAP_RESPONSE  16

#define SA_CHAP_OP_CHAP_ENABLED                  0x1
#define SA_CHAP_OP_RECEIVE_ALGORITHM             0x2
#define SA_CHAP_OP_ALGORITHM_DECIDED             0x100
#define SA_CHAP_OP_RECEIVE_TARGET_CHALLENGE      0x4
#define SA_CHAP_OP_RECEIVE_TARGET_ID             0x8
#define SA_CHAP_OP_SEND_RESPONSE                 0x10

#define SA_CHAP_OP_SEND_MUTUAL_CHALLENGE_AND_ID  0x20
#define SA_CHAP_OP_RECEIVE_MUTAL_TARGET_RESPONSE 0x80
#define SA_CHAP_OP_VERIFY_MUTUAL_RESPONSE        0x40

#define SA_CHAP_DONE (SA_CHAP_OP_CHAP_ENABLED |              \
                      SA_CHAP_OP_RECEIVE_ALGORITHM |         \
                      SA_CHAP_OP_ALGORITHM_DECIDED |         \
                      SA_CHAP_OP_RECEIVE_TARGET_CHALLENGE |  \
                      SA_CHAP_OP_RECEIVE_TARGET_ID |         \
                      SA_CHAP_OP_SEND_RESPONSE)              \

#define SA_MUTUAL_CHAP_DONE (SA_CHAP_DONE |                           \
                          SA_CHAP_OP_SEND_MUTUAL_CHALLENGE_AND_ID |   \
                          SA_CHAP_OP_RECEIVE_MUTAL_TARGET_RESPONSE |  \
                          SA_CHAP_OP_VERIFY_MUTUAL_RESPONSE)          \

typedef struct _SA_CHAP_CONTEXT {    

    u16 op;          // bitmask of operations completed
    u8 id;           // ID for this exchange used to compute response
    u8 algorithm;    // Algorithm from the target, such as CHAP_MD5_ALGORITHM.

    u32 challengeLength;              // Length of received/generated challege (likely 1024)
    u8 challenge[SA_MAX_CHAP_CHALLENGE]; 

    u32 responseLength;               // Always MAX_CHAP_RESPONSE bytes after response is computed
    u8 response[SA_MAX_CHAP_RESPONSE];

} SA_CHAP_CONTEXT, *PSA_CHAP_CONTEXT;


void sa_chap_generate_challenge (PSA_CHAP_CONTEXT chap, u32 len);
void sa_chap_calculate_response (PSA_CHAP_CONTEXT chap, const c8* secret);

#if !defined(SA_OMIT_TEST_CODE)
    boolean sa_chap_test (void);
#endif

SA_CPP_TRAILER
#endif // __sa_chap_h__

