/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   $RCSfile: sa_object_allocator.h,v $

@brief
    Implements generic index and object allocators that use a 
    free list to efficiently allocate/deallocate items.  Supports
    both multi-thread safe, and non-safe methods.

CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/runtime/sa_object_allocator.h,v 1.3 2005/11/01 20:48:27 menz Exp $
*/
#ifndef __sa_object_allocator_h__
#define __sa_object_allocator_h__

#include "sa_index_allocator.h"
SA_CPP_HEADER

// ----------------------------------------------------------
// Object Allocator (Index Allocater Based)
// ----------------------------------------------------------

//
// Use this to declare the objects for a ring allocator. 
//
#define SA_OBJECT_ALLOCATOR_DECLARE(_type_,_name_,_num_)    \
        SA_OBJECT_ALLOCATOR _name_;                         \
        SA_INDEX_ALLOCATOR_DECLARE (_name_##_ia, _num_);    \
        _type_  _name_##Objects [_num_]                     \
        

//
// Use this to init ring allocator. 
//
#define SA_OBJECT_ALLOCATOR_CREATE SA_OA_CREATE
#define SA_OA_CREATE(_name_,_num_)                       \
    {                                                    \
        SA_INDEX_ALLOCATOR_CREATE (_name_##_ia, _num_);  \
        sa_objalloc_create (                             \
            &_name_,                                     \
            _num_,                                       \
            sizeof (_name_##Objects[0]),                 \
            _name_##Objects,                             \
            &_name_##_ia);                               \
    }                                                    \


typedef struct _SA_OBJECT_ALLOCATOR {

    SA_INDEX_ALLOCATOR *ia; //! Index allocator
    u32 object_size;        //! Bytes per object 
    PVOID object_array;     //! Pointer to beginning of array of objects to allocate

} SA_OBJECT_ALLOCATOR, *PSA_OBJECT_ALLOCATOR;


PSA_OBJECT_ALLOCATOR
sa_objalloc_allocate (u32 num,
                      u32 objectSize,
                      u32 alloc_flags,
                      const c8* optional_name);
void 
sa_objalloc_free (PSA_OBJECT_ALLOCATOR oa);



SA_STATUS
sa_objalloc_create (PSA_OBJECT_ALLOCATOR oa, 
                    u32 num, 
                    u32 objectSize, 
                    PVOID objectArray, 
                    PSA_INDEX_ALLOCATOR ia); // previously created index allocator

void
sa_objalloc_destroy (PSA_OBJECT_ALLOCATOR oa);


PVOID
sa_objalloc_allocate_object (PSA_OBJECT_ALLOCATOR oa);

void
sa_objalloc_free_object (PSA_OBJECT_ALLOCATOR oa, PVOID object);

u32 
sa_objalloc_index (PSA_OBJECT_ALLOCATOR oa, PVOID object);

PVOID
sa_objalloc_interlocked_allocate_object (PSA_OBJECT_ALLOCATOR oa);

void
sa_objalloc_interlocked_free_object (PSA_OBJECT_ALLOCATOR oa, PVOID object);

STATIC
INLINE
u32 sa_objalloc_num_free (PSA_OBJECT_ALLOCATOR oa) {
    ASSERT (oa);
    return sa_indexalloc_num_free (oa->ia);
}

STATIC
INLINE
u32 sa_objalloc_num_allocated (PSA_OBJECT_ALLOCATOR oa) {
    ASSERT (oa);    
    return sa_indexalloc_num_allocated (oa->ia);
}

SA_CPP_TRAILER
#endif

