/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   $RCSfile: sa_ring_index_allocator.h,v $

@brief
    Implements generic index allocator using a ring.
    

CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/runtime/sa_ring_index_allocator.h,v 1.8 2006/03/10 19:16:12 menz Exp $
*/
#ifndef __sa_ring_index_allocator_h__
#define __sa_ring_index_allocator_h__

#include "system/sa_system.h"
#include "runtime/sa_int_math.h"

SA_CPP_HEADER

// ----------------------------------------------------------
// Ring Index Allocator
// ----------------------------------------------------------

typedef struct _SA_RING_ALLOCATOR {

    u32 magic;
    u32 num;                 // num of elements that can be allocated
    u32 allocate_index;       // index increments with allocates
    u32 free_index;           // index increments with frees
    u32 *ring;               // buffer with num+1 elements
    const c8* dbg_name;        // Name of allocator for debug prints
    boolean *dbg_vector;         // optional dbg vector with num elements

} SA_RING_ALLOCATOR, *PSA_RING_ALLOCATOR;

#define SA_RA_BAD_INDEX                 (0xFFFFFFFFUL)
#define SA_RA_MAGIC                     (0xFD12390DUL)

PSA_RING_ALLOCATOR 
sa_ringalloc_allocate (u32 num,
                       u32 alloc_flags,
                       const c8* optional_name);
void
sa_ringalloc_free (PSA_RING_ALLOCATOR ra);


SA_STATUS
sa_ringalloc_create (PSA_RING_ALLOCATOR ra, 
                     u32 *ring, 
                     u32 num, 
                     boolean *optional_dbg_vector, 
                     const c8* optional_name);
void
sa_ringalloc_destroy (PSA_RING_ALLOCATOR ra);

u32
sa_ringalloc_allocate_index (PSA_RING_ALLOCATOR ra);

void
sa_ringalloc_free_index (PSA_RING_ALLOCATOR ra, u32 index);

STATIC INLINE
u32 sa_ringalloc_num_free (PSA_RING_ALLOCATOR ra) {
    return sa_subc (ra->free_index, ra->allocate_index, ra->num+1);
}

STATIC INLINE
u32 sa_ringalloc_num_allocated (PSA_RING_ALLOCATOR ra) {
    return ra->num - sa_ringalloc_num_free (ra);
}

void sa_ringalloc_test (void);

void sa_indexalloc_test (void);


// ------------------------------------------------------------------------
// Macro wrappers for allocating a ring allocator object and the associated
// ring within a struct instance.  The wrappers allow using the name of the
// struct 
// ------------------------------------------------------------------------

#define SA_RING_ALLOCATOR_DECLARE(_name_,_num_)       \
        struct {                                      \
            SA_RING_ALLOCATOR ra;                     \
            SA_DBG_ONLY (boolean dbgVector [_num_];)  \
            u32 ring [_num_ + 1];                     \
        } _name_                                      \
                                                      \

#define SA_RING_ALLOCATOR_CREATE SA_RA_CREATE
#define SA_RA_CREATE(_name_, _num_)      \
    sa_ringalloc_create (& _name_ .ra,   \
        _name_ .ring,                    \
        _num_,                            \
        SA_DBG_PTR (_name_ .dbgVector),  \
        SA_DBG_CSTR (#_name_))            \

#define RA_ALLOCATE(_name_)          sa_ringalloc_allocate_index (& _name_  .ra)
#define RA_FREE(_name_, _index_)     sa_ringalloc_free_index (& _name_  .ra, _index_)
#define RA_NUM_FREE(_name_)          sa_ringalloc_num_free (& _name_  .ra)
#define RA_NUM_ALLOCATED(_name_)     sa_ringalloc_num_allocated (& _name_  .ra)

SA_CPP_TRAILER

#endif
