/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

#ifndef __sa_text_h__
#define __sa_text_h__

#include "system/sa_system.h"
#include "runtime/sa_network_packet.h"

SA_CPP_HEADER

// ------------------------------------------------------------------
//  P U B L I C   I N T E R F A C E S
// ------------------------------------------------------------------

// Converst one character to upper case
STATIC INLINE c8 sa_text_toupper(c8 c) {
    return (c <='z' && c >='a' ? c - 0x20 : c);
}

STATIC INLINE c8 sa_text_toprintable(c8 c) {
    return (c >= 32 && c <= 126 ? c : '.');
}

// Converst one character to upper case
STATIC INLINE c16 sa_wtext_toupper(c16 c) {
    return (c <= L'z' && c >= L'a' ? c - 0x20 : c);
}


#define SA_IP_FORMAT "%d.%d.%d.%d"
#define SA_IP_LFORMAT L"%d.%d.%d.%d"
#define SA_IP_ARGS(_ip_)              \
        (u32)(((_ip_) >> 24) & 0xff),  \
        (u32)(((_ip_) >> 16) & 0xff),  \
        (u32)(((_ip_) >> 8) & 0xff),   \
        (u32)(((_ip_)) & 0xff)         \

#define SA_MAC_FORMAT   "%02X:%02X:%02X:%02X:%02X:%02X"
#define SA_MAC_LFORMAT L"%02X:%02X:%02X:%02X:%02X:%02X"
#define SA_MAC_ARGS(_mac_address_)  \
        (u32)((_mac_address_).bytes[0]),     \
        (u32)((_mac_address_).bytes[1]),     \
        (u32)((_mac_address_).bytes[2]),     \
        (u32)((_mac_address_).bytes[3]),     \
        (u32)((_mac_address_).bytes[4]),     \
        (u32)((_mac_address_).bytes[5])      \


#define SA_IPV6_FORMAT   "%04X:%04X:%04X:%04X:%04X:%04X:%04X:%04X"
#define SA_IPV6_ARGS(_ipv6_address_)  \
        (u32)(sa_swap16 ((_ipv6_address_).addr16[0])),     \
        (u32)(sa_swap16 ((_ipv6_address_).addr16[1])),     \
        (u32)(sa_swap16 ((_ipv6_address_).addr16[2])),     \
        (u32)(sa_swap16 ((_ipv6_address_).addr16[3])),     \
        (u32)(sa_swap16 ((_ipv6_address_).addr16[4])),     \
        (u32)(sa_swap16 ((_ipv6_address_).addr16[5])),     \
        (u32)(sa_swap16 ((_ipv6_address_).addr16[6])),     \
        (u32)(sa_swap16 ((_ipv6_address_).addr16[7]))      \


#define SA_UNICODE_NULL ((c16)0)

STATIC INLINE u32 sa_strlen (const c8* str) {
    ASSERT (str);
    return (u32)strlen (str);
}

u32
sa_text_copy (
    c8* dest, 
    const c8* src, 
    u32 destLength);

u32 
sa_wtext_copy (
    c16*  dest,
    const c16* src,
    u32   destLength);

u32 
sa_text_cat (
    c8* dest, 
    const c8* src, 
    u32 destLength);


boolean
sa_text_str_to_int (
    const c8* str, 
    u32 size, 
    u32 *value);

boolean
sa_wtext_str_to_int (
    const c16* str,
    u32   size,
    u32*  value);

boolean
sa_text_hex_str_to_int (
    const c8* str, 
    u32 size, 
    u32 *value);

boolean
sa_wtext_hex_str_to_int (
    const c16*   str,
    u32   size,
    u32*  value);

u32
sa_text_hex_str_to_byte_array (
    const c8* str, 
    u32 strLength, 
    u8 *value, 
    u32 valueLength);

boolean
sa_text_dec_str_to_int (
    const c8* str, 
    u32 size, 
    u32 *value);

boolean
sa_wtext_dec_str_to_int (
    const c16*   str,
    u32   size,
    u32*  value);

boolean
sa_text_base_str_to_int (
    const c8* str, 
    u32 size, 
    u32 base, 
    u32 *value);

boolean
sa_wtext_base_str_to_int (
    const c16*   str,
    u32   size,
    u32   base,
    u32*  value);

u32
sa_text_int_to_hex_str (
    c8* str, 
    u32 strLength, 
    u32 number, 
    u32 fixedDigits);

u32
sa_wtext_int_to_hex_str (
    c16* str,
    u32  strLength,
    u32  number,
    u32  fixedDigits);

u32
sa_text_byte_array_to_hex_str (
    c8* str, 
    u32 strLength, 
    u8 *value, 
    u32 valueLength);

u32
sa_wtext_byte_array_to_hex_str (
    c16* str,
    u32  strLength,
    u8*  value,
    u32  valueLength);

u32
sa_text_int_to_dec_str (
    c8* str, 
    u32 strLength, 
    u32 number, 
    u32 fixedDigits) ;

u32
sa_wtext_int_to_dec_str (
    c16* str,
    u32  strLength,
    u32  number,
    u32  fixedDigits);

u32
sa_text_int_to_base_str (
    c8* str, 
    u32 strLength, 
    u32 number, 
    u32 base, 
    u32 fixedDigits);

u32
sa_wtext_int_to_base_str (
    c16* str,
    u32  strLength,
    u32  base,
    u32  number,
    u32  fixedDigits);

boolean 
sa_text_str_to_mac (
    const c8* str, 
    u32 strLength, 
    SA_MAC_ADDRESS* mac);

boolean 
sa_wtext_str_to_mac (
    c16*            wstr,
    u32             strLength,
    SA_MAC_ADDRESS* mac);

u32
sa_text_mac_to_str (
    c8* str, 
    u32 strLength, 
    SA_MAC_ADDRESS* mac);

u32 
sa_wtext_mac_to_str (
    c16*            str, // unicode chars
    u32             size,
    SA_MAC_ADDRESS* mac);

boolean 
sa_text_str_to_ip (
    const c8* str, 
    u32 strLength, 
    u32* ip);

u32
sa_text_ip_to_str (
    c8* str, 
    u32 strLength, 
    u32 ip);


u32 
sa_text_key_str_copy (
    c8* dest, 
    const c8* key, 
    const c8* value, 
    u32 destLength);

u32 
sa_text_key_int_copy (
    c8* dest, 
    const c8* key, 
    u32 value, 
    u32 destLength);

const c8*
sa_text_find_end (
    const c8* input, 
    const c8* find, 
    u32 inputLength);

SA_STATUS
sa_text_wmi_string_to_ansi (
    c16* wmiStr, 
    PVOID  ansiStr, 
    u32* ansiLen);
i32 
sa_text_unicode_to_ansi_str (
    c16* wStr, 
    c8*  aStr, 
    i32   aLen, 
    i32   wSiz);

SA_STATUS
sa_text_ansi_to_unicode (
    c8*  ansiStr, 
    c16* wmiStr, 
    u32    wmiStrLen);

u32 
sa_text_wmi_string_length (
    c16* wmiStr);

boolean
sa_text_wmi_str_equal (
    c16* one, 
    c16* two);

u32
sa_text_hash (
    c8* str, 
    u32 len);

u32
sa_text_len (
    const c8* str, 
    u32 maxLength);

u32
sa_wtext_len (
    const c16* str, 
    u32 maxLength);

u32 
sa_text_find_char (
    const c8* str, 
    c8 c, 
    u32 maxLength);

u32 
sa_wtext_find_char (
    const c16* str, 
    c16 c, 
    u32 maxLength);

boolean
sa_text_compare_no_case (
    const c8* a, 
    const c8 *b, 
    u32 maxLength);

boolean
sa_wtext_compare_no_case (
    const c16* a, 
    const c16 *b, 
    u32 maxLength);

u32
sa_text_str_to_byte_array  (
    const c8* str, 
    u32 strLength, 
    u8 *value, 
    u32 valueLength);

u32
sa_text_byte_array_to_base64_str (
    c8* str, 
    u32 strLength, 
    u8 *value, 
    u32 valueLength);

u32
sa_text_base64_str_to_byte_array (
    const c8* str, 
    u32 strLength, 
    u8 *value, 
    u32 valueLength);

SA_CPP_TRAILER

#endif
