/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_defines.h

@brief

    Standard defines from the OS.
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/system/sa_defines.h,v 1.14 2006/03/14 19:46:47 menz Exp $
*/
#ifndef __sa_defines_h__
#define __sa_defines_h__
SA_CPP_HEADER

// Page size
#if defined(SA_WIN_KERNEL)
    #define SA_PAGE_SHIFT  PAGE_SHIFT
    #define SA_PAGE_SIZE   PAGE_SIZE
#else
    #define SA_PAGE_SHIFT  12
    #define SA_PAGE_SIZE   4096 // BUG
#endif

// Standard MIN and MAX macros
#if !defined(MIN) && !defined(NOMINMAX)
   #define MIN(_a_,_b_) (((_a_)<(_b_)) ? (_a_) : (_b_))
#endif
#if !defined(MAX) && !defined(NOMINMAX)
   #define MAX(_a_, _b_) (((_a_)>(_b_)) ? (_a_) : (_b_))
#endif
#define SA_MIN MIN
#define SA_MAX MAX

#ifndef UNALIGNED
    #define UNALIGNED
#endif

// Constants.  
// These are primarily for compatibility between 32 and 64-bit machines.
// On 32-bit machines the suffix UL denotes 32-bit unsigned long.
// On 64-bit machines the suffix UL denotes 64-bit unsigned long.  
// Hence, it is not very useful when you are dealing with exact sized types.
#define SA_U8_0  ((u8)0)
#define SA_U16_0 ((u16)0)
#define SA_U32_0 ((u32)0)
#define SA_U64_0 ((u64)0)
#define SA_I8_0  ((i8)0)
#define SA_I16_0 ((i16)0)
#define SA_I32_0 ((i32)0)
#define SA_I64_0 ((i64)0)

#define SA_U8_F  (~SA_U8_0)
#define SA_U16_F (~SA_U16_0)
#define SA_U32_F (~SA_U32_0)
#define SA_U64_F (~SA_U64_0)
#define SA_I8_F  (~SA_I8_0)
#define SA_I16_F (~SA_I16_0)
#define SA_I32_F (~SA_I32_0)
#define SA_I64_F (~SA_I64_0)


// Appropriate casts to avoid compiler warnings about casting between
// integers and pointers.
#define SA_PTR_TO_INT(_p_) ((SA_SIZE_T)(_p_))
#define SA_PTR_TO_U32(_p_) ((u32)SA_PTR_TO_INT(_p_))
#define SA_PTR_TO_U64(_p_) ((u64)SA_PTR_TO_INT(_p_))

#define SA_U64_TO_PTR(_u_) ((UNALIGNED PVOID)(SA_SIZE_T)(_u_))
#define SA_U32_TO_PTR(_u_) ((UNALIGNED PVOID)(SA_SIZE_T)(_u_))
#define SA_INT_TO_PTR(_u_) ((UNALIGNED PVOID)(SA_SIZE_T)(_u_))

// With enough casts, anything can be a u32.
#define SA_TOU32(_struct_)  (*((u32*)(&(_struct_))))  

// Conversion from 64 bit values to hi and lo 32-bit values.
STATIC FORCEINLINE u32 sa_hi (u64 quad) { return  (u32)(quad >> 32); }
STATIC FORCEINLINE u32 sa_lo (u64 quad) { return  (u32)(quad); }
STATIC FORCEINLINE u64 sa_make_u64 (u32 hi, u32 lo) { return (((u64)hi) << 32) | lo; }

STATIC FORCEINLINE u32 sa_pvoid_hi (PVOID p) { return sa_hi (SA_PTR_TO_U64 (p)); }
STATIC FORCEINLINE u32 sa_pvoid_lo (PVOID p) { return sa_lo (SA_PTR_TO_U64 (p)); }
STATIC FORCEINLINE PVOID sa_make_pvoid (u32 hi, u32 lo) { 
    return SA_U64_TO_PTR (sa_make_u64(hi, lo)); 
}

// Convert macro token to a string.
#define SA_STRINGIFY(_token_) #_token_ 

// Returns byte offset of field within given struct type.
#define SA_FIELD_OFFSET(_type_, _field_)              \
    (SA_PTR_TO_U32 ((u8*)&(((_type_ *)0)->_field_)))  \

// Returns byte size of given field with a structure.
#define SA_SIZEOF_FIELD( _type_, _field_ )  \
    sizeof((( _type_*)0)->_field_)          \

// Returns the number of items in the field array.
#define SA_NUMBER_OF_FIELD(_type_, _field_)                                \
    (SA_SIZEOF_FIELD(_type_, _field_)/sizeof((((_type_*)0)->_field_[0])))  \

// Returns number of bytes in the struct from start, up thru and including the
// given field.
#define SA_SIZEOF_THROUGH_FIELD(_type_, _field_)                           \
    (SA_FIELD_OFFSET(_type_, _field_) + SA_SIZEOF_FIELD(_type_, _field_))  \

// Explicitly acknowledges that an input parameter is ignored.
#define SA_UNREFERENCED_PARAMETER(_p_)  ((_p_) = (_p_))
#define SA_NOT_USED SA_UNREFERENCED_PARAMETER  // abbreviation

//  Returns number of pages spanned by the size of data starting at the given address.
#define SA_PAGES_SPANNED(_address, _size)                                                                    \
   ((u32)(((SA_PTR_TO_INT(_address) & (SA_PAGE_SIZE -1)) + (_size) + (SA_PAGE_SIZE - 1)) >> SA_PAGE_SHIFT))  \

// number of elements in an array
//  e.g.  u32 array[128];
//  SA_NUMBER_OF(array) == 128
#define SA_NUMBER_OF(_x_) (sizeof(_x_)/sizeof((_x_)[0]))

// An expression that yields the _type_ of a field in a struct.
#define SA_FIELD_TYPE(_type_, _field_) (((_type_*)0)->_field_)

// Calculate the address of the base of the structure given its type, and an
// address of a field within the structure.
#define SA_CONTAINING_RECORD(_address_, _type_, _field_) \
    ((_type_ *)((u8*)(_address_) - (u8*)(&((_type_ *)0)->_field_)))

// Gives the pointers page offset as an integer, from 0 to pagesize-1.
#define SA_PAGE_OFFSET(_addr_) (SA_PTR_TO_INT(_addr_) & (SA_PAGE_SIZE-1))


// Returns the next higher aligned address to the given byte multiple.
STATIC INLINE
PVOID SA_ALIGN_PTR (PVOID va, u32 multiple) {
    u64 addr = SA_PTR_TO_U64 (va);

    if (multiple == 0) {
        return va;
    }

    addr = (addr / multiple) + 1;
    addr *= multiple;

    return SA_U64_TO_PTR (addr);
}

// Returns the next higher aligned address to the given byte multiple.
STATIC INLINE
u64 SA_ALIGN64 (u64 value, u32 multiple) {
    if (multiple == 0) {
        return value;
    }
    value = (value / multiple) + 1;
    value *= multiple;
    return value;
}


// Returns the next higher page aligned address.
STATIC INLINE
PVOID SA_PAGE_ALIGN (PVOID _addr_) {
    u64 addr = SA_PTR_TO_U64 (_addr_);

    if( addr & (SA_PAGE_SIZE - 1)) {
        addr = (addr & ~(SA_PAGE_SIZE - 1)) + SA_PAGE_SIZE ;
        return SA_U64_TO_PTR (addr);
    }
    return _addr_ ;
}

// Returns the next higher page aligned address.
STATIC INLINE
u64 SA_PAGE_ALIGN64 (u64 addr) {
    if( addr & (SA_PAGE_SIZE - 1ULL)) {
        addr = (addr & ~(SA_PAGE_SIZE - 1ULL)) + SA_PAGE_SIZE ;
        return addr;
    }
    return addr ;
}


// Units.
#define SA_TB (1024*1024*1024*1024ULL)
#define SA_GB (1024*1024*1024)
#define SA_MB (1024*1024)
#define SA_KB (1024)


// TODO ---
#if defined(SA_MICROSOFT) && defined(SA_KERNEL)
//    #if SA_WIN_VERSION >= 2600
        #define  SA_NUMBER_PROCESSORS    KeNumberProcessors
//    #else
//        #define  SA_NUMBER_PROCESSORS    *KeNumberProcessors
//    #endif
//#else
//    #define  SA_NUMBER_PROCESSORS 1
//    // TODO ?
#endif
    

SA_CPP_TRAILER    
#endif

