/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_dev_storport.h

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/system/sa_dev_sport.h,v 1.2 2006/03/14 19:55:32 menz Exp $
*/
#ifndef __sa_dev_storport_h__
#define __sa_dev_storport_h__

#include "sa_types.h"
#include "sa_assert.h"

// The device object uses a "magic" field to help catch memory
// corruption bugs.  It is initialized to the magic number defined here
// and ASSERTS verify this number in all other functions.
#define SA_DEV_STORPORT_MAGIC ((u32)'sdev')
#define SA_DEV_ASSERT(_pdev_) ASSERT ((_pdev_)->magic == SA_DEV_STORPORT_MAGIC)

// The storage miniport platform supports up to this number of BARs on the device.
#define SA_DEV_MAX_NUM_BARS (8)

//
// Context structure passed to the sa_dev_get_bus_address function.  It
// is an optional SRB pointer.
//
typedef SCSI_REQUEST_BLOCK    SA_DEV_BUS_ADDRESS_CONTEXT, *PSA_DEV_BUS_ADDRESS_CONTEXT;

//
// Context structure passed to create the device object.  This is the
// device extension for the storage miniport.
//
typedef PVOID                 SA_DEV_OS_HANDLE,           *PSA_DEV_OS_HANDLE;

//
// Context structure for logging an error message.
//
typedef struct _SA_DEV_LOG_ERROR_CONTEXT {
    
    PSCSI_REQUEST_BLOCK optional_srb;
    u8 path_id;
    u8 target_id;
    u8 lun;

} SA_DEV_LOG_ERROR_CONTEXT,  *PSA_DEV_LOG_ERROR_CONTEXT;



//
// Device Object
//
typedef struct _SA_DEV {

    u32                          magic;
    u32                          num_bars;                       // number of register bars
    SA_DEV_OS_HANDLE             dev_ext;                        // device extension pointer
    SA_DEV_BAR_LOCATIONS         bars[SA_DEV_MAX_NUM_BARS];      // info per bar (from create function)    

    u32 csr_bar_num, pd_bar_num, pci_bar_num;

} SA_DEV, *PSA_DEV;

#endif

