/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_mem_storport.h

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/system/sa_mem_sport.h,v 1.3 2006/03/14 19:47:20 menz Exp $
*/
#ifndef __sa_mem_storport_h__
#define __sa_mem_storport_h__

#include "sa_mem.h"

typedef struct _SA_SGL {
    PVOID va;
    SA_PHYSICAL_ADDRESS pa;
    u32 length;
} SA_SGL, *PSA_SGL;

STATIC INLINE
u32
sa_sgl_get_byte_count(
    IN PSA_SGL sgl )
{
    ASSERT (sgl);
    return  sgl->length;
}

STATIC INLINE u32
sa_sgl_get_page_count (PSA_SGL sgl) {
    ASSERT (sgl);
    return SA_PAGES_SPANNED (sgl->va, sgl->length);
}

STATIC INLINE PVOID
sa_sgl_get_base_va (
    IN PSA_SGL sgl)
{
    ASSERT (sgl);
    return sgl->va;
}

STATIC INLINE u32
sa_sgl_get_offset (
    PSA_SGL sgl)
{
    ASSERT (sgl);
    return (u32)SA_PAGE_OFFSET (sgl->va);
}

STATIC INLINE 
SA_PHYSICAL_ADDRESS
sa_sgl_get_page (
    IN  PSA_SGL  sgl, 
    IN  u32      page_index)
{
    SA_PHYSICAL_ADDRESS pa_aligned;    

    ASSERT (sgl);
    ASSERT (page_index < sa_sgl_get_page_count (sgl));

    pa_aligned = sgl->pa & ~(SA_PAGE_SIZE-1);
    return pa_aligned + page_index * SA_PAGE_SIZE;
}



#endif // __sa_mem_storport_h__

