/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*
 * Universal Error Code definitions
 */
/*   3 3 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0
 *   1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0  
 *  |-------|-------|-----|-----|---|-------------------------------|
 *  |   M   |   S   |  L  |  P  | A |          Error Code           |
 *  |-------|-------|-----|-----|---|-------------------------------|
 *
 *  M - Module (4-bits)
 *  S - Sub module (4-bits)
 *  L - Error Level (3-bits)
 *	A - Additional Length (2-bits)
 *  Error Code (16-bits)
 *
 */

typedef enum _UNI_MODULE               UNI_MODULE;
typedef unsigned int                   UNI_SUBMODULE;
typedef unsigned int                   UNI_ERROR_LEVEL;
typedef enum _UNI_PHASE                UNI_PHASE;

typedef enum _UNI_SUBMODULE_ACIT       UNI_SUBMODULE_ACIT;
typedef enum _UNI_SUBMODULE_WINOSM     UNI_SUBMODULE_WINOSM;
typedef enum _UNI_SUBMODULE_LNXOSM     UNI_SUBMODULE_LNXOSM;
typedef enum _UNI_SUBMODULE_SOLOSM     UNI_SUBMODULE_SOLOSM;

enum _UNI_MODULE {
    UNI_MODULE_ACIT = 1,
    UNI_MODULE_FIRMWARE,
	UNI_MODULE_OSM,
	UNI_MODULE_MILI,
	UNI_MODULE_CLI,
	UNI_MODULE_GUI
};

enum _UNI_SUBMODULE_ACIT {
    UNI_SUBMODULE_ACIT_GENERIC = 1,
    UNI_SUBMODULE_ACIT_ISCSI = 2,
    UNI_SUBMODULE_ACIT_MGMT = 3,
    UNI_SUBMODULE_ACIT_HWI = 4,
};

enum _UNI_SUBMODULE_WINOSM {
    UNI_SUBMODULE_WINOSM_GENERIC = 1,
    UNI_SUBMODULE_WINOSM_IOCTL = 2,
    UNI_SUBMODULE_WINSOM_PORT_DRIVER_API = 3,
    UNI_SUBMODULE_WINOSM_WMI = 4,
};

enum _UNI_SUBMODULE_LNXOSM {
    UNI_SUBMODULE_LNXOSM_GENERIC = 1,
    UNI_SUBMODULE_LNXOSM_IOCTL = 2,
};

enum _UNI_SUBMODULE_SOLOSM {
    UNI_SUBMODULE_SOLOSM_GENERIC = 1,
    UNI_SUBMODULE_SOLOSM_IOCTL = 2,
};

#define UNI_CFG_LOG_LEVEL_NONE      0
#define UNI_CFG_LOG_LEVEL_INFO      1
#define UNI_CFG_LOG_LEVEL_WARN      2
#define UNI_CFG_LOG_LEVEL_ERROR     3
#define UNI_CFG_LOG_LEVEL_CRITICAL  4

enum _UNI_PHASE {
    UNI_PHASE_DONT_CARE = 0,
    UNI_PHASE_INIT = 1,
    UNI_PHASE_POWER_MGMT = 2,
    UNI_PHASE_IOCTL = 3,
    UNI_PHASE_IO = 4,
    UNI_PHASE_TMF = 5,
    UNI_PHASE_RECOVERY = 6,
    UNI_PHASE_SHUTDOWN = 7,
};

/* Bit Masks and shift macros */
#define UNI_MODULE_BIT_MASK         (0xf0000000)
#define UNI_MODULE_BIT_SHIFT        (28)

#define UNI_SUBMODULE_BIT_MASK      (0x0f000000)
#define UNI_SUBMODULE_BIT_SHIFT     (24)

#define UNI_ERROR_LEVEL_BIT_MASK    (0x00e00000)
#define UNI_ERROR_LEVEL_BIT_SHIFT   (21)

#define UNI_PHASE_BIT_MASK          (0x001c0000)
#define UNI_PHASE_BIT_SHIFT         (18)

#define UNI_EXTENSION_BIT_MASK	    (0x00030000)
#define UNI_EXTENSION_BIT_SHIFT	    (16) 

#define UNI_ERROR_CODE_BIT_MASK     (0x0000ffff)
#define UNI_ERROR_CODE_BIT_SHIFT    (0)

/* Error code definitions for each module */

typedef enum _ACIT_ERROR_CODE {

    /*
     * ACIT uses the following scheme for error codes.
     * The 16 bits are divided into groups of 4 bits,
     * one for each submodule.
     * Generic  - bits 0-3
     * iSCSI    - bits 4-7
     * MGMT     - bits 8-11
     * HWI      - bits 12-15
     */

    ACIT_EC_CONTROLLER_FAILURE = 0x1,
    ACIT_EC_FIRMWARE_MISMATCH,
    ACIT_EC_LINK_DOWN,
    ACIT_EC_LINK_UP,
    ACIT_EC_LDTO_EXPIRY,
    ACIT_EC_TARGET_ETO_EXPIRY,
    ACIT_EC_UNHANDLED_TMF_REQUEST,
    ACIT_EC_TASK_INVALID_FOR_OSM_ABORT,
    ACIT_EC_FW_CONFIG_FAILED,
    ACIT_EC_LOGICAL_LINK_DOWN,
    ACIT_EC_FW_CFG_BAD_ISCSI_IPL,

    ACIT_EC_ISCSI_SESSION_RESET = 0x0010,
    ACIT_EC_TCP_CONNECTION_RESET,
    ACIT_EC_ISCSI_CMDSN_ERROR,
    ACIT_EC_PING_WAIT_TIMEOUT,

    ACIT_EC_ARM_INIT_STATUS = 0x0100,
    ACIT_EC_OFFLOAD_ERROR,
    ACIT_EC_IOCTL_HANDLE_STATUS,
    ACIT_EC_MCC_SGE_EXCEEDED,


    ACIT_EC_HWI_UE = 0x1000,
    ACIT_EC_HWI_INIT_FAILED,

} ACIT_ERROR_CODE, *PACIT_ERROR_CODE;

typedef enum _OSM_ERROR_CODE {

    OSM_EC_HW_INIT_FAILED = 0x1,
    OSM_EC_GET_SGLIST_FAILED = 0x2,
    OSM_EC_OSM_API_TABLE_INITIALIZE_FAILED = 0x3,
    OSM_EC_ACIT_API_FAILED = 0x4,
    OSM_EC_UNSUPPORTED_HARDWARE = 0x5,
    OSM_EC_RESOURCE_ALLOCATION_FAILURE = 0x6,
    OSM_EC_WMI_DRIVER_ERROR_CODE = 0x7,
    OSM_EC_WMI_IOCTL_ERROR_CODE = 0x8,
    OSM_EC_CONFIG_PARAMETER_OUTSIDE_RANGE = 0x9,
    OSM_EC_CONFIG_PARAMETER_NOT_ADJUSTABLE = 0xa,
    OSM_EC_CONFIG_PARAMETER_INVALID = 0xb,
    OSM_EC_MAP_REGISTERS_FAILED = 0xc,
    OSM_EC_REDUCED_CAPABILITY_CONFIG = 0xd,
    OSM_EC_DRIVER_LOAD_SUCCESS = 0xe,
    OSM_EC_TARGET_RECONNECTED = 0xf,
    OSM_EC_INTR_RED_DY_FY = 0x10,
    OSM_EC_INTR_RED_DY_FN = 0x11,
    OSM_EC_UNSUPPORTED_HW_WIN7 = 0x12,
    OSM_EC_INTR_RED_DY_HN = 0x13

} OSM_ERROR_CODE, *POSM_ERROR_CODE;


/* current log level set by default */
#define UNI_CFG_LOG_LEVEL UNI_CFG_LOG_LEVEL_ERROR

// sub error codes

// For ACIT_EC_HWI_INIT_FAILED

typedef enum _ACIT_EC_HWI_INIT_FAILED_SUBERROR {

    HWI_ERR_SA_DEV_CREATE = 0x0,
    HWI_ERR_BECLIB_INIT = 0x1,
    HWI_ERR_FUN_CHIP_CREATE = 0x2,
    HWI_ERR_SOFT_RESET = 0x3,
    HWI_ERR_EQ_CREATE = 0x4,
    HWI_ERR_SUPPORTED_FW_CHECK = 0x5,
    HWI_ERR_FW_CONFIG_GET = 0x6,
    HWI_ERR_MCC_CQ_CREATE = 0x7,
    HWI_ERR_MCC_RING_CREATE = 0x8,
    HWI_ERR_ISCSI_CQ_CREATE = 0x9,
    HWI_ERR_DPDU_HDR_RING_CREATE = 0xa,
    HWI_ERR_DPDU_DATA_RING_CREATE = 0xb,
    HWI_ERR_FRAGNUM_SET = 0xc,
    HWI_ERR_SGL_PAGES_POST = 0xd,
    HWI_ERR_WRB_QUEUE_CREATE = 0xe,
    HWI_ERR_FUNCTION_RESET = 0xf,
    HWI_ERR_TEMPLATE_HEADER_POST = 0x10

} ACIT_EC_HWI_INIT_FAILED_SUBERROR;

